﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Geometry.GGraphic.Controls.ElementaryControls;
using System.Windows.Forms;
using Geometry.GGraphic.Forms.Detail.Layout;
using System.Drawing;

namespace Geometry.GGraphic.Controls.GroupBoxes
{
    public partial class DynamicGroupBox : GroupBox, IFakeable
    {
        private GroupBoxLayoutStyle _layoutStyle;
        private Size _minimumSize;

        public int DynamicWidth { get; set; }
        public int DynamicHeight { get; set; }
        public Point DynamicLocation { get; set; }


        /// <summary>
        /// DynamicTabPage nastaví všem svým groupBoxům a jejim kontrolům velikosti 
        /// a pozice hned po prvním vytvoření a pak už je nemění.
        /// 
        /// Až se přidají všechny groupboxi s controly, zavolat SetConstantSize.
        /// - Tato metoda projde každý groupBox a zavolá na něm SetMinimumSize, zárověň jim nastaví pozici.
        /// - Potom nastaví všem groupBoxům stejnou velikost zavoláním SetConstantSize.
        /// Nakonec si nastaví Width a Height, které čte detail.
        /// </summary>
        public DynamicGroupBox()
        {
            _minimumSize = new Size(300, 100);
        }
        /// <summary>
        /// Potomci této třídy nastaví všem svým groupBoxům a jejim kontrolům velikosti 
        /// a pozice hned po prvním vytvoření a pak už je nemění.
        /// 
        /// Až se přidají všechny groupboxi s controly, zavolat SetConstantSize.
        /// - Tato metoda projde každý groupBox a zavolá na něm SetMinimumSize, zárověň jim nastaví pozici.
        /// - Potom nastaví všem groupBoxům stejnou velikost zavoláním SetConstantSize.
        /// Nakonec si nastaví Width a Height, které čte detail.
        /// </summary>
        public DynamicGroupBox(Size minimumSize)
            : base()
        {
            this._minimumSize = minimumSize;
        }

        public void SetContents(List<Control> contents, GroupBoxLayoutStyle layoutStyle)
        {
            this._layoutStyle = layoutStyle;
            foreach (Control c in contents)
                this.Controls.Add(c);
        }

        /// <summary>
        /// Předpokládá uložení prvků jako při psaní!
        /// </summary>
        public void SetDynamicMinimumSize()
        {
            if (this._layoutStyle == GroupBoxLayoutStyle.Equations)
                SetEquationsMinimumSize();
            if (this._layoutStyle == GroupBoxLayoutStyle.ControlOnLeft)
                SetControlOnLeftMinimumSize();
            if (this._layoutStyle == GroupBoxLayoutStyle.ControlInCenter)
                SetControlInCenterMinimumSize();
        }

        private void SetControlInCenterMinimumSize()
        {
            if (this.Controls.Count != 1)
                throw new Exception("Špatný počet prvků v groupBoxu.");

            IFakeable control = (IFakeable)this.Controls[0];
            int height = control.DynamicHeight;
            this.DynamicWidth = LayoutProperties.GroupBoxSideSpace * 2 + control.DynamicWidth;
            this.DynamicHeight = LayoutProperties.GroupBoxTopSpace + LayoutProperties.GroupBoxBottomSpace + height;
            this.DynamicWidth = Math.Max(_minimumSize.Width, this.DynamicWidth);
        }
        private void SetControlOnLeftMinimumSize()
        {
            if (this.Controls.Count != 1)
                throw new Exception("Špatný počet prvků v groupBoxu.");

            IFakeable control = (IFakeable)this.Controls[0];
            this.DynamicWidth = LayoutProperties.GroupBoxSideSpace * 2 + control.DynamicWidth;
            this.DynamicHeight = LayoutProperties.GroupBoxTopSpace + LayoutProperties.GroupBoxBottomSpace + control.DynamicHeight;
            this.DynamicWidth = Math.Max(_minimumSize.Width, this.DynamicWidth);
        }
        private void SetEquationsMinimumSize()
        {
            if (this.Controls.Count != 3)
                throw new Exception("Špatný počet prvků v groupBoxu.");

            IFakeable A = (IFakeable)this.Controls[0];
            IFakeable B = (IFakeable)this.Controls[1];
            IFakeable C = (IFakeable)this.Controls[2];
            this.DynamicWidth =
                LayoutProperties.GroupBoxSideSpace * 2 +
                A.DynamicWidth + B.DynamicWidth + C.DynamicWidth +
                LayoutProperties.ControlsSideSpace * 2;
            this.DynamicHeight = Math.Max(Math.Max(A.DynamicHeight, B.DynamicHeight), C.DynamicHeight) +
                LayoutProperties.GroupBoxTopSpace + LayoutProperties.GroupBoxBottomSpace;
            this.DynamicWidth = Math.Max(_minimumSize.Width, this.DynamicWidth);
        }



        private int ComputeYCoordinateToBeInCenter(IFakeable control)
        {
            int height = control.DynamicHeight;
            return (this.DynamicHeight
                - LayoutProperties.GroupBoxTopSpace
                - LayoutProperties.GroupBoxBottomSpace) / 2
                - height / 2
                + LayoutProperties.GroupBoxTopSpace;
        }
        private int ComputeXCoordinateToBeInCenter(IFakeable control)
        {
            return (this.DynamicWidth - control.DynamicWidth) / 2;
        }


        public void SetDynamicWidth(int width)
        {
            this.DynamicWidth = width;
            if (this._layoutStyle == GroupBoxLayoutStyle.Equations)
                SetEquationsDynamicSize();
            if (this._layoutStyle == GroupBoxLayoutStyle.ControlOnLeft)
                SetOnLeftDynamicSize();
            if (this._layoutStyle == GroupBoxLayoutStyle.ControlInCenter)
                SetInCenterDynamicSize();
        }
        private void SetInCenterDynamicSize()
        {
            if (this.Controls.Count != 1)
                throw new Exception("Špatný počet prvků v groupBoxu.");

            IFakeable control = (IFakeable)this.Controls[0];
            int x = (this.DynamicWidth - control.DynamicWidth) / 2;
            control.DynamicLocation = new Point(x, LayoutProperties.GroupBoxTopSpace);
        }
        private void SetOnLeftDynamicSize()
        {
            if (this.Controls.Count != 1)
                throw new Exception("Špatný počet prvků v groupBoxu.");

            IFakeable control = (IFakeable)this.Controls[0];
            control.DynamicLocation = new Point(LayoutProperties.GroupBoxSideSpace, LayoutProperties.GroupBoxTopSpace);
        }
        private void SetEquationsDynamicSize()
        {
            if (this.Controls.Count != 3)
                throw new Exception("Špatný počet prvků v groupBoxu.");

            IFakeable A = (IFakeable)this.Controls[0];
            IFakeable B = (IFakeable)this.Controls[1];
            IFakeable C = (IFakeable)this.Controls[2];

            A.DynamicLocation = new Point(LayoutProperties.GroupBoxSideSpace, ComputeYCoordinateToBeInCenter(A));
            C.DynamicLocation = new Point(this.DynamicWidth - C.DynamicWidth - LayoutProperties.GroupBoxSideSpace, ComputeYCoordinateToBeInCenter(C));
            B.DynamicLocation = new Point(C.DynamicLocation.X - B.DynamicWidth - LayoutProperties.ControlsSideSpace, ComputeYCoordinateToBeInCenter(B));
        }

        public void SetRealDynamicSize()
        {
            this.Bounds = new Rectangle(DynamicLocation.X, DynamicLocation.Y, DynamicWidth, DynamicHeight);
            foreach (IFakeable fakeable in this.Controls)
                fakeable.SetRealDynamicSize();
        }
    }
}
