﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Geometry;

namespace GMath
{
    /// <summary>
    /// |ux uy uz 0|
    /// |vx vy vz 0|
    /// |wx wy wz 0|
    /// |m  n  p  1|
    /// </summary>
    public class TransformationMatrix:Matrix
    {

        private TransformationMatrix(Matrix src) :base(src)
        {
        }

        public TransformationMatrix(int size)
            : base(size)
        {
        }

        public TransformationMatrix(Vector x, Vector y, Vector z) 
        {
            _width = 4;
            _height = 4;
            _values = new double[]
            {
                x.GetX(),x.GetY(),x.GetZ(),0,
                y.GetX(),y.GetY(),y.GetZ(),0,
                z.GetX(),z.GetY(),z.GetZ(),0,
                0,0,0,1
            };
            double determinant = this.Determinant4(); // pravotociva baze.
            if (determinant < 0)
                this.MultIthRow(1, -1);
        }

        public TransformationMatrix(int width, int height, double[] values) :base(width, height, values)
        {
        }


        public TransformationMatrix(TransformationMatrix src):base(src) 
        {
        }

        public TransformationMatrix(RotationAngles angles):base(4) 
        {
            TransformationMatrix xMatrix = TransformationMatrix.GetXRotationMatrix(angles.GetXAngle());
            TransformationMatrix yMatrix = TransformationMatrix.GetYRotationMatrix(angles.GetYAngle());
            TransformationMatrix zMatrix = TransformationMatrix.GetZRotationMatrix(angles.GetZAngle());
            _values = (zMatrix * yMatrix * xMatrix)._values;
            //TransformationMatrix temp = zMatrix * yMatrix * xMatrix;
            //MatrixComparer mc = new MatrixComparer(this, temp.TransponMatrix());
            //mc.ShowDialog();
        }

        public TransformationMatrix(RotationAngles angles, TransformationMatrix rotationMatrix)
            : base(4) 
        {
            TransformationMatrix xMatrix = new TransformationMatrix(4);
            TransformationMatrix yMatrix = new TransformationMatrix(4);
            TransformationMatrix zMatrix = new TransformationMatrix(4);
            if (angles.GetXAngle() != 0)
            {
                xMatrix = GetRotationMatrixAroundRotatedXAxe(rotationMatrix, angles.GetXAngle());
            }
            if (angles.GetYAngle() != 0)
            {
                yMatrix = GetRotationMatrixAroundRotatedYAxe(rotationMatrix, angles.GetYAngle());                
            }
            if (angles.GetZAngle() != 0)
            {
                zMatrix = GetRotationMatrixAroundRotatedZAxe(rotationMatrix, angles.GetZAngle());
            }
            _values = (zMatrix * yMatrix * xMatrix)._values;
        }



        public static TransformationMatrix operator *(TransformationMatrix a, TransformationMatrix b) 
        {
            Matrix matrixA = new Matrix(a);
            Matrix matrixB = new Matrix(b);
            return new TransformationMatrix(matrixA * matrixB);
        }

        public new TransformationMatrix ComputeInverseMatrix() 
        {
            Matrix matrix = new Matrix(_width, _height, _values).ComputeInverseMatrix();
            return new TransformationMatrix(matrix);
        }

        public new TransformationMatrix TransponMatrix() 
        {
            Matrix matrix = new Matrix(_width, _height, _values).TransponMatrix();
            return new TransformationMatrix(matrix);
        }

        
        
        public void SetTranslation(Vector translation) 
        {
            _values[12] = translation.GetX();
            _values[13] = translation.GetY();
            _values[14] = translation.GetZ();
        }

        public Vector GetTranslation() 
        {
            return new Vector(_values[12], _values[13], _values[14], 1);
        }



        public static TransformationMatrix GetXRotationMatrix(double degrees)
        {
            double fi = (degrees / 180) * Math.PI;
            double cf = Math.Cos(fi);
            double sf = Math.Sin(fi);
            double[] values = new double[]
            {
                1,0,0,0,
                0,cf,-sf,0,
                0,sf,cf,0,
                0,0,0,1
            };
            return new TransformationMatrix(4, 4, values);
        }

        public static TransformationMatrix GetYRotationMatrix(double degrees)
        {
            double fi = (degrees / 180) * Math.PI;
            double cf = Math.Cos(fi);
            double sf = Math.Sin(fi);
            double[] values = new double[]
            {
                cf,0,sf,0,
                0,1,0,0,
                -sf,0,cf,0,
                0,0,0,1
            };
            return new TransformationMatrix(4, 4, values);
        }

        public static TransformationMatrix GetZRotationMatrix(double degrees)
        {
            double fi = (degrees / 180) * Math.PI;
            double cf = Math.Cos(fi);
            double sf = Math.Sin(fi);
            double[] values = new double[]
            {
                cf,-sf,0,0,
                sf,cf,0,0,
                0,0,1,0,
                0,0,0,1
            };
            return new TransformationMatrix(4, 4, values);
        }



        public RotationAngles DecomposeRotationMatrixOnAngles()
        {
            TransformationMatrix copy = this;
            RotationAngles angles = copy.DecomposeRotationMatrixHelp();
            //if (angles.GetXAngle() > 180 || angles.GetXAngle() < -180 ||
            //    angles.GetYAngle() > 180 || angles.GetYAngle() < -180 ||
            //    angles.GetZAngle() > 180 || angles.GetZAngle() < -180)
                //throw new Exception("");
            return angles;
        }

        private RotationAngles DecomposeRotationMatrixHelp()
        {
            TransformationMatrix copy = this;
            double r11 = copy._values[0];
            double r12 = copy._values[1];
            double r13 = copy._values[2];
            double r21 = copy._values[4];
            double r31 = copy._values[8];
            double r32 = copy._values[9];
            double r33 = copy._values[10];

            if (Math.Abs(r31) < 1 && Math.Asin(r31) != double.NaN)
            {
                double y1 = -Math.Asin(r31);
                double y2 = Math.PI - y1;
                double cosy1 = Math.Cos(y1);
                double cosy2 = Math.Cos(y2);

                double x1 = Math.Atan2(r32 / cosy1, r33 / cosy1);
                double x2 = Math.Atan2(r32 / cosy2, r33 / cosy2);

                double z1 = Math.Atan2(r21 / cosy1, r11 / cosy1);
                double z2 = Math.Atan2(r21 / cosy2, r11 / cosy2);
                return new RotationAngles(x1 * 180 / Math.PI, y1 * 180 / Math.PI, z1 * 180 / Math.PI);
            }
            else
            {
                if (r31 < 0) // == -1 
                {
                    double x1 = 0;
                    double y1 = Math.PI / 2.0;
                    double z1 = x1 + Math.Atan2(r12, r13);
                    return new RotationAngles(x1 * 180 / Math.PI, y1 * 180 / Math.PI, z1 * 180 / Math.PI);
                }
                else // == 1w
                {
                    double x2 = 0;
                    double y2 = -Math.PI / 2.0;
                    double z2 = -x2 + Math.Atan2(-r12, -r13);
                    return new RotationAngles(x2 * 180 / Math.PI, y2 * 180 / Math.PI, z2 * 180 / Math.PI);
                }
            }
        }
        
        public List<RotationAngles> DecomposeRotationMatrix2()
        {
            TransformationMatrix copy = this;
            double r11 = copy._values[0];
            double r12 = copy._values[1];
            double r13 = copy._values[2];
            double r21 = copy._values[4];
            double r31 = copy._values[8];
            double r32 = copy._values[9];
            double r33 = copy._values[10];

            if (Math.Abs(r31) < 1 && Math.Asin(r31) != double.NaN)
            {
                double y1 = -Math.Asin(r31);
                double y2 = Math.PI - y1;
                double cosy1 = Math.Cos(y1);
                double cosy2 = Math.Cos(y2);

                double x1 = Math.Atan2(r32 / cosy1, r33 / cosy1);
                double x2 = Math.Atan2(r32 / cosy2, r33 / cosy2);

                double z1 = Math.Atan2(r21 / cosy1, r11 / cosy1);
                double z2 = Math.Atan2(r21 / cosy2, r11 / cosy2);
                //if (x2 == 0)
                //return new RotationAngles(x2 * 180 / Math.PI, y2 * 180 / Math.PI, z2 * 180 / Math.PI);
                return new List<RotationAngles>(){
                    new RotationAngles(x1 * 180 / Math.PI, y1 * 180 / Math.PI, z1 * 180 / Math.PI),
                    new RotationAngles(x2 * 180 / Math.PI, y2 * 180 / Math.PI, z2 * 180 / Math.PI)};
            }
            else
            {
                if (r31 < 0) // == -1 
                {
                    double x1 = 0;
                    double y1 = Math.PI / 2.0;
                    double z1 = x1 + Math.Atan2(r12, r13);
                    return new List<RotationAngles>(){
                        new RotationAngles(x1 * 180 / Math.PI, y1 * 180 / Math.PI, z1 * 180 / Math.PI)};
                }
                else // == 1w
                {
                    double x2 = 0;
                    double y2 = -Math.PI / 2.0;
                    double z2 = -x2 + Math.Atan2(-r12, -r13);
                    return new List<RotationAngles>(){
                        new RotationAngles(x2 * 180 / Math.PI, y2 * 180 / Math.PI, z2 * 180 / Math.PI)};
                }
            }
        }

        public TransformationMatrix GetRotationMatrixAroundRotatedXAxe(TransformationMatrix rotationMatrix, double angle)
        {
            Vector xAxe = new Vector(1, 0, 0, 1);
            return GetRotationMatrixAroundAAxe(xAxe * rotationMatrix.ComputeInverseMatrix(), angle);
        }

        public TransformationMatrix GetRotationMatrixAroundRotatedYAxe(TransformationMatrix rotationMatrix, double angle)
        {
            Vector yAxe = new Vector(0, 1, 0, 1);
            return GetRotationMatrixAroundAAxe(yAxe * rotationMatrix.ComputeInverseMatrix(), angle);
        }

        public TransformationMatrix GetRotationMatrixAroundRotatedZAxe(TransformationMatrix rotationMatrix, double angle)
        {
            Vector zAxe = new Vector(0, 0, 1, 1);
            return GetRotationMatrixAroundAAxe(zAxe * rotationMatrix.ComputeInverseMatrix(), angle);
        }


        public static TransformationMatrix GetRotationMatrixAroundAAxe(Vector rotatedAxe, double angle)
        {
            double sina = Math.Sin(angle * Math.PI / 180);
            double cosa = Math.Cos(angle * Math.PI / 180);
            Vector rotatedAxeCopy = new Vector(rotatedAxe).NornVector();
            double n1 = rotatedAxeCopy.GetX();
            double n2 = rotatedAxeCopy.GetY();
            double n3 = rotatedAxeCopy.GetZ();


            Vector v1 =
                new Vector(cosa + n1 * n1 * (1 - cosa), n1 * n2 * (1 - cosa) - n3 * sina, n1 * n3 * (1 - cosa) + n2 * sina).NornVector();
            Vector v2 =
                new Vector(n1 * n2 * (1 - cosa) + n3 * sina, cosa + n2 * n2 * (1 - cosa), n2 * n3 * (1 - cosa) - n1 * sina).NornVector();
            Vector v3 =
                new Vector(n1 * n3 * (1 - cosa) - n2 * sina, n2 * n3 * (1 - cosa) + n1 * sina, cosa + n3 * n3 * (1 - cosa)).NornVector();
            double[] values = new double[] 
            {
                v1.GetX(), v1.GetY(), v1.GetZ(), 0,
                v2.GetX(), v2.GetY(), v2.GetZ(), 0,
                v3.GetX(), v3.GetY(), v3.GetZ(), 0,
                0,0,0,1
            };
            return new TransformationMatrix(4, 4, values);
        }



        public TransformationMatrix NormalizeRows()
        {
            Vector a = new Vector(this._values[0], this._values[1], this._values[2]).NornVector();
            Vector b = new Vector(this._values[4], this._values[5], this._values[6]).NornVector();
            Vector c = new Vector(this._values[8], this._values[9], this._values[10]).NornVector();
            double[] values = new double[] 
            {
                a.GetX(), a.GetY(), a.GetZ(), 0,
                b.GetX(), b.GetY(), b.GetZ(), 0,
                c.GetX(), c.GetY(), c.GetZ(), 0,
                _values[12],_values[13],_values[14],1
            };
            return new TransformationMatrix(4, 4, values);
        }

        public TransformationMatrix NormalizeColumns() 
        {
            Vector a = new Vector(this._values[0], this._values[4], this._values[8]).NornVector();
            Vector b = new Vector(this._values[1], this._values[5], this._values[9]).NornVector();
            Vector c = new Vector(this._values[2], this._values[6], this._values[10]).NornVector();
            double[] values = new double[] 
            {
                a.GetX(), b.GetX(), c.GetX(), 0,
                a.GetY(), b.GetY(), c.GetY(), 0,
                a.GetZ(), b.GetZ(), c.GetZ(), 0,
                _values[12],_values[13],_values[14],1
            };
            return new TransformationMatrix(4, 4, values);
        }


        public TransformationMatrix MultIthColumn(int i, double koeficient)
        {
            TransformationMatrix copy = new TransformationMatrix(this);
            for (int k = 0; k < _height; k++)
            {
                copy._values[i + k * _width] = koeficient * copy._values[i + k * _width];
            }
            return copy;
        }

        private void MultIthRow(int i, double koeficient)
        {
            for (int k = 0; k < _width; k++)
            {
                this._values[k + i * _height] = koeficient * this._values[k + i * _height];
            }
        }

        public static bool IsTransformationMatrix(Matrix m)
        {
            if (m.GetHeight() == m.GetWidth() && m.GetWidth() == 3)
                return TransformationMatrix.IsTransformationMatrix2D(m);
            if (m.GetHeight() == m.GetWidth() && m.GetWidth() == 4)
                return TransformationMatrix.IsTransformationMatrix3D(m);
            throw new Exception();

        }
        private static bool IsTransformationMatrix3D(Matrix m)
        {
            Vector v1 = new Vector(m.Get(0), m.Get(4), m.Get(8));
            Vector v2 = new Vector(m.Get(1), m.Get(5), m.Get(9));
            Vector v3 = new Vector(m.Get(2), m.Get(6), m.Get(10));
            Vector v4 = new Vector(m.Get(3), m.Get(7), m.Get(11));

            return
                Program.IsEqualInTolerance(Vector.Size(v1), 1) &&
                Program.IsEqualInTolerance(Vector.Size(v2), 1) &&
                Program.IsEqualInTolerance(Vector.Size(v3), 1) &&
                Program.IsEqualInTolerance(Vector.Size(v4), 0) &&
                Program.IsEqualInTolerance(v1 * v2, 0) &&
                Program.IsEqualInTolerance(v1 * v3, 0) &&
                Program.IsEqualInTolerance(v2 * v3, 0);
        }
        private static bool IsTransformationMatrix2D(Matrix m)
        {
            Vector v1 = new Vector(m.Get(0), m.Get(3), 0);
            Vector v2 = new Vector(m.Get(1), m.Get(4), 0);
            Vector v3 = new Vector(m.Get(2), m.Get(3), 0);
            double s1 = Vector.Size(v1);
            double s2 = Vector.Size(v2);
            double s3 = Vector.Size(v3);
            double sk = v1 * v2;
            return
                Program.IsEqualInTolerance(s1, 1) &&
                Program.IsEqualInTolerance(s2, 1) &&
                Program.IsEqualInTolerance(s3, 0) &&
                Program.IsEqualInTolerance(sk, 0);
        }
    }
}
