﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Geometry;

namespace GMath
{
    public class Equation
    {
        private double[] _row;
        public const int _homogenous = 0;
        public const int _nonHomogenous = 1;
        private  int _version;  

        public Equation(double x,double y, double z) 
        {
            _version = _homogenous;
            _row = new double[3];
            _row[0] = x;
            _row[1] = y;
            _row[2] = z;
        }
        public Equation(double x, double y, double z, double constant)
        {
            _version = _nonHomogenous;
            _row = new double[4];
            _row[0] = x;
            _row[1] = y;
            _row[2] = z;
            _row[3] = constant;
        }
        public Equation(Equation equation) 
        {
            if (equation.GetVersion() == _nonHomogenous)
            {
                _version = _nonHomogenous;
                _row = new double[4];
                _row[0] = equation.GetX();
                _row[1] = equation.GetY();
                _row[2] = equation.GetZ();
                _row[3] = equation.GetConstant();
            }
            else 
            {
                _version = _homogenous;
                _row = new double[3];
                _row[0] = equation.GetX();
                _row[1] = equation.GetY();
                _row[2] = equation.GetZ();
            }
        }

        public int GetVersion() 
        {
            return _version;
        }

        public double GetX() 
        {
            return _row[0];
        }
        public double GetY()
        {
            return _row[1];
        }
        public double GetZ()
        {
            return _row[2];
        }
        public double GetConstant() 
        {
            return _row[3];
        }

        public double Get(int index)
        {
            if (_version == _nonHomogenous && index < 4)
                return _row[index];
            if (_version == _homogenous && index < 3) 
            {
                return _row[index];
            }
            throw new Exception("Get(int) z rovnice byl zavolan se spatnym argumentem.");
        }
        public double GetAbsolutelyMaximalElement() 
        {
            double max = 0;
            foreach (double d in _row)
            {
                max = Math.Max(Math.Abs(d), Math.Abs(max));
            }
            return max;
        }

        public static Equation operator*(Equation equation,double koeficient)
        {
            switch (equation._version) 
            {
                case _homogenous: 
                    return new Equation(equation.GetX() * koeficient,
                        equation.GetY() * koeficient, 
                        equation.GetZ() * koeficient);
                default:
                    return new Equation(equation.GetX() * koeficient,
                        equation.GetY() * koeficient,
                        equation.GetZ() * koeficient,
                        equation.GetConstant() * koeficient);
            }
        }
        public static Equation operator *(double koeficient, Equation equation) 
        {
            return equation * koeficient;
        }
        public static Equation operator +(Equation a, Equation b) 
        {
            switch (a._version) 
            {
                case _homogenous:
                    return new Equation(a.GetX() + b.GetX(),
                                        a.GetY() + b.GetY(),
                                        a.GetZ() + b.GetZ());
                default:
                    return new Equation(a.GetX() + b.GetX(),
                                        a.GetY() + b.GetY(),
                                        a.GetZ() + b.GetZ(),
                                        a.GetConstant() + b.GetConstant());
            }
        }
        public static Equation operator -(Equation a, Equation b) 
        {
            switch (a._version)
            {
                case _homogenous:
                    return new Equation(a.GetX() - b.GetX(),
                                        a.GetY() - b.GetY(),
                                        a.GetZ() - b.GetZ());
                default:
                    return new Equation(a.GetX() - b.GetX(),
                                        a.GetY() - b.GetY(),
                                        a.GetZ() - b.GetZ(),
                                        a.GetConstant() - b.GetConstant());
            }
        }

        public void SwitchColumns(int firstColumnIndex, int secondColumnIndex) 
        {
            double temp = _row[firstColumnIndex];
            _row[firstColumnIndex] = _row[secondColumnIndex];
            _row[secondColumnIndex] = temp;
        }
        public int GetVariablesCount() 
        {
            if (_version == _homogenous)
                return 3;
            else return 4;
        }

        public ParametrizationRST Substitute(OwnVectorParametrization owp)
        {
            return
                GetX() * owp.GetX() +
                GetY() * owp.GetY() +
                GetZ() * owp.GetZ();
        }
        public double Substitute(Vector owp)
        {
            return
                GetX() * owp.GetX() +
                GetY() * owp.GetY() +
                GetZ() * owp.GetZ();
        }

        public bool IsZero()
        {
            bool isZero = true;
            int values;
            if (_version == _homogenous)
                values = 3;
            else values = 4;
            for (int i = 0; i < values; i++)
                isZero = isZero && Program.IsEqualInTolerance(this._row[i], 0);
            return isZero;
        }
        public override string ToString()
        {
            string xText = "";
            string yText = "";
            string zText = "";
            string constant = "";

            if (!Program.IsEqualInTolerance(this._row[0], 0))
                xText = Program.GetOutputFormatedForTextBox(this._row[0]) + "x";
            if (!Program.IsEqualInTolerance(this._row[1], 0))
                yText = Program.GetOutputFormatedForTextBox(this._row[1]) + "y";
            if (!Program.IsEqualInTolerance(this._row[2], 0))
                zText = Program.GetOutputFormatedForTextBox(this._row[2]) + "z";
            if (this._version == _nonHomogenous)
                constant = Program.GetOutputFormatedForTextBox(this._row[3]);

            return xText + yText + zText + constant;
        }
    }
}
