﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Geometry;

namespace GMath
{
    public class OwnVectorEquationSystem:EquationSystem
    {
        private int _ownNumberMultiplicity;

        public OwnVectorEquationSystem(Equation a, Equation b, Equation c, int ownNumberMultiplicity)
            :base(a,b,c) 
        {
            _ownNumberMultiplicity = ownNumberMultiplicity;
        }

        public OwnVectorEquationSystem(OwnVectorEquationSystem src)
            : base(new Equation(src._rows[0]),
                  new Equation(src._rows[1]),
                  new Equation(src._rows[2])) 
        {
            _ownNumberMultiplicity = src._ownNumberMultiplicity;
        }

        ////////////////////////////////////////////////////////////////////////////////////
        /////////////////  Reseni homogenni soustavy rovnic  ///////////////////////////////
        ////////////////////////////////////////////////////////////////////////////////////

        public void MakeTriangleSystemCustom()
        {
            EliminateZeroesRows();
            base.MakeTriangleSystem();
        }
        public void EliminateZeroesRows()
        {
            List<double> biggestMembers = new List<double>();
            foreach (Equation equation in _rows)
                biggestMembers.Add(equation.GetAbsolutelyMaximalElement());
            int smallestElementIndex;
            for (int i = 0; i < _ownNumberMultiplicity; i++)
            {
                smallestElementIndex = GetSmallestElementIndex(biggestMembers);
                _rows[smallestElementIndex] = new Equation(0, 0, 0);
                biggestMembers[smallestElementIndex] = double.MaxValue;
            }
        }
        private int GetSmallestElementIndex(List<double> list)
        {
            int minIndex = 0;
            double min = double.MaxValue;
            for (int i = 0; i < list.Count; i++)
                if (Math.Abs(list[i]) < min)
                {
                    min = list[i];
                    minIndex = i;
                }
            return minIndex;
        }

        /// <summary>
        /// Z trojuhelnikove matice soustavy 3 rovnic o 3 neznamych vyextrahuje řešení.
        /// (navzajem kolme vektory)
        /// </summary>
        /// <param name="eqations"></param>
        /// <returns></returns>
        public OwnVectorParametrization Solve()
        {
            OwnVectorEquationSystem copy = new OwnVectorEquationSystem(this);
            MakeTriangleSystem();
            OwnVectorParametrization solution = SolveTriangleMatrix();
            ParametrizationRST substitution1 = copy._rows[0].Substitute(solution);
            ParametrizationRST substitution2 = copy._rows[1].Substitute(solution);
            ParametrizationRST substitution3 = copy._rows[2].Substitute(solution);
            if (!substitution1.IsZero() || !substitution2.IsZero() || !substitution3.IsZero())
            {
                this._rows = new OwnVectorEquationSystem(copy)._rows;
                MakeTriangleSystemCustom();
                solution = SolveReducedTriangleMatrix();
                substitution1 = copy._rows[0].Substitute(solution);
                substitution2 = copy._rows[1].Substitute(solution);
                substitution3 = copy._rows[2].Substitute(solution);
                if (!substitution1.IsZero() || !substitution2.IsZero() || !substitution3.IsZero())
                    _rows = _rows;
            }
            return solution;
        }
        /// <summary>
        /// Vraci reseni soustavy 3 rovnic o trech neznamich.
        /// </summary>
        /// <param name="matrix"></param>
        /// <returns></returns>
        private OwnVectorParametrization SolveTriangleMatrix() 
        {
            switch (_ownNumberMultiplicity) 
            {
                case 1: return GetOneParameterSolutionExpected();
                case 2: return GetTwoParameterSolutionExpected();
                default: return GetThreeParameterSolutionExpected();
            }
        }
        private OwnVectorParametrization SolveReducedTriangleMatrix() 
        {
            if (Program.IsEqualInTolerance(_rows[0].GetX(), 0))
            {
                if (Program.IsEqualInTolerance(_rows[1].GetY(), 0))
                {
                    if (Program.IsEqualInTolerance(_rows[2].GetZ(), 0))
                        return CaseNNN();
                    else return CaseNNP();
                }
                else
                {
                    if (Program.IsEqualInTolerance(_rows[2].GetZ(), 0))
                        return CaseNPN();
                    else return CaseNPP();
                }
            }
            else
            {
                if (Program.IsEqualInTolerance(_rows[1].GetY(), 0))
                {
                    if (Program.IsEqualInTolerance(_rows[2].GetZ(), 0))
                        return CasePNN();
                    else return CasePNP();
                }
                else
                {
                    if (Program.IsEqualInTolerance(_rows[2].GetZ(), 0))
                        return CasePPN();
                    else return CasePPP();
                }
            }
        }

        private OwnVectorParametrization GetOneParameterSolutionExpected() 
        {
            int probablyZeroRowIndex = 0;
            List<double> rowsBiggestElements = new List<double>();
            foreach (Equation e in _rows) 
            {
                rowsBiggestElements.Add(e.GetAbsolutelyMaximalElement()); // vyberu nejvetsi prvek kazde rovnice
            }
            for (int i = 1; i < rowsBiggestElements.Count; i++)
                if (Math.Abs(rowsBiggestElements[i]) < Math.Abs(rowsBiggestElements[probablyZeroRowIndex]))
                    probablyZeroRowIndex = i;                            // zjistim ktera rovnice ma nejmensiho zastupce
            _rows[probablyZeroRowIndex] = new Equation(0, 0, 0); // znuluju tu nejnulovejsi rovnici
            return SolveReducedTriangleMatrix();
        }
        private OwnVectorParametrization GetTwoParameterSolutionExpected()
        {
            int probablyZeroRowIndex = 0;
            List<double> rowsBiggestElements = new List<double>();
            foreach (Equation e in _rows)
            {
                rowsBiggestElements.Add(e.GetAbsolutelyMaximalElement());// vyberu nejvetsi prvek kazde rovnice
            }
            for (int i = 1; i < rowsBiggestElements.Count; i++)
                if (Math.Abs(rowsBiggestElements[i]) < Math.Abs(rowsBiggestElements[probablyZeroRowIndex]))
                    probablyZeroRowIndex = i;                           // zjistim ktera rovnice ma nejmensiho zastupce
            _rows[probablyZeroRowIndex] = new Equation(0, 0, 0);        // znuluju tu nejnulovejsi rovnici
            rowsBiggestElements[probablyZeroRowIndex] = double.MaxValue; // vyradim znulovanou rovnici z dalsiho zpracovani

            probablyZeroRowIndex = 0;
            for (int i = 1; i < rowsBiggestElements.Count; i++)
                if (Math.Abs(rowsBiggestElements[i]) < Math.Abs(rowsBiggestElements[probablyZeroRowIndex]))
                    probablyZeroRowIndex = i;                   // zjistim ktera rovnice ma druheho nejmensiho zastupce
            _rows[probablyZeroRowIndex] = new Equation(0, 0, 0); // znuluju druhou nejnulovejsi rovnici
            return SolveReducedTriangleMatrix();        
        }
        private OwnVectorParametrization GetThreeParameterSolutionExpected() 
        {
            ParametrizationRST x = new ParametrizationRST(0, 1, 0, 0);
            ParametrizationRST y = new ParametrizationRST(0, 0, 1, 0);
            ParametrizationRST z = new ParametrizationRST(0, 0, 0, 1);
            return new OwnVectorParametrization(x, y, z);
        }

        public bool IsRoot(double x, double y, double z)
        {
            double r1 = _rows[0].GetX() * x + _rows[0].GetY() * y + _rows[0].GetZ() * z;
            double r2 = _rows[1].GetX() * x + _rows[1].GetY() * y + _rows[1].GetZ() * z;
            double r3 = _rows[2].GetX() * x + _rows[2].GetY() * y + _rows[2].GetZ() * z;
            return Program.IsEqualInTolerance(r1, 0) &&
                Program.IsEqualInTolerance(r2, 0) &&
                Program.IsEqualInTolerance(r3, 0);
        }

        /// <summary>
        /// 0 0 0
        /// 0 0 0
        /// 0 0 0
        /// </summary>
        /// <returns></returns>
        private OwnVectorParametrization CaseNNN() 
        {
            ParametrizationRST x = new ParametrizationRST(0, 1, 0, 0);
            ParametrizationRST y = new ParametrizationRST(0, 0, 1, 0);
            ParametrizationRST z = new ParametrizationRST(0, 0, 0, 1);
            return new OwnVectorParametrization(x,y,z);
        }
        /// <summary>
        /// 0 0 0
        /// 0 0 0
        /// 0 0 1
        /// -----
        /// (p,q,0)
        /// </summary>
        /// <returns></returns>
        private OwnVectorParametrization CaseNNP() 
        {
            ParametrizationRST x = new ParametrizationRST(0, 1, 0, 0);
            ParametrizationRST y = new ParametrizationRST(0, 0, 1, 0);
            ParametrizationRST z = new ParametrizationRST(0, 0, 0, 0);
            return new OwnVectorParametrization(x,y,z);
        }
        /// <summary>
        /// 0 0 0
        /// 0 1 x
        /// 0 0 0
        /// -----
        /// (p,x,r)
        /// </summary>
        /// <returns></returns>
        private OwnVectorParametrization CaseNPN() 
        {
            double xx = -_rows[1].GetZ() / _rows[1].GetY();
            ParametrizationRST x = new ParametrizationRST(0, 1, 0, 0);
            ParametrizationRST y = new ParametrizationRST(0, 0, 0, xx);
            ParametrizationRST z = new ParametrizationRST(0, 0, 0, 1);
            return new OwnVectorParametrization(x,y,z);
        }
        /// <summary>
        /// 0 0 0
        /// 0 1 0
        /// 0 0 1
        /// -----
        /// (p,0,0)
        /// </summary>
        /// <returns></returns>
        private OwnVectorParametrization CaseNPP()
        {
            ParametrizationRST x = new ParametrizationRST(0, 1, 0, 0);
            ParametrizationRST y = new ParametrizationRST(0, 0, 0, 0);
            ParametrizationRST z = new ParametrizationRST(0, 0, 0, 0);
            return new OwnVectorParametrization(x,y,z);
        }
        /// <summary>
        /// 1 x x
        /// 0 0 0
        /// 0 0 0
        /// -----
        /// (x,q,r)
        /// </summary>
        /// <returns></returns>
        private OwnVectorParametrization CasePNN()
        {
            double x1 = -_rows[0].GetY() / _rows[0].GetX();
            double x2 = -_rows[0].GetZ() / _rows[0].GetX();
            ParametrizationRST x = new ParametrizationRST(0, 0, x1, x2);
            ParametrizationRST y = new ParametrizationRST(0, 0, 1, 0);
            ParametrizationRST z = new ParametrizationRST(0, 0, 0, 1);
            return new OwnVectorParametrization(x,y,z);
        }
        /// <summary>
        /// 1 x 0
        /// 0 0 0
        /// 0 0 1
        /// -----
        /// (x,q,0)
        /// </summary>
        /// <returns></returns>
        private OwnVectorParametrization CasePNP()
        {
            double xx = -_rows[0].GetY() / _rows[0].GetX();
            ParametrizationRST x = new ParametrizationRST(0, 0, xx, 0);
            ParametrizationRST y = new ParametrizationRST(0, 0, 1, 0);
            ParametrizationRST z = new ParametrizationRST(0, 0, 0, 0);
            return new OwnVectorParametrization(x,y,z);
        }
        /// <summary>
        /// 1 0 x
        /// 0 1 x
        /// 0 0 0
        /// -----
        /// (x,x,r)
        /// </summary>
        /// <returns></returns>
        private OwnVectorParametrization CasePPN()
        {
            double x1 = -_rows[1].GetZ() / _rows[1].GetY();
            double x2 = -_rows[0].GetZ() / _rows[0].GetX();
            ParametrizationRST x = new ParametrizationRST(0, 0, 0, x2);
            ParametrizationRST y = new ParametrizationRST(0, 0, 0, x1);
            ParametrizationRST z = new ParametrizationRST(0, 0, 0, 1);
            return new OwnVectorParametrization(x,y,z);
        }
        /// <summary>
        /// 1 0 0
        /// 0 1 0
        /// 0 0 1
        /// -----
        /// (0,0,0)
        /// </summary>
        /// <returns></returns>
        private OwnVectorParametrization CasePPP()
        {
            ParametrizationRST x = new ParametrizationRST(0, 0, 0, 0);
            ParametrizationRST y = new ParametrizationRST(0, 0, 0, 0);
            ParametrizationRST z = new ParametrizationRST(0, 0, 0, 0);
            return new OwnVectorParametrization(x,y,z);
        }

        ///////////////////////////////////////////////////////////////////////////////
        ///////////////////////////////////////////////////////////////////////////////
    }
}
