﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Geometry.GraphicObjects.Equations;
using Geometry;

namespace GMath
{
    /// <summary>
    /// ax3+bx2+cx+d=0;
    /// </summary>
    public class CubicEquation:SuperEquation
    {
        private double _a;
        private double _b;
        private double _c;
        private double _d;

        public CubicEquation(double a, double b, double c, double d) 
        {
            _a = a;
            _b = b;
            _c = c;
            _d = d;
        }

        public double GetA() { return _a; }
        public double GetB() { return _b; }
        public double GetC() { return _c; }
        public double GetD() { return _d; }


        public static CubicEquation operator *(CubicEquation a, double d) 
        {
            return new CubicEquation(
                a._a * d, 
                a._b * d, 
                a._c * d, 
                a._d * d);
        }
        public static CubicEquation operator+(CubicEquation a, CubicEquation b)
        {
            return new CubicEquation(
                a._a + b._a, 
                a._b + b._b, 
                a._c + b._c, 
                a._d + b._d);
        }
        public static CubicEquation operator -(CubicEquation a, CubicEquation b)
        {
            return new CubicEquation(
                a._a - b._a,
                a._b - b._b,
                a._c - b._c,
                a._d - b._d);
        }
        public static QuarticEquation operator *(CubicEquation a, LinearEquation b) 
        {
            QuarticEquation quartic = new QuarticEquation(
                a._a * b.GetA(), 
                a._b * b.GetA(), 
                a._c * b.GetA(), 
                a._d * b.GetA(), 
                0);
            CubicEquation cubic = a * b.GetB();
            return quartic + cubic;
        }

        public override bool ContainsVariable()
        {
            return !Program.IsEqualInTolerance(_a, 0) ||
                !Program.IsEqualInTolerance(_b, 0) ||
                !Program.IsEqualInTolerance(_c, 0);
        }

        public override Complex Substitute(Complex x)
        {
            return _a * (x * x * x) + _b * (x * x) + _c * (x) + _d;
        }
        public override double Substitute(double x)
        {
            return _a * (x * x * x) + _b * (x * x) + _c * (x) + _d;
        }

        public QuadraticEquation Derivate() 
        {
            return new QuadraticEquation(3 * _a, 2 * _b, _c);
        }

        protected override List<double> GetListOfKoeficients()
        {
            return new List<double>() { _a, _b, _c, _d };
        }


        ////////////////////////////////////////////////////////
        ////////////////////////////////////////////////////////
        ////////////////////////////////////////////////////////



        /// <summary>
        /// a = 9.8273570086442223
        /// b = 3599384.4887428568
        /// c = 9698.2062588910376
        /// d = 0
        /// Failuje
        /// 
        /// a = 1
        /// b = -1018609.1479667751
        /// c = -1172470283241.1858
        /// d = -2.4372092228801808E+17
        /// </summary>
        /// <returns></returns>
        public override List<List<Complex>> SolveGetAllSetsOfRoots()
        {
            if (_a == 0)// || Program.IsEqualInToleranceHard(_a / _b, 0))
                return new QuadraticEquation(_b, _c, _d).SolveGetAllSetsOfRoots();
            List<List<Complex>> asQuadraticSolutions = new QuadraticEquation(_b, _c, _d).SolveGetAllSetsOfRoots();
            List<List<Complex>> alternativeSolutions = CheckForSpecialCases();
            List<List<Complex>> roots1 = SolveCardan();
            List<List<Complex>> roots2 = Solve1728();

            List<List<Complex>> retval = Append(asQuadraticSolutions, alternativeSolutions);
            retval = Append(retval, roots1);
            retval = Append(retval, roots2);
            return retval;
        }


        protected List<List<Complex>> CheckForSpecialCases()
        {
            List<List<Complex>> alternativeSolutions = new List<List<Complex>>();
            if (_d == 0)
            {
                if (_c == 0)
                {
                    if (_b == 0)
                    {
                        alternativeSolutions.Add(new List<Complex>() { new Complex(0), new Complex(0), new Complex(0) });
                        return alternativeSolutions;
                    }
                    alternativeSolutions = new LinearEquation(_a, _b).SolveGetAllSetsOfRoots();
                    alternativeSolutions[0].Add(new Complex(0));
                    alternativeSolutions[0].Add(new Complex(0));
                    return alternativeSolutions;
                }
                alternativeSolutions = new QuadraticEquation(_a, _b, _c).SolveGetAllSetsOfRoots();
                alternativeSolutions[0].Add(new Complex(0));
                return alternativeSolutions;
            }
            return alternativeSolutions;
        }

        ////////////////////////////////////////////////////////
        ////////////////////////////////////////////////////////
        ////////////////////////////////////////////////////////


        /// <summary>
        /// First make substitution x = t - a/3;
        /// Then substitution       t = y - p/3y, where p is koef. in linear member after first substitution;
        /// Then solve quadratic equation for y^3.
        /// Resubtitute and find roots.
        /// 
        /// Metoda z wikipedie, neni to decartova metoda
        /// </summary>
        /// <returns></returns>
        protected List<List<Complex>> SolveCardan()
        {
            double b = _b / _a;
            double c = _c / _a;
            double d = _d / _a;

            double p = c - (b * b) / 3.0;
            double q = d + (2 * b * b * b - 9 * b * c) / 27.0;

            double aa = 1;
            double bb = q;
            double cc = -(p * p * p) / 27.0;

            Complex D = new Complex(bb * bb - 4 * aa * cc);
            List<Complex> disks = Complex.SecondSqrt1728(D);
            List<List<Complex>> roots = new List<List<Complex>>();
            for (int i = 0; i < disks.Count; i++)
            {
                Complex y31 = new Complex((-bb + disks.ElementAt(i)) / (2 * aa));
                List<Complex> u3 = Complex.ThirdSqrt(y31);
                Complex t1 = ComputeTi(u3.ElementAt(0), p);
                Complex t2 = ComputeTi(u3.ElementAt(1), p);
                Complex t3 = ComputeTi(u3.ElementAt(2), p);
                Complex x1 = ComputeXi(t1, b);
                Complex x2 = ComputeXi(t2, b);
                Complex x3 = ComputeXi(t3, b);
                roots.Add(new List<Complex>() { x1, x2, x3 });
            }
            return roots;
        }

        protected Complex ComputeTi(Complex Yi, double p)
        {
            if (Program.IsEqualInTolerance(Yi._z, 0))
            {
                return new Complex(0);
            }
            return new Complex(Yi - p / (3 * Yi));
        }
        protected Complex ComputeXi(Complex Ti, double a)
        {
            if (Program.IsEqualInTolerance(a, 0))
            {
                return Ti;
            }
            return new Complex(Ti - (a / 3.0));
        }

        ////////////////////////////////////////////////////////
        ////////////////////////////////////////////////////////
        ////////////////////////////////////////////////////////

         ///// 2x3   - 4x2   - 22x + 24 = 0
            //_a = 2;
            //_b = -4;
            //_c = -22;
            //_d = 24;
            //Solve2();
            ///// 3x3   - 10x2   + 14x + 27 = 0
            //_a = 3;
            //_b = -10;
            //_c = 14;
            //_d = 27;
            //Solve2();
            ///// x3   + 6x2   + 12x + 8 = 0
            //_a = 1;
            //_b = 6;
            //_c = 12;
            //_d = 8;
            //Solve2();

        protected List<List<Complex>> Solve1728() 
        {
            double a = _a;
            double b = _b;
            double c = _c;
            double d = _d;
            /// f = ((3c/a) - (b²/a²))/3
            double f = ((3 * c) / a - (b * b / (a * a))) / 3; 
            /// g = ((2b³/a³) - (9bc/a²) + (27d/a))/ 27
            double g = ((2 * b * b * b / (a * a * a)) - (9 * b * c / (a * a)) + 27 * d / a) / 27;
            /// h = (g²/4) + (f³/27)
            double h = (g * g / 4) + (f * f * f / 27);

            if (f == 0 && g == 0 && h == 0)
                return new List<List<Complex>> { OneTrippleRoot() };
            if (h <= 0)
                return new List<List<Complex>> { AllThreeRootsAreReal(f, g, h) };
            else
                return new List<List<Complex>>() { OnlyOneRootIsReal(f, g, h) };
        }

        private List<Complex> AllThreeRootsAreReal(double f, double g, double h)
        {
            double a = _a;
            double b = _b;
            double c = _c;
            double d = _d;
            /// i = ((g²/4) - h) ½
            double i = Math.Sqrt(-f * f * f / 27); // určitě je to kladné..
            /// j = (i) 1/3 
            double j = Program.Cbroot(i);
            /// k = arc cosine (- (g / 2i)) 
            double K = Math.Acos(Math.Min(Math.Max(-g / (2 * i), -1), 1));
            /// L = j * -1
            double L = -j;
            /// M = cosine (K/3)
            double M = Math.Cos(K / 3.0);
            /// N = (Square Root of 3) * sine (K/3)
            double N = Math.Sqrt(3) * Math.Sin(K / 3.0);
            /// P = (b/3a) * -1
            double P = -(b / (3 * a));
            /// x1 = 2j * cosine(k/3)   -(b/3a)
            Complex x1 = new Complex(2 * j * Math.Cos(K / 3.0) - (b / (3 * a)), 0);
            /// x2 = L * (M + N) + P
            Complex x2 = new Complex(L * (M + N) + P, 0);
            /// x3 = L * (M - N) + P
            Complex x3 = new Complex(L * (M - N) + P, 0);
            return new List<Complex>() { x1, x2, x3 };
        }
        private List<Complex> OnlyOneRootIsReal(double f, double g, double h) 
        {
            double a = _a;
            double b = _b;
            double c = _c;
            double d = _d;
            /// R = -(g/2) + (h)½
            double R = -(g / 2) + Math.Sqrt(h); // H je kladné
            /// S = (R)1/3 
            double S = Program.Cbroot(R);
            /// T = -(g/2) - (h)½ 
            double T = -(g / 2) - Math.Sqrt(h);
            /// U = (T)1/3
            double U = Program.Cbroot(T);
            
            /// X1 = (S + U) - (b/3a)
            Complex x1 = new Complex((S + U) - (b / (3 * a)), 0);
            /// X2 = -(S + U)/2 - (b/3a) + i*(S-U)*(3)½/2
            Complex x2 = new Complex(-(S + U) / 2 - (b / (3 * a)), (S - U) * Math.Sqrt(3) / 2);
            /// X3 = -(S + U)/2 - (b/3a) - i*(S-U)*(3)½/2
            Complex x3 = new Complex(-(S + U) / 2 - (b / (3 * a)), -(S - U) * Math.Sqrt(3) / 2);
            return new List<Complex>() { x1, x2, x3 };
        }
        private List<Complex> OneTrippleRoot() 
        {
            double d = _d;
            double a = _a;
            /// x1 = x2 = x3= (d/a)1/3 * -1
            double x = -Program.Cbroot(d/a);
            return new List<Complex>() { new Complex(x, 0), new Complex(x), new Complex(x) };
        }
    }
}
