﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Geometry.GraphicObjects.Equations;
using Geometry;

namespace GMath
{
    public class QuarticEquation : SuperEquation
    {
        private double _a;
        private double _b;
        private double _c;
        private double _d;
        private double _e;

        public QuarticEquation(double a, double b, double c, double d, double e)
        {
            _a = a;
            _b = b;
            _c = c;
            _d = d;
            _e = e;
        }

        public double GetA() { return _a; }
        public double GetB() { return _b; }
        public double GetC() { return _c; }
        public double GetD() { return _d; }
        public double GetE() { return _e; }


        public static QuarticEquation operator +(QuarticEquation a, CubicEquation b) 
        {
            return new QuarticEquation(
                a._a, 
                a._b + b.GetA(), 
                a._c + b.GetB(), 
                a._d + b.GetC(), 
                a._e + b.GetD());
        }
        public static QuarticEquation operator +(QuarticEquation a, QuarticEquation b)
        {
            return new QuarticEquation(
                a._a + b._a,
                a._b + b._b,
                a._c + b._c,
                a._d + b._d,
                a._e + b._e);
        }
        public static QuarticEquation operator -(QuarticEquation a, QuarticEquation b)
        {
            return new QuarticEquation(
                a._a - b._a,
                a._b - b._b,
                a._c - b._c,
                a._d - b._d,
                a._e - b._e);
        }

        public override bool ContainsVariable()
        {
            return !Program.IsEqualInTolerance(_a, 0) ||
                !Program.IsEqualInTolerance(_b, 0) ||
                !Program.IsEqualInTolerance(_c, 0) ||
                !Program.IsEqualInTolerance(_d, 0);
        }

        public override double Substitute(double x)
        {
            return (_a * x * x * x * x + _b * x * x * x + _c * x * x + _d * x + _e);
        }

        public void DivideByA()
        {
            if (_a == 0)
                return;
            _b /= _a;
            _c /= _a;
            _d /= _a;
            _e /= _a;
            _a = 1;
        }
        public CubicEquation Derivate() 
        {
            return new CubicEquation(4 * _a, 3 * _b, 2 * _c, _d);
        }
        protected override List<double> GetListOfKoeficients()
        {
            return new List<double>() { _a, _b, _c, _d, _e };
        }

        //////////////////////////////////////////////
        //////////////////////////////////////////////
        //////////////////////////////////////////////


        public override List<List<Complex>> SolveGetAllSetsOfRoots()
        {
            if (_a == 0)
                return new CubicEquation(_b, _c, _d, _e).SolveGetAllSetsOfRoots();
            List<List<Complex>> asCubicSolutions = new CubicEquation(_b, _c, _d, _e).SolveGetAllSetsOfRoots();
            List<List<Complex>> alternativeSolutions = CheckForSpecialCases();
            List<List<Complex>> rootsWiki = SolveWikiAllSets(); // wiki (decartes)
            List<Complex> roots3 = Solve1728();
            //List<Complex> roots4 = SolveCardan(); // cardan

            List<List<Complex>> retval = Append(asCubicSolutions, alternativeSolutions);
            retval = Append(retval, rootsWiki);
            retval.Add(roots3);
            //retval.Add(roots4);
            return retval;
        }

        /// <summary>
        /// Definiční obor je dán množinou hodnot parametru U, 
        /// pro které je jistá až kvartická rovnice (obecně je utvářena jako kvartická rovnice) kladná, 
        /// protože se z ní bere odmocnina.
        /// 
        /// Vypočítané aproximace kořenů otestuji, zda je funkční hodnota kladná.
        /// Pokud ne najdu v okolí jinou aproximaci, která kladná bude.
        /// 
        /// Když toto udělám, velmi se zmenší šance, že bude na krajích nebo spojích křivky průniku část chybět.
        /// </summary>
        /// <returns></returns>
        public List<double> GetPositiveRootsForDefinitionInterval() 
        {
            return null;
        }

        private List<List<Complex>> CheckForSpecialCases()
        {
            List<List<Complex>> alternativeSolutions = new List<List<Complex>>();
            if (_e == 0)
            {
                if (_d == 0)
                {
                    if (_c == 0)
                    {
                        if (_b == 0)
                        {
                            alternativeSolutions.Add(
                                new List<Complex>() { new Complex(0), new Complex(0), new Complex(0), new Complex(0) });
                            return alternativeSolutions;                            
                        }
                        alternativeSolutions = new LinearEquation(_a, _b).SolveGetAllSetsOfRoots();
                        alternativeSolutions[0].Add(new Complex(0));
                        alternativeSolutions[0].Add(new Complex(0));
                        alternativeSolutions[0].Add(new Complex(0));
                        return alternativeSolutions;
                    }
                    alternativeSolutions = new QuadraticEquation(_a, _b, _c).SolveGetAllSetsOfRoots();
                    alternativeSolutions[0].Add(new Complex(0));
                    alternativeSolutions[0].Add(new Complex(0));
                    return alternativeSolutions;
                }
                alternativeSolutions = new CubicEquation(_a, _b, _c, _d).SolveGetAllSetsOfRoots();
                alternativeSolutions[0].Add(new Complex(0));
                return alternativeSolutions;
            }
            return alternativeSolutions;
        }

        //////////////////////////////////////////////
        //////////////////////////////////////////////
        //////////////////////////////////////////////


        private List<List<Complex>> SolveWikiAllSets()
        {
            DivideByA();
            double B = _b;
            double C = _c;
            double D = _d;
            double E = _e;
            double P = (-6 * B * B + 16 * C) / 16;
            double Q = (4 * B * B * B - 16 * B * C + 32 * D) / 32;
            double R = (-3 * B * B * B * B + 16 * B * B * C - 64 * B * D + 16 * 16 * E) / (16 * 16);
            List<Complex> S = new CubicEquation(1, 2 * P, P * P - 4 * R, -Q * Q).SolveComplex();
            List<List<Complex>> solutions = new List<List<Complex>>();
            Complex K;
            foreach (Complex s in S)
            {
                K = Complex.SecondSqrt1728(s)[0];// muze byt vice reseni?
                List<Complex> sols;
                if (Q == 0)
                    sols = SolveAsBiQuadratic(B, P, Q, R, K);
                else sols = SolveAsTwoQuadratics(B, P, Q, R, K);
                solutions.Add(sols);
            }
            return solutions;
        }

        /// <summary>
        /// y4 + Py2 + R = 0;
        /// </summary>
        /// <returns></returns>
        private List<Complex> SolveAsBiQuadratic(double B, double P, double Q, double R, Complex K)
        {
            List<Complex> z1 = new QuadraticEquation(1, P, R).SolveComplex();
            List<Complex> y1 = new List<Complex>();
            foreach (Complex z in z1)
            {
                y1.Add(Complex.SecondSqrt1728(z)[0]);
                y1.Add(Complex.SecondSqrt1728(z)[1]);
            }
            List<Complex> retval = new List<Complex>();
            foreach (Complex y in y1)
                retval.Add(y - B / 4);
            return retval;
        }

        /// <summary>
        /// (y2 + Ky + L)(y2 -Ky + N) = 0;
        /// </summary>
        /// <returns></returns>
        private List<Complex> SolveAsTwoQuadratics(double B, double P, double Q, double R, Complex K)
        {
            Complex N = (Q / K + K * K + P) / 2;
            Complex L;
            if (N._z == 0)
            {
                if (Program.IsEqualInTolerance(R, 0))
                    L = new Complex(0);
                else return new List<Complex>(); // nema reseni.
            }
            else { L = R / N; }
            // k = 3.1600257999303509
            List<Complex> y1 = new ComplexQuadraticEquation(new Complex(1), K, L).SolveComplex();
            List<Complex> y2 = new ComplexQuadraticEquation(new Complex(1), -K, N).SolveComplex();
            List<Complex> retval = new List<Complex>();
            foreach (Complex y in y1)
                retval.Add(y - B / 4);
            foreach (Complex y in y2)
                retval.Add(y - B / 4);
            return retval;
        }

        //////////////////////////////////////////////
        //////////////////////////////////////////////
        //////////////////////////////////////////////

        private List<Complex> Solve1728() 
        {
            double b = _b / _a;
            double c = _c / _a;
            double d = _d / _a;
            double e = _e / _a;
            /// f = c - (3b2/8)
            double f = c - 3 * (b * b) / 8;
            /// g = d + (b3 / 8) - (b*c/2)            
            double g = d + (b * b * b) / 8 - (b * c) / 2;
            /// h = e - (3*b4/256) + (b 2 * c/16) - ( b*d/4)
            double h = e - 3 * (b * b * b * b) / 256 + (b * b * c) / 16 - (b * d) / 4;
            /// Y3 + (f/2)*Y2 + ((f2 -4*h)/16)*Y -g2/64 = 0
            CubicEquation cubic = new CubicEquation(1, f / 2, (f * f - 4 * h) / 16, -g * g / 64);
            List<Complex> complexRoots = cubic.SolveComplex();
            for (int i = 0; i < complexRoots.Count; i++)
                if (Math.Abs(complexRoots[i]._real) < 1E-13)
                    complexRoots[i]._real = 0;
            int realRootsCount = 0;
            foreach (Complex co in complexRoots)
                if (Program.IsEqualInTolerance(co._i, 0))
                    realRootsCount++;
            List<Complex> ps = new List<Complex>();
            if (realRootsCount > 1) // vezmeme dva reálné kořeny.
            {
                foreach (Complex co in complexRoots)
                    if (Program.IsEqualInTolerance(co._i, 0) && co._real != 0)
                        ps.Add(Complex.SecondSqrt1728(co)[0]);// je jedno jakou ze dvou odmocnin vezmu
            }
            else //vezmeme dva komplexní kořeny
            {
                foreach (Complex co in complexRoots)
                    if (!(Program.IsEqualInTolerance(co._i, 0)) && co._real != 0) 
                    {
                        ps.Add(Complex.SecondSqrt1728(co)[0]); // je jedno jakou ze dvou odmocnin vezmu
                    }
            }
            if (ps.Count == 0)
                return new List<Complex>();
            Complex p;
            Complex q;
            Complex r;
            if (ps.Count == 1)
            {
                p = ps[0];
                q = new Complex(0);
                r = new Complex(0);
            }
            else
            {
                p = ps[0];
                q = ps[1];
                r = -g / (8 * p * q);
            }
            double s = b/(4);
            Complex x1 = p + q + r - s;
            Complex x2 = p - q - r - s;
            Complex x3 = -p + q - r - s;
            Complex x4 = -p - q + r - s;
            List<Complex> retval = new List<Complex>() { x1, x2, x3, x4 };
            return retval;
        }

        //////////////////////////////////////////////
        //////////////////////////////////////////////
        //////////////////////////////////////////////

        private List<Complex> SolveCardan()
        {
            double b3 = _b / _a;
            double b2 = _c / _a;
            double b1 = _d / _a;
            double b0 = _e / _a;

            double y3 = GetCubicRezoluentRoot(b3, b2, b1, b0);
            double A = b3 / 2;
            double B = y3 / 2;
            double D = Math.Sqrt(B * B - b0);
            double C;
            if (D == 0)
                C = Math.Sqrt(A * A - b2 + y3);
            else C = (A * B - b1 / 2) / D;
            return GetCardanRoots(A, B, C, D);
        }
        private double GetCubicRezoluentRoot(double b3, double b2, double b1, double b0)
        {
            double c2 = -b2;
            double c1 = b3 * b1 - 4 * b0;
            double c0 = b0 * (4 * b2 - b3 * b3) - b1 * b1;

            double Q = (3 * c1 - c2 * c2) / 9;
            double R = (9 * c2 * c1 - 27 * c0 - 2 * c2 * c2 * c2) / 54;

            double Qr = Q * Q * Q + R * R;

            double S = Program.Cbroot(R + Math.Sqrt(Qr));
            double T = Program.Cbroot(R - Math.Sqrt(Qr));
            if (Qr >= 0)
            {
                double y3 = S + T - c2 / 3;
                return y3;
            }
            else
            {
                double y3 = 2 * Math.Sqrt(-Q) * Math.Cos(1.0 / 3.0 * Math.Acos(R / Math.Pow(-Q, 1.5))) - c2 / 3;
                return y3;
            }
        }
        private List<Complex> GetCardanRoots(double A, double B, double C, double D)
        {
            QuadraticEquation e1 = new QuadraticEquation(1, A + C, B + D);
            QuadraticEquation e2 = new QuadraticEquation(1, A - C, B - D);
            List<Complex> retval = new List<Complex>();
            List<Complex> e1roots = e1.SolveComplex();
            List<Complex> e2roots = e2.SolveComplex();
            foreach (Complex d in e1roots)
                retval.Add(d);
            foreach (Complex d in e2roots)
                retval.Add(d);
            return retval;
        }
    }
}
