﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Geometry;

namespace GMath
{
    public class ParametrizationRST
    {
        private double _constant;
        private double _r;
        private double _s;
        private double _t;

        public ParametrizationRST(ParametrizationRST src) 
        {
            _constant = src._constant;
            _r = src._r;
            _s = src._s;
            _t = src._t;
        }
        public ParametrizationRST(double constant, double r, double s, double t) 
        {
            _constant = constant;
            _r = r;
            _s = s;
            _t = t;
        }

        public double GetConstant() {return _constant;}
        public double GetR() {return _r;}
        public double GetS() { return _s; }
        public double GetT() { return _t; }

        public void SetConstant(double constant) { _constant = constant; }
        public void SetR(double r) { _r = r; }
        public void SetS(double s) { _s = s; }
        public void SetT(double t) { _t = t; }

        public static ParametrizationRST operator*(ParametrizationRST parametrization, double a)
        {
            return new ParametrizationRST
                (
                    parametrization._constant * a,
                    parametrization._r * a,
                    parametrization._s * a,
                    parametrization._t * a
                );
        }
        public static ParametrizationRST operator *(double a, ParametrizationRST parametrization) 
        {
            return parametrization * a;
        }
        public static ParametrizationRST operator +(ParametrizationRST a, ParametrizationRST b) 
        {
            return new ParametrizationRST(
                a._constant + b._constant, 
                a._r + b._r, 
                a._s + b._s, 
                a._t + b._t);
        } 
        public static bool operator==(ParametrizationRST a, ParametrizationRST b)
        {
            return Program.IsEqualInTolerance(a._constant, b._constant) &&
                Program.IsEqualInTolerance(a._r, b._r) &&
                Program.IsEqualInTolerance(a._s, b._s) &&
                Program.IsEqualInTolerance(a._t, b._t);
        }
        public static bool operator !=(ParametrizationRST a, ParametrizationRST b)
        {
            return !(Program.IsEqualInTolerance(a._constant, b._constant) &&
                Program.IsEqualInTolerance(a._r, b._r) &&
                Program.IsEqualInTolerance(a._s, b._s) &&
                Program.IsEqualInTolerance(a._t, b._t));
        }

        public void SubstituteForR(ParametrizationRST r) 
        {
            _s += _r * r._s;
            _t += _r * r._t;
            _constant += _r * r._constant;
            _r = 0;
        }
        public void SubstituteForS(ParametrizationRST s)
        {
            _r += _s * s._r;
            _t += _s * s._t;
            _constant += _s * s._constant;
            _s = 0;
        }
        public void SubstituteForT(ParametrizationRST t)
        {
            _r += _t * t._r;
            _s += _t * t._s;
            _constant += _t * t._constant;
            _t = 0;
        }

        public void ChooseForT(double t) 
        {
            this._constant = t * this._t + this._constant;
            this._t = 0;
        }
        public void ChooseForR(double r)
        {
            this._constant = r * this._r + this._constant;
            this._r = 0;
        }
        public void ChooseForS(double s)
        {
            this._constant = s * this._s + this._constant;
            this._s = 0;
        }

        public bool ContainsConstant()
        {
            return !Program.IsEqualInTolerance(_constant, 0);
        }
        public bool ContainsRParameter() 
        {
            return !Program.IsEqualInTolerance(_r, 0);
        }
        public bool ContainsSParameter()
        {
            return !Program.IsEqualInTolerance(_s, 0);
        }
        public bool ContainsTParameter()
        {
            return !Program.IsEqualInTolerance(_t, 0);
        }
        public bool ContainsParameter() 
        {
            return ContainsRParameter() || ContainsSParameter() || ContainsTParameter();
        }
        public int GetParametersCount() 
        {
            int parametersCount = 0;
            if (ContainsRParameter())
                parametersCount++;
            if (ContainsSParameter())
                parametersCount++;
            if (ContainsTParameter())
                parametersCount++;
            return parametersCount;
        }

        public ParametrizationRST ExpressR() 
        {
            if (!ContainsRParameter())
                throw new Exception("Unable to express R parameter.");
            ParametrizationRST expression = new ParametrizationRST(this);
            expression = expression * (-1/GetR());
            expression.SetR(0);
            return expression;
        }
        public ParametrizationRST ExpressS()
        {
            if (!ContainsSParameter())
                throw new Exception("Unable to express S parameter.");
            ParametrizationRST expression = new ParametrizationRST(this);
            expression = expression * (-1 / GetS());
            expression.SetS(0);
            return expression;
        }
        public ParametrizationRST ExpressT()
        {
            if (!ContainsTParameter())
                throw new Exception("Unable to express T parameter.");
            ParametrizationRST expression = new ParametrizationRST(this);
            expression = expression * (-1 / GetT());
            expression.SetT(0);
            return expression;
        }

        public bool IsZero()
        {
            return Program.IsEqualInTolerance(this._constant, 0) &&
                Program.IsEqualInTolerance(this._r, 0) &&
                Program.IsEqualInTolerance(this._s, 0) &&
                Program.IsEqualInTolerance(this._t, 0);
        }
        public override bool Equals(object obj)
        {
            return base.Equals(obj);
        }
        public override int GetHashCode()
        {
            return base.GetHashCode();
        }
        public override string ToString()
        {
            string rText = Program.GetOutputFormatedForTextBox(this._r);
            string sText = Program.GetOutputFormatedForTextBox(this._s);
            string tText = Program.GetOutputFormatedForTextBox(this._t);

            string retval = Program.GetOutputFormatedForTextBox(this._constant);
            if (!Program.IsEqualInTolerance(this._r, 0))
                retval += "+" + rText + "r";
            if (!Program.IsEqualInTolerance(this._s, 0))
                retval += "+" + sText + "s";
            if (!Program.IsEqualInTolerance(this._t, 0))
                retval += "+" + tText + "t";
            return retval;
        }
    }
}
