﻿using System;
using System.Collections.Generic;
using GMath;
using Geometry.GraphicObjects.Equations;
using Geometry.GraphicObjects.Matrixes;

namespace Geometry.GraphicObjects.ConicSections
{
    /// <summary>
    /// Matice kuzelosecky je symetricka matice 3x3.
    /// Pro prevedeni na kanonicky tvar pouzivam stejny postup jako u kvadrik.
    /// Proto pracuji s transformacnimi maticemi typu 4x4.
    /// Pri aplikaci transformacni matice na kuzelosecku nejdrive orezu matici na 3x3.
    /// Konstruktory kuzelosecek jsou volany s transformacnimi maticemi typu 4x4.
    /// </summary>
    public class ConicRecognizer
    {
        private ConicEquation _originalEquation;
        private string _name;
        private GColor _color;
        private int _ID;

        private struct OwnNumberStruct
        {
            public double ownNumber; // samotne vlastni cislo
            public int multiplicity; // nasobnost vlastniho cisla
            public bool error;      // indikuje selhani pri hledani vlastnich cisel
            public OwnNumberStruct(double ownNumber, int multiplicity)
            {
                this.ownNumber = ownNumber;
                this.multiplicity = multiplicity;
                error = false;
            }
            public OwnNumberStruct(bool error)
            {
                ownNumber = 0;
                multiplicity = 3;
                this.error = error;
            }
        }
        private struct OwnVectorStruct
        {
            public OwnVectorParametrization parametrization; // parametrizace vlastniho vektoru
            public bool extended; // udává, zda byla parametrizace upravena
        }
        private struct OwnPairStruct
        {
            public OwnNumberStruct ownNumber; // vlastni cislo
            public List<OwnVectorStruct> parametrizations; // parametrizace příslušné danému číslu
        }
        private struct ParametersValuesStruct
        {
            public double rValue;
            public double sValue;
            public double tValue;

            public ParametersValuesStruct(double r, double s, double t)
            {
                rValue = r;
                sValue = s;
                tValue = t;
            }
        }

        public ConicRecognizer(ConicEquation equation, GColor color, string name, int ID)
        {
            _originalEquation = equation;
            _name = name;
            _color = color;
            _ID = ID;
        }

        private ConicCharacteristicEquation Det(ConicMatrix conicMatrix)
        {
            double a = conicMatrix.GetA();
            double b = conicMatrix.GetB();
            double d = conicMatrix.GetD();

            double u = -(a + b);
            double v = (a * b) - (d * d);

            return new ConicCharacteristicEquation(1, u, v);
        }
        /// <summary>
        /// Resi kubickou rovnici urcenou equationem ax3 + bx2 +cx +d = 0; (x3 povazuji za 1).
        /// </summary>
        /// <param name="equation"></param>
        /// <returns></returns>
        private List<OwnNumberStruct> GetOwnNumbers(ConicMatrix conicMatrix)
        {
            ConicCharacteristicEquation conicCharacteristicEquation = Det(conicMatrix);
            List<double> simpleOwnNumbers = conicCharacteristicEquation.Solve();
            if (simpleOwnNumbers.Count != 2)
            {
                return new List<OwnNumberStruct>() { new OwnNumberStruct(true) }; //indikuje chybovy stav
            }
            //for (int i = 0; i < simpleOwnNumbers.Count; i++)
            //{
            //    if (Program.IsEqualInToleranceLow(simpleOwnNumbers[i], 0)) 
            //    {
            //        simpleOwnNumbers[i] = 0;                    
            //    }   
            //}
            List<OwnNumberStruct> ownNumbers = new List<OwnNumberStruct>();
            while (simpleOwnNumbers.Count != 0)
            {
                OwnNumberStruct o = new OwnNumberStruct(simpleOwnNumbers[0], 0);
                for (int i = 0; i < simpleOwnNumbers.Count; i++)
                {
                    if (Program.IsEqualInTolerance(o.ownNumber, simpleOwnNumbers[i]))
                    {
                        simpleOwnNumbers.RemoveAt(i--);
                        o.multiplicity++;
                    }
                }
                ownNumbers.Add(o);
            }
            return ownNumbers;
        }
        private List<OwnNumberStruct> SortOwnNumbers(List<OwnNumberStruct> ownNumbers)
        {
            // netreba tridit, zavoven je osetren chybovy stav. 
            // OwnNumbers sice nemuze byt null, ale muze obsahovat jedno neplatne vlastni cislo.
            if (ownNumbers.Count == 1)
                return ownNumbers;
            List<OwnNumberStruct> retval = new List<OwnNumberStruct>();
            foreach (OwnNumberStruct ownNumber in ownNumbers)
                if (ownNumber.ownNumber > 0 && !Program.IsEqualInTolerance(ownNumber.ownNumber, 0))
                    retval.Add(ownNumber); // nejprve vlozim kladna vlastni cisla
            foreach (OwnNumberStruct ownNumber in ownNumbers)
                if (ownNumber.ownNumber < 0 && !Program.IsEqualInTolerance(ownNumber.ownNumber, 0))
                    retval.Add(ownNumber); // potom vlozim zaporna vlastni cisla
            foreach (OwnNumberStruct ownNumber in ownNumbers)
                if (Program.IsEqualInTolerance(ownNumber.ownNumber, 0))
                    retval.Add(ownNumber); // nakonec vlozim nulova vlastni cisla
            return retval;
        }

        /// <summary>
        /// Z transformacni matice vyctu transformaci roviny kuzelosecky, stejne jako v ActualizeEquations.
        /// Pak tuto transformaci slozim s rozpoznanou transformací. Je-li nejak ulozena transformace modifikovana,
        /// projevi se to jen na rovnici roviny. Ale posunuti ve smerech x nebo y jsou ignorovany.
        /// </summary>
        public ConicSection Recognize(TransformationMatrix transformationMatrix)
        {
            ConicSection cs = Recognize();
            ConicSection.Matrixes matrixes = ConicSection.DecomposeTransformationMatrix(transformationMatrix);
            TransformationMatrix planTransformation = matrixes.planTransformation;
            TransformationMatrix newTransformation = planTransformation * cs.GetTransformationMatrixAlgebraic();

            cs.SetTransformationMatrixAlgebraic(newTransformation);
            return cs;
        }        

        public ConicSection Recognize()
        {
            List<double> numbers = new List<double>() { 1, 1000, 0.001, 1000000, 0.000001 };
            List<ConicSection> recognizedConics = new List<ConicSection>();            
            ConicSection temp;
            foreach (double d in numbers)
            {
                try
                {
                    temp = MakeCanonicConicSection(new ConicEquation(d * _originalEquation));
                    if (temp == null)
                        temp = MakeCanonicConicSection(new ConicEquation(-d * _originalEquation));
                }
                catch (Exception)
                {
                    continue;
                }
                if (temp != null)
                {
                    return temp;
                    //recognizedConics.Add(temp);
                }
            }
            return GetBestOutputConic(recognizedConics);
        }

        private ConicSection GetBestOutputConic(List<ConicSection> recognizedQuadrics)
        {
            if (recognizedQuadrics.Count == 0)
                return new EmptySet(_color, _name, _ID);
            try
            {
                Type quadricType = GetTypeOfRecognizedConic(recognizedQuadrics);
                int index = 0;
                double minWorseDifference = double.MaxValue;
                double temp, temp2;
                for (int i = 0; i < recognizedQuadrics.Count; i++)
                {
                    if (recognizedQuadrics[i].GetType() != quadricType)
                        continue;
                    temp = GetWorseDifference(recognizedQuadrics[i].GetCommonEquation(), _originalEquation);
                    temp2 = GetSubstitutionsDifference(recognizedQuadrics[i].GetCommonEquation(), _originalEquation);
                    if (temp2 < minWorseDifference)
                    {
                        index = i;
                        minWorseDifference = temp2;
                    }
                }
                return recognizedQuadrics[index];
            }
            catch (Exception) 
            {
                return recognizedQuadrics[0];
            }
        }
        private Type GetTypeOfRecognizedConic(List<ConicSection> recognizedConics)
        {
            List<KeyValuePair<Type, int>> list = new List<KeyValuePair<Type, int>>();
            for (int i = 0; i < recognizedConics.Count; i++)
            {
                Type type = recognizedConics[i].GetType();
                int j = 0;
                for (j = 0; j < list.Count; j++)
                    if (list[j].Key == type)
                    {
                        list[j] = new KeyValuePair<Type, int>(type, list[j].Value + 1);
                        break;
                    }
                if (j == list.Count)
                    list.Add(new KeyValuePair<Type, int>(type, 1));
            }
            Type temp = null;
            int maxCount = 0;
            for (int j = 0; j < list.Count; j++)
                if (list[j].Value > maxCount)
                {
                    maxCount = list[j].Value;
                    temp = list[j].Key;
                }
            return temp;
        }
        private double GetWorseDifference(ConicEquation a, ConicEquation inputConic)
        {
            double d = 0;
            int index = 0;
            List<double> koefs = inputConic.GetValues();
            for (int i = 0; i < koefs.Count; i++)
            {
                index = i;
                d = Math.Max(Math.Abs(koefs[i]), d);
            }
            ConicEquation aCopy = a.DivideBy(a.Get(index) / d);
            ConicEquation bCopy = inputConic.DivideBy(Math.Sign(inputConic.Get(index)));//inputQuadric.Get(index));
            double worseDifference = 0;
            for (int i = 0; i < aCopy.GetValues().Count; i++)
            {
                if (Math.Abs(aCopy.Get(i) - bCopy.Get(i)) > worseDifference)
                    worseDifference = Math.Abs(aCopy.Get(i) - bCopy.Get(i));
            }
            return worseDifference;
        }
        private double GetSubstitutionsDifference(ConicEquation a, ConicEquation inputConic)
        {
            double d = 0;
            List<double> koefs = inputConic.GetValues();
            int index = 0;
            for (int i = 0; i < koefs.Count; i++)
            {
                d = Math.Max(Math.Abs(koefs[i]), d);
                index = i;
            }
            ConicEquation aCopy = a.DivideBy(a.Get(index) / d);
            ConicEquation bCopy = inputConic.DivideBy(Math.Sign(inputConic.Get(index)));//inputQuadric.Get(index));
            double substitutionToA = 0;
            double substitutionToInput = 0;
            Vector v = new Vector(0, 0, 0);
            for (int x = -10; x <= 10; x += 10)
                for (int y = -10; y <= 10; y += 10)
                    for (int z = -10; z <= 10; z += 10)
                    {
                        v.SetValues(x, y, z);
                        substitutionToA += aCopy.Substitute(v);
                        substitutionToInput += bCopy.Substitute(v);
                    }
            return Math.Abs(substitutionToInput - substitutionToA);
        }

        private ConicSection MakeCanonicConicSection(ConicEquation originalEquation)
        {
            ConicMatrix conicMatrix = new ConicMatrix(originalEquation);
            List<OwnNumberStruct> ownNumbers = SortOwnNumbers(GetOwnNumbers(conicMatrix));
            if (ownNumbers.Count == 1 && (ownNumbers[0].error || Program.IsEqualInTolerance(ownNumbers[0].ownNumber, 0)))
                return null;
            List<Vector> ownVectorsPermutation = GetOwnVectors(ownNumbers, conicMatrix);
            TransformationMatrix rotationMatrix = GetRotationMatrix(ownVectorsPermutation, conicMatrix);
            TransformationMatrix toCanonicMatrix;
            ConicEquation canonizedConicEquation;
            if (rotationMatrix != null)
            {
                CenterParametrization center = GetCenterParametrization(conicMatrix);
                if (center == null)
                {
                    center = GetPeakParametrization(conicMatrix, rotationMatrix);
                    toCanonicMatrix = GetTransformationMatrix(rotationMatrix, center);
                    canonizedConicEquation = new ConicEquation(GetConicInBasis(conicMatrix, toCanonicMatrix));
                }
                else
                {
                    TransformationMatrix toCanonicComputed = GetTransformationMatrix(new TransformationMatrix(rotationMatrix), center);
                    CenterParametrization centerSquared = GetCenterParametrization(conicMatrix, rotationMatrix);
                    TransformationMatrix toCanonicSquared = GetTransformationMatrix(new TransformationMatrix(rotationMatrix), centerSquared);
                    ConicEquation computed = new ConicEquation(GetConicInBasis(conicMatrix, toCanonicComputed));
                    ConicEquation squared = new ConicEquation(GetConicInBasis(conicMatrix, toCanonicSquared));
                    computed = (ConicEquation)computed.Normalize();
                    squared = (ConicEquation)squared.Normalize();
                    double d1 = Math.Max(Math.Abs(computed.GetG()), Math.Abs(computed.GetH()));
                    double d2 = Math.Max(Math.Abs(squared.GetG()), Math.Abs(squared.GetH()));
                    // pokud jsou oba koeficienty L desetinna cisla, ale computed je dokonce i v toleranci nulovosti,
                    // a je vice nulovy jako squared nebo je nejmensi koeficient squared moc velky a zaroven je 
                    // nejmensi koeficient computed mensi.
                    if ((Math.Abs(squared.GetL()) < 0 && Program.IsEqualInTolerance(computed.GetL(), 0) &&
                        Math.Abs(computed.GetL()) < Math.Abs(squared.GetL()) || (!Program.IsEqualInToleranceLow(d2, 0) && d1 < d2)))
                    {
                        toCanonicMatrix = toCanonicComputed;
                        canonizedConicEquation = computed;
                    }
                    else
                    {
                        toCanonicMatrix = toCanonicSquared;
                        canonizedConicEquation = squared;
                    }
                }
                return ClasifyConicEquation(canonizedConicEquation, toCanonicMatrix);
            }
            else
            {
                return null;
            }
        }

        private ConicMatrix GetConicInBasis(ConicMatrix conicMatrix, TransformationMatrix transformationMatrix)
        {
            return conicMatrix.ApplyTransformation(transformationMatrix);
        }




        /// <summary>
        /// 
        /// </summary>        
        private List<Vector> GetOwnVectors(List<OwnNumberStruct> ownNumbers, ConicMatrix conicMatrix)
        {
            List<OwnPairStruct> ownPairs = new List<OwnPairStruct>();
            // pro kazde vlastni cislo vypocitam prislusnou parametrizaci vlastniho vektoru.
            foreach (OwnNumberStruct ownNumber in ownNumbers)
            {
                OwnVectorParametrization owp = GetOwnVectorParametrization(conicMatrix, ownNumber);
                OwnPairStruct ownPairStruct = new OwnPairStruct();
                ownPairStruct.ownNumber = ownNumber;
                ownPairStruct.parametrizations = new List<OwnVectorStruct>() { };
                // podle nasobnosti vlastniho cisla, pridam prislusny pocet kopii vlastniho vektoru
                for (int i = 0; i < ownNumber.multiplicity; i++)
                {
                    OwnVectorStruct ownVectorStruct = new OwnVectorStruct();
                    ownVectorStruct.extended = false;
                    ownVectorStruct.parametrization = new OwnVectorParametrization(owp);
                    ownPairStruct.parametrizations.Add(ownVectorStruct);
                }
                ownPairs.Add(ownPairStruct);
            }
            bool existsCenter = GetCenterParametrization(conicMatrix) != null;
            ExtendOwnPairs(ownPairs, conicMatrix, existsCenter);
            List<Vector> ownVectors = GetOwnVectorsFromParametrizations(conicMatrix, ownPairs, existsCenter);
            return ownVectors;
        }
        /// <summary>
        /// 
        /// </summary>
        private void ExtendOwnPairs(List<OwnPairStruct> pairs, ConicMatrix conicMatrix, bool existsCenter)
        {
            int count = 0;
            for (int i = 0; i < pairs.Count; i++)
            {
                if (Program.IsEqualInTolerance(pairs[i].ownNumber.ownNumber, 0))
                    for (int j = 0; j < pairs[i].ownNumber.multiplicity; j++)
                    {
                        // pokud se nejedna o PVP nebo jedna ale nejde o posledni vektor.
                        if (false || (!existsCenter && count < 1))
                        {
                            pairs[i].parametrizations[j] = ExtendOwnVector(conicMatrix, pairs[i].parametrizations[j]);
                        }
                        count++;
                    }
                else count += pairs[i].ownNumber.multiplicity;
            }
        }
        /// <summary>
        /// 
        /// </summary>
        private OwnVectorParametrization GetOwnVectorParametrization(ConicMatrix conicMatrix, OwnNumberStruct ownNumber)
        {
            double on = ownNumber.ownNumber;
            Equation row1 = new Equation(conicMatrix.GetA() - on, conicMatrix.GetD(), 0, 0);
            Equation row2 = new Equation(conicMatrix.GetD(), conicMatrix.GetB() - on, 0, 0);
            Equation row3 = new Equation(0, 0, 0, 0);

            return new OwnVectorEquationSystem(row1, row2, row3, 0).Solve();
        }
        /// <summary>
        /// 
        /// </summary>        
        private OwnVectorStruct ExtendOwnVector(ConicMatrix conicMatrix, OwnVectorStruct ownVectorStruct)
        {
            //return own;
            if (ownVectorStruct.parametrization == null)
                return ownVectorStruct;
            OwnVectorParametrization own = ownVectorStruct.parametrization;
            double a = conicMatrix.GetA();
            double b = conicMatrix.GetB();
            double d = conicMatrix.GetD();
            double g = conicMatrix.GetG();
            double h = conicMatrix.GetH();
            ParametrizationRST M = a * own.GetX() + d * own.GetY();
            ParametrizationRST N = d * own.GetX() + b * own.GetY();
            if (!M.IsZero() || !N.IsZero())
                return ownVectorStruct;
            ParametrizationRST constant = g * own.GetX() + h * own.GetY();
            if (constant.ContainsRParameter())
            {
                constant = constant * (-1 / constant.GetR());
                constant.SetR(0);
                own.SubstituteForR(constant);
                ownVectorStruct.extended = true;
            }
            else
                if (constant.ContainsSParameter())
                {
                    constant = constant * (-1 / constant.GetS());
                    constant.SetS(0);
                    own.SubstituteForS(constant);
                    ownVectorStruct.extended = true;
                }
                else
                    if (constant.ContainsTParameter())
                    {
                        constant = constant * (-1 / constant.GetT());
                        constant.SetT(0);
                        own.SubstituteForT(constant);
                        ownVectorStruct.extended = true;
                    }
            return ownVectorStruct;
        }
        /// <summary>
        /// Zarucim aby vektory nebyly linearne zavisle, pak je orthonormalizuji.
        /// </summary>
        /// <param name="parametrization"></param>
        /// <returns></returns>
        private List<Vector> GetOwnVectorsFromParametrizations(ConicMatrix conicMatrix, List<OwnPairStruct> parametrizations, bool existsCenter)
        {
            List<Vector> ownVectors = new List<Vector>();

            for (int i = 0; i < parametrizations.Count; i++)
            {
                if (parametrizations[i].ownNumber.multiplicity == 1)
                    ownVectors.Add(SubstituteSoThatVectorIsNotZero(parametrizations[i].parametrizations[0].parametrization).NornVector());
                if (parametrizations[i].ownNumber.multiplicity == 2)
                {
                    OwnVectorParametrization theFirst;
                    OwnVectorParametrization theSecond;
                    if (parametrizations[i].parametrizations[0].extended) // prednost dostane ten upraveny vektor
                    {
                        theFirst = parametrizations[i].parametrizations[0].parametrization;
                        theSecond = parametrizations[i].parametrizations[1].parametrization;
                    }
                    else
                    {
                        theFirst = parametrizations[i].parametrizations[1].parametrization;
                        theSecond = parametrizations[i].parametrizations[0].parametrization;
                    }
                    Vector v1 = SubstituteSoThatVectorIsNotZero(theFirst).NornVector();
                    Vector v2 = SubstituteSoThatVectorIsNotZeroAndOrtogonal(v1, theSecond).NornVector();
                    if (parametrizations[i].parametrizations[0].extended)
                    {
                        ownVectors.Add(v1);
                        ownVectors.Add(v2);
                    }
                    else
                    {
                        ownVectors.Add(v2);
                        ownVectors.Add(v1);
                    }
                }
            }
            return Vector.OrthoNormalize(ownVectors);
        }
        /// <summary>
        /// Předpokládá 1 nebo 2 parametry.
        /// </summary>
        private Vector SubstituteSoThatVectorIsNotZero(OwnVectorParametrization owp)
        {
            ParametrizationRST x = owp.GetX();
            ParametrizationRST y = owp.GetY();
            ParametrizationRST z = owp.GetZ();

            ParametrizationRST parametrizationWithParameter = null;

            if (x.ContainsParameter())
                parametrizationWithParameter = new ParametrizationRST(x);
            if (y.ContainsParameter())
                parametrizationWithParameter = new ParametrizationRST(y);
            // Z ignoruji, domnivam se ze je vzdy parametr.

            ParametersValuesStruct values = GetValuesOfParametersSoThatBiggestOneRests(parametrizationWithParameter);
            owp.SubstituteForR(values.rValue);
            owp.SubstituteForS(values.sValue);
            owp.SubstituteForT(values.tValue);
            return new Vector(owp.GetX().GetConstant(), owp.GetY().GetConstant(), owp.GetZ().GetConstant());
        }
        /// <summary>
        /// 
        /// </summary>
        private Vector SubstituteSoThatVectorIsNotZeroAndOrtogonal(Vector vector, OwnVectorParametrization owp)
        {
            ParametrizationRST x = owp.GetX() * vector.GetX();
            ParametrizationRST y = owp.GetY() * vector.GetY();
            ParametrizationRST z = owp.GetZ() * vector.GetZ();

            ParametrizationRST equation = x + y + z;
            OwnVectorParametrization copy = new OwnVectorParametrization(owp);

            if (equation.ContainsRParameter())
                copy.SubstituteForR(equation.ExpressR());
            else if (equation.ContainsSParameter())
                copy.SubstituteForS(equation.ExpressS());
            else if (equation.ContainsTParameter())
                copy.SubstituteForT(equation.ExpressT());
            return SubstituteSoThatVectorIsNotZero(copy);
        }
        /// <summary>
        /// 
        /// </summary>        
        private ParametersValuesStruct GetValuesOfParametersSoThatBiggestOneRests(ParametrizationRST parametrization)
        {
            double rValue = Math.Abs(parametrization.GetR());
            double sValue = Math.Abs(parametrization.GetS());
            double tValue = Math.Abs(parametrization.GetT());
            if (rValue > sValue)
                if (rValue > tValue) // r je nejvetsi
                    return new ParametersValuesStruct(
                        GetOneOrMinusOne(-parametrization.GetConstant() / parametrization.GetR()), 0, 0);
                else // t je nejvetsi
                    return new ParametersValuesStruct(
                        0, 0, GetOneOrMinusOne(-parametrization.GetConstant() / parametrization.GetT()));
            else
                if (sValue > tValue) // s je nejvetsi
                    return new ParametersValuesStruct(
                        0, GetOneOrMinusOne(-parametrization.GetConstant() / parametrization.GetS()), 0);
                else // t je nejvetsi
                    return new ParametersValuesStruct(
                        0, 0, GetOneOrMinusOne(-parametrization.GetConstant() / parametrization.GetT()));
        }
        /// <summary>
        /// Z cisel 1 a -1 vraci to, ktere je vzdalenejsi od cisla number;
        /// </summary>
        private double GetOneOrMinusOne(double number)
        {
            // z cisel 1 a -1 vraci to vzdalenejsi od number
            if (Math.Abs(number - 1) > Math.Abs(number + 1))
                return 1;
            else return -1;
        }



        private TransformationMatrix GetRotationMatrix(List<Vector> ownVectors, ConicMatrix conicMatrix)
        {
            foreach (Vector v in ownVectors)
                if (v == null)
                    return null;
            return new TransformationMatrix(ownVectors[0], ownVectors[1], new Vector(0, 0, 1));
        }
        private TransformationMatrix GetTransformationMatrix(TransformationMatrix rotationMatrix, CenterParametrization centerParametrization)
        {
            Vector center = centerParametrization.GetReprezentant();
            rotationMatrix.SetTranslation(center);
            return rotationMatrix;
        }




        private CenterParametrization GetCenterParametrization(ConicMatrix conicMatrix)
        {
            Equation row1 = new Equation(conicMatrix.GetA(), conicMatrix.GetD(), 0, conicMatrix.GetG());
            Equation row2 = new Equation(conicMatrix.GetD(), conicMatrix.GetB(), 0, conicMatrix.GetH());
            Equation row3 = new Equation(0, 0, 0, 0);

            return new CenterEquationSystem(row1, row2, row3).Solve();
        }
        private CenterParametrization GetCenterParametrization(ConicMatrix conicMatrix, TransformationMatrix rotation)
        {
            ConicMatrix matrixRotated = conicMatrix.ApplyTransformation(rotation);
            try
            {
                double a = matrixRotated.GetA();
                double b = matrixRotated.GetB();
                double g = matrixRotated.GetG() * 2.0;
                double h = matrixRotated.GetH() * 2.0;

                double m = 0;
                double n = 0;
                double o = 0;
                ParametrizationRST x, y, z;
                if (!Program.IsEqualInTolerance(a, 0))
                    m = (g / (2 * a));
                if (!Program.IsEqualInTolerance(b, 0))
                    n = (h / (2 * b));
                x = new ParametrizationRST(-m, 0, 0, 0);
                y = new ParametrizationRST(-n, 0, 0, 0);
                z = new ParametrizationRST(-o, 0, 0, 0);
                return new CenterParametrization(x, y, z) * rotation;
            }
            catch (Exception) { return new CenterParametrization(); };
        }
        private CenterParametrization GetPeakParametrization(ConicMatrix conicMatrix, TransformationMatrix rotation)
        {
            ConicMatrix matrixRotated = conicMatrix.ApplyTransformation(rotation);
            try
            {
                double a = matrixRotated.GetA();
                double g = matrixRotated.GetG() * 2.0;
                double h = matrixRotated.GetH() * 2.0;
                double l = matrixRotated.GetL();
                double m, n, o;
                ParametrizationRST x, y, z;
                if (Program.IsEqualInTolerance(a, 0) || Program.IsEqualInTolerance(h, 0))
                    return new CenterParametrization();
                m = (g / (2 * a));
                n = (l - (g * g) / (4 * a)) / h;
                o = 0;
                x = new ParametrizationRST(-m, 0, 0, 0);
                y = new ParametrizationRST(-n, 1, 0, 0);
                z = new ParametrizationRST(-o, 0, 0, 0);
                return new CenterParametrization(x, y, z) * rotation;
            }
            catch (Exception) { return new CenterParametrization(); };
        }





        private ConicSection ClasifyConicEquation(ConicEquation equation, TransformationMatrix matrix)
        {
            ConicSection temp = null;
            bool isCanonized = true;
            //for (int i = 0; i < equations.Count; i++)
            {
                isCanonized = true;
                isCanonized = isCanonized && Program.IsEqualInToleranceLow(equation.GetD(), 0);
                if (!isCanonized)
                {
                    return null;
                }
                temp = IsParabola(equation, matrix);
                if (temp != null)
                    return temp;
                //equation = NormalizeEquation(equation);
                temp = ClasificationHelp(equation, matrix);
                if (temp != null)
                    return temp;
            }
            return null;
        }
        /// <summary>
        /// Podelim equation koeficientem -L pokud neni nulovy
        /// </summary>
        /// <returns></returns>
        private ConicEquation NormalizeEquation(ConicEquation equation)
        {
            if (!Program.IsEqualInTolerance(equation.GetL(), 0)) // deleni absolutnim clenem 
            {
                for (int j = 0; j < ConicEquation.MembersCount; j++)
                    equation.Set(j, equation.Get(j) / -equation.GetL());
            }
            return equation;
        }
        private ConicSection ClasificationHelp(ConicEquation equation, TransformationMatrix transformation)
        {
            ConicSection temp;

            if ((temp = IsElipse(equation, transformation)) != null)
                return temp;
            if ((temp = IsImaginarElipse(equation, transformation)) != null)
                return temp;
            if ((temp = IsParabola(equation, transformation)) != null)
                return temp;
            if ((temp = IsHyperbola(equation, transformation)) != null)
                return temp;
            if ((temp = IsParallelLines(equation, transformation)) != null)
                return temp;
            if ((temp = IsImaginarParallelLines(equation, transformation)) != null)
                return temp;
            if ((temp = IsCrosingLines(equation, transformation)) != null)
                return temp;
            if ((temp = IsImaginarCrosingLines(equation, transformation)) != null)
                return temp;
            if ((temp = IsDoubleLine(equation, transformation)) != null)
                return temp;
            return null;
        }




        /// <summary>
        /// x2 + y2 = 1
        /// </summary>
        private ConicSection IsElipse(ConicEquation conicEquation, TransformationMatrix transformation)
        {
            bool isElipse = true;
            double a = conicEquation.GetA();
            double b = conicEquation.GetB();
            isElipse = isElipse && a > 0 && !IsCloseToZero(a);
            isElipse = isElipse && b > 0 && !IsCloseToZero(b);
            //isElipse = isElipse && Program.IsEqualInTolerance(conicEquation.GetG(), 0);
            //isElipse = isElipse && Program.IsEqualInTolerance(conicEquation.GetH(), 0);
            isElipse = isElipse && !IsCloseToZero(conicEquation.GetL()) && conicEquation.GetL() < 0;
            if (isElipse)
            {
                conicEquation = NormalizeEquation(conicEquation);
                a = conicEquation.GetA();
                b = conicEquation.GetB();

                a = Math.Sqrt(1.0 / a);
                b = Math.Sqrt(1.0 / b);
                return new Elipse(_originalEquation, new ConicParametricEquation(a, b, 0),
                    transformation, _color, _name, _ID);
            }
            return null;
        }
        /// <summary>
        /// x2 + y2 = -1
        /// </summary>
        private ConicSection IsImaginarElipse(ConicEquation conicEquation, TransformationMatrix transformation)
        {
            bool isElipse = true;
            double a = conicEquation.GetA();
            double b = conicEquation.GetB();
            isElipse = isElipse && a > 0 && !IsCloseToZero(a);
            isElipse = isElipse && b > 0 && !IsCloseToZero(b);
            //isElipse = isElipse && Program.IsEqualInTolerance(conicEquation.GetG(), 0);
            //isElipse = isElipse && Program.IsEqualInTolerance(conicEquation.GetH(), 0);
            isElipse = isElipse && !IsCloseToZero(conicEquation.GetL()) && conicEquation.GetL() > 0;
            if (isElipse)
            {
                conicEquation = NormalizeEquation(conicEquation);
                a = conicEquation.GetA();
                b = conicEquation.GetB();

                a = Math.Sqrt(1.0 / -a);
                b = Math.Sqrt(1.0 / -b);
                return new ImaginarElipse(_originalEquation, new ConicParametricEquation(a, b, 0),
                    transformation, _color, _name, _ID);
            }
            return null;
        }
        /// <summary>
        /// x2 - y2 = 1
        /// </summary>
        private ConicSection IsHyperbola(ConicEquation conicEquation, TransformationMatrix transformation)
        {
            bool isHyperbola = true;
            double a = conicEquation.GetA();
            double b = conicEquation.GetB();
            isHyperbola = isHyperbola && a > 0 && !IsCloseToZero(a);
            isHyperbola = isHyperbola && b < 0 && !IsCloseToZero(b);
            //isHyperbola = isHyperbola && Program.IsEqualInTolerance(conicEquation.GetG(), 0);
            //isHyperbola = isHyperbola && Program.IsEqualInTolerance(conicEquation.GetH(), 0);
            isHyperbola = isHyperbola && !IsCloseToZero(conicEquation.GetL()) && conicEquation.GetL() < 0;
            if (isHyperbola)
            {
                conicEquation = NormalizeEquation(conicEquation);
                a = conicEquation.GetA();
                b = conicEquation.GetB();

                a = Math.Sqrt(1.0 / a);
                b = Math.Sqrt(1.0 / -b);
                return new Hyperbola(_originalEquation, new ConicParametricEquation(a, b, 0),
                    transformation, _color, _name, _ID);
            }
            return null;
        }
        /// <summary>
        /// x2 + 2py = 0
        /// </summary>
        private ConicSection IsParabola(ConicEquation conicEquation, TransformationMatrix transformation)
        {
            bool isParabola = true;
            double a = conicEquation.GetA();
            double p = conicEquation.GetH() / 2.0;
            isParabola = isParabola && a > 0 && !IsCloseToZero(a);
            isParabola = isParabola && !IsCloseToZero(p);
            isParabola = isParabola && Program.IsEqualInTolerance(conicEquation.GetB(), 0);
            //isParabola = isParabola && Program.IsEqualInTolerance(conicEquation.GetG(), 0);
            //isParabola = isParabola && Program.IsEqualInTolerance(conicEquation.GetL(), 0);
            if (isParabola)
            {
                a = Math.Sqrt(1.0 / a);
                return new Parabola(_originalEquation, new ConicParametricEquation(a, 0, p),
                    transformation, _color, _name, _ID);
            }
            return null;
        }
        /// <summary>
        /// x2 - y2 = 0
        /// </summary>
        private ConicSection IsCrosingLines(ConicEquation conicEquation, TransformationMatrix transformation)
        {
            bool isCrossingLines = true;
            double a = conicEquation.GetA();
            double b = conicEquation.GetB();
            isCrossingLines = isCrossingLines && a > 0 && !IsCloseToZero(a);
            isCrossingLines = isCrossingLines && b < 0 && !IsCloseToZero(b);
            //isCrossingLines = isCrossingLines && Program.IsEqualInTolerance(conicEquation.GetG(), 0);
            //isCrossingLines = isCrossingLines && Program.IsEqualInTolerance(conicEquation.GetH(), 0);
            isCrossingLines = isCrossingLines && Program.IsEqualInTolerance(conicEquation.GetL(), 0);
            if (isCrossingLines)
            {
                a = Math.Sqrt(1.0 / a);
                b = Math.Sqrt(-1.0 / b);
                return new CrossingLines(_originalEquation, new ConicParametricEquation(a, b, 0),
                    transformation, _color, _name, _ID);
            }
            return null;
        }
        /// <summary>
        /// x2 + y2 = 0
        /// </summary>
        private ConicSection IsImaginarCrosingLines(ConicEquation conicEquation, TransformationMatrix transformation)
        {
            bool isCrossingLines = true;
            double a = conicEquation.GetA();
            double b = conicEquation.GetB();
            isCrossingLines = isCrossingLines && a > 0 && !IsCloseToZero(a);
            isCrossingLines = isCrossingLines && b > 0 && !IsCloseToZero(b);
            //isCrossingLines = isCrossingLines && Program.IsEqualInTolerance(conicEquation.GetG(), 0);
            //isCrossingLines = isCrossingLines && Program.IsEqualInTolerance(conicEquation.GetH(), 0);
            isCrossingLines = isCrossingLines && Program.IsEqualInTolerance(conicEquation.GetL(), 0);
            if (isCrossingLines)
            {
                a = Math.Sqrt(1.0 / a);
                b = Math.Sqrt(1.0 / b);
                return new ImaginarCrossingLines(_originalEquation, new ConicParametricEquation(a, b, 0),
                    transformation, _color, _name, _ID);
            }
            return null;
        }
        /// <summary>
        /// x2 - c = 0
        /// </summary>
        private ConicSection IsParallelLines(ConicEquation conicEquation, TransformationMatrix transformation)
        {
            bool isParallelLines = true;
            double a = conicEquation.GetA();
            isParallelLines = isParallelLines && a > 0 && !IsCloseToZero(a);
            isParallelLines = isParallelLines && Program.IsEqualInTolerance(conicEquation.GetB(), 0);
            //isParallelLines = isParallelLines && Program.IsEqualInTolerance(conicEquation.GetG(), 0);
            isParallelLines = isParallelLines && Program.IsEqualInTolerance(conicEquation.GetH(), 0);
            isParallelLines = isParallelLines && !IsCloseToZero(conicEquation.GetL()) && conicEquation.GetL() < 0;
            if (isParallelLines)
            {
                conicEquation = NormalizeEquation(conicEquation);
                a = conicEquation.GetA();
                
                a = Math.Sqrt(1.0 / a);
                return new ParallelLines(_originalEquation, new ConicParametricEquation(a, 0, 0),
                    transformation, _color, _name, _ID);
            }
            return null; ;
        }
        /// <summary>
        /// x2 + c = 0
        /// </summary>
        private ConicSection IsImaginarParallelLines(ConicEquation conicEquation, TransformationMatrix transformation)
        {
            bool isParallelLines = true;
            double a = conicEquation.GetA();
            isParallelLines = isParallelLines && a > 0 && !IsCloseToZero(a);
            isParallelLines = isParallelLines && Program.IsEqualInTolerance(conicEquation.GetB(), 0);
            //isParallelLines = isParallelLines && Program.IsEqualInTolerance(conicEquation.GetG(), 0);
            isParallelLines = isParallelLines && Program.IsEqualInTolerance(conicEquation.GetH(), 0);
            isParallelLines = isParallelLines && !IsCloseToZero(conicEquation.GetL()) && conicEquation.GetL() > 0;
            if (isParallelLines)
            {
                conicEquation = NormalizeEquation(conicEquation);
                a = conicEquation.GetA();

                a = Math.Sqrt(1.0 / -a);
                return new ImaginarParallelLines(_originalEquation, new ConicParametricEquation(a, 0, 0),
                    transformation, _color, _name, _ID);
            }
            return null; ;
        }
        /// <summary>
        /// x2 = 0
        /// </summary>
        private ConicSection IsDoubleLine(ConicEquation conicEquation, TransformationMatrix transformation)
        {
            bool isDoubleLine = true;
            double a = conicEquation.GetA();
            isDoubleLine = isDoubleLine && a > 0 && !IsCloseToZero(a);
            isDoubleLine = isDoubleLine && Program.IsEqualInTolerance(conicEquation.GetB(), 0);
            //isDoubleLine = isDoubleLine && Program.IsEqualInTolerance(conicEquation.GetG(), 0);
            //isDoubleLine = isDoubleLine && Program.IsEqualInTolerance(conicEquation.GetH(), 0);
            isDoubleLine = isDoubleLine && Program.IsEqualInTolerance(conicEquation.GetL(), 0);
            if (isDoubleLine)
            {
                a = Math.Sqrt(1.0 / a);
                return new DoubleLine(_originalEquation, new ConicParametricEquation(a, 0, 0),
                    transformation, _color, _name, _ID);
            }
            return null;
        }


        private bool IsCloseToZero(double d)
        {
            double tolerance = 0.00001;
            return Math.Abs(d) < tolerance;
        }

    }
}
