﻿using System;
using Geometry.GGraphic.Forms.Detail.FormatString;
using Geometry.GraphicObjects.Planar;
using Geometry.GraphicObjects.Equations;
using Geometry.GraphicObjects.Matrixes;
using GMath;
using Tao.OpenGl;
using System.Collections.Generic;
using Geometry.BaseClasses;

namespace Geometry.GraphicObjects.ConicSections
{
    public class ConicSection : WithCanonicEquationObject
    {
        protected static CurveCylinder _curveCylinder = new CurveCylinder();

        private Plan _conicPlan;

        public static string _conicCommonFormatString =
            "L[!Ax" + U.Exp2 + " + !By" + U.Exp2 +
            " + !Dxy + !Gx + !Hy + !L = 0]";

        public struct Matrixes 
        {
            public TransformationMatrix planTransformation;
            public TransformationMatrix conicTransformation;

            public Matrixes(TransformationMatrix planTransformationArg, TransformationMatrix conicTransformationArg) 
            {
                planTransformation = planTransformationArg;
                conicTransformation = conicTransformationArg;
            }
        }

        public ConicSection(CommonEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix, GColor color, string name,
            string wordDescriptionFirstCase, string wordDescriptionSecondCase, string wordDescriptionFourthCase,
            string canonicFormatString,
            string parametricXFormatString, string parametricYFormatString,
            int ID)
        {
            base._ID = ID;
            base._commonEquation = commonEquation;
            base._parametricEquation = parametricEquation;
            base._color = color;
            base._name = name;
            base.SetWordDescriptionFirstCase(wordDescriptionFirstCase);
            base.SetWordDescriptionSecondCase(wordDescriptionSecondCase);
            base.SetWordDescriptionFourthCase(wordDescriptionFourthCase);
            base._commonEquationFormatSring = new CommonEquationFormatString(_conicCommonFormatString);
            base._canonicEquationFormatString = new CanonicEquationFormatString(canonicFormatString);
            base._parametricEquationFormatString = new ParametricEquationsFormatString(parametricXFormatString, parametricYFormatString);
            _conicPlan = new Plan(color, name, ID);
            SetTransformationMatrixAlgebraic(toCanonicTransformationMatrix.ComputeInverseMatrix());
        }


        public void ApplyTransformation(TransformationMatrix transformation)
        {
            base.SetTransformationMatrixAlgebraic(transformation * _transformationAlgebraic);
        }



        /// <summary>
        /// Tato rovina je otocena i posunuta.
        /// Normálně tedy při jednotkové rotační matici, je rovinou, rovina Z=0.
        /// </summary>
        private void ActualizeConicPlan(TransformationMatrix transformationMatrixAlgebraic)
        {
            _conicPlan.SetTransformationMatrixAlgebraic(transformationMatrixAlgebraic);
        }
        public Plan GetConicPlan()
        {
            return _conicPlan;
        }


        /// <summary>
        /// Vytvoří kopii kuželošečky s novým ID.
        /// Aplikuje na ni transformaci aplikovanou na rovinu původní kuželosečky  
        /// </summary>
        public override WithCommonEquationObject MakeCopyOfObject(CommonEquation commonEquation, int ID)
        {
            ConicRecognizer recognizer = new ConicRecognizer((ConicEquation)commonEquation, _color, _name, ID);
            ConicSection newConic = recognizer.Recognize();
            newConic.CopyDrawableSlots(newConic, this);
            TransformationMatrix newConicTransformationMatrix =
                this.GetConicPlan().GetTransformationMatrixAlgebraic() *
                newConic.GetTransformationMatrixAlgebraic();
            newConic.SetTransformationMatrixAlgebraic(newConicTransformationMatrix);
            return newConic;
        }
        /// <summary>
        /// Kanonicka nebo parametricka rovnice editovana.
        /// Vytvori se novy object kvadriky.
        /// Zkopiruji se ji drawable slots.
        /// </summary>
        public override WithParametricEquationObject MakeCopyOfObject(KoefsAndCenterPair newKoeficientsAndCenter, int ID)
        {
            ParametricEquation newParametricEquation = newKoeficientsAndCenter.GetParametricEquation();
            CenterParametrization newCenterInCanonicBasis = newKoeficientsAndCenter.GetCenterParametrization();
            CenterParametrization newCenterParametrization = ConvertCanonicToStandardBasis(newCenterInCanonicBasis);
            TransformationMatrix newToCanonicMatrix =
                GetToCanonicTransformationMatrix(GetToCanonicTransformationMatrix(), newCenterParametrization);
            ConicMatrix canonicMatrix = GetCanonicMatrix();
            ConicMatrix commonMatrix = canonicMatrix.ApplyTransformation(newToCanonicMatrix.ComputeInverseMatrix());
            ConicEquation newCommonEquation = (ConicEquation)NormalizeEquation(new ConicEquation(commonMatrix));

            ConicSection conic =
                CallConstructor(newCommonEquation, newParametricEquation, newToCanonicMatrix, _color, _name, ID);
            conic.CopyDrawableSlots(conic, this);
            return conic;
        }
        /// <summary>
        /// Vytvoří kopii kuželosečky s novým ID
        /// </summary>
        public override WithParametricEquationObject MakeCopyOfObject(int ID)
        {
            ConicParametricEquation equationCopy =
                new ConicParametricEquation((ConicParametricEquation)_parametricEquation);
            CenterParametrization centerCopy =
                new CenterParametrization(GetCenterInCanonicBasis());
            return MakeCopyOfObject(new KoefsAndCenterPair(equationCopy, centerCopy), ID);
        }


        protected virtual ConicSection CallConstructor(ConicEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix, GColor color, string name, int ID)
        {
            throw new Exception("Tato metoda musí být přepsána v potomkovi.");
        }

        public virtual void BuildAsIntersection(Scene scene, RelativeQuality relativeQuality) { }

        public override void Draw(Scene scene)
        {
            base.Draw(scene);
            //_conicPlan.BuildDisplayList(scene, scene.GetBaseQuality(), _relativeQuality);
            //_conicPlan.ChangeColor(new Color(0.75, 0.75, 0.75));
            //_conicPlan.Draw(scene);
        }

        protected void BuildAsCylinder(List<Vector> points)
        {
            //Gl.glPolygonMode(Gl.GL_FRONT_AND_BACK, Gl.GL_FILL);
            for (int i = 1; i < points.Count; i++)
            {
                _curveCylinder.MakeCylinderComplete(points[i - 1], points[i]);
            }
        }
        protected void BuildAsLineStrip(List<Vector> points)
        {
            Gl.glDisable(Gl.GL_LIGHTING);
            Gl.glBegin(Gl.GL_LINE_STRIP);

            foreach (Vector point in points)
                Gl.glVertex3d(point.GetX(), point.GetY(), point.GetZ());

            Gl.glEnd();
            Gl.glEnable(Gl.GL_LIGHTING);
        }

        public new ConicEquation GetCommonEquation()
        {
            return (ConicEquation)_commonEquation;
        }
        public ConicParametricEquation GetCanonicEquation()
        {
            return (ConicParametricEquation)_parametricEquation;
        }


        /// <summary>
        /// Na rovinu v kanonickém tvaru aplikuji transformační matici kuželosečky.
        /// Zjistím úhel normálového vektoru této roviny a normálového vektoru roviny v kanonickém tvaru;
        /// Poté stanovím transformační matici převádějící kanonickou rovinu na otočenou rovinu.
        /// Tuto transformaci aplikuji na směrové vektory kanonické roviny.
        /// Dále aplikuji původní transformaci na směrové vektory kanonické roviny.
        /// Nyní jsou roviny určeny dvojicemi směrových vektorů opticky totožné...
        /// Úhly obou směrových vektorů otočených transformační maticí kuželosečky a druhou transformační maticí jsou stejné.
        /// Vezmu libovolný z těchto dvou úhlů.
        /// --- Jako transformaci kuželosečky vezmu rotaci okolo osy Z o tento uhel.
        /// Jako transformaci roviny nelze nevím proč vzít rotaci, která otočila kanonickou rovinu do potřebné pozice.
        /// --- Místo toho jako transformaci roviny vezmu původní transformaci a odeberu od ní rotaci kuželosečky.
        /// </summary>
        /// <param name="transformationMatrix"></param>
        /// <returns></returns>
        public static Matrixes DecomposeTransformationMatrix(TransformationMatrix transformationMatrix) 
        {
            TransformationMatrix rotationMatrix = new TransformationMatrix(transformationMatrix);
            rotationMatrix.SetTranslation(new Vector());

            Vector rotatedNormal = Plan.canonicalNormalVector * rotationMatrix;
            double angl = Vector.Angle(Plan.canonicalNormalVector, rotatedNormal) * 180 / Math.PI;
            Vector axe = Vector.CrossProduct(Plan.canonicalNormalVector, rotatedNormal);

            TransformationMatrix planTransformation = TransformationMatrix.GetRotationMatrixAroundAAxe(axe, -angl);

            Vector fdvOrig = new Vector(1, 0, 0) * planTransformation;
            Vector sdvOrig = new Vector(0, 1, 0) * planTransformation;

            Vector fdvNew = new Vector(1, 0, 0) * rotationMatrix;
            Vector sdvNew = new Vector(0, 1, 0) * rotationMatrix;
            double fdva = Vector.OrientedAngle(fdvOrig, fdvNew, rotatedNormal);// *180 / Math.PI;
            double sdva = Vector.OrientedAngle(sdvOrig, sdvNew, rotatedNormal);// *180 / Math.PI;

            Vector translation = transformationMatrix.GetTranslation();
            Vector planTranslation = new Vector(0, 0, translation.GetZ());
            Vector canonicConicTranslation = new Vector(translation.GetX(), translation.GetY(), 0);

            TransformationMatrix newConicTransformation = TransformationMatrix.GetZRotationMatrix(fdva);
            newConicTransformation.SetTranslation(canonicConicTranslation);            
            TransformationMatrix newPlanTransformation = transformationMatrix * newConicTransformation.ComputeInverseMatrix();
            newPlanTransformation.SetTranslation(planTranslation);
            return new Matrixes(newPlanTransformation, newConicTransformation);
        }
        public static Matrixes DecomposeTransformationMatrixOld(TransformationMatrix transformationMatrix)
        {
            TransformationMatrix rotationMatrix = new TransformationMatrix(transformationMatrix);
            rotationMatrix.SetTranslation(new Vector());

            Vector planNormal = Plan.canonicalNormalVector * rotationMatrix;
            RotationAngles angles = rotationMatrix.DecomposeRotationMatrixOnAngles();

            TransformationMatrix conicTransformation = TransformationMatrix.GetZRotationMatrix(angles.GetZAngle());
            TransformationMatrix m = TransformationMatrix.GetRotationMatrixAroundAAxe(planNormal, -angles.GetZAngle());
            Vector translation = transformationMatrix.GetTranslation();
            conicTransformation.SetTranslation(new Vector(translation.GetX(), translation.GetY(), 0));
            TransformationMatrix planTransformationMatrix = m * rotationMatrix;
            planTransformationMatrix.SetTranslation(new Vector(0, 0, translation.GetZ()));
            return new Matrixes(planTransformationMatrix, conicTransformation);
        }

        protected override void ActualizeEquations()
        {
            Matrixes matrixes = DecomposeTransformationMatrix(GetTransformationMatrixAlgebraic());
            TransformationMatrix planTransformationMatrix = matrixes.planTransformation;
            TransformationMatrix conicTransformation = matrixes.conicTransformation;

            ActualizeConicPlan(planTransformationMatrix);

            CenterParametrization canonicCenter = GetCenterInCanonicBasis();
            ConicMatrix canonicMatrix = GetCanonicMatrix();
            ConicMatrix commonMatrix = canonicMatrix.ApplyTransformation(conicTransformation);

            ConicEquation equation = new ConicEquation(commonMatrix);
            _commonEquation = NormalizeEquation(equation);
            _commonEquationFormatSring.SubstituteToFormatString(_commonEquation);
            _canonicEquationFormatString.SubstituteToFormatString(_parametricEquation, canonicCenter);
            _parametricEquationFormatString.SubstituteToFormatStrings(_parametricEquation, canonicCenter);
        }
        public virtual ConicMatrix GetCanonicMatrix()
        {
            throw new Exception("Tato metoda musí být přepsána v potomkovi.");
        }

    }
}
