﻿using System;
using Tao.OpenGl;
using Geometry.GraphicObjects.Matrixes;
using Geometry.GraphicObjects.Equations;
using Geometry.BaseClasses;
using GMath;

namespace Geometry.GraphicObjects.ConicSections
{
    public class CrossingLines : SimpleRuledConic
    {
        private double _a;
        private double _b;

        public static string canonicFormatString = "E[x-!m;!a]L[-]E[y-!n;!b]L[ = 0]";
        public static string parametricXFormatString = "L[x = !m + u]";
        public static string parametricYFormatString = "L[y = !n " + U.PM + "]F[L[u" + U.Times + "!b];L[!a]]";

        public static string _wordDescriptionFirstCase = "Různoběžné přímky";
        public static string _wordDescriptionSecondCase = "Různoběžných přímek";
        public static string _wordDescriptionFourthCase = "Různoběžné přímky";

        public CrossingLines(ConicEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix,
            GColor color, string name, int ID)
            : base(commonEquation, parametricEquation, toCanonicTransformationMatrix,
            color, name,
            _wordDescriptionFirstCase, _wordDescriptionSecondCase, _wordDescriptionFourthCase,
            canonicFormatString, parametricXFormatString, parametricYFormatString, ID)
        {
            _a = parametricEquation.GetA();
            _b = parametricEquation.GetB();
        }


        /// x = m + u.
        public double ComputeXCoordinateAt(double v)
        {
            return v;
        }

        /// y = n + u*b/a. 
        public double ComputeYCoordinateAt(double v)
        {
            return (v * _b) / _a;
        }

        protected override void BuildParametricDisplayList(Scene scene, RelativeQuality relativeQuality)
        {
            double cleft = GetClippingLeft(scene);
            double cright = GetClippingRight(scene);
            double ctop = GetClippingTop(scene);
            double cbottom = GetClippingBottom(scene);

            double uMin;
            double uMax;

            _displayListIdentificator = Gl.glGenLists(1);
            Gl.glNewList(_displayListIdentificator, Gl.GL_COMPILE);

            uMin = Math.Max(cleft, cbottom * _a / _b);
            uMax = Math.Min(cright, ctop * _a / _b);
            BuildAsLineStrip(uMin, uMax, 1);


            uMin = Math.Max(cleft, -ctop * _a / _b);
            uMax = Math.Min(cright,  -cbottom * _a / _b);
            BuildAsLineStrip(uMin, uMax, -1);

            Gl.glEndList();
        }

        public override void BuildAsIntersection(Scene scene, RelativeQuality relativeQuality)
        {
            double cleft = GetClippingLeft(scene);
            double cright = GetClippingRight(scene);
            double ctop = GetClippingTop(scene);
            double cbottom = GetClippingBottom(scene);

            double uMin;
            double uMax;

            _displayListIdentificator = Gl.glGenLists(1);
            Gl.glNewList(_displayListIdentificator, Gl.GL_COMPILE);

            uMin = Math.Max(cleft, cbottom * _a / _b);
            uMax = Math.Min(cright, ctop * _a / _b);
            BuildAsCylinder(uMin, uMax, 1);


            uMin = Math.Max(cleft, -ctop * _a / _b);
            uMax = Math.Min(cright, -cbottom * _a / _b);
            BuildAsCylinder(uMin, uMax, -1);

            Gl.glEndList();
        }

        
        protected void BuildAsCylinder(double uMin, double uMax, int sign) 
        {
            Vector a, b;
            a = new Vector(ComputeXCoordinateAt(uMin), ComputeYCoordinateAt(sign*uMin), 0);
            b = new Vector(ComputeXCoordinateAt(uMax), ComputeYCoordinateAt(sign*uMax), 0);
            Planar.Line.BuildLineAsCylinder(a, b);
        }
        protected void BuildAsLineStrip(double uMin, double uMax, int sign)
        {
            Vector a, b;
            a = new Vector(ComputeXCoordinateAt(uMin), ComputeYCoordinateAt(sign * uMin), 0);
            b = new Vector(ComputeXCoordinateAt(uMax), ComputeYCoordinateAt(sign * uMax), 0);
            Planar.Line.BuildLineAsLineStrip(a, b);
        }


        //////////////////////////////////////////////////////////////////////////////////////////
        //////////////////////////////////////////////////////////////////////////////////////////



        protected override ConicSection CallConstructor(ConicEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix, GColor color, string name, int ID)
        {
            return new CrossingLines(commonEquation, parametricEquation, toCanonicTransformationMatrix, color, name, ID);
        }

        public override ConicMatrix GetCanonicMatrix()
        {
            double A = 1 / (_parametricEquation.GetA() * _parametricEquation.GetA());
            double B = -1 / (_parametricEquation.GetB() * _parametricEquation.GetB());
            ConicEquation newCanonicEquation = new ConicEquation(A, B, 0, 0, 0, 0);
            return new ConicMatrix(newCanonicEquation);
        }
    }
}
