﻿using System;
using Tao.OpenGl;
using Geometry.GraphicObjects.Matrixes;
using Geometry.GraphicObjects.Equations;
using Geometry.BaseClasses;
using GMath;
using System.Collections.Generic;

namespace Geometry.GraphicObjects.ConicSections
{
    /// <summary>
    /// x2/a2 + y2/b2 = 1;
    /// </summary>
    public class Elipse:ConicSection
    {

        private double _a;
        private double _b;
        private double _baseQuality = 30;

        public static string canonicFormatString = "E[x-!m;!a]L[ + ]E[y-!n;!b]L[ = 1]";
        public static string parametricXFormatString = "L[x = !m + !a" + U.Times + "cos(v)]";
        public static string parametricYFormatString = "L[y = !n + !b" + U.Times + "sin(v)]";

        public static string _wordDescriptionFirstCase = "Elipsa";
        public static string _wordDescriptionSecondCase = "Elipsy";
        public static string _wordDescriptionFourthCase = "Elipsu";

        public Elipse(ConicEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix, GColor color, string name, int ID)
            : base(commonEquation, parametricEquation, toCanonicTransformationMatrix,
            color, name,
            _wordDescriptionFirstCase, _wordDescriptionSecondCase, _wordDescriptionFourthCase,
            canonicFormatString, parametricXFormatString, parametricYFormatString, ID)
        {
            _a = parametricEquation.GetA();
            _b = parametricEquation.GetB();
        }

        /// x = m + a*cos(v).
        public double ComputeXCoordinateAt(double v)
        {
            return _a * Math.Cos(v);
        }

        /// y = n + b*sin(v).
        public double ComputeYCoordinateAt(double v)
        {
            return _b * Math.Sin(v);
        }

        /// <summary>
        /// Dosadi do parametricke rovnice v a vraci souradnice bodu.
        /// </summary>
        /// <param name="u"></param>
        /// <param name="v"></param>
        /// <returns></returns>
        public Vector ComputeCoordinatesAt(double v) 
        {
            double x = ComputeXCoordinateAt(v);
            double y = ComputeYCoordinateAt(v);

            return new Vector(x, y, 0);
        }

        protected override void BuildParametricDisplayList(Scene scene, RelativeQuality relativeQuality)
        {
            List<Vector> points = new List<Vector>();
            double step = Math.PI / (_baseQuality * (scene.GetBaseQuality() + relativeQuality));
            for (double v = 0; v < Math.PI * 2 + step; v += step)
            {
                if (v > Math.PI * 2)
                    v = Math.PI * 2;
                points.Add(new Vector(ComputeXCoordinateAt(v), ComputeYCoordinateAt(v), 0));
                if (v == Math.PI * 2)
                    break;
            }

            _displayListIdentificator = Gl.glGenLists(1);
            Gl.glNewList(_displayListIdentificator, Gl.GL_COMPILE);
            BuildAsLineStrip(points);
            Gl.glEndList();
        }
        public override void BuildAsIntersection(Scene scene, RelativeQuality relativeQuality)
        {
            List<Vector> points = new List<Vector>();
            double step = Math.PI / (_baseQuality * (scene.GetBaseQuality() + relativeQuality));
            for (double v = 0; v < Math.PI * 2 + step; v += step)
            {
                if (v > Math.PI * 2)
                    v = Math.PI * 2;
                points.Add(new Vector(ComputeXCoordinateAt(v), ComputeYCoordinateAt(v), 0));
                if (v == Math.PI * 2)
                    break;
            }

            _displayListIdentificator = Gl.glGenLists(1);
            Gl.glNewList(_displayListIdentificator, Gl.GL_COMPILE);
            BuildAsCylinder(points);
            Gl.glEndList();
        }


        protected override ConicSection CallConstructor(ConicEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix, GColor color, string name, int ID)
        {
            return new Elipse(commonEquation, parametricEquation, toCanonicTransformationMatrix, color, name, ID);
        }


        public override ConicMatrix GetCanonicMatrix()
        {
            double A = 1 / (_parametricEquation.GetA() * _parametricEquation.GetA());
            double B = 1 / (_parametricEquation.GetB() * _parametricEquation.GetB());
            ConicEquation newCanonicEquation = new ConicEquation(A, B, 0, 0, 0, -1);
            return new ConicMatrix(newCanonicEquation);
        }
    }
}
