﻿using System;
using System.Collections.Generic;
using Tao.OpenGl;
using Geometry.GraphicObjects.Matrixes;
using Geometry.GraphicObjects.Equations; 
using Geometry.BaseClasses;
using GMath;

namespace Geometry.GraphicObjects.ConicSections
{
    /// <summary>
    /// x2/a2 - y2/b2 = 1;
    /// </summary>
    public class Hyperbola : SimpleRuledConic
    {

        private double _a;
        private double _b;
        private int _linesPerInterval = 50;

        public static string canonicFormatString = "E[x-!m;!a]L[ - ]E[y-!n;!b]L[ = 1]";
        public static string parametricXFormatString = "L[x = !m "+U.PM +" !a" + U.Times + "cosh(v)]";
        public static string parametricYFormatString = "L[y = !n + !b" + U.Times + "sinh(v)]";

        public static string _wordDescriptionFirstCase = "Hyperbola";
        public static string _wordDescriptionSecondCase = "Hyperboly";
        public static string _wordDescriptionFourthCase = "Hyperbolu";

        const int LEFT = 0;
        const int RIGHT = 1; 
        public struct Pair 
        {
            public Interval Interval;
            public int Part;
            public Pair(Interval interval, int part) 
            {
                Interval = interval;
                Part = part;
            }
        }

        public Hyperbola(ConicEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix,
            GColor color, string name, int ID)
            : base(commonEquation, parametricEquation, toCanonicTransformationMatrix,
            color, name,
            _wordDescriptionFirstCase, _wordDescriptionSecondCase, _wordDescriptionFourthCase,
            canonicFormatString, parametricXFormatString, parametricYFormatString, ID)
        {
            _a = parametricEquation.GetA();
            _b = parametricEquation.GetB();
        }

        /// x = m + a*cosh(v).
        public double ComputeXCoordinateAt(double v)
        {
            return _a*Math.Cosh(v);
        }
        /// y = n + b*sinh(v).
        public double ComputeYCoordinateAt(double v)
        {
            return _b * Math.Sinh(v);
        }

        ///// x = m + a*sqrt(b*b+u*u).
        //public double ComputeXCoordinateAtEasier(double v)
        //{
        //    return _a * Math.Sqrt(1 + v * v);
        //}
        ///// y = n + v*b.
        //public double ComputeYCoordinateAtEasier(double v)
        //{
        //    return _b*v;
        //}



        ///// <summary>
        ///// Dosadi do parametricke rovnice v a vraci souradnice bodu.
        ///// </summary>
        ///// <param name="u"></param>
        ///// <param name="v"></param>
        ///// <returns></returns>
        //public Vector ComputeCoordinatesAt(double v)
        //{
        //    double x = ComputeXCoordinateAtEasier(v);
        //    double y = ComputeYCoordinateAtEasier(v);

        //    return new Vector(x, y, 0);
        //}

        private List<List<Vector>> GetPoints(Scene scene, RelativeQuality relativeQuality) 
        {
            List<Pair> clippingIntervalsPairs = GetClippingIntervals(scene);
            List<List<Vector>> points = new List<List<Vector>>();
            double uMin;
            double uMax;
            int sign = 1;
            double step; // = Math.PI / (_baseQuality * (scene.GetBaseQuality() + relativeQuality));
            int linesPerInterval = _linesPerInterval * (scene.GetBaseQuality() + relativeQuality);
            foreach (Pair p in clippingIntervalsPairs)
            {
                if (p.Interval.IsEmpty())
                    continue;
                step = (p.Interval.Size() / linesPerInterval);
                uMin = p.Interval.GetA();
                uMax = p.Interval.GetB();
                if (p.Part == LEFT)
                    sign = -1;
                else sign = 1;
                List<Vector> hyperbolaPart = new List<Vector>();
                for (double v = uMin; v <= uMax; v += 0)
                {
                    hyperbolaPart.Add(new Vector(sign*ComputeXCoordinateAt(v), ComputeYCoordinateAt(v), 0));
                    if (v == uMax)
                        break;
                    v = Math.Min(v + step, uMax);
                }
                points.Add(hyperbolaPart);
            }
            return points;
        }

        private List<Pair> GetClippingIntervals(Scene scene) 
        {
            double cleft = GetClippingLeft(scene);
            double cright = GetClippingRight(scene);
            double ctop = GetClippingTop(scene);
            double cbottom = GetClippingBottom(scene);

            double temp = ArcCosh(cright / _a);
            Interval right = new Interval(-temp, temp); // "u" ve kterych hyperbola protina pravou hranici
            temp = ArcCosh(cleft / _a); // muze byt NAN
            Interval left = new Interval(-temp, temp); // "u" ve kterych hyperbola protina levou hranici

            Interval bottomTop = new Interval(ArcSinh(cbottom / _b), ArcSinh(ctop / _b)); // interval u pro ktere je hyperbola mezi up a down

            double leftPeak = -ComputeXCoordinateAt(0);
            double rightPeak = ComputeXCoordinateAt(0);

            List<Pair> pairs = new List<Pair>();

            // leva hyperbola je uvnitr
            if (left.GetA() != double.NaN || (bottomTop.GetA() != double.NaN && bottomTop.GetB() != double.NaN))
            {
                if (new Interval(cleft, cright).Belongs(leftPeak))
                {
                    pairs.Add(new Pair(Interval.Intersect(left, bottomTop), LEFT)); // vrchol leve hyperboly je uvnitr
                }
                else if (leftPeak > cright)
                {
                    List<Interval> tempInterval = Interval.MinusSmallerFromBigger(left, right);
                    foreach (Interval i in tempInterval)
                        pairs.Add(new Pair(Interval.Intersect(i, bottomTop), LEFT)); // leva hyperbola je rozdelena na dve
                } 
            }
            if(right.GetA() != double.NaN || (bottomTop.GetA() != double.NaN && bottomTop.GetB() != double.NaN))
            {
                if (new Interval(cleft, cright).Belongs(rightPeak))
                {
                    pairs.Add(new Pair(Interval.Intersect(right, bottomTop), RIGHT)); // vrchol prave hyperbooly je uvnitr
                }
                else if (rightPeak < cleft)
                {
                    List<Interval> tempInterval = Interval.MinusSmallerFromBigger(left, right);
                    foreach (Interval i in tempInterval)
                        pairs.Add(new Pair(Interval.Intersect(i, bottomTop), RIGHT)); // prava hyperbola je rozdelena na dve
                }
            }
            return pairs;
        }

        protected override void BuildParametricDisplayList(Scene scene, RelativeQuality relativeQuality)
        {
            List<List<Vector>> points = GetPoints(scene, relativeQuality);

            _displayListIdentificator = Gl.glGenLists(1);
            Gl.glNewList(_displayListIdentificator, Gl.GL_COMPILE);
            foreach (List<Vector> subPoints in points)
                BuildAsLineStrip(subPoints);
            Gl.glEndList();
        }
        public override void BuildAsIntersection(Scene scene, RelativeQuality relativeQuality)
        {
            List<List<Vector>> points = GetPoints(scene, relativeQuality);
            
            _displayListIdentificator = Gl.glGenLists(1);
            Gl.glNewList(_displayListIdentificator, Gl.GL_COMPILE);
            foreach (List<Vector> subPoints in points)
                BuildAsCylinder(subPoints);
            foreach (List<Vector> subPoints in points)
                BuildAsLineStrip(subPoints);
            Gl.glEndList();
        }




        protected override ConicSection CallConstructor(ConicEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix, GColor color, string name, int ID)
        {
            return new Hyperbola(commonEquation, parametricEquation, toCanonicTransformationMatrix, color, name, ID);
        }


        public override ConicMatrix GetCanonicMatrix()
        {
            double A = 1 / (_parametricEquation.GetA() * _parametricEquation.GetA());
            double B = -1 / (_parametricEquation.GetB() * _parametricEquation.GetB());
            ConicEquation newCanonicEquation = new ConicEquation(A, B, 0, 0, 0, -1);
            return new ConicMatrix(newCanonicEquation);
        }
    }
}
