﻿using System;
using Tao.OpenGl;
using Geometry.GraphicObjects.Matrixes;
using Geometry.GraphicObjects.Equations;
using Geometry.BaseClasses;
using GMath;
using System.Collections.Generic;

namespace Geometry.GraphicObjects.ConicSections
{
    public class Parabola : SimpleRuledConic
    {
        private double _a;
        private double _p;
        private int _linesPerInterval = 100;

        public static string canonicFormatString = "E[x-!m;!a]L[+]L[2" + U.Times + "!p" + U.Times + "(y-!n) = 0]";
        public static string parametricXFormatString = "L[x = !m + u]";
        public static string parametricYFormatString = "L[y = !n - ]P[u;!a;!p]";

        public static string _wordDescriptionFirstCase = "Parabola";
        public static string _wordDescriptionSecondCase = "Paraboly";
        public static string _wordDescriptionFourthCase = "Parabolu";

        public Parabola(ConicEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix,
            GColor color, string name, int ID)
            : base(commonEquation, parametricEquation, toCanonicTransformationMatrix,
            color, name,
            _wordDescriptionFirstCase, _wordDescriptionSecondCase, _wordDescriptionFourthCase,
            canonicFormatString, parametricXFormatString, parametricYFormatString, ID)
        {
            _a = parametricEquation.GetA();
            _p = parametricEquation.GetP();
        }


        /// x = m + v.
        public double ComputeXCoordinateAt(double v)
        {
            return v;
        }
        /// y = n - u2/(a2*2p) 
        public double ComputeYCoordinateAt(double v)
        {
            return -(v*v)/(_a*_a*2*_p);
        }
        /// <summary>
        /// Dosadi do parametricke rovnice v a vraci souradnice bodu.
        /// </summary>
        /// <param name="u"></param>
        /// <param name="v"></param>
        /// <returns></returns>
        public Vector ComputeCoordinatesAt(double v)
        {
            double x = ComputeXCoordinateAt(v);
            double y = ComputeYCoordinateAt(v);

            return new Vector(x, y, 0);
        }

        protected override void BuildParametricDisplayList(Scene scene, RelativeQuality relativeQuality)
        {
            List<List<Vector>> points = GetPoints(scene, relativeQuality); 

            _displayListIdentificator = Gl.glGenLists(1);
            Gl.glNewList(_displayListIdentificator, Gl.GL_COMPILE);
            foreach (List<Vector> subPoints in points)
                BuildAsLineStrip(subPoints);
            Gl.glEndList();
        }
        public override void BuildAsIntersection(Scene scene, RelativeQuality relativeQuality)
        {
            List<List<Vector>> points = GetPoints(scene, relativeQuality);

            _displayListIdentificator = Gl.glGenLists(1);
            Gl.glNewList(_displayListIdentificator, Gl.GL_COMPILE);
            foreach (List<Vector> subPoints in points)
                BuildAsCylinder(subPoints);
            Gl.glEndList();
        }

        private List<List<Vector>> GetPoints(Scene scene, RelativeQuality relativeQuality)
        {
            List<Interval> uIntervals = GetUIntervals(scene, relativeQuality);
            double step;
            double uMin;
            double uMax;
            List<List<Vector>> points = new List<List<Vector>>();
            int linesPerInterval = _linesPerInterval * (scene.GetBaseQuality() + relativeQuality);
            foreach (Interval uInterval in uIntervals)
            {
                if (uInterval.IsEmpty())
                    continue;
                step = uInterval.GetSize() / linesPerInterval;
                uMin = uInterval.GetA();
                uMax = uInterval.GetB();
                List<Vector> subPoints = new List<Vector>();
                for (double v = uMin; v < uMax + step; v += step)
                {
                    subPoints.Add(new Vector(ComputeXCoordinateAt(v), ComputeYCoordinateAt(v), 0));
                }
                points.Add(subPoints);
            }
            return points;
        }
        private List<Interval> GetUIntervals(Scene scene, RelativeQuality relativeQuality) 
        {
            double cleft = GetClippingLeft(scene);
            double cright = GetClippingRight(scene);
            double ctop = GetClippingTop(scene);
            double cbottom = GetClippingBottom(scene);
            double c = _a * _a * 2 * _p;

            Interval xInterval = new Interval(cleft, cright);
            List<Interval> uIntervals = new List<Interval>();

            QuadraticEquation topIntervalEquation = new QuadraticEquation(1, 0, ctop * c);
            QuadraticEquation bottomIntervalEquation = new QuadraticEquation(1, 0, cbottom * c);
            List<double> topIntervalSolutions = topIntervalEquation.Solve();
            List<double> bottomIntervalSolutions = bottomIntervalEquation.Solve();
            if (topIntervalSolutions.Count != 2 && bottomIntervalSolutions.Count != 2) // parabola je mimo ctverec
            {
            }
            if (bottomIntervalSolutions.Count == 2 && topIntervalSolutions.Count == 2) // parabola je uvnitr ale vrchol ne
            {
                Interval topInterval = new Interval(topIntervalSolutions[0], topIntervalSolutions[1]);
                Interval bottomInterval = new Interval(bottomIntervalSolutions[0], bottomIntervalSolutions[1]);
                List<Interval> yIntervals = Interval.MinusSmallerFromBigger(topInterval, bottomInterval);
                uIntervals = Interval.Intersect(yIntervals, xInterval);
            }
            if (bottomIntervalSolutions.Count == 2 && topIntervalSolutions.Count != 2) // parabola uvnitr (kladne _p)
            {
                Interval yInterval = new Interval(bottomIntervalSolutions[0], bottomIntervalSolutions[1]);
                uIntervals = new List<Interval>() { Interval.Intersect(yInterval, xInterval) };
            }
            if (topIntervalSolutions.Count == 2 && bottomIntervalSolutions.Count != 2) // parabola uvnitr (zaporne _p)
            {
                Interval yInterval = new Interval(topIntervalSolutions[0], topIntervalSolutions[1]);
                uIntervals = new List<Interval>() { Interval.Intersect(yInterval, xInterval) };
            }
            return uIntervals;
        }



        protected override ConicSection CallConstructor(ConicEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix, GColor color, string name, int ID)
        {
            return new Parabola(commonEquation, parametricEquation, toCanonicTransformationMatrix, color, name, ID);
        }


        public override ConicMatrix GetCanonicMatrix()
        {
            double A = 1 / (_parametricEquation.GetA() * _parametricEquation.GetA());
            double P = 2 * _parametricEquation.GetP();
            ConicEquation newCanonicEquation = new ConicEquation(A, 0, 0, 0, P, 0);
            return new ConicMatrix(newCanonicEquation);
        }
    }
}
