﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Geometry.GraphicObjects.ConicSections;
using Geometry.GraphicObjects.Equations;
using GMath;
using Geometry.GraphicObjects.Matrixes;
using Geometry.GraphicObjects.Planar;

namespace Geometry.GraphicObjects.Intersections
{
    public class ConicsSimpleRuledParametrizations
    {

        public static List<ParametrizationVector> GetParametrization(ConicSection conicSection)
        {
            if (conicSection is DoubleLine)
                return GetParametrization((DoubleLine)conicSection);
            if (conicSection is ParallelLines)
                return GetParametrization((ParallelLines)conicSection);
            if (conicSection is CrossingLines)
                return GetParametrization((CrossingLines)conicSection);
            if (conicSection is Parabola)
                return GetParametrization((Parabola)conicSection);
            if (conicSection is Hyperbola)
                return GetParametrization((Hyperbola)conicSection);
            return null;
        }

        /// <summary>
        /// (0, u, 0)
        /// </summary>
        private static List<ParametrizationVector> GetParametrization(DoubleLine line)
        {
            Expression xParametrization = new Expression(new UV(0, 0, 0));
            Expression yParametrization = new Expression(new UV(1, 1, 0));
            Expression zParametrization = new Expression(new UV(0, 0, 0));
            Expression oneParametrization = new Expression(new UV(1, 0, 0));
            return new List<ParametrizationVector>()
            {
                new ParametrizationVector(xParametrization, yParametrization, zParametrization, oneParametrization)
            };
        }
        /// <summary>
        /// (u, -u2/a2*2*p, 0)
        /// </summary>
        private static List<ParametrizationVector> GetParametrization(Parabola parabola)
        {
            ConicParametricEquation canonicEquation = (ConicParametricEquation)parabola.GetParametricEquation();
            double a = canonicEquation.GetA();
            double p = canonicEquation.GetP();
            Expression xParametrization = new Expression(new UV(1, 1, 0));
            Expression yParametrization = new Expression(new UV(1 / (a * a * 2 * -p), 2, 0));
            Expression zParametrization = new Expression(new UV(0, 0, 0));
            Expression oneParametrization = new Expression(new UV(1, 0, 0));
            return new List<ParametrizationVector>()
            {
                new ParametrizationVector(xParametrization, yParametrization, zParametrization, oneParametrization)
            };
        }
        /// <summary>
        /// x = _a/2(u + 1/u),
        /// y = _b/2(u + 1/u),
        /// z = 0.
        /// </summary>
        private static List<ParametrizationVector> GetParametrization(Hyperbola hyperbola)
        {
            ConicParametricEquation canonicEquation = (ConicParametricEquation)hyperbola.GetCanonicEquation();
            double a = canonicEquation.GetA() / 2.0;
            double b = canonicEquation.GetB() / 2.0;
            Expression xParametrization = new Expression(new UV(a, 1, 0), new UV(a, -1, 0));
            Expression yParametrization = new Expression(new UV(b, 1, 0), new UV(-b, -1, 0));
            Expression zParametrization = new Expression(new UV(0, 0, 0));
            Expression oneParametrization = new Expression(new UV(1, 0, 0));

            return new List<ParametrizationVector>()
            {
                new ParametrizationVector(xParametrization, yParametrization, zParametrization, oneParametrization)
            };
        }
        /// <summary>
        /// (+-a, u, 0)
        /// </summary>
        private static List<ParametrizationVector> GetParametrization(ParallelLines lines)
        {
            ConicParametricEquation canonicEquation = lines.GetCanonicEquation();
            double a = canonicEquation.GetA();
            Expression xParametrization1 = new Expression(new UV(a, 0, 0));
            Expression yParametrization1 = new Expression(new UV(1, 1, 0));
            Expression zParametrization1 = new Expression(new UV(0, 0, 0));
            Expression oneParametrization1 = new Expression(new UV(1, 0, 0));

            Expression xParametrization2 = new Expression(new UV(-a, 0, 0));
            Expression yParametrization2 = new Expression(new UV(1, 1, 0));
            Expression zParametrization2 = new Expression(new UV(0, 0, 0));
            Expression oneParametrization2 = new Expression(new UV(1, 0, 0));
            return new List<ParametrizationVector>()
            {
                new ParametrizationVector(xParametrization1, yParametrization1, zParametrization1, oneParametrization1),
                new ParametrizationVector(xParametrization2, yParametrization2, zParametrization2, oneParametrization2)
            };
        }
        /// <summary>
        /// (au, +-bu, 0)
        /// </summary>
        private static List<ParametrizationVector> GetParametrization(CrossingLines lines)
        {
            ConicParametricEquation canonicEquation = lines.GetCanonicEquation();
            double a = canonicEquation.GetA();
            double b = canonicEquation.GetB();
            Expression xParametrization1 = new Expression(new UV(a, 1, 0));
            Expression yParametrization1 = new Expression(new UV(b, 1, 0));
            Expression zParametrization1 = new Expression(new UV(0, 0, 0));
            Expression oneParametrization1 = new Expression(new UV(1, 0, 0));

            Expression xParametrization2 = new Expression(new UV(a, 1, 0));
            Expression yParametrization2 = new Expression(new UV(-b, 1, 0));
            Expression zParametrization2 = new Expression(new UV(0, 0, 0));
            Expression oneParametrization2 = new Expression(new UV(1, 0, 0));
            return new List<ParametrizationVector>()
            {
                new ParametrizationVector(xParametrization1, yParametrization1, zParametrization1, oneParametrization1),
                new ParametrizationVector(xParametrization2, yParametrization2, zParametrization2, oneParametrization2)
            };
        }

        public static DoubleLine GetDoubleLineFromParametrizationOfCompoundConicsection(ParametrizationVector pv, ConicSection conic) 
        {
            if (conic is CrossingLines)
                return GetDoubleLineFromParametrizationOfCompoundConicsection(pv, (CrossingLines)conic);
            if (conic is ParallelLines)
                return GetDoubleLineFromParametrizationOfCompoundConicsection(pv, (ParallelLines)conic);
            if (conic is DoubleLine)
                return (DoubleLine)conic;
            return null;
        }
        /// <summary>
        /// x =   au;
        /// y = +-bu;
        /// </summary>
        private static DoubleLine GetDoubleLineFromParametrizationOfCompoundConicsection(ParametrizationVector pv, CrossingLines conic) 
        {
            Expression x = (Expression)pv.GetXParametrization();
            Expression y = (Expression)pv.GetYParametrization();
            double a = x.GetValueAt(1, 1);
            double b = y.GetValueAt(1, 1);
            double angle = -Math.Atan2(a, b) * 180.0 / Math.PI;
            TransformationMatrix matrix = TransformationMatrix.GetZRotationMatrix(angle);
            ConicEquation commonEquation = new ConicEquation(new ConicMatrix(new ConicEquation(1, 0, 0, 0, 0, 0)).ApplyTransformation(matrix));
            return new DoubleLine(commonEquation, new ConicParametricEquation(1,0,0), 
                matrix.ComputeInverseMatrix(), conic.GetColor(), conic.GetName(), conic.GetID()); 

        }
        /// <summary>
        /// x = +-a;
        /// y =   u;
        /// </summary>
        private static DoubleLine GetDoubleLineFromParametrizationOfCompoundConicsection(ParametrizationVector pv, ParallelLines conic)
        {
            Expression x = (Expression)pv.GetXParametrization();
            Expression y = (Expression)pv.GetYParametrization();
            double a = x.GetValueAt(1, 1);
            TransformationMatrix matrix = new TransformationMatrix(4);
            ConicEquation commonEquation = new ConicEquation(new ConicMatrix(new ConicEquation(1, 0, 0, 0, 0, 0)));
            DoubleLine line =  new DoubleLine(commonEquation, new ConicParametricEquation(1,0,0),
                matrix.ComputeInverseMatrix(), conic.GetColor(), conic.GetName(), conic.GetID());
            line.SetCenterInStandardBasis(new CenterParametrization(new Vector(a, 0, 0)));
            return line;
        }
    }
}
