﻿using System.Collections.Generic;
using Geometry.GraphicObjects.ConicSections;
using Geometry.GraphicObjects.Planar;
using GMath;
using Geometry.GraphicObjects.Equations;
using Geometry.GraphicObjects.Intersections;
using System;

namespace Geometry.GraphicObjects.Intersections
{
    class ConicAndLineIntersectionRecognizer
    {
        private ConicSection _conicSection;
        private Line _line;
        private GColor _color;
        private string _name;
        private int _ID;



        public ConicAndLineIntersectionRecognizer(ConicSection conicSection, Line line, GColor color, string name, int ID) 
        {
            _conicSection = conicSection;
            _line = line;
            _color = color;
            _name = name;
            _ID = ID;
        }

        public ConicAndLineIntersection GetIntersection()
        {
            try
            {
                Plan conicPlan = _conicSection.GetConicPlan();

                PlanAndLineIntersection planAndLineIntersection =
                    new PlanAndLineIntersectionRecognizer(conicPlan, _line, _color, _name, _ID).GetIntersection();
                List<GraphicObject> intersections = new List<GraphicObject>();
                if (planAndLineIntersection.GetTheOnlyIntersection() is Line)
                    intersections = GetConicAndLineIntersectionOnLine();
                if (planAndLineIntersection.GetTheOnlyIntersection() is GPoint)
                    intersections = GetPointAsIntersection((GPoint)planAndLineIntersection.GetTheOnlyIntersection()); // prunikem je bod

                for (int i = intersections.Count - 1; i > -1; i--) // odstraním možné prázdné průniky
                {
                    if (intersections[i] is EmptyIntersection)
                        intersections.RemoveAt(i);
                }
                if (intersections.Count == 0) // pokud nemám zádný průnik, přidám prázdný průnik
                    intersections.Add(new EmptyIntersection());
                return new ConicAndLineIntersection(_conicSection, _line, intersections, _color, _name, _ID);
            }
            catch (Exception) 
            {
                return new ConicAndLineIntersection(_conicSection, _line, new List<GraphicObject>() {new EmptyIntersection() }, 
                    _color, _name, _ID);
            }
        }

        private List<GraphicObject> GetPointAsIntersection(GPoint point)
        {
            Vector pointInLine = 
                point.GetCenterInStandardBasis().GetReprezentant();
            CommonEquation equation = _conicSection.GetCommonEquation();
            double substitution = equation.Substitute(pointInLine);
            if (Program.IsEqualInToleranceIntersection(substitution, 0))
                return new List<GraphicObject>() { point };
            return new List<GraphicObject>();
        }

        private List<GraphicObject> GetConicAndLineIntersectionOnLine() 
        {
            Line lineInConicCanonicFrame = (Line)_line.MakeCopyOfObject(_ID);
            lineInConicCanonicFrame.ApplyTransformation(_conicSection.GetToCanonicTransformationMatrix());


            Vector direction = new Vector(
                lineInConicCanonicFrame.GetParametricEquation().GetUx(),
                lineInConicCanonicFrame.GetParametricEquation().GetUy(),
                lineInConicCanonicFrame.GetParametricEquation().GetUz());
            Vector center = lineInConicCanonicFrame.GetCenterInStandardBasis().GetReprezentant();

            CommonEquation conicEquation = new ConicEquation(_conicSection.GetCanonicMatrix());
            double a11 = conicEquation.GetA();
            double a22 = conicEquation.GetB();
            double a12 = conicEquation.GetD()/2.0;
            double a13 = conicEquation.GetG()/2.0;
            double a23 = conicEquation.GetH()/2.0;
            double a33 = conicEquation.GetL();
            double u = direction.GetX();
            double v = direction.GetY();
            double m = center.GetX();
            double n = center.GetY();
            QuadraticEquation equation;
            double A =
                a11 * u * u +
                a22 * v * v +
                2 * a12 * u * v;
            double B =
                u * (a11 * m + a12 * n + a13) +
                v * (a12 * m + a22 * n + a23);
            double C =
                a11 * m * m +
                a22 * n * n +
                2 * a12 * m * n +
                2 * a13 * m +
                2 * a23 * n + a33;

            equation = new QuadraticEquation(A, 2*B, C);
            List<double> solutions = equation.Solve();
            if (Program.IsEqualInToleranceIntersection(A, 0) &&
                Program.IsEqualInToleranceIntersection(B, 0) &&
                Program.IsEqualInToleranceIntersection(C, 0))
                return GetLineAsIntersection();
            return GetIntersectionPoints(lineInConicCanonicFrame, solutions);
        }

        private List<GraphicObject> GetLineAsIntersection() 
        {
            Line line = (Line)_line.MakeCopyOfObject(_ID);
            return new List<GraphicObject>() { line };
        }



        private List<GraphicObject> GetIntersectionPoints(Line lineInConicCanonicFrame, List<double> parameters)
        {
            if (parameters.Count == 0)
                return new List<GraphicObject>();
            List<GraphicObject> intersections = new List<GraphicObject>();
            Vector lCenter = lineInConicCanonicFrame.GetCenterInStandardBasis().GetReprezentant();
            LineParametricEquation lParametricEquation = 
                (LineParametricEquation)lineInConicCanonicFrame.GetParametricEquation();
            Vector jointPoint;
            if (parameters.Count == 1 || 
                (parameters.Count == 2 && (Program.IsEqualInTolerance(parameters[0], parameters[1]))))
            { // jeden prunik nebo tecna
                jointPoint = lCenter + lParametricEquation.Substitute(parameters[0]);
                jointPoint = jointPoint * _conicSection.GetTransformationMatrixAlgebraic().ComputeInverseMatrix();
                intersections.Add(new GPoint(jointPoint, _color, _name, _ID));
            }
            else
            {
                jointPoint = lCenter + lParametricEquation.Substitute(parameters[0]);
                jointPoint = jointPoint * _conicSection.GetTransformationMatrixAlgebraic().ComputeInverseMatrix();                
                intersections.Add(new GPoint(jointPoint, _color, _name, _ID));

                jointPoint = lCenter + lParametricEquation.Substitute(parameters[1]);
                jointPoint = jointPoint * _conicSection.GetTransformationMatrixAlgebraic().ComputeInverseMatrix();
                intersections.Add(new GPoint(jointPoint, _color, _name, _ID));
            }
            return intersections;
        }

        private Vector MakePermutationOnVariables(Vector point, int[] variablesOrder)
        {
            return new Vector(point.Get(variablesOrder[0]), point.Get(variablesOrder[1]), 0, 1);
        }
    }
}
