﻿using System;
using System.Collections.Generic;
using Geometry.GraphicObjects.Planar;
using Geometry.GraphicObjects.Equations;
using GMath;
 

namespace Geometry.GraphicObjects.Intersections
{
    /// <summary>
    /// http://maths.cz/clanky/analyticka-geometrie-vzajemna-poloha-dvou-primek-a-primky-s-rovinou.html
    /// </summary>
    public class LineAndLineIntersectionRecognizer
    {
        private Line _lineA;
        private Line _lineB;
        private GColor _color;
        private string _name;
        private int _ID;

        public LineAndLineIntersectionRecognizer(Line lineA, Line lineB, GColor color, string name, int ID)
        {
            _lineA = lineA;
            _lineB=lineB;
            _color = color;
            _name = name;
            _ID = ID;
        }

        /// <summary>
        /// Tri moznosti bud jsou primky totozne nebo se protinaji, nebo jsou mimobezne.
        /// </summary>
        /// <returns></returns>
        public LineAndLineIntersetion GetIntersection()
        {
            try
            {
                ParametricEquation lineAEquation = _lineA.GetParametricEquation();
                Vector lineACenter = _lineA.GetCenterInStandardBasis().GetReprezentant();
                Vector lineADirection = new Vector(lineAEquation.GetUx(), lineAEquation.GetUy(), lineAEquation.GetUz());

                ParametricEquation lineBEquation = _lineB.GetParametricEquation();
                Vector lineBCenter = _lineB.GetCenterInStandardBasis().GetReprezentant();
                Vector lineBDirection = new Vector(lineBEquation.GetUx(), lineBEquation.GetUy(), lineBEquation.GetUz());

                if (Vector.AreLinearyBounded(lineADirection, new List<Vector>() { lineBDirection }) &&
                    BelongsPointToLine(lineADirection, lineACenter, lineBCenter))
                {   // primky jsou totozne
                    Line copy = (Line)_lineA.MakeCopyOfObject(_ID);
                    return new LineAndLineIntersetion(_lineA, _lineB, copy, _color, _name, _ID);
                }
                Vector joitPont;
                if ((joitPont = GetJointPoint(lineADirection, lineACenter, lineBDirection, lineBCenter)) != null)
                {
                    GPoint intersection = new GPoint(joitPont, _color, _name, _ID);
                    return new LineAndLineIntersetion(_lineA, _lineB, intersection, _color, _name, _ID);
                }
                return new LineAndLineIntersetion(_lineA, _lineB, new EmptyIntersection(), _color, _name, _ID);
            }
            catch (Exception) 
            {
                // Toto by nemelo nikdy nastat, protože jedine prekpoklady jsou, že přímky nemají nulový směrový vektor, 
                // což je vždy pravda 
                return new LineAndLineIntersetion(_lineA, _lineB, new EmptyIntersection(), _color, _name, _ID);
            }
        }

        /// <summary>
        /// x = m + vt
        /// y = n + ut
        /// z = o + wt
        /// 
        /// Vyjadrim t z jedne z rovnic.
        /// Pak dosadim do vsech a porovnam x,y,z, se souradnicemi bodu.
        /// </summary>
        private bool BelongsPointToLine(Vector lineDirection, Vector lineCenter, Vector point) 
        {
            double tValue = 1;
            if (!Program.IsEqualInTolerance(lineDirection.GetX(), 0))
                tValue = (point.GetX() - lineCenter.GetX()) / lineDirection.GetX();
            else if (!Program.IsEqualInTolerance(lineDirection.GetY(), 0))
                tValue = (point.GetY() - lineCenter.GetY()) / lineDirection.GetY();
            else if (!Program.IsEqualInTolerance(lineDirection.GetZ(), 0))
                tValue = (point.GetZ() - lineCenter.GetZ()) / lineDirection.GetZ();
            else throw new Exception("Vsechny koeficienty smeroveho vektoru primky jsou nulove.");

            bool belongs = true;
            belongs = belongs && (Program.IsEqualInTolerance(point.GetX(), lineCenter.GetX() + tValue * lineDirection.GetX()));
            belongs = belongs && (Program.IsEqualInTolerance(point.GetY(), lineCenter.GetY() + tValue * lineDirection.GetY()));
            belongs = belongs && (Program.IsEqualInTolerance(point.GetZ(), lineCenter.GetZ() + tValue * lineDirection.GetZ()));
            return belongs;
        }

        /// <summary>
        /// m1 + v1t = m2 + v2s
        /// n1 + u1t = n2 + u2s
        /// o1 + w1t = o2 + w2s
        /// </summary>
        private Vector GetJointPoint(Vector lineADirection, Vector lineACenter, Vector lineBDirection, Vector lineBCenter) 
        {
            Equation a = new Equation(lineADirection.GetX(), -lineBDirection.GetX(), 0, lineACenter.GetX() - lineBCenter.GetX());
            Equation b = new Equation(lineADirection.GetY(), -lineBDirection.GetY(), 0, lineACenter.GetY() - lineBCenter.GetY());
            Equation c = new Equation(lineADirection.GetZ(), -lineBDirection.GetZ(), 0, lineACenter.GetZ() - lineBCenter.GetZ());
            CenterEquationSystem equation= new CenterEquationSystem(a, b, c);
            CenterParametrization center = equation.Solve();
            if (center == null)
                return null;
            Vector parameters = center.GetReprezentant();
            double t = parameters.GetX();
            return new Vector(
                lineACenter.GetX() + lineADirection.GetX() * t,
                lineACenter.GetY() + lineADirection.GetY() * t,
                lineACenter.GetZ() + lineADirection.GetZ() * t);
        }
    }
}
