﻿using System;
using Geometry.GraphicObjects.Planar;
using Geometry.GraphicObjects.Equations;
using GMath;
using System.Collections.Generic;

namespace Geometry.GraphicObjects.Intersections
{
    /// <summary>
    /// http://maths.cz/clanky/analyticka-geometrie-vzajemna-poloha-dvou-primek-a-primky-s-rovinou.html
    /// </summary>
    public class PlanAndLineIntersectionRecognizer
    {
        private Plan _plan;
        private Line _line;
        private GColor _color;
        private string _name;
        private int _ID;

        public PlanAndLineIntersectionRecognizer(Plan plan, Line line, GColor color, string name, int ID) 
        {
            _plan = plan;
            _line = line;
            _color = color;
            _name = name;
            _ID = ID;
        }

        public PlanAndLineIntersection GetIntersection() 
        {
            try
            {
                CommonEquation planCommonEquation = _plan.GetCommonEquation();
                Vector planNormalVector =
                    new Vector(planCommonEquation.GetA(), planCommonEquation.GetB(), planCommonEquation.GetC());
                ParametricEquation lineParametricEquation = _line.GetParametricEquation();
                Vector lineDirection =
                    new Vector(lineParametricEquation.GetUx(), lineParametricEquation.GetUy(), lineParametricEquation.GetUz());
                Vector lineCenter = _line.GetCenterInStandardBasis().GetReprezentant();
                double scalar = lineDirection * planNormalVector;
                double angle = (Vector.Angle(lineDirection, planNormalVector) * 180) / Math.PI;
                GraphicObject intersection;
                if (Program.IsEqualInToleranceIntersection(angle, 90))
                { // jsou bud rovnobezne, nebo primka lezi v rovine.
                    double substitution =
                        lineCenter.GetX() * planCommonEquation.GetA() +
                        lineCenter.GetY() * planCommonEquation.GetB() +
                        lineCenter.GetZ() * planCommonEquation.GetC() + planCommonEquation.GetD();
                    if (Program.IsEqualInToleranceLow(substitution, 0))
                    { // primka lezi v rovine
                        intersection = GetLineAsIntersection(lineDirection);
                    }
                    else
                    { // jsou rovnobezne
                        intersection = null;
                    }
                }
                else
                { // jsou ruznobezne
                    intersection = GetPointAsIntersection(planCommonEquation, lineDirection, lineCenter);
                }
                if (intersection == null)
                    intersection = new EmptyIntersection();
                return new PlanAndLineIntersection(_plan, _line, intersection, _color, _name, _ID);
            }
            catch (Exception) 
            {
                return new PlanAndLineIntersection(_plan, _line, new EmptyIntersection(), _color, _name, _ID);
            }
        }

        private GPoint GetPointAsIntersection(CommonEquation planCommonEquation, Vector lineDirection, Vector lineCenter)
        {
            double m = lineCenter.GetX();
            double n = lineCenter.GetY();
            double o = lineCenter.GetZ();

            double a = planCommonEquation.GetA();
            double b = planCommonEquation.GetB();
            double c = planCommonEquation.GetC();
            double d = planCommonEquation.GetD();

            double left = a * lineDirection.GetX() + b * lineDirection.GetY() + c * lineDirection.GetZ();
            double right = -(d + a * m + b * n + c * o);
            if (left == 0)
                throw new Exception("Delitel je nulovy");
            double tValue = right / left;
            Vector jointPoint = new Vector(
                m + tValue * lineDirection.GetX(),
                n + tValue * lineDirection.GetY(),
                o + tValue * lineDirection.GetZ());
            return new GPoint(jointPoint, _color, _name, _ID);
        }

        private Line GetLineAsIntersection(Vector lineDirection) 
        {
            return (Line)_line.MakeCopyOfObject(_ID);
        }
    }
}
