﻿using Geometry.GraphicObjects.Planar;
using Geometry.GraphicObjects.Equations;
using GMath;
using System; 

namespace Geometry.GraphicObjects.Intersections
{
    class PlanAndPlanIntersectionRecognizer
    {
        private Plan _planA;
        private Plan _planB;
        private GColor _color;
        private string _name;
        private int _ID;

        public PlanAndPlanIntersectionRecognizer(Plan planA, Plan planB, GColor color, string name, int ID) 
        {
            _planA = planA;
            _planB = planB;
            _color = color;
            _name = name;
            _ID = ID;
        }

        public PlanAndPlanIntersection GetIntersection() 
        {
            try
            {
                CommonEquation planAEquation = _planA.GetCommonEquation();
                CommonEquation planBEquation = _planB.GetCommonEquation();
                Vector planANormal = new Vector(planAEquation.GetA(), planAEquation.GetB(), planAEquation.GetC());
                Vector planBNormal = new Vector(planBEquation.GetA(), planBEquation.GetB(), planBEquation.GetC());

                GraphicObject intersection;
                if (Vector.AreTwoVectorsInIntersectionToleranceLinearyBounded(planANormal, planBNormal))
                {
                    intersection = GetPlanOrEmptySetAsIntersection();
                }
                else
                {
                    intersection = GetLineAsIntersection();
                }
                if (intersection == null)
                    intersection = new EmptyIntersection();
                return new PlanAndPlanIntersection(_planA, _planB, intersection, _color, _name, _ID);
            }
            catch (Exception) 
            {
                return new PlanAndPlanIntersection(_planA, _planB, new EmptyIntersection(), _color, _name, _ID);
            }
        }


        private GraphicObject GetPlanOrEmptySetAsIntersection() 
        {
            CommonEquation planAEquation = _planA.GetCommonEquation();
            Vector planBCenter = _planB.GetCenterInStandardBasis().GetReprezentant();
            double substitution =
                planAEquation.GetA() * planBCenter.GetX() +
                planAEquation.GetB() * planBCenter.GetY() +
                planAEquation.GetC() * planBCenter.GetZ() + planAEquation.GetD();
            if (!Program.IsEqualInToleranceIntersection(substitution, 0))
                return new EmptyIntersection();
            return GetPlanAsIntersection();
        }

        private Plan GetPlanAsIntersection() 
        {
            return (Plan)_planA.MakeCopyOfObject(_ID);
        }

        private Line GetLineAsIntersection()
        {
            PlanEquation planACopy = new PlanEquation((PlanEquation)_planA.GetCommonEquation());
            PlanEquation planBCopy = new PlanEquation((PlanEquation)_planB.GetCommonEquation());
            int[] variablesOrder = new int[] { 0, 1, 2 };
            double b, c, d;
            double tempDouble;
            int tempInt;
            if (Program.IsEqualInTolerance(planACopy.GetA(), 0)) // prohozeni v obou rovnicich, aby v prvni byl x nenulovy
            {
                if (!Program.IsEqualInTolerance(planACopy.GetB(), 0))
                {
                    tempDouble = planACopy.GetA();
                    planACopy.SetA(planACopy.GetB());
                    planACopy.SetB(tempDouble);
                    tempDouble = planBCopy.GetA();
                    planBCopy.SetA(planBCopy.GetB());
                    planBCopy.SetB(tempDouble);
                    variablesOrder[0] = 1; // prohozeno x a y.
                    variablesOrder[1] = 0;
                }
                else
                {
                    tempDouble = planACopy.GetA();
                    planACopy.SetA(planACopy.GetC());
                    planACopy.SetC(tempDouble);
                    tempDouble = planBCopy.GetA();
                    planBCopy.SetA(planBCopy.GetC());
                    planBCopy.SetC(tempDouble);
                    variablesOrder[0] = 2; // prohozeno x a z.
                    variablesOrder[2] = 0;
                }
            }
            b = planBCopy.GetB() - (planBCopy.GetA() * planACopy.GetB()) / planACopy.GetA();
            c = planBCopy.GetC() - (planBCopy.GetA() * planACopy.GetC()) / planACopy.GetA();
            d = planBCopy.GetD() - (planBCopy.GetA() * planACopy.GetD()) / planACopy.GetA();
            double substituent1 = 1; // do rovnice by+cz+d=0, za z dosadim.
            double substituent2 = -1; // do rovnice by+cz+d=0, za z dosadim.
            double y1, y2;
            if (Program.IsEqualInTolerance(b, 0))
            {
                tempDouble = planACopy.GetB();
                planACopy.SetB(planACopy.GetC());
                planACopy.SetC(tempDouble);
                tempDouble = planBCopy.GetB();
                planBCopy.SetB(planBCopy.GetC());
                planBCopy.SetC(tempDouble);
                tempDouble = b;
                b = c;
                c = tempDouble;
                tempInt = variablesOrder[1];
                variablesOrder[1] = variablesOrder[2]; // zamena y a z
                variablesOrder[2] = tempInt;
            }
            y1 = (-c * substituent1 - d) / b;
            y2 = (-c * substituent2 - d) / b;
            double x1 = -(planACopy.GetB() * y1 + planACopy.GetC() * substituent1 + planACopy.GetD()) / planACopy.GetA();
            double x2 = -(planACopy.GetB() * y2 + planACopy.GetC() * substituent2 + planACopy.GetD()) / planACopy.GetA();

            Vector point1 = new Vector(x1, y1, substituent1);
            Vector point2 = new Vector(x2, y2, substituent2);

            point1 = MakePermutationOnVariables(point1, variablesOrder);
            point2 = MakePermutationOnVariables(point2, variablesOrder);

            Vector u = point2 - point1;
            LineParametricEquation lineEquation = new LineParametricEquation(u.GetX(), u.GetY(), u.GetZ());
            return new Line(lineEquation, new CenterParametrization(point1), _color, _name, _ID);
        }

        private Vector MakePermutationOnVariables(Vector point, int[] variablesOrder)
        {
            Vector retval = new Vector();
            retval.Set(variablesOrder[0], point.Get(0));
            retval.Set(variablesOrder[1], point.Get(1));
            retval.Set(variablesOrder[2], point.Get(2));
            return retval;
        }
    }
}
