﻿using System.Collections.Generic;
using Geometry.GraphicObjects.Quadrics;
using Geometry.GraphicObjects.Planar;
using Geometry.GraphicObjects.Equations;
using GMath;
using System;

namespace Geometry.GraphicObjects.Intersections
{
    /// <summary>
    /// http://home.pf.jcu.cz/~kvadriky/kapitola1/Kap_1_3.pdf
    /// </summary>
    public class QuadricAndLineIntersectionRecognizer
    {
        private Quadric _quadric;
        private Line _line;
        private GColor _color;
        private string _name;
        private int _ID;

        public QuadricAndLineIntersectionRecognizer(Quadric quadric, Line line, GColor color, string name, int ID) 
        {
            _quadric = quadric;
            _line = line;
            _color = color;
            _name = name;
            _ID = ID;
        }

        /// <summary>
        /// http://home.pf.jcu.cz/~kvadriky/kapitola1/Kap_1_3.pdf
        /// </summary>
        /// <returns></returns>
        public QuadricAndLineIntersection GetIntersection()
        {
            try
            {
                List<GraphicObject> intersections = new List<GraphicObject>();

                QuadraticEquation equation = SubstituteToQuadricEquation();
                double A = equation.GetA();
                double B = equation.GetB();
                double C = equation.GetC();
                if (Program.IsEqualInTolerance(A, 0) && Program.IsEqualInTolerance(B, 0) && Program.IsEqualInTolerance(C, 0))
                    intersections = GetLineAsIntersection();
                else
                {
                    List<double> solutions = equation.Solve();
                    if (solutions.Count == 2)
                        intersections = GetTwoPointsOrDoublePointAsIntersection(solutions);
                    if (solutions.Count == 1)
                        intersections = GetOnePointAsIntersection(solutions);
                }

                for (int i = intersections.Count - 1; i > -1; i--) // odstraním možné prázdné průniky
                {
                    if (intersections[i] is EmptyIntersection)
                        intersections.RemoveAt(i);
                }
                //intersections.Add(graphicObject);
                if (intersections.Count == 0) // pokud nemám zádný průnik, přidám prázdný průnik
                    intersections.Add(new EmptyIntersection());
                return new QuadricAndLineIntersection(_quadric, _line, intersections, _color, _name, _ID);
            }
            catch (Exception) 
            {
                return new QuadricAndLineIntersection(_quadric, _line, new List<GraphicObject>() { new EmptyIntersection() },
                    _color, _name, _ID);
            }
        }

        private QuadraticEquation SubstituteToQuadricEquation() 
        {
            CommonEquation quadricEquation = _quadric.GetCommonEquation();

            double a11 = quadricEquation.GetA();
            double a22 = quadricEquation.GetB();
            double a33 = quadricEquation.GetC();
            double a12 = quadricEquation.GetD() / 2.0;
            double a13 = quadricEquation.GetE() / 2.0;
            double a23 = quadricEquation.GetF() / 2.0;
            double a14 = quadricEquation.GetG() / 2.0;
            double a24 = quadricEquation.GetH() / 2.0;
            double a34 = quadricEquation.GetK() / 2.0;
            double a44 = quadricEquation.GetL();

            ParametricEquation lineParametricEquation = _line.GetParametricEquation();
            double u = lineParametricEquation.GetUx();
            double v = lineParametricEquation.GetUy();
            double w = lineParametricEquation.GetUz();
            Vector lineCenter = _line.GetCenterInStandardBasis().GetReprezentant();
            double m = lineCenter.GetX();
            double n = lineCenter.GetY();
            double p = lineCenter.GetZ();

            double A =
                a11 * u * u +
                a22 * v * v +
                a33 * w * w +
                2 * a12 * u * v +
                2 * a13 * u * w +
                2 * a23 * v * w;
            double B =
                u * (a11 * m + a12 * n + a13 * p + a14) +
                v * (a12 * m + a22 * n + a23 * p + a24) +
                w * (a13 * m + a23 * n + a33 * p + a34);
            double C =
                a11 * m * m +
                a22 * n * n +
                a33 * p * p +
                2 * a12 * m * n +
                2 * a13 * m * p +
                2 * a23 * n * p +
                2 * a14 * m +
                2 * a24 * n +
                2 * a34 * p + a44;
            return new QuadraticEquation(A, 2.0 * B, C);
        }


        private List<GraphicObject> GetLineAsIntersection() 
        {
            Line lineCopy = (Line)_line.MakeCopyOfObject(_ID);
            return new List<GraphicObject>() { lineCopy };
        }

        private List<GraphicObject> GetTwoPointsOrDoublePointAsIntersection(List<double> t) 
        {
            ParametricEquation equation = _line.GetParametricEquation();
            Vector lineCenter = _line.GetCenterInStandardBasis().GetReprezentant();
            List<GraphicObject> intersections = new List<GraphicObject>();
            double x;
            double y;
            double z;
            if (Program.IsEqualInTolerance(t[0], t[1])) // primka je tecnou.
            {
                x = lineCenter.GetX() + equation.GetUx() * t[0];
                y = lineCenter.GetY() + equation.GetUy() * t[0];
                z = lineCenter.GetZ() + equation.GetUz() * t[0];
                GPoint point = new GPoint(new Vector(x, y, z), _color, _name, _ID);
                intersections.Add(point);
            }
            else 
            {
                x = lineCenter.GetX() + equation.GetUx() * t[0];
                y = lineCenter.GetY() + equation.GetUy() * t[0];
                z = lineCenter.GetZ() + equation.GetUz() * t[0];
                GPoint point1 = new GPoint(new Vector(x, y, z), _color, _name, _ID);
                intersections.Add(point1);

                x = lineCenter.GetX() + equation.GetUx() * t[1];
                y = lineCenter.GetY() + equation.GetUy() * t[1];
                z = lineCenter.GetZ() + equation.GetUz() * t[1];
                GPoint point2 = new GPoint(new Vector(x, y, z), _color, _name, _ID);
                intersections.Add(point2);
            }
            return intersections;
        }

        private List<GraphicObject> GetOnePointAsIntersection(List<double> t)
        {
            ParametricEquation equation = _line.GetParametricEquation();
            Vector lineCenter = _line.GetCenterInStandardBasis().GetReprezentant();
            List<GraphicObject> intersections = new List<GraphicObject>();
            double x;
            double y;
            double z;
            x = lineCenter.GetX() + equation.GetUx() * t[0];
            y = lineCenter.GetY() + equation.GetUy() * t[0];
            z = lineCenter.GetZ() + equation.GetUz() * t[0];
            GPoint point = new GPoint(new Vector(x, y, z), _color, _name, _ID);
            intersections.Add(point);
            return intersections;
        }
    }
}
