﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Geometry.GraphicObjects.Quadrics;
using GMath;

namespace Geometry.GraphicObjects.Intersections
{
    public class ConstantUClipManager:IClipManager
    {
        private ParametrizationVector _parametrizationVector;

        public ConstantUClipManager(CurveAsQuadricsIntersection curve, Scene scene, Interval interval, List<double> minusValues)
            :base(curve, scene, 1, interval, minusValues) 
        {
            this._parametrizationVector = curve.GetParametrizationVectorInCanonicBasis();
        }

        /// <summary>
        /// x = 0
        /// y = 0
        /// z = c
        /// </summary>
        private List<Interval> GetLinearClipIntervals(ImaginarCrossingPlans imaginarCrossingPlans,
            Interval x, Interval y, Interval z)
        {
            return new List<Interval>() { new Interval(z.GetA(), z.GetB()) };
        }
        /// <summary>
        /// x = 0
        /// y = c
        /// z = v
        /// </summary>
        private List<Interval> GetLinearClipIntervals(DoublePlan doublePlan,
            Interval x, Interval y, Interval z)
        {
            Interval xClip = new Interval(double.MinValue, double.MaxValue);
            Interval yClip = new Interval(double.MinValue, double.MaxValue);
            Interval zClip = new Interval(z.GetA(), z.GetB());
            return new List<Interval>() { Interval.Intersect(zClip, Interval.Intersect(xClip, yClip)) };
        }
        /// <summary>
        /// x = -a
        /// y = c
        /// z = v
        /// </summary>
        private List<Interval> GetLinearClipIntervals(ParallelPlans parallelPlans,
            Interval x, Interval y, Interval z)
        {
            Interval xClip = new Interval(double.MinValue, double.MaxValue);
            Interval yClip = new Interval(double.MinValue, double.MaxValue);
            Interval zClip = new Interval(z.GetA(), z.GetB());
            return new List<Interval>() { Interval.Intersect(zClip, Interval.Intersect(xClip, yClip)) };
        }
        /// <summary>
        /// x = ac
        /// y = +-bc
        /// z = v
        /// </summary>
        private List<Interval> GetLinearClipIntervals(CrossingPlans crossingPlans,
            Interval x, Interval y, Interval z)
        {
            Interval xClip = new Interval(double.MinValue, double.MaxValue);
            Interval yClip = new Interval(double.MinValue, double.MaxValue);
            Interval zClip = new Interval(z.GetA(), z.GetB());
            return new List<Interval>() { Interval.Intersect(zClip, Interval.Intersect(xClip, yClip)) };
        }
        /// <summary>
        /// x = c
        /// y = -(c*c)/(constantA)
        /// z = v
        /// </summary>
        private List<Interval> GetLinearClipIntervals(ParabolicCylinder parabolicCylinder,
            Interval x, Interval y, Interval z)
        {
            Interval xClip = new Interval(double.MinValue, double.MaxValue);
            Interval yClip = new Interval(double.MinValue, double.MaxValue);
            Interval zClip = new Interval(z.GetA(), z.GetB());
            return new List<Interval>() { Interval.Intersect(zClip, Interval.Intersect(xClip, yClip)) };
        }
        /// <summary>
        /// x = _a/2*(u + 1/u)
        /// y = _b/2*(u - 1/u)
        /// z = v
        /// </summary>
        private List<Interval> GetLinearClipIntervals(HyperbolicCylinder hyperbolicCylinder,
            Interval x, Interval y, Interval z)
        {
            Interval xClip = new Interval(double.MinValue, double.MaxValue);
            Interval yClip = new Interval(double.MinValue, double.MaxValue);
            Interval zClip = new Interval(z.GetA(), z.GetB());
            return new List<Interval>() { Interval.Intersect(zClip, Interval.Intersect(xClip, yClip)) };
        }
        /// <summary>
        /// x = a(c + v)
        /// y = b(c - v)
        /// z = -pcv
        /// </summary>
        private List<Interval> GetLinearClipIntervals(HyperbolicParaboloid hyperbolicParaboloid,
            Interval x, Interval y, Interval z)
        {
            LinearEquation xEquation = GetVLinearEquation((Expression)_parametrizationVector.GetXParametrization());
            LinearEquation yEquation = GetVLinearEquation((Expression)_parametrizationVector.GetYParametrization());
            LinearEquation zEquation = GetVLinearEquation((Expression)_parametrizationVector.GetZParametrization());

            Interval xClip = new Interval((x.GetA() - xEquation.GetB()) / xEquation.GetA(), (x.GetB() - xEquation.GetB()) / xEquation.GetA());
            Interval yClip = new Interval((y.GetA() - yEquation.GetB()) / yEquation.GetA(), (y.GetB() - yEquation.GetB()) / yEquation.GetA());
            Interval zClip = new Interval((z.GetA() - zEquation.GetB()) / zEquation.GetA(), (z.GetB() - zEquation.GetB()) / zEquation.GetA());
            
            return new List<Interval>() { Interval.Intersect(zClip, Interval.Intersect(xClip, yClip)) };
        }

        public List<Interval> GetParameterIntervals()
        {
            return GetClipIntervals(_scene, _simpleRuled);
        }

        private List<Interval> GetClipIntervals(Scene scene, SimpleRulledQuadric simpleRuled)
        {
            Interval x = new Interval(simpleRuled.GetClippingLeft(scene), simpleRuled.GetClippingRight(scene));
            Interval y = new Interval(simpleRuled.GetClippingBottom(scene), simpleRuled.GetClippingTop(scene));
            Interval z = new Interval(simpleRuled.GetClippingBack(scene), simpleRuled.GetClippingFront(scene));

            if (simpleRuled is HyperbolicParaboloid)
                return GetLinearClipIntervals((HyperbolicParaboloid)simpleRuled, x, y, z);
            if (simpleRuled is HyperbolicCylinder)
                return GetLinearClipIntervals((HyperbolicCylinder)simpleRuled, x, y, z);
            if (simpleRuled is ParabolicCylinder)
                return GetLinearClipIntervals((ParabolicCylinder)simpleRuled, x, y, z);
            if (simpleRuled is CrossingPlans)
                return GetLinearClipIntervals((CrossingPlans)simpleRuled, x, y, z);
            if (simpleRuled is ParallelPlans)
                return GetLinearClipIntervals((ParallelPlans)simpleRuled, x, y, z);
            if (simpleRuled is DoublePlan)
                return GetLinearClipIntervals((DoublePlan)simpleRuled, x, y, z);
            if (simpleRuled is ImaginarCrossingPlans)
                return GetLinearClipIntervals((ImaginarCrossingPlans)simpleRuled, x, y, z);
            throw new Exception();
        }

        private LinearEquation GetVLinearEquation(Expression equation)
        {
            double A = 0;
            double B = 0;
            foreach (UV uv in equation.GetList())
            {
                if (uv.GetVexponent() == 1)
                    A += uv.GetValue();
                else
                    if (uv.GetVexponent() == 0)
                        B += uv.GetValue();
                    else throw new Exception();
            }
            return new LinearEquation(A, B);
        }
    }
}
