﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Geometry.GraphicObjects.Quadrics;
using GMath;

namespace Geometry.GraphicObjects.Intersections
{
    public class ConstantVClipManager : IClipManager
    {
        private ParametrizationVector _parametrizationVector;

        public ConstantVClipManager(CurveAsQuadricsIntersection curve, Scene scene, Interval interval, List<double> minusValues)
            : base(curve, scene, 1, interval, minusValues)
        {
            this._parametrizationVector = curve.GetParametrizationVectorInCanonicBasis();
        }

        /// <summary>
        /// x = 0
        /// y = 0
        /// z = c
        /// </summary>
        private List<Interval> GetLinearClipIntervals(ImaginarCrossingPlans imaginarCrossingPlans,
            Interval x, Interval y, Interval z)
        {
            return new List<Interval>() { new Interval(double.MinValue, double.MaxValue) };
        }
        /// <summary>
        /// x = 0
        /// y = u
        /// z = c
        /// </summary>
        private List<Interval> GetLinearClipIntervals(DoublePlan doublePlan,
            Interval x, Interval y, Interval z)
        {
            Interval xClip = new Interval(double.MinValue, double.MaxValue);
            Interval yClip = new Interval(y.GetA(), y.GetB());
            Interval zClip = new Interval(double.MinValue, double.MaxValue);            
            return new List<Interval>() { Interval.Intersect(zClip, Interval.Intersect(xClip, yClip)) };
        }
        /// <summary>
        /// x = -a
        /// y = u
        /// z = c
        /// </summary>
        private List<Interval> GetLinearClipIntervals(ParallelPlans parallelPlans,
            Interval x, Interval y, Interval z)
        {
            Interval xClip = new Interval(double.MinValue, double.MaxValue);
            Interval yClip = new Interval(y.GetA(), y.GetB());
            Interval zClip = new Interval(double.MinValue, double.MaxValue);
            return new List<Interval>() { Interval.Intersect(zClip, Interval.Intersect(xClip, yClip)) };
        }
        /// <summary>
        /// x = au
        /// y = +-bu
        /// z = c
        /// </summary>
        private List<Interval> GetLinearClipIntervals(CrossingPlans crossingPlans,
            Interval x, Interval y, Interval z)
        {
            double a = crossingPlans.GetParametricEquation().GetA();
            double b = crossingPlans.GetParametricEquation().GetB();
            Interval xClip = new Interval(x.GetA() / a, x.GetB() / a);
            Interval y1Clip = new Interval(y.GetA() / b, x.GetB() / b);
            Interval y2Clip = new Interval(y.GetA() / -b, x.GetB() / -b);
            Interval zClip = new Interval(double.MinValue, double.MaxValue);
            return new List<Interval>() 
            {   
                Interval.Intersect(zClip, Interval.Intersect(xClip, y1Clip)),
                Interval.Intersect(zClip, Interval.Intersect(xClip, y2Clip))
            };
        }
        /// <summary>
        /// x = u
        /// y = -(u*u)/(constantA)
        /// z = c
        /// </summary>
        private List<Interval> GetLinearClipIntervals(ParabolicCylinder parabolicCylinder,
            Interval x, Interval y, Interval z)
        {
            double constantA = QuadricsSimpleRuledParametrizations.GetConstantA(parabolicCylinder);
            Interval xClip = new Interval(x.GetA(), x.GetB());
            Interval yClip = new Interval(Math.Sqrt(y.GetA() * constantA), Math.Sqrt(y.GetB() * constantA));
            Interval zClip = new Interval(double.MinValue, double.MaxValue);
            return new List<Interval>() { Interval.Intersect(zClip, Interval.Intersect(xClip, yClip)) };
        }
        /// <summary>
        /// x = _a/2*(u + 1/u)
        /// y = _b/2*(u - 1/u)
        /// z = c
        /// 
        /// au2 - xu + a = 0
        /// bu2 - yu - b = 0
        /// </summary>
        private List<Interval> GetLinearClipIntervals(HyperbolicCylinder hyperbolicCylinder,
            Interval x, Interval y, Interval z)
        {
            double a = hyperbolicCylinder.GetParametricEquation().GetA() / 2.0;
            double b = hyperbolicCylinder.GetParametricEquation().GetB() / 2.0;

            QuadraticEquation xEquationLeft = new QuadraticEquation(a, -x.GetA(), a);
            QuadraticEquation xEquationRight = new QuadraticEquation(a, -x.GetB(), a);
            List<double> xSolutionsLeft = xEquationLeft.Solve();
            List<double> xSolutionsRight = xEquationRight.Solve();
            QuadraticEquation yEquationBottom = new QuadraticEquation(b, -x.GetA(), -b);
            QuadraticEquation yEquationTop = new QuadraticEquation(b, -x.GetB(), -b);
            List<double> ySolutionsBottom = yEquationBottom.Solve();
            List<double> ySolutionsTop = yEquationTop.Solve();
            List<double> xySolutions = Merge(Merge(Merge(xSolutionsLeft, xSolutionsRight), ySolutionsBottom), ySolutionsTop);
            xySolutions.Sort();
            List<Interval> xyIntervals = new List<Interval>();
            for (int i = 0; i < xySolutions.Count - 1; i++)
            {
                double u = (xySolutions[i] + xySolutions[i + 1]) / 2.0;
                if (!Program.IsEqualInToleranceHard(u, 0) && x.Belongs(a * (u + 1 / u)) && y.Belongs(b * (u - 1 / u)))
                    xyIntervals.Add(new Interval(xySolutions[i], xySolutions[i + 1]));
            }
            Interval zClip = new Interval(double.NegativeInfinity, double.NegativeInfinity);
            return Interval.Intersect(xyIntervals, zClip);
        }
        /// <summary>
        /// x = a(u + v)
        /// y = b(u - v)
        /// z = -puc
        /// </summary>
        private List<Interval> GetLinearClipIntervals(HyperbolicParaboloid hyperbolicParaboloid,
            Interval x, Interval y, Interval z)
        {
            LinearEquation xEquation = GetULinearEquation((Expression)_parametrizationVector.GetXParametrization());
            LinearEquation yEquation = GetULinearEquation((Expression)_parametrizationVector.GetYParametrization());
            LinearEquation zEquation = GetULinearEquation((Expression)_parametrizationVector.GetZParametrization());

            Interval xClip = new Interval((x.GetA() - xEquation.GetB()) / xEquation.GetA(), (x.GetB() - xEquation.GetB()) / xEquation.GetA());
            Interval yClip = new Interval((y.GetA() - yEquation.GetB()) / yEquation.GetA(), (y.GetB() - yEquation.GetB()) / yEquation.GetA());
            Interval zClip = new Interval((z.GetA() - zEquation.GetB()) / zEquation.GetA(), (z.GetB() - zEquation.GetB()) / zEquation.GetA());

            return new List<Interval>() { Interval.Intersect(zClip, Interval.Intersect(xClip, yClip)) };
        }

        public List<Interval> GetParameterIntervals()
        {
            return GetClipIntervals(_scene, _simpleRuled);
        }

        private List<Interval> GetClipIntervals(Scene scene, SimpleRulledQuadric simpleRuled)
        {
            Interval x = new Interval(simpleRuled.GetClippingLeft(scene), simpleRuled.GetClippingRight(scene));
            Interval y = new Interval(simpleRuled.GetClippingBottom(scene), simpleRuled.GetClippingTop(scene));
            Interval z = new Interval(simpleRuled.GetClippingBack(scene), simpleRuled.GetClippingFront(scene));

            if (simpleRuled is HyperbolicParaboloid)
                return GetLinearClipIntervals((HyperbolicParaboloid)simpleRuled, x, y, z);
            if (simpleRuled is HyperbolicCylinder)
                return GetLinearClipIntervals((HyperbolicCylinder)simpleRuled, x, y, z);
            if (simpleRuled is ParabolicCylinder)
                return GetLinearClipIntervals((ParabolicCylinder)simpleRuled, x, y, z);
            if (simpleRuled is CrossingPlans)
                return GetLinearClipIntervals((CrossingPlans)simpleRuled, x, y, z);
            if (simpleRuled is ParallelPlans)
                return GetLinearClipIntervals((ParallelPlans)simpleRuled, x, y, z);
            if (simpleRuled is DoublePlan)
                return GetLinearClipIntervals((DoublePlan)simpleRuled, x, y, z);
            if (simpleRuled is ImaginarCrossingPlans)
                return GetLinearClipIntervals((ImaginarCrossingPlans)simpleRuled, x, y, z);
            throw new Exception();
        }

        private LinearEquation GetVLinearEquation(Expression equation)
        {
            double A = 0;
            double B = 0;
            foreach (UV uv in equation.GetList())
            {
                if (uv.GetVexponent() == 1)
                    A += uv.GetValue();
                else
                    if (uv.GetVexponent() == 0)
                        B += uv.GetValue();
                    else throw new Exception();
            }
            return new LinearEquation(A, B);
        }
        private LinearEquation GetULinearEquation(Expression equation)
        {
            double A = 0;
            double B = 0;
            foreach (UV uv in equation.GetList())
            {
                if (uv.GetUexponent() == 1)
                    A += uv.GetValue();
                else
                    if (uv.GetUexponent() == 0)
                        B += uv.GetValue();
                    else throw new Exception();
            }
            return new LinearEquation(A, B);
        }
    }
}
