﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using GMath;
using Geometry.GraphicObjects.Intersections;
using Geometry.GraphicObjects.Quadrics; 

namespace Geometry.GraphicObjects.Intersections
{
    public class IClipManager
    {

        protected IParameterVParametrization _vParametrization;
        protected IParameterVParametrization _vParametrizationZerosEliminated;
        protected SimpleRulledQuadric _simpleRuled;
        protected Scene _scene;
        protected int _sign;
        protected CurveAsQuadricsIntersection _curve;
        private Interval _partOfDefinitionInterval;
        protected List<double> _minusValues;

        public IClipManager(CurveAsQuadricsIntersection curve, Scene scene, int sign, Interval interval, List<double> minusValues)
        {
            _curve = curve;
            _vParametrization = curve.GetVParametrization();
            _vParametrizationZerosEliminated = _vParametrization.EliminateZeros();
            _simpleRuled = curve.GetSimpleRuledQuadric();
            _scene = scene;
            _sign = sign;
            _partOfDefinitionInterval = interval;
            _minusValues = minusValues;
        }


        protected List<Interval> GetClipInterval(Func<double, IParameterVParametrization, double> function, Arguments arguments)
        {
            List<Interval> retval = new List<Interval>();
            double temp, temp2, middleValue;
            SuperEquation firstBorderEquation = arguments.firstSideEquation;
            SuperEquation secondBorderEquation = arguments.secondSideEquation;
            List<double> firstBorderSolutions = firstBorderEquation.SolveGetAllSetsOfRootsReal();
            List<double> secondBorderSolutions = secondBorderEquation.SolveGetAllSetsOfRootsReal();
            List<double> list = GetCopy(Merge(firstBorderSolutions, secondBorderSolutions));
            list.Add(_partOfDefinitionInterval.GetA());
            list.Add(_partOfDefinitionInterval.GetB());
            list = Sort(list);
            for (int i = 0; i < list.Count - 1; i++)
            {
                middleValue = GetMiddleValueFromIntervalAndDefinitionInterval(new Interval(list[i], list[i + 1]));
                temp = function(middleValue, _vParametrization);
                temp2 = function(middleValue, _vParametrizationZerosEliminated);
                if (arguments.interval.Belongs(temp) || arguments.interval.Belongs(temp2))
                    retval = Interval.Union(retval, new Interval(list[i], list[i + 1]));
            }
            return retval;
        }

        /// <summary>
        /// Funkce substituující do složky jejíž parametrizace je 
        /// 
        /// x = a*u + a*v,
        /// 
        /// u je parametr, v je vyjádření parametru v jako řesení kvadratické rovnice, obsahující u.
        /// </summary>
        protected double SubstitutionAFunction(double u, IParameterVParametrization vParametrization)
        {
            double v;
            if (vParametrization is QuadraticV)
                v = ((QuadraticV)vParametrization).GetVvalueConvertDiskriminantNegativity(u, _sign);
            else v = vParametrization.GetVvalue(u, _sign);
            double constantA = QuadricsSimpleRuledParametrizations.GetConstantA((HyperbolicParaboloid)_simpleRuled);
            return constantA * u + constantA * v;
        }
        /// <summary>
        /// Funkce substituující do složky jejíž parametrizace je 
        /// 
        /// y = b*u - b*v,
        /// 
        /// u je parametr, v je vyjádření parametru v jako řesení kvadratické rovnice, obsahující u.
        /// </summary>
        protected double SubstitutionBFunction(double u, IParameterVParametrization vParametrization)
        {
            double v;
            if (vParametrization is QuadraticV)
                v = ((QuadraticV)vParametrization).GetVvalueConvertDiskriminantNegativity(u, _sign);
            else v = vParametrization.GetVvalue(u, _sign);
            double constantB = QuadricsSimpleRuledParametrizations.GetConstantB((HyperbolicParaboloid)_simpleRuled); 
            return constantB * u - constantB * v;
        }
        /// <summary>
        /// Funkce substituující do složky jejíž parametrizace je 
        /// 
        /// z = -sign(p)*u*v,
        /// 
        /// u je parametr, v je vyjádření parametru v jako řesení kvadratické rovnice, obsahující u.
        /// </summary>
        protected double SubstitutionPFunction(double u, IParameterVParametrization vParametrization)
        {
            double v;
            if (vParametrization is QuadraticV)
                v = ((QuadraticV)vParametrization).GetVvalueConvertDiskriminantNegativity(u, _sign);
            else v = vParametrization.GetVvalue(u, _sign);
            double p = ((HyperbolicParaboloid)_simpleRuled).GetParametricEquation().GetP();
            double v2 = -Math.Sign(p) * u * v;
            return v2;
        }
        /// <summary>
        /// Funkce substituující do složky jejíž parametrizace je 
        /// 
        /// y = -(u*u)/(a*a*p),
        /// 
        /// u je parametr
        /// </summary>
        protected double SubstitutionPCFunction(double u, IParameterVParametrization vParametrization)
        {
            double constantA = QuadricsSimpleRuledParametrizations.GetConstantA((ParabolicCylinder)_simpleRuled);
            return -(u * u) / constantA;
        }
        /// <summary>
        /// Funkce substituující do složky jejíž parametrizace je 
        /// 
        /// x = v,
        /// 
        /// v je vyjádření parametru v jako řesení kvadratické rovnice, obsahující u.
        /// </summary>
        protected double SubstitutionVFunction(double u, IParameterVParametrization vParametrization)
        {
            if (vParametrization is QuadraticV)
                return ((QuadraticV)vParametrization).GetVvalueConvertDiskriminantNegativity(u, _sign);
            return vParametrization.GetVvalue(u, _sign);
        }


        protected List<double> Merge(List<double> a, List<double> b)
        {
            List<double> retval = new List<double>();
            foreach (double d in a)
                retval.Add(d);
            foreach (double d in b)
                retval.Add(d);
            return retval;
        }
        protected List<double> Merge(List<List<double>> firstBorderSolutions, List<List<double>> secondBorderSolutions)
        {
            List<double> retval = new List<double>();
            for (int i = 0; i < firstBorderSolutions.Count; i++)
                for (int j = 0; j < firstBorderSolutions[i].Count; j++)
                    retval.Add(firstBorderSolutions[i][j]);
            for (int i = 0; i < secondBorderSolutions.Count; i++)
                for (int j = 0; j < secondBorderSolutions[i].Count; j++)
                    retval.Add(secondBorderSolutions[i][j]);
            return retval;
        } 
        protected List<double> Sort(List<double> list)
        {
            list.Sort();
            return list;
        }
        protected List<double> GetCopy(List<double> src)
        {
            List<double> retval = new List<double>();
            foreach (double d in src)
                retval.Add(d);
            return retval;
        }

        protected double GetMiddleValueFromIntervalAndDefinitionInterval(Interval interval) 
        {
            double a = interval.GetA();
            double b = interval.GetB();
            List<double> values = new List<double>(){a, b};
            // interval rozdělím minus hodnotami
            // Pokud nyní vezmu libovolný vnitřní bod, pak to bude z intervalu
            // vyjimka pokud a == b a a je v minus, pak vrácená hodnota není z definičního oboru.
            foreach(double d in _minusValues)
                if (interval.Belongs(d) && d != a && d != b)
                {
                    values.Add(d);
                }
            values.Sort();
            // žádná minus hodnota, nebyla v intervalu.
            if (values.Count == 2)
            {
                if (a == double.NegativeInfinity && b == double.PositiveInfinity)
                    return 0;
                if (a == double.NegativeInfinity)
                    return b - 10;
                if (b == double.PositiveInfinity)
                    return a + 10;
                return (a + b) / 2.0;
            }
            else // alespoň jedna minus hodnota byla v intervalu
            {
                if (a == double.NegativeInfinity) // vrátím druhý zmenšený o deset. (-infinity x y b).
                    return values[1] - 10;
                if (b == double.PositiveInfinity)
                    return values[values.Count - 2] + 10; // vrátím předposlední zvetseny o deset. (a y z infinity)
                return (values[0] + values[1]) / 2.0;
            }
        }
    }
}
