﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using GMath;
using Geometry.GraphicObjects.Quadrics;
using Geometry.GraphicObjects.Intersections;

namespace Geometry.GraphicObjects.Intersections
{
    public class LinearClipManager:IClipManager
    {
        public LinearClipManager(CurveAsQuadricsIntersection curve, Scene scene, Interval interval, List<double> minusValues)
            :base(curve, scene, 1, interval, minusValues) 
        {
        }

        public List<Interval> GetParameterIntervals()
        {
            return GetClipIntervals(_scene, _simpleRuled);
        }

        private List<Interval> GetClipIntervals(Scene scene, SimpleRulledQuadric simpleRuled)
        {
            Interval x = new Interval(simpleRuled.GetClippingLeft(scene), simpleRuled.GetClippingRight(scene));
            Interval y = new Interval(simpleRuled.GetClippingBottom(scene), simpleRuled.GetClippingTop(scene));
            Interval z = new Interval(simpleRuled.GetClippingBack(scene), simpleRuled.GetClippingFront(scene));

            if (simpleRuled is HyperbolicParaboloid)
                return GetLinearClipIntervals((HyperbolicParaboloid)simpleRuled, x, y, z);
            if (simpleRuled is HyperbolicCylinder)
                return GetLinearClipIntervals((HyperbolicCylinder)simpleRuled, x, y, z);
            if (simpleRuled is ParabolicCylinder)
                return GetLinearClipIntervals((ParabolicCylinder)simpleRuled, x, y, z);
            if (simpleRuled is CrossingPlans)
                return GetLinearClipIntervals((CrossingPlans)simpleRuled, x, y, z);
            if (simpleRuled is ParallelPlans)
                return GetLinearClipIntervals((ParallelPlans)simpleRuled, x, y, z);
            if (simpleRuled is DoublePlan)
                return GetLinearClipIntervals((DoublePlan)simpleRuled, x, y, z);
            if (simpleRuled is ImaginarCrossingPlans)
                return GetLinearClipIntervals((ImaginarCrossingPlans)simpleRuled, x, y, z);
            throw new Exception();
        }


        /// <summary>
        /// x = 0
        /// y = 0
        /// z = u
        /// </summary>
        private List<Interval> GetLinearClipIntervals(ImaginarCrossingPlans imaginarCrossingPlans,
            Interval x, Interval y, Interval z)
        {
            Interval xClip = new Interval(double.MinValue, double.MaxValue);
            Interval yClip = new Interval(double.MinValue, double.MaxValue);
            Interval zClip = new Interval(z.GetA(), z.GetB());
            return new List<Interval>() { Interval.Intersect(Interval.Intersect(xClip, yClip), zClip) };
        }
        /// <summary>
        /// x = 0
        /// y = u
        /// z = v
        /// </summary>
        private List<Interval> GetLinearClipIntervals(DoublePlan doublePlan,
            Interval x, Interval y, Interval z)
        {
            Interval xClip = new Interval(double.MinValue, double.MaxValue);
            Interval yClip = new Interval(y.GetA(), y.GetB());
            List<Interval> zClips = GetClipIntervalV(z);
            return Interval.Intersect(zClips, Interval.Intersect(xClip, yClip));
        }
        /// <summary>
        /// x = -a
        /// y = u
        /// z = v
        /// </summary>
        private List<Interval> GetLinearClipIntervals(ParallelPlans parallelPlans,
            Interval x, Interval y, Interval z)
        {
            Interval xClip = new Interval(double.MinValue, double.MaxValue);
            Interval yClip = new Interval(y.GetA(), y.GetB());
            List<Interval> zClips = GetClipIntervalV(z);
            return Interval.Intersect(zClips, Interval.Intersect(xClip, yClip));
        }
        /// <summary>
        /// x = au
        /// y = +-bu
        /// z = v
        /// </summary>
        private List<Interval> GetLinearClipIntervals(CrossingPlans crossingPlans,
            Interval x, Interval y, Interval z)
        {
            double a = _simpleRuled.GetParametricEquation().GetA();
            double b = _simpleRuled.GetParametricEquation().GetB();
            Interval xClip = new Interval(x.GetA() / a, x.GetB() / a);
            List<Interval> yClips = new List<Interval>();
            yClips.Add(new Interval(y.GetA() / b, y.GetB() / b));
            yClips.Add(new Interval(y.GetA() / -b, y.GetB() / -b));
            List<Interval> zClips = GetClipIntervalV(z);
            return Interval.Intersect(Interval.Intersect(yClips, xClip), zClips);
        }
        /// <summary>
        /// x = u
        /// y = -(u*u)/(constantA)
        /// z = v
        /// </summary>
        private List<Interval> GetLinearClipIntervals(ParabolicCylinder parabolicCylinder,
            Interval x, Interval y, Interval z)
        {
            double constantA = QuadricsSimpleRuledParametrizations.GetConstantA(parabolicCylinder);
            Interval xClip = x;
            QuadraticEquation topEquation = new QuadraticEquation(1, 0, y.GetA() * constantA);
            QuadraticEquation bottomEquation = new QuadraticEquation(1, 0, y.GetB());
            Arguments arguments = new Arguments(bottomEquation, topEquation, y);
            List<Interval> yClips = GetClipInterval(SubstitutionPCFunction, arguments);
            List<Interval> zClips = GetClipIntervalV(z);
            return Interval.Intersect(Interval.Intersect(yClips, xClip), zClips);
        }
        /// <summary>
        /// x = _a/2*(u + 1/u)
        /// y = _b/2*(u - 1/u)
        /// z = v
        /// 
        /// au2 - xu + a = 0
        /// bu2 - yu - b = 0
        /// </summary>
        private List<Interval> GetLinearClipIntervals(HyperbolicCylinder hyperbolicCylinder,
            Interval x, Interval y, Interval z)
        {
            double a = hyperbolicCylinder.GetParametricEquation().GetA() / 2.0;
            double b = hyperbolicCylinder.GetParametricEquation().GetB() / 2.0;

            QuadraticEquation xEquationLeft = new QuadraticEquation(1, -x.GetA() / a, 1);
            QuadraticEquation xEquationRight = new QuadraticEquation(1, -x.GetB() / a, 1);
            List<double> xSolutionsLeft = xEquationLeft.Solve();
            List<double> xSolutionsRight = xEquationRight.Solve();
            QuadraticEquation yEquationBottom = new QuadraticEquation(1, -x.GetA() / b, -1);
            QuadraticEquation yEquationTop = new QuadraticEquation(1, -x.GetB() / b, -1);
            List<double> ySolutionsBottom = yEquationBottom.Solve();
            List<double> ySolutionsTop = yEquationTop.Solve();
            List<double> xySolutions = Merge(Merge(Merge(xSolutionsLeft, xSolutionsRight), ySolutionsBottom), ySolutionsTop);
            xySolutions.Sort();
            List<Interval> xyIntervals = new List<Interval>();
            for (int i = 0; i < xySolutions.Count -1; i++) 
            {
                double u = (xySolutions[i] + xySolutions[i + 1]) / 2.0;
                if (!Program.IsEqualInToleranceHard(u, 0) && x.Belongs(a * (u + 1 / u)) && y.Belongs(b * (u - 1 / u)))
                    xyIntervals.Add(new Interval(xySolutions[i], xySolutions[i + 1]));
            }
            List<Interval> zClip = GetClipIntervalV(z);
            return Interval.Intersect(xyIntervals, zClip);
        }
        /// <summary>
        /// x = a(u + v)
        /// y = b(u - v)
        /// z = -puv
        /// </summary>
        private List<Interval> GetLinearClipIntervals(HyperbolicParaboloid hyperbolicParaboloid,
            Interval x, Interval y, Interval z)
        {
            double constantA = QuadricsSimpleRuledParametrizations.GetConstantA(hyperbolicParaboloid);
            double constantB = QuadricsSimpleRuledParametrizations.GetConstantB(hyperbolicParaboloid);
            double p = _simpleRuled.GetParametricEquation().GetP();

            List<Interval> xClips = GetClipIntervalA(x, constantA);
            List<Interval> yClips = GetClipIntervalB(y, constantB);
            List<Interval> zClip = GetClipIntervalP(z, p);
            return Interval.Intersect(zClip, Interval.Intersect(xClips, yClips));
        }

        

        /// <summary>
        /// Vyřeším intervaly hodnot, pro které je hodnota x v intervalu "interval".
        /// 
        ///              a     
        /// x = du + d * -
        ///              b
        /// </summary>
        private List<Interval> GetClipIntervalA(Interval interval, double a)
        {
            Func<double, IParameterVParametrization, double> testFunction = new Func<double, IParameterVParametrization, double>(SubstitutionAFunction);
            QuarticEquation leftEquation = GetQuarticEquationA(interval.GetA(), a);
            QuarticEquation rightEquation = GetQuarticEquationA(interval.GetB(), a);
            return GetClipInterval(testFunction, new Arguments(leftEquation, rightEquation, interval));
        }
        /// <summary>
        /// Vyřeším intervaly hodnot, pro které je hodnota x v intervalu "interval".
        /// 
        ///              a     
        /// x = du - d * -
        ///              b
        /// </summary>
        private List<Interval> GetClipIntervalB(Interval interval, double b)
        {
            Func<double, IParameterVParametrization, double> testFunction = new Func<double, IParameterVParametrization, double>(SubstitutionBFunction);
            QuarticEquation bottomEquation = GetQuarticEquationB(interval.GetA(), b);
            QuarticEquation topEquation = GetQuarticEquationB(interval.GetB(), b);
            return GetClipInterval(testFunction, new Arguments(bottomEquation, topEquation, interval));
        }
        /// <summary>
        /// Vyřeším intervaly hodnot, pro které je hodnota x v intervalu "interval".
        /// 
        ///           a    
        /// z = -pu * -
        ///           b
        /// </summary>
        private List<Interval> GetClipIntervalP(Interval interval, double p)
        {
            Func<double, IParameterVParametrization, double> testFunction = new Func<double, IParameterVParametrization, double>(SubstitutionPFunction);
            QuarticEquation backEquations = GetQuarticEquationP(interval.GetA(), p);
            QuarticEquation frontEquations = GetQuarticEquationP(interval.GetB(), p);
            return GetClipInterval(testFunction, new Arguments(backEquations, frontEquations, interval));
        }
        /// <summary>
        /// Vyřeším intervaly hodnot, pro které je hodnota vyjádření parametru "v" v intervalu "interval".
        /// 
        ///     a     
        /// v = -
        ///     b
        /// </summary>
        private List<Interval> GetClipIntervalV(Interval interval)
        {
            Func<double, IParameterVParametrization, double> testFunction = new Func<double, IParameterVParametrization, double>(SubstitutionVFunction);
            QuarticEquation bottomEquation = GetQuarticEquationV(interval.GetA());
            QuarticEquation topEquation = GetQuarticEquationV(interval.GetB());
            return GetClipInterval(testFunction, new Arguments(bottomEquation, topEquation, interval));
        }


        /// <summary>
        ///              a     
        /// x = du + d * -
        ///              b
        ///__________________________            
        /// (x*b - d*u*b - d*a) = 0
        /// </summary>
        private QuarticEquation GetQuarticEquationA(double x, double a)
        {
            Expression aC = (Expression)((LinearV)_vParametrization).GetNumerator();
            Expression bC = (Expression)((LinearV)_vParametrization).GetDenominator();
            Expression xC = new Expression(new UV(x, 0, 0));
            Expression du = new Expression(new UV(a, 1, 0));
            Expression equation = (xC * bC) + (-du * bC) + (-a * aC);
            return equation.GetQuarticEquationInU();

        }
        /// <summary>
        ///              a     
        /// y = du - d * -
        ///              b        
        ///__________________________            
        /// (y*b - d*u*b + d*a) = 0
        /// </summary>
        private QuarticEquation GetQuarticEquationB(double y, double b)
        {
            Expression aC = ((LinearV)_vParametrization).GetNumerator();
            Expression bC = ((LinearV)_vParametrization).GetDenominator();
            Expression yC = new Expression(new UV(y, 0, 0));
            Expression du = new Expression(new UV(b, 1, 0));
            Expression equation = (yC * bC) + (-du * bC) + (b * aC);
            return equation.GetQuarticEquationInU();
        }
        /// <summary>
        ///                 a    
        /// z = -Sign(p)u * -
        ///                 b
        ///__________________________            
        /// (z*b + p*u*a) = 0
        /// --- 
        /// a i b jsou v případě paraboloidu maximálně kvadratické.
        /// </summary>
        private QuarticEquation GetQuarticEquationP(double z, double p)
        {
            Expression aC = ((LinearV)_vParametrization).GetNumerator();
            Expression bC = ((LinearV)_vParametrization).GetDenominator();
            Expression zC = new Expression(new UV(z, 0, 0));
            Expression pu = new Expression(new UV(Math.Sign(p), 1, 0));
            Expression equation = (zC * bC) + (pu * aC);
            return equation.GetQuarticEquationInU();
        }
        /// <summary>
        ///     a     
        /// v = -
        ///     b
        ///__________________________            
        /// bv - a = 0
        /// </summary>
        private QuarticEquation GetQuarticEquationV(double x)
        {
            Expression aC = ((LinearV)_vParametrization).GetNumerator();
            Expression bC = ((LinearV)_vParametrization).GetDenominator();
            Expression equation = (bC * x) + (-aC);
            return equation.GetQuarticEquationInU();
        }
    }


}
