﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Geometry.GraphicObjects.Quadrics;
using GMath;


namespace Geometry.GraphicObjects.Intersections
{
    public class QuadraticClipManager : IClipManager
    {

        public QuadraticClipManager(CurveAsQuadricsIntersection curve, Scene scene, int sign, Interval interval, List<double> minusValues)
            : base(curve, scene, sign, interval, minusValues)
        {
        }

        public List<Interval> GetParameterIntervals()
        {
            return GetClipIntervals(_scene, _simpleRuled, _sign);
        }
    
        private List<Interval> GetClipIntervals(Scene scene, SimpleRulledQuadric simpleRuled, int sign)
        {
            //return _definitionInterval.GetIntervals();
            Interval x = new Interval(simpleRuled.GetClippingLeft(scene), simpleRuled.GetClippingRight(scene));
            Interval y = new Interval(simpleRuled.GetClippingBottom(scene), simpleRuled.GetClippingTop(scene));
            Interval z = new Interval(simpleRuled.GetClippingBack(scene), simpleRuled.GetClippingFront(scene));

            if (simpleRuled is HyperbolicParaboloid)
                return GetQuadraticClipIntervals((HyperbolicParaboloid)simpleRuled, x, y, z, sign);
            if (simpleRuled is HyperbolicCylinder)
                return GetQuadraticClipIntervals((HyperbolicCylinder)simpleRuled, x, y, z, sign);
            if (simpleRuled is ParabolicCylinder)
                return GetQuadraticClipIntervals((ParabolicCylinder)simpleRuled, x, y, z, sign);
            if (simpleRuled is CrossingPlans)
                return GetQuadraticClipIntervals((CrossingPlans)simpleRuled, x, y, z, sign);
            if (simpleRuled is ParallelPlans)
                return GetQuadraticClipIntervals((ParallelPlans)simpleRuled, x, y, z, sign);
            if (simpleRuled is DoublePlan)
                return GetQuadraticClipIntervals((DoublePlan)simpleRuled, x, y, z, sign);
            if (simpleRuled is ImaginarCrossingPlans)
                return GetQuadraticClipIntervals((ImaginarCrossingPlans)simpleRuled, x, y, z, sign);
            throw new Exception();
        }

        /// <summary>
        /// x = constantA(u + v)
        /// y = constantB(u - v)
        /// z = -sign(p)puv
        /// </summary>
        private List<Interval> GetQuadraticClipIntervals(HyperbolicParaboloid simpleRuled,
            Interval x, Interval y, Interval z, int sign)
        {
            double contantA = QuadricsSimpleRuledParametrizations.GetConstantA(simpleRuled);
            double constantB = QuadricsSimpleRuledParametrizations.GetConstantB(simpleRuled);
            double p = simpleRuled.GetParametricEquation().GetP();

            List<Interval> xClips = GetClipIntervalA(x, sign, contantA);
            List<Interval> yClips = GetClipIntervalB(y, sign, constantB);
            List<Interval> zClip = GetClipIntervalP(z, sign, p);
            return Interval.Intersect(zClip, Interval.Intersect(xClips, yClips));
        }
        /// <summary>
        /// x = _a/2*(u + 1/u)
        /// y = _b/2*(u - 1/u)
        /// z = v
        /// 
        /// au2 - xu + a = 0
        /// bu2 - yu - b = 0
        /// </summary>
        private List<Interval> GetQuadraticClipIntervals(HyperbolicCylinder simpleRuled,
            Interval x, Interval y, Interval z, int sign)
        {
            double a = simpleRuled.GetParametricEquation().GetA() / 2.0;
            double b = simpleRuled.GetParametricEquation().GetB() / 2.0;

            QuadraticEquation xEquationLeft = new QuadraticEquation(1, -x.GetA() / a, 1);
            QuadraticEquation xEquationRight = new QuadraticEquation(1, -x.GetB() / a, 1);
            List<double> xSolutionsLeft = xEquationLeft.Solve();
            List<double> xSolutionsRight = xEquationRight.Solve();
            QuadraticEquation yEquationBottom = new QuadraticEquation(1, -x.GetA() / b, -1);
            QuadraticEquation yEquationTop = new QuadraticEquation(1, -x.GetB() / b, -1);
            List<double> ySolutionsBottom = yEquationBottom.Solve();
            List<double> ySolutionsTop = yEquationTop.Solve();
            List<double> xySolutions = Merge(Merge(Merge(xSolutionsLeft, xSolutionsRight), ySolutionsBottom), ySolutionsTop);
            xySolutions.Sort();
            List<Interval> xyIntervals = new List<Interval>();
            for (int i = 0; i < xySolutions.Count - 1; i++)
            {
                double u = (xySolutions[i] + xySolutions[i + 1]) / 2.0;
                if (!Program.IsEqualInToleranceHard(u, 0) && x.Belongs(a * (u + 1 / u)) && y.Belongs(b * (u - 1 / u)))
                    xyIntervals.Add(new Interval(xySolutions[i], xySolutions[i + 1]));
            }
            List<Interval> zClip = GetClipIntervalV(z, sign);
            return Interval.Intersect(xyIntervals, zClip);
        }
        /// <summary>
        /// x = u
        /// y = -(u*u)/(constantA)
        /// z = v
        /// </summary>
        private List<Interval> GetQuadraticClipIntervals(ParabolicCylinder simpleRuled,
            Interval x, Interval y, Interval z, int sign)
        {
            double constantA = QuadricsSimpleRuledParametrizations.GetConstantA(simpleRuled);
            Interval xClip = x;
            QuadraticEquation topEquation = new QuadraticEquation(1, 0, y.GetA() * constantA);
            QuadraticEquation bottomEquation = new QuadraticEquation(1, 0, y.GetB() * constantA);
            Arguments arguments = new Arguments(bottomEquation, topEquation, y);
            List<Interval> yClips = GetClipInterval(SubstitutionPCFunction, arguments);
            List<Interval> zClips = GetClipIntervalV(z, sign);
            return Interval.Intersect(Interval.Intersect(yClips, xClip), zClips);
        }
        /// <summary>
        /// x = au
        /// y = +-bu
        /// z = v
        /// </summary>
        private List<Interval> GetQuadraticClipIntervals(CrossingPlans simpleRuled,
            Interval x, Interval y, Interval z, int sign)
        {
            double a = simpleRuled.GetParametricEquation().GetA();
            double b = simpleRuled.GetParametricEquation().GetB();
            Interval xClip = new Interval(x.GetA() / a, x.GetB() / a);
            Interval i1 = new Interval(y.GetA() / b, y.GetB() / b);
            Interval i2 = new Interval(y.GetA() / -b, y.GetB() / -b);
            List<Interval> yClips = Interval.Union(i1, i2);
            List<Interval> zClips = GetClipIntervalV(z, sign);
            return Interval.Intersect(Interval.Intersect(yClips, xClip), zClips);
        }
        /// <summary>
        /// x = +-a
        /// y = u
        /// z = v
        /// </summary>
        private List<Interval> GetQuadraticClipIntervals(ParallelPlans simpleRuled,
            Interval x, Interval y, Interval z, int sign)
        {
            Interval xClip = new Interval(double.MinValue, double.MaxValue);
            Interval yClip = new Interval(y.GetA(), y.GetB());
            List<Interval> zClips = GetClipIntervalV(z, sign);
            return Interval.Intersect(zClips, Interval.Intersect(xClip, yClip));
        }
        /// <summary>
        /// x = 0
        /// y = u
        /// z = v
        /// </summary>
        private List<Interval> GetQuadraticClipIntervals(DoublePlan simpleRuled,
            Interval x, Interval y, Interval z, int sign)
        {
            Interval xClip = new Interval(double.MinValue, double.MaxValue);
            Interval yClip = new Interval(y.GetA(), y.GetB());
            List<Interval> zClips = GetClipIntervalV(z, sign);
            return Interval.Intersect(zClips, Interval.Intersect(xClip, yClip));
        }
        /// <summary>
        /// x = 0
        /// y = 0
        /// z = u
        /// </summary>
        private List<Interval> GetQuadraticClipIntervals(ImaginarCrossingPlans simpleRuled,
            Interval x, Interval y, Interval z, int sign)
        {
            Interval xClip = new Interval(double.MinValue, double.MaxValue);
            Interval yClip = new Interval(double.MinValue, double.MaxValue);
            Interval zClip = new Interval(z.GetA(), z.GetB());
            return new List<Interval>() { Interval.Intersect(Interval.Intersect(xClip, yClip), zClip) };
        }

        /// <summary>
        /// Vyřeším intervaly hodnot, pro které je hodnota x v intervalu "interval".
        /// 
        ///              -b +- Sqrt(b2 - 4ac)     
        /// x = du + d * -------------------
        ///                       2a
        /// </summary>
        private List<Interval> GetClipIntervalA(Interval interval, int sign, double a)
        {
            Func<double, IParameterVParametrization, double> testFunction = new Func<double, IParameterVParametrization, double>(SubstitutionAFunction);
            QuarticEquation leftEquation = GetQuarticEquationA(interval.GetA(), a);
            QuarticEquation rightEquation = GetQuarticEquationA(interval.GetB(), a);
            return GetClipInterval(testFunction, new Arguments(leftEquation, rightEquation, interval));
        }
        /// <summary>
        /// Vyřeším intervaly hodnot, pro které je hodnota x v intervalu "interval".
        /// 
        ///              -b +- Sqrt(b2 - 4ac)     
        /// x = du - d * -------------------
        ///                       2a
        /// </summary>
        private List<Interval> GetClipIntervalB(Interval interval, int sign, double b)
        {
            Func<double, IParameterVParametrization, double> testFunction = new Func<double, IParameterVParametrization, double>(SubstitutionBFunction);
            QuarticEquation bottomEquation = GetQuarticEquationB(interval.GetA(), b);
            QuarticEquation topEquation = GetQuarticEquationB(interval.GetB(), b);
            return GetClipInterval(testFunction, new Arguments(bottomEquation, topEquation, interval));
        }
        /// <summary>
        /// Vyřeším intervaly hodnot, pro které je hodnota x v intervalu "interval".
        /// 
        ///                 -b +- Sqrt(b2 - 4ac)     
        /// z = -sign(p)u * -------------------
        ///                         2a
        /// </summary>
        private List<Interval> GetClipIntervalP(Interval interval, int sign, double p)
        {
            Func<double, IParameterVParametrization, double> testFunction = new Func<double, IParameterVParametrization, double>(SubstitutionPFunction);
            QuarticEquation backEquations = GetQuarticEquationP(interval.GetA(), p);
            QuarticEquation frontEquations = GetQuarticEquationP(interval.GetB(), p);
            return GetClipInterval(testFunction, new Arguments(backEquations, frontEquations, interval));
        }
        /// <summary>
        /// Vyřeším intervaly hodnot, pro které je hodnota vyjádření parametru "v" v intervalu "interval".
        /// 
        ///    -b +- Sqrt(b2 - 4ac)     
        /// v = -------------------
        ///             2a
        /// </summary>
        private List<Interval> GetClipIntervalV(Interval interval, int sign)
        {
            Func<double, IParameterVParametrization, double> testFunction = new Func<double, IParameterVParametrization, double>(SubstitutionVFunction);
            QuarticEquation bottomEquation = GetQuarticEquationV(interval.GetA());
            QuarticEquation topEquation = GetQuarticEquationV(interval.GetB());
            return GetClipInterval(testFunction, new Arguments(bottomEquation, topEquation, interval));
        }

        /// <summary>
        ///              -b +- Sqrt(b2 - 4ac)     
        /// x = du + d * -------------------
        ///                       2a
        ///__________________________            
        /// (x2 - 2xdu + d2*u2)*a/d2 + (bx - dbu)/d + c = 0
        /// </summary>
        private QuarticEquation GetQuarticEquationA(double x, double a)
        {
            Expression aC = ((QuadraticV)_vParametrization).GetA();
            Expression bC = ((QuadraticV)_vParametrization).GetB();
            Expression cC = ((QuadraticV)_vParametrization).GetC();
            Expression xC = new Expression(new UV(x, 0, 0));
            Expression du = new Expression(new UV(a, 1, 0));
            Expression equation = (xC * xC * aC / (a * a)) + (-2 * du * aC * xC / (a * a));
            equation += (du * du * aC / (a * a));
            equation += (bC * xC / a) + (-bC * du / a) + (cC);
            return equation.GetQuarticEquationInU();

        }
        /// <summary>
        ///              -b +- Sqrt(b2 - 4ac)     
        /// y = du - d * -------------------
        ///                       2a
        ///__________________________            
        /// (y2 - 2ydu + d2*u2)*a/d2 - (by - dbu)/d + c = 0
        /// </summary>
        private QuarticEquation GetQuarticEquationB(double y, double b)
        {
            Expression aC = (Expression)((QuadraticV)_vParametrization).GetA();
            Expression bC = (Expression)((QuadraticV)_vParametrization).GetB();
            Expression cC = (Expression)((QuadraticV)_vParametrization).GetC();
            Expression yC = new Expression(new UV(y, 0, 0));
            Expression du = new Expression(new UV(b, 1, 0));
            Expression equation = (yC * yC * aC / (b * b)) + (-2 * du * aC * yC / (b * b));
            equation += (du * du * aC / (b * b));
            equation += (-bC * yC / b) + (bC * du / b) + (cC);
            return equation.GetQuarticEquationInU();
        }
        /// <summary>
        ///                 -b +- Sqrt(b2 - 4ac)     
        /// z = -sign(p)u * -------------------
        ///                         2a
        ///__________________________            
        /// (z2*a - zbu*sign(p) + cp2u2 = 0
        /// </summary>
        private QuarticEquation GetQuarticEquationP(double z, double p)
        {
            Expression aC = ((QuadraticV)_vParametrization).GetA();
            Expression bC = ((QuadraticV)_vParametrization).GetB();
            Expression cC = ((QuadraticV)_vParametrization).GetC();
            Expression zC = new Expression(new UV(z, 0, 0));
            Expression pu = new Expression(new UV(Math.Sign(p), 1, 0));
            Expression equation = (zC * zC * aC) + (-zC * bC * pu) + (cC * pu * pu);
            return equation.GetQuarticEquationInU();
        }
        /// <summary>
        ///    -b +- Sqrt(b2 - 4ac)     
        /// v = -------------------
        ///             2a
        ///__________________________            
        /// ax2 + bx + c = 0
        /// </summary>
        private QuarticEquation GetQuarticEquationV(double x)
        {
            Expression aC = ((QuadraticV)_vParametrization).GetA();
            Expression bC = ((QuadraticV)_vParametrization).GetB();
            Expression cC = ((QuadraticV)_vParametrization).GetC();
            Expression equation = (aC * x * x) + (bC * x) + (cC);
            return equation.GetQuarticEquationInU();
        }
    }
}
