﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using GMath;

namespace Geometry.GraphicObjects.Intersections
{
    public class Expression
    {
        private List<UV> _list;

        public Expression()
        {
            _list = new List<UV>() { new UV(0, 0, 0) };
        }
        public Expression(double d) 
        {
            _list = new List<UV>() { new UV(d, 0, 0) };
        }
        public Expression(UV uv) 
        {
            _list = new List<UV>() { uv };
        }
        /// <summary>
        /// Automaticky se vola SumAndSimplify 
        /// </summary>
        public Expression(UV uv1, UV uv2)
        {
            _list = new List<UV>() { uv1, uv2 };
            _list = SumAndSimplify(_list);
        }
        /// <summary>
        /// Automaticky se vola SumAndSimplify 
        /// </summary>
        public Expression(Expression expression)
        {
            _list = new List<UV>();
            foreach (UV uv in expression.GetList())
                _list.Add(new UV(uv));
            _list = SumAndSimplify(_list);
        }
        /// <summary>
        /// Automaticky se vola SumAndSimplify 
        /// </summary>
        public Expression(List<UV> list)
        {
            if (list == null || list.Count < 1)
                _list = new List<UV>() { new UV(0, 0, 0) };
            else
            {
                _list = SumAndSimplify(list);
            }
        }

        public List<UV> GetList() { return _list; }
        public int Count() { return _list.Count; }
        public bool IsZero()
        {
            foreach (UV uv in this._list)
                if (!Program.IsEqualInTolerance(uv.GetValue(), 0))
                    return false;
            return true;
        }

        public static Expression operator *(Expression a, UV b)
        {
            return b * a;
        }
        public static Expression operator *(UV a, Expression b)
        {
            List<UV> retval = new List<UV>();
            for (int i = 0; i < b._list.Count; i++)
            {
                retval.Add(a * b._list[i]);
            }
            return new Expression(retval);
        }
        public static Expression operator *(Expression a, Expression b)
        {
            List<UV> retval = new List<UV>();
            foreach (UV uvA in a._list)
                foreach (UV uvB in b._list)
                    retval.Add(uvA * uvB);
            return new Expression(retval);
        }
        public static Expression operator *(Expression a, double b)
        {
            return b * a;
        }
        public static Expression operator *(double a, Expression b)
        {
            List<UV> retval = new List<UV>();
            for (int i = 0; i < b._list.Count; i++)
            {
                retval.Add(a * b._list[i]);
            }
            return new Expression(retval);
        }


        public static Expression operator +(Expression a, Expression b)
        {
            List<UV> retval = new List<UV>();
            foreach (UV uv in a._list)
                retval.Add(uv);
            foreach (UV uv in b._list)
                retval.Add(uv);
            return new Expression(retval);
        }
        public static Expression operator +(Expression a, UV b)
        {
            List<UV> retval = new List<UV>();
            foreach (UV uv in a._list)
                retval.Add(uv);
            retval.Add(b);
            return new Expression(retval);
        }

        public static Expression operator -(Expression a, Expression b)
        {
            List<UV> retval = new List<UV>();
            foreach (UV uv in a._list)
                retval.Add(uv);
            foreach (UV uv in b._list)
                retval.Add(-uv);
            return new Expression(retval);
        }
        public static Expression operator -(Expression a)
        {
            Expression retval = new Expression();
            foreach (UV uv in a._list)
                retval._list.Add(-uv);
            return retval;
        }

        public static Expression operator /(Expression a, double b)
        {
            List<UV> retval = new List<UV>();
            for (int i = 0; i < a._list.Count; i++)
            {
                retval.Add(a._list[i] / b);
            }
            return new Expression(retval);
        }


        private static List<UV> SumAndSimplify(List<UV> listUv)
        {
            int uMinExponent = -2;
            int vMinExponent = 0;
            int uMaxExponent = 4;
            int vMaxExponent = 2;
            double sum = 0;
            List<UV> retval = new List<UV>();
            for (int uExponent = uMinExponent; uExponent <= uMaxExponent; uExponent++)
                for (int vExponent = vMinExponent; vExponent <= vMaxExponent; vExponent++)
                {
                    sum = 0;
                    foreach (UV uv in listUv)
                    {
                        if (uv.GetUexponent() == uExponent && uv.GetVexponent() == vExponent)
                            sum += uv.GetValue();
                    }
                    if (sum != 0)
                    {
                        retval.Add(new UV(sum, uExponent, vExponent));
                    }
                }
            retval.Reverse();
            return retval;
        }

        public bool ContainsUWithExponent(int exponent)
        {
            foreach (UV uv in this._list)
            {
                if (uv.ContainsUWithExponent(exponent))
                    return true;
            }
            return false;
        }
        public bool ContainsVWithExponent(int exponent)
        {
            foreach (UV uv in this._list)
            {
                if (uv.ContainsVWithExponent(exponent))
                    return true;
            }
            return false;
        }
        public bool ContainsUTimesVOnly()
        {
            List<int> uExponents = GetListOfUExponents();
            List<int> vExponents = GetListOfVExponents();
            if (_list.Count == 1 && uExponents[0] > 0 && vExponents[0] > 0)
                return true;
            return false;
        }
        public bool ContainsUDivideVOnly()
        {
            List<int> uExponents = GetListOfUExponents();
            List<int> vExponents = GetListOfVExponents();
            if (_list.Count == 1 && uExponents[0] < 0 && vExponents[0] > 0)
                return true;
            return false;
        }

        private List<int> GetListOfUExponents() 
        {
            List<int> retval = new List<int>();
            foreach (UV uv in _list)
                retval.Add(uv.GetUexponent());
            return retval;
        }
        private List<int> GetListOfVExponents()
        {
            List<int> retval = new List<int>();
            foreach (UV uv in _list)
                retval.Add(uv.GetVexponent());
            return retval;
        }

        public Expression SubstituteForU(double u)
        {
            List<UV> retvalValues = new List<UV>();
            foreach (UV uv in this._list)
            {
                retvalValues.Add(uv.SubstituteForU(u));
            }
            return new Expression(retvalValues);
        }
        public Expression SubstituteForV(double v)
        {
            List<UV> retvalValues = new List<UV>();
            foreach (UV uv in this._list)
            {
                retvalValues.Add(uv.SubstituteForV(v));
            }
            return new Expression(retvalValues);
        }


        public double GetValueAt(double u)
        {
            double retval = 0;
            foreach (UV uv in this._list)
            {
                retval += Math.Pow(u, uv.GetUexponent()) * uv.GetValue();
            }
            return retval;
        }
        public double GetValueAt(double u, double v)
        {
            double retval = 0;
            foreach (UV uv in this._list)
            {
                double a = Math.Pow(u, uv.GetUexponent());
                double b = Math.Pow(v, uv.GetVexponent());
                double c = uv.GetValue();
                retval += a * b * c;
            }
            return retval;
        }

        public override string ToString()
        {
            string retval = "";
            foreach (UV uv in this._list)
                retval +=  uv.ToString() + " ";
            return retval;
        }
        public string ToStringIgnoreV()
        {
            string retval = "";
            foreach (UV uv in this._list)
                retval += uv.ToStringIgnoreV() + " ";
            return retval;
        }
        public string ToFormatString()
        {
            string retval = "L[";
            foreach (UV uv in this._list)
                retval += uv.ToString() + " ";
            return retval + "]";
        }
        public string ToFormatStringIgnoreV()
        {
            string retval = "L[";
            foreach (UV uv in this._list)
                retval += uv.ToStringIgnoreV() + " ";
            return retval + "]";
        }
        

        public QuarticEquation GetQuarticEquationInU()
        {
            double qA = 0;
            double qB = 0;
            double qC = 0;
            double qD = 0;
            double qE = 0;
            double qF = 0;
            double qG = 0;
            foreach (UV uv in _list)
            {
                if (uv.GetUexponent() == 4)
                    qA += uv.GetValue();
                if (uv.GetUexponent() == 3)
                    qB += uv.GetValue(); 
                if (uv.GetUexponent() == 2)
                    qC += uv.GetValue(); 
                if (uv.GetUexponent() == 1)
                    qD += uv.GetValue(); 
                if (uv.GetUexponent() == 0)
                    qE += uv.GetValue();
                if (uv.GetUexponent() == -1)
                    qF += uv.GetValue();
                if (uv.GetUexponent() == -2)
                    qG += uv.GetValue();
            }
            if (qF != 0 || qG != 0)
                return new QuarticEquation(qC, qD, qE, qF, qG); // rovnice je vynasobena x2
            return new QuarticEquation(qA, qB, qC, qD, qE);
        }
        public List<Expression> GetLinearEquationInV()
        {
            Expression A = new Expression(0);
            Expression B = new Expression(0);
            foreach (UV uv in _list)
            {
                if (uv.GetVexponent() == 1)
                    A += uv;
                if (uv.GetVexponent() == 0)
                    B += uv;
            }
            return new List<Expression>() { A, B};
        }
        public List<Expression> GetQuadraticEquationInV()
        {
            Expression A = new Expression(0);
            Expression B = new Expression(0);
            Expression C = new Expression(0);
            foreach (UV uv in _list)
            {
                if (uv.GetVexponent() == 2)
                    A += uv;
                if (uv.GetVexponent() == 1)
                    B += uv;
                if (uv.GetVexponent() == 0)
                    C += uv;
            }
            return new List<Expression>() { A, B, C };
        }

        public bool IsOne()
        {
            foreach (UV uv in _list)
                if (uv.GetValue() != 1 || uv.GetUexponent() != 0 || uv.GetVexponent() != 0)
                    return false;
            return true;
        }

        public Expression EliminateZeros() 
        {
            List<UV> newlist = new List<UV>();
            foreach (UV uv in this._list) 
            {
                if (!Program.IsEqualInTolerance(uv.GetValue(), 0))
                    newlist.Add(new UV(uv));
            }
            return new Expression(newlist);
        }

        public Expression Normalize() 
        {
            List<double> values = ToListDouble();
            return this * Program.GetNormalizingKoeficient(values);
        }

        internal List<double> ToListDouble()
        {
            List<double> values = new List<double>();
            foreach (UV uv in _list)
                values.Add(uv.GetValue());
            return values;
        }
    }
}
