﻿using System;
using System.Collections.Generic;
using Geometry.BaseClasses;
using Geometry.GraphicObjects.Quadrics;
using Tao.OpenGl;
using GMath;
using System.Threading;

namespace Geometry.GraphicObjects.Intersections
{
    public class CurveAsQuadricsIntersection:GraphicObject
    {
        protected IParameterVParametrization _vParametrization;
        protected DefinitionInterval _definitionIntervals;
        private CurveCylinder _curveCylinder = new CurveCylinder();
        private CurveSphere _curveSphere = new CurveSphere();
        protected ParametrizationVector _parametrizationVectorInCanonicBasis;
        protected ParametrizationVector _parametrizationVectorInStandardBasis;
        protected SimpleRulledQuadric _simpleRulled;
        protected const double PointQuality = 20;
        protected const double PointRadius = 1;


        private struct Pair
        {
            public double u;
            public Vector coordinates;
            public Pair(double u, Vector coordinates) 
            {
                this.u = u;
                this.coordinates = coordinates;
            }
        }

        public CurveAsQuadricsIntersection(ParametrizationVector parametrizationVector,
            IParameterVParametrization vParametrization, DefinitionInterval definitionIntervals, SimpleRulledQuadric simpleRulled)
            :base()
        {
            this._vParametrization = vParametrization;
            definitionIntervals.Sort();
            this._definitionIntervals = definitionIntervals;
            this._parametrizationVectorInCanonicBasis = parametrizationVector;
            this._parametrizationVectorInStandardBasis =
                (parametrizationVector * simpleRulled.GetToCanonicTransformationMatrix());
            this._simpleRulled = simpleRulled;
        }

        public SimpleRulledQuadric GetSimpleRuledQuadric() { return _simpleRulled; }

        public IParameterVParametrization GetVParametrization() { return _vParametrization; }
        public DefinitionInterval GetDefinitionIntervals() { return _definitionIntervals; }

        public ParametrizationVector GetParametrizationVectorInStadardBasis()
        {
            return _parametrizationVectorInStandardBasis;
        }
        public ParametrizationVector GetParametrizationVectorInCanonicBasis()
        {
            return _parametrizationVectorInCanonicBasis;
        }


        private Vector GetVertexAtConvertDiskrininantNegativity(double u, int sign) 
        {
            double valueOfV;
            if (_vParametrization is QuadraticV)
                valueOfV = ((QuadraticV)_vParametrization).GetVvalueConvertDiskriminantNegativity(u, sign);
            else valueOfV = _vParametrization.GetVvalue(u, sign);
            return _parametrizationVectorInCanonicBasis.GetValueAt(u, valueOfV);
        }
        private Vector GetVertexAt(double u, int sign)
        {
            //return new Vector(u, sign * u * u, 0);
            double valueOfV = _vParametrization.GetVvalue(u, sign);
            return _parametrizationVectorInCanonicBasis.GetValueAt(u, valueOfV);
        }

        public override void BuildDisplayListIfNecessary(Scene scene, RelativeQuality relativeQuality)
        {
            //Test(scene);
            //return;
            if (Gl.glIsList(_displayListIdentificator) == Gl.GL_TRUE)
                Gl.glDeleteLists(_displayListIdentificator, 1);
            _displayListIdentificator = Gl.glGenLists(1);
            Gl.glNewList(_displayListIdentificator, Gl.GL_COMPILE);
            //Gl.glPolygonMode(Gl.GL_FRONT_AND_BACK, Gl.GL_FILL);            
            foreach (Interval interval in _definitionIntervals.GetIntervals())
            {
                //Gl.glColor3d(1, 1, 0);
                BuildCurvatureCylinder(scene, interval, 1, relativeQuality);
                if (_vParametrization is QuadraticV)
                {
                    //Gl.glColor3d(1, 0, 0);
                    BuildCurvatureCylinder(scene, interval, -1, relativeQuality);
                }
            }
            foreach (double d in _definitionIntervals.GetPlusValues()) 
            {
                BuildPoint(scene, d);
            } 
            Gl.glEndList();
        }

        /// <summary>
        /// Signum je v pripade LinearVP a LinearUP ignorovano.
        /// Vyznam ma u QuadraticVP.
        /// </summary>
        protected void BuildCurvatureCylinder(Scene scene, Interval interval, int sign, RelativeQuality relativeQuality)
        {
            try
            {
                List<Interval> tempI = GetClippingIntervals(scene, _simpleRulled, sign, interval);
                List<Interval> finalIntervals = Interval.Intersect(tempI, interval);

                List<Pair> points = new List<Pair>();
                double qualityKoeficient = (scene.GetBaseQuality() + relativeQuality) / 2.0;
                double maxAltitude = 0.1 / qualityKoeficient;
                double maxAngle = (180 - (15 / qualityKoeficient)) * Math.PI / 180;
                foreach (Interval finalInterval in finalIntervals)
                {
                    if (finalInterval.IsEmpty())
                        continue;
                    double size = finalInterval.Size();
                    double u;
                    points = new List<Pair>();
                    int steps = 10;
                    int side = 1;
                    for (int i = 0; i <= steps; i++)
                    {
                        u = finalInterval.GetA() + (i / (double)steps * size);
                        points.Add(GetPairCheckNan(side, u, size / steps, sign));
                        side = -1;
                    }

                    for (int i = steps - 2; i >= 0; i -= 2)
                    {
                        GetPoints(maxAltitude, maxAngle, points, i, sign);
                    }

                    for (int i = 1; i < points.Count; i++)
                    {
                        BuildCylinder(points[i - 1].coordinates, points[i].coordinates);
                    }
                    Gl.glPushMatrix();
                    Vector v = points[points.Count - 1].coordinates;
                    Gl.glTranslated(v.GetX(), v.GetY(), v.GetZ());
                    _curveSphere.MakeDisplayList();
                    Gl.glPopMatrix();
                }
            }
            catch (Exception) 
            {
            }
        }

        /// <summary>
        /// Mám hoddnotu U, která je mezi sousednímy dvěma U.
        /// Prostou iterací přes UMin + i*step, můžu dostat hodnotu U v níž není sekce křivky definovaná.
        /// Proto zkusím v okolí zkoumaného U v několika iteracích najít jiné U.
        /// - Rozdelením definicniho oboru minus hodnotami neni dobre protože:
        /// - pokud vezmu lehce mensi a lehce vetsi hodnotu, pak vzdy vznikne mezi krivkami pripadajicim temto podintervalum mezera, muze byt i velka.
        /// - pokud vezmu lehce vetsi a lehce mensi, pak je porad minus hodnota v techto podintervalech. 
        /// </summary>
        private Pair GetPairCheckNan(int side, double u, double size, int sign)
        {
            int iterations = 10; // maximalni pocet minus hodnot je 4, ja zkusim 10
            double delta = size / iterations * side;
            double localU;
            Vector temp;
            //int j = 0;
            for (int j = 0; j < iterations; j++)
            {
                localU = u + delta * j;
                temp = GetVertexAtConvertDiskrininantNegativity(localU, sign);
                if (!temp.ContainsNaN())
                    return new Pair(localU, temp);
            }
            u = u + size * side;
            return new Pair(u, GetVertexAt(u, sign));
        }


        private void GetPoints(double maxAltitude, double maxAngleRadians, List<Pair> points, int position, int sign) 
        {
            Vector a = points[position].coordinates;
            Vector b = points[position + 1].coordinates;
            Vector c = points[position + 2].coordinates;
            double angle = GetAngle(a, b, c);
            double altitude = GetAltitudeFromB(a, b, c);
            double distance = Vector.Distance(a, c);
            double middleValue;
            if (altitude > maxAltitude ||
                (Vector.Distance(a, c) > 1 && altitude > maxAltitude / 2.0) ||
                (Vector.Distance(a, c) > 5 && altitude > maxAltitude / 8.0))
            {
                middleValue = (points[position + 1].u + points[position + 2].u) / 2;
                points.Insert(position + 2, new Pair(middleValue, GetVertexAtConvertDiskrininantNegativity(middleValue, sign)));
                GetPoints(maxAltitude, maxAngleRadians, points, position + 1, sign); // nutno v tomto poradi

                middleValue = (points[position].u + points[position + 1].u) / 2;
                points.Insert(position + 1, new Pair(middleValue, GetVertexAtConvertDiskrininantNegativity(middleValue, sign)));
                GetPoints(maxAltitude, maxAngleRadians, points, position, sign); // jinak toto volani posune position
            }
            else
            {
                double distance1 = Vector.Distance(b, c);
                double distance2 = Vector.Distance(a, b);
                if (double.IsNaN(angle))
                    return;
                if (angle < maxAngleRadians && distance1 > 1 && distance1 < 30)
                {
                    middleValue = (points[position + 1].u + points[position + 2].u) / 2;
                    points.Insert(position + 2, new Pair(middleValue, GetVertexAtConvertDiskrininantNegativity(middleValue, sign)));
                    GetPoints(maxAltitude, maxAngleRadians, points, position + 1, sign); // nutno v tomto poradi
                }
                if (angle < maxAngleRadians && distance2 > 1 && distance2 < 30)
                {
                    middleValue = (points[position].u + points[position + 1].u) / 2;
                    points.Insert(position + 1, new Pair(middleValue, GetVertexAtConvertDiskrininantNegativity(middleValue, sign)));
                    GetPoints(maxAltitude, maxAngleRadians, points, position, sign); // jinak toto volani posune position
                }
            }
        }
        private double GetAngle(Vector a, Vector b, Vector c) 
        {
            Vector v1 = b - a;
            Vector v2 = b - c;
            double angle = Vector.Angle(v1, v2);
            return angle;
        }
        private double GetAltitudeFromB(Vector a, Vector b, Vector c) 
        {
            Vector v1 = c - a;
            Vector v2 = b - a;
            double angle = Vector.Angle(v1, v2);
            double B = Vector.Size(v2);
            return Math.Sin(angle) * B;
        }
        private void BuildCylinder(Vector previousCoords, Vector currentCoords)
        {
            Vector direction = currentCoords - previousCoords;
            Vector x = new Vector(1, 0, 0);
            Vector axe = Vector.CrossProduct(direction, x).NornVector();
            double angle = Vector.Angle(direction, x) * 180 / Math.PI;
            TransformationMatrix rotation = TransformationMatrix.GetRotationMatrixAroundAAxe(axe, angle);
            Gl.glPushMatrix();
            Gl.glTranslated(previousCoords.GetX(), previousCoords.GetY(), previousCoords.GetZ());
            _curveSphere.MakeDisplayList();
            Gl.glMultMatrixd(rotation.ToDoubleArray());
            _curveCylinder.MakeCylinder(previousCoords, currentCoords);
            Gl.glPopMatrix();
        }
        private List<Interval> GetClippingIntervals(Scene scene, SimpleRulledQuadric simpleRuled, int sign, Interval interval) 
        {
            if (_vParametrization is QuadraticV)
                return new QuadraticClipManager(this, scene, sign, interval, _definitionIntervals.GetMinusValues()).GetParameterIntervals();
            if (_vParametrization is LinearV)
                return new LinearClipManager(this, scene, interval, _definitionIntervals.GetMinusValues()).GetParameterIntervals();
            if (_vParametrization is ConstantU)
                return new ConstantUClipManager(this, scene, interval, _definitionIntervals.GetMinusValues()).GetParameterIntervals();
            if (_vParametrization is ConstantV)
                return new ConstantVClipManager(this, scene, interval, _definitionIntervals.GetMinusValues()).GetParameterIntervals();
            throw new Exception();
        }



        public override void DrawAsIntersection(Scene scene)
        {
            try
            {
                Gl.glPushMatrix();
                Gl.glMultMatrixd(_simpleRulled.GetToCanonicTransformationMatrix().ToDoubleArray());
                Gl.glCallList(_displayListIdentificator);
                Gl.glPopMatrix();
            }
            catch (Exception e) 
            {
                ManipulationManager.FailOccured(e);
            }
        }
        public override void Draw(Scene scene)
        {
            DrawAsIntersection(scene);
        }
        public override void DrawIgnoreVisibilityAndAxes(Scene scene)
        {
            DrawAsIntersection(scene);
        }

        private Interval GetU(Interval x, Interval y, Interval z)
        {
            if (_simpleRulled is HyperbolicParaboloid)
                return GetU((HyperbolicParaboloid)_simpleRulled, x, y, z);
            if (_simpleRulled is HyperbolicCylinder)
                return GetU((HyperbolicCylinder)_simpleRulled, x, y, z);
            if (_simpleRulled is ParabolicCylinder)
                return GetU((ParabolicCylinder)_simpleRulled, x, y, z);
            if (_simpleRulled is CrossingPlans)
                return GetU((CrossingPlans)_simpleRulled, x, y, z);
            if (_simpleRulled is ParallelPlans)
                return GetU((ParallelPlans)_simpleRulled, x, y, z);
            if (_simpleRulled is DoublePlan)
                return GetU((DoublePlan)_simpleRulled, x, y, z);
            if (_simpleRulled is ImaginarCrossingPlans)
                return GetU((ImaginarCrossingPlans)_simpleRulled, x, y, z);
            throw new Exception("Neznámá simpleRulled");
        }
        private Interval GetU(HyperbolicParaboloid simpleRulled, Interval x, Interval y, Interval z)
        {
            double a = simpleRulled.GetParametricEquation().GetA();
            double b = simpleRulled.GetParametricEquation().GetB();
            double p = simpleRulled.GetParametricEquation().GetP();

            double f1 = p * (a * y.GetA() - b * x.GetA());
            double f2 = p * (a * y.GetA() - b * x.GetB());
            double f3 = p * (a * y.GetB() - b * x.GetA());
            double f4 = p * (a * y.GetB() - b * x.GetB());

            double min = Math.Min(f1, Math.Min(f2, Math.Min(f3, f4)));
            double max = Math.Max(f1, Math.Max(f2, Math.Max(f3, f4)));

            double d1 = a * b * z.GetA();
            double d2 = a * b * z.GetB();

            if (Program.IsEqualInTolerance(min, 0) || Program.IsEqualInTolerance(max, 0))
            {
                if (a < b)
                    return y / b;
                else return x / a;
            }

            return new Interval(-30, 30);
            //return new Interval(d1 / max, d2 / min);
        }
        private Interval GetU(HyperbolicCylinder simpleRulled, Interval x, Interval y, Interval z)
        {
            return y;
        }
        private Interval GetU(ParabolicCylinder simpleRulled, Interval x, Interval y, Interval z)
        {
            return x;
        }
        private Interval GetU(CrossingPlans simpleRulled, Interval x, Interval y, Interval z)
        {
            double a = simpleRulled.GetParametricEquation().GetA();
            double b = _parametrizationVectorInCanonicBasis.GetYParametrization().GetValueAt(1);
            Interval x1 = a * x;
            Interval y1 = b * y;
            double umin = Math.Max(x1.GetA(), y1.GetA());
            double umax = Math.Min(x1.GetB(), y1.GetB());
            return new Interval(umin, umax);
        }
        private Interval GetU(ParallelPlans simpleRulled, Interval x, Interval y, Interval z)
        {
            return y;
        }
        private Interval GetU(DoublePlan simpleRulled, Interval x, Interval y, Interval z)
        {
            return y;
        }
        private Interval GetU(ImaginarCrossingPlans simpleRulled, Interval x, Interval y, Interval z)
        {
            return z;
        }

        private void Test(Scene scene)
        {
            Test(scene, 1);
            if (_vParametrization is QuadraticV)
                Test(scene, -1);
        }
        private void Test(Scene scene, int sign)
        {
            try
            {
                Interval x = new Interval(_simpleRulled.GetClippingLeft(scene), _simpleRulled.GetClippingRight(scene));
                Interval y = new Interval(_simpleRulled.GetClippingBottom(scene), _simpleRulled.GetClippingTop(scene));
                Interval z = new Interval(_simpleRulled.GetClippingBack(scene), _simpleRulled.GetClippingFront(scene));

                double delta = 5;
                Interval x2 = new Interval(_simpleRulled.GetClippingLeft(scene) + delta, _simpleRulled.GetClippingRight(scene) - delta);
                Interval y2 = new Interval(_simpleRulled.GetClippingBottom(scene) + delta, _simpleRulled.GetClippingTop(scene) - delta);
                Interval z2 = new Interval(_simpleRulled.GetClippingBack(scene) + delta, _simpleRulled.GetClippingFront(scene) - delta);
                Interval uInterval = GetU(x, y, z);
                double step = 0.01;
                List<Interval> clipIntervals = new List<Interval>();
                List<Interval> temp;
                foreach (Interval defInterval in _definitionIntervals.GetIntervals()) // pro každej definiční interval spočítám clipintervals s sezbírám je.
                {
                    temp = GetClippingIntervals(scene, _simpleRulled, sign, defInterval);
                    foreach (Interval i in temp)
                        clipIntervals.Add(i);
                }
                double min = double.MaxValue;
                double max = double.MinValue;
                for (int i = 0; i < clipIntervals.Count; i++)
                {
                    min = Math.Min(min, clipIntervals[i].GetA());
                    max = Math.Max(max, clipIntervals[i].GetB());
                }
                Vector v;
                double d;
                for (double u = uInterval.GetA(); u < uInterval.GetB(); u += step)
                {
                    d = _vParametrization.GetVvalue(u, sign);
                    if (!double.IsNaN(d)
                        && !BelongsToIntervals(u, _definitionIntervals.GetIntervals())
                        && !double.IsNaN(_vParametrization.GetVvalue(u + 0.1, sign))
                        && !double.IsNaN(_vParametrization.GetVvalue(u - 0.1, sign)))
                    {
                        //FailOccured();
                        _vParametrization.GetVvalue(u, sign);
                    }
                    v = GetVertexAt(u, sign);
                    if (!double.IsNaN(v.GetX()) && !double.IsNaN(v.GetY()) && !double.IsNaN(v.GetZ())
                        && x2.Belongs(v.GetX()) && y2.Belongs(v.GetY()) && z2.Belongs(v.GetZ())
                        && !BelongsToIntervals(u, clipIntervals)
                        && BelongsToIntervals(u, _definitionIntervals.GetIntervals()))
                    {
                        //FailOccured(); 
                    }
                }
            }
            catch (Exception e) 
            {
                string s = e.Message;
            }
        }

        private bool BelongsToIntervals(double u, List<Interval> intervals)
        {
            foreach (Interval i in intervals)
                if (i.Belongs(u))
                    return true;
            return false;
        }

        protected void BuildPoint(Scene scene, double d)
        {
            Vector coordinates = GetVertexAtConvertDiskrininantNegativity(d, 1);
            if (coordinates.ContainsNaN())
                return;
            Gl.glPushMatrix();
            Gl.glTranslated(coordinates.GetX(), coordinates.GetY(), coordinates.GetZ());
            _curveSphere.MakeDisplayList();
            Gl.glPopMatrix();
        }
        private void DrawCube()
        {
            double a = 30 * 4;

            Gl.glBegin(Gl.GL_LINE_STRIP);
            Gl.glVertex3d(-a, a, a);
            Gl.glVertex3d(a, a, a);
            Gl.glVertex3d(a, -a, a);
            Gl.glVertex3d(-a, -a, a);
            Gl.glVertex3d(-a, a, a);

            Gl.glVertex3d(-a, a, -a);
            Gl.glVertex3d(a, a, -a);
            Gl.glVertex3d(a, -a, -a);
            Gl.glVertex3d(-a, -a, -a);
            Gl.glVertex3d(-a, a, -a);
            Gl.glEnd();

            Gl.glBegin(Gl.GL_LINES);
            Gl.glVertex3d(-a, -a, a);
            Gl.glVertex3d(-a, -a, -a);

            Gl.glVertex3d(a, a, a);
            Gl.glVertex3d(a, a, -a);

            Gl.glVertex3d(a, -a, a);
            Gl.glVertex3d(a, -a, -a);
            Gl.glEnd();
        }
    }
}
