﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Geometry.GraphicObjects.Quadrics;
using Geometry.GraphicObjects.Equations;
using GMath;
using Geometry.GraphicObjects.Matrixes;

namespace Geometry.GraphicObjects.Intersections
{
    public class QuadricsSimpleRuledParametrizations
    {
        public QuadricsSimpleRuledParametrizations() { }


        public List<ParametrizationVector> GetParametrization(Quadric quadric)
        {
            if (quadric is ImaginarCrossingPlans)
                return GetParametrization((ImaginarCrossingPlans)quadric);
            if (quadric is HyperbolicParaboloid)
                return GetParametrization((HyperbolicParaboloid)quadric);
            if (quadric is DoublePlan)
                return GetParametrization((DoublePlan)quadric);
            if (quadric is ParallelPlans)
                return GetParametrization((ParallelPlans)quadric);
            if (quadric is CrossingPlans)
                return GetParametrization((CrossingPlans)quadric);
            if (quadric is HyperbolicCylinder)
                return GetParametrization((HyperbolicCylinder)quadric);
            if (quadric is ParabolicCylinder)
                return GetParametrization((ParabolicCylinder)quadric);
            return null;
        }

        /// <summary>
        /// (0,0,u)
        /// </summary>
        private List<ParametrizationVector> GetParametrization(ImaginarCrossingPlans line)
        {
            Expression xParametrization = new Expression(0);
            Expression yParametrization = new Expression(0);
            Expression zParametrization = new Expression(new UV(1, 1, 0));
            Expression oneParametrization = new Expression(1);
            return new List<ParametrizationVector>()
            {
                new ParametrizationVector(xParametrization, yParametrization,zParametrization, oneParametrization)
            };
        }
        /// <summary>
        /// (a(u+v),b(u-v),-sign(p)uv)
        /// </summary>
        private List<ParametrizationVector> GetParametrization(HyperbolicParaboloid hyperbolicParaboloid)
        {
            QuadricParametricEquation canonicEquation = hyperbolicParaboloid.GetCanonicEquation();
            double constantA = GetConstantA(hyperbolicParaboloid);
            double constantB = GetConstantB(hyperbolicParaboloid);
            double p = canonicEquation.GetP();
            Expression xParametrization = new Expression(new UV(constantA, 1, 0), new UV(constantA, 0, 1)); //au+av
            Expression yParametrization = new Expression(new UV(constantB, 1, 0), new UV(-constantB, 0, 1)); //bu-bv 
            Expression zParametrization = new Expression(new List<UV>{new UV(Math.Sign(-p), 1, 1)});
            Expression oneParametrization = new Expression(new UV(1, 0, 0));
            return new List<ParametrizationVector>()
            {
                new ParametrizationVector(xParametrization, yParametrization,zParametrization, oneParametrization)
            };
        }
        /// <summary>
        /// (0,u,v)
        /// </summary>
        private List<ParametrizationVector> GetParametrization(DoublePlan doublePlan)
        {
            Expression xParametrization = new Expression(new UV(0, 0, 0));
            Expression yParametrization = new Expression(new UV(1, 1, 0));
            Expression zParametrization = new Expression(new UV(1, 0, 1));
            Expression oneParametrization = new Expression(new UV(1, 0, 0));
            return new List<ParametrizationVector>()
            {
                new ParametrizationVector(xParametrization, yParametrization,zParametrization, oneParametrization)
            };
        }
        /// <summary>
        /// (a,u,v), (-a,u,v)
        /// </summary>
        private List<ParametrizationVector> GetParametrization(ParallelPlans parallelPlans)
        {
            QuadricParametricEquation canonicEquation = parallelPlans.GetCanonicEquation();
            double a = canonicEquation.GetA();
            Expression xParametrization1 = new Expression(new UV(a, 0, 0));
            Expression yParametrization1 = new Expression(new UV(1, 1, 0));
            Expression zParametrization1 = new Expression(new UV(1, 0, 1));
            Expression oneParametrization1 = new Expression(new UV(1, 0, 0));

            Expression xParametrization2 = new Expression(new UV(-a, 0, 0));
            Expression yParametrization2 = new Expression(new UV(1, 1, 0));
            Expression zParametrization2 = new Expression(new UV(1, 0, 1));
            Expression oneParametrization2 = new Expression(new UV(1, 0, 0));
            return new List<ParametrizationVector>()
            {
                new ParametrizationVector(xParametrization1, yParametrization1, zParametrization1, oneParametrization1),
                new ParametrizationVector(xParametrization2, yParametrization2, zParametrization2, oneParametrization2)
            };
        }
        /// <summary>
        /// (au,bu,v), (au,-bu,v)
        /// </summary>
        private List<ParametrizationVector> GetParametrization(CrossingPlans crossingPlans)
        {
            QuadricParametricEquation canonicEquation = crossingPlans.GetCanonicEquation();
            double a = canonicEquation.GetA();
            double b = canonicEquation.GetB();
            Expression xParametrization1 = new Expression(new UV(a, 1, 0));
            Expression yParametrization1 = new Expression(new UV(b, 1, 0));
            Expression zParametrization1 = new Expression(new UV(1, 0, 1));
            Expression oneParametrization1 = new Expression(new UV(1, 0, 0));

            Expression xParametrization2 = new Expression(new UV(a, 1, 0));
            Expression yParametrization2 = new Expression(new UV(-b, 1, 0));
            Expression zParametrization2 = new Expression(new UV(1, 0, 1));
            Expression oneParametrization2 = new Expression(new UV(1, 0, 0));
            return new List<ParametrizationVector>()
            {
                new ParametrizationVector(xParametrization1, yParametrization1, zParametrization1, oneParametrization1),
                new ParametrizationVector(xParametrization2, yParametrization2, zParametrization2, oneParametrization2)
            };
        }
        /// <summary>
        /// (2*a(u+1/u), b(u-1/u),v)
        /// </summary>
        private List<ParametrizationVector> GetParametrization(HyperbolicCylinder hyperbolicCylinder)
        {
            QuadricParametricEquation canonicEquation = hyperbolicCylinder.GetCanonicEquation();
            double a = canonicEquation.GetA() / 2.0;
            double b = canonicEquation.GetB() / 2.0;
            //double c = Math.Sqrt(2); 
            Expression xParametrization = new Expression(new UV(a, 1, 0), new UV(a, -1, 0));
            Expression yParametrization = new Expression(new UV(b, 1, 0), new UV(-b, -1, 0));
            Expression zParametrization = new Expression(new UV(1, 0, 1));
            Expression oneParametrization = new Expression(new UV(1, 0, 0));

            return new List<ParametrizationVector>()
            {
                new ParametrizationVector(xParametrization, yParametrization, zParametrization, oneParametrization)
            };
        }
        /// <summary>
        /// (u,-u2/a2*2*p,v)
        /// </summary>
        private List<ParametrizationVector> GetParametrization(ParabolicCylinder parabolicCylinder)
        {
            QuadricParametricEquation canonicEquation = parabolicCylinder.GetCanonicEquation();
            double a = canonicEquation.GetA();
            double p = canonicEquation.GetP();
            Expression xParametrization = new Expression(new UV(1, 1, 0));
            Expression yParametrization = new Expression(new UV(-1 / GetConstantA(parabolicCylinder), 2, 0));
            Expression zParametrization = new Expression(new UV(1, 0, 1));
            Expression oneParametrization = new Expression(new UV(1, 0, 0));
            return new List<ParametrizationVector>()
            {
                new ParametrizationVector(xParametrization, yParametrization, zParametrization, oneParametrization),
            };
        }


        public static double GetConstantA(HyperbolicParaboloid hp) 
        {
            double a = hp.GetParametricEquation().GetA();
            double p = hp.GetParametricEquation().GetP();
            return Math.Sqrt(Math.Abs(a * a * p * 2)) / 2.0;
        }
        public static double GetConstantB(HyperbolicParaboloid hp)
        {
            double b = hp.GetParametricEquation().GetB();
            double p = hp.GetParametricEquation().GetP();
            return Math.Sqrt(Math.Abs(b * b * p * 2)) / 2.0;
        }

        public static double GetConstantA(ParabolicCylinder pc) 
        {
            double a = pc.GetParametricEquation().GetA();
            double p = pc.GetParametricEquation().GetP();
            return (a * a * 2 * p);
        }



        public static DoublePlan GetPlanFromParametrizationOfCompoundQuadric(ParametrizationVector pv, Quadric quadric)
        {
            if (quadric is CrossingPlans)
                return GetPlanFromParametrizationOfCompoundQuadric(pv, (CrossingPlans)quadric);
            if (quadric is ParallelPlans)
                return GetPlanFromParametrizationOfCompoundQuadric(pv, (ParallelPlans)quadric);
            if (quadric is DoublePlan)
                return (DoublePlan)quadric;
            return null;
        }
        /// <summary>
        /// x =   au;
        /// y = +-bu;
        /// </summary>
        private static DoublePlan GetPlanFromParametrizationOfCompoundQuadric(ParametrizationVector pv, CrossingPlans quadric)
        {
            Expression x = (Expression)pv.GetXParametrization();
            Expression y = (Expression)pv.GetYParametrization();
            double a = x.GetValueAt(1, 1);
            double b = y.GetValueAt(1, 1);
            double angle = -Math.Atan2(a, b) * 180.0 / Math.PI;
            TransformationMatrix matrix = TransformationMatrix.GetZRotationMatrix(angle);
            TransformationMatrix simpleTr = quadric.GetTransformationMatrixAlgebraic();
            matrix = simpleTr * matrix;
            QuadricEquation commonEquation = new QuadricEquation(
                new QuadricMatrix(new QuadricEquation(1, 0, 0, 0, 0, 0, 0, 0, 0, 0)).ApplyTransformation(matrix));
            return new DoublePlan(commonEquation, new QuadricParametricEquation(1, 0, 0, 0),
                matrix.ComputeInverseMatrix(), quadric.GetColor(), quadric.GetName(), quadric.GetID());

        }
        /// <summary>
        /// x = +-a;
        /// y =   u;
        /// </summary>
        private static DoublePlan GetPlanFromParametrizationOfCompoundQuadric(ParametrizationVector pv, ParallelPlans quadric)
        {
            Expression x = (Expression)pv.GetXParametrization();
            Expression y = (Expression)pv.GetYParametrization();
            double a = x.GetValueAt(1, 1);
            TransformationMatrix matrix = new TransformationMatrix(4);
            matrix.SetTranslation(new Vector(-a, 0, 0));
            TransformationMatrix simpleTr = quadric.GetTransformationMatrixAlgebraic();
            matrix = simpleTr * matrix;
            QuadricEquation commonEquation = new QuadricEquation(
                new QuadricMatrix(new QuadricEquation(1, 0, 0, 0, 0, 0, 0, 0, 0, 0)).ApplyTransformation(matrix));
            return new DoublePlan(commonEquation, new QuadricParametricEquation(1, 0, 0, 0),
                matrix.ComputeInverseMatrix(), quadric.GetColor(), quadric.GetName(), quadric.GetID());
        }
    }
}
