﻿using System;
using Tao.OpenGl;
using Geometry.GraphicObjects.Equations;
using Geometry.GGraphic.Forms.Detail.FormatString;
using Geometry.BaseClasses;
using GMath;

namespace Geometry.GraphicObjects.Planar
{
    /// <summary>
    /// canonicky tvar y = 0, tedy osa x.
    /// </summary>
    public class Line : WithParametricEquationObject
    {
        protected static CurveCylinder _curveCylinder = new CurveCylinder();
        private Plan _linePlan;
        public static string _parametricXFormatString = "L[x = !m + !u" + U.Index1Char + U.Times + "t]";
        public static string _parametricYFormatString = "L[y = !n + !u" + U.Index2Char + U.Times + "t]";
        public static string _parametricZFormatString = "L[y = !o + !u" + U.Index3Char + U.Times + "t]";

        public static string _canonicParametricXFormatString = "L[x = 1" + U.Times + "t]";
        public static string _canonicParametricYFormatString = "L[y = 0]";
        public static string _canonicParametricZFormatString = "L[y = 0]";

        public static string _wordDescriptionFirstCase = "Přímka";
        public static string _wordDescriptionSecondCase = "Přímky";
        public static string _wordDescriptionFourthCase = "Přímku";

        public Line(LineParametricEquation parametricEquation,
            CenterParametrization centerInSceneBasis, GColor color, string name, int ID)
        {
            base._color = color;
            base._name = name;
            base._ID = ID;
            base._parametricEquation = parametricEquation;
            base.SetWordDescriptionFirstCase(_wordDescriptionFirstCase);
            base.SetWordDescriptionSecondCase(_wordDescriptionSecondCase);
            base.SetWordDescriptionFourthCase(_wordDescriptionFourthCase);
            base._parametricEquationFormatString = new ParametricEquationsFormatString(_parametricXFormatString,
                _parametricYFormatString, _parametricZFormatString);
            _linePlan = new Plan(color, name, ID);
            SetTransformationMatrixAlgebraic(GetTransformationMatrixAlgebraic(
                ComputeRotationMatrix((LineParametricEquation)_parametricEquation), centerInSceneBasis));            
        }

        /// <summary>
        /// Primka v rovine z = 0;
        /// </summary>
        public Line(LineEquation commonEquation, GColor color, string name, int ID) 
        {
            CenterParametrization centerParametrization = GetCenterFromCommonEquation(commonEquation);
            _parametricEquation = 
                GetParametricEquationFromCommonEquation(commonEquation, centerParametrization.GetReprezentant());
            _color = color;
            _name = name;
            base._parametricEquationFormatString = new ParametricEquationsFormatString(_parametricXFormatString,
                _parametricYFormatString, _parametricZFormatString);
            _linePlan = new Plan(color, name, ID);
            SetTransformationMatrixAlgebraic(GetTransformationMatrixAlgebraic(
                ComputeRotationMatrix((LineParametricEquation)_parametricEquation), centerParametrization));
        }

        public Plan GetLinePlan() { return _linePlan; }
        public void ApplyTransformation(TransformationMatrix transformation)
        {
            base.SetTransformationMatrixAlgebraic(transformation * _transformationAlgebraic);
        }
        /// <summary>
        /// Tato rovina je otocena i posunuta.
        /// Normálně tedy při jednotkové rotační matici, je rovinou, rovina Z=0.
        /// </summary>
        private void ActualizeLinePlan(TransformationMatrix transformationMatrixAlgebraic)
        {
            _linePlan.SetTransformationMatrixAlgebraic(transformationMatrixAlgebraic);
        }
        private CenterParametrization GetCenterFromCommonEquation(LineEquation equation) 
        {
            if (!Program.IsEqualInTolerance(equation.GetA(), 0))
            {
                return new CenterParametrization(-equation.GetC() / equation.GetA(), 0, 0);
            }
            else {
                return new CenterParametrization(0, -equation.GetC() / equation.GetB(), 0);
            }
        }
        public LineParametricEquation GetParametricEquationFromCommonEquation(LineEquation equation, Vector center) 
        {
            Vector direction = null;
            Vector pointOnLine;

            if (!Program.IsEqualInTolerance(equation.GetA(), 0))
            {
                Vector x = new Vector(1, 0, 0);
                pointOnLine = new Vector(-(equation.GetC() + equation.GetB()) / equation.GetA(), 1, 0);
                direction = pointOnLine - center;
            }
            else 
            {
                Vector y = new Vector(0, 1, 0);
                pointOnLine = new Vector(1, -(equation.GetC() +equation.GetA()) / equation.GetB(), 0);
                direction = pointOnLine - center;
            }
            return new LineParametricEquation(direction.GetX(), direction.GetY(), direction.GetZ());
        }
        private TransformationMatrix ComputeRotationMatrix(LineParametricEquation parametricEquation) 
        {
            //throw new Exception("Potreba dodelat. objevuji se hodnoty intMax");
            Vector u = new Vector(parametricEquation.GetUx(), parametricEquation.GetUy(), parametricEquation.GetUz());
            Vector x = new Vector(1, 0, 0);
            double alpha = Vector.Angle(x, u);
            Vector v = Vector.CrossProduct(x, u);
            if (v.IsZero())
                return new TransformationMatrix(4);
            return TransformationMatrix.GetRotationMatrixAroundAAxe(v.NornVector(), alpha * 180 / Math.PI);
        }

        public void BuildAsIntersection(Scene scene, RelativeQuality relativeQuality)
        {
            _displayListIdentificator = Gl.glGenLists(1);
            Gl.glNewList(_displayListIdentificator, Gl.GL_COMPILE);

            double left = GetClippingLeft(scene);
            double right = GetClippingRight(scene);
            BuildLineAsCylinder(new Vector(left, 0, 0), new Vector(right, 0, 0));
            Gl.glEndList();
        }
        protected override void BuildParametricDisplayList(Scene scene, RelativeQuality relativeQuality)
        {
            _displayListIdentificator = Gl.glGenLists(1);
            Gl.glNewList(_displayListIdentificator, Gl.GL_COMPILE);

            double left = GetClippingLeft(scene);
            double right = GetClippingRight(scene);
            BuildLineAsLineStrip(new Vector(left, 0, 0), new Vector(right, 0, 0));
            Gl.glEndList();
        }

        public static void BuildLineAsCylinder(Vector a, Vector b) 
        {
            int parts = 20;
            Vector delta = (b - a) / parts;
            Vector accum = a;
            //Gl.glPolygonMode(Gl.GL_FRONT_AND_BACK, Gl.GL_FILL);            
            for (int i = 0; i < parts; i++)
            {
                _curveCylinder.MakeCylinderComplete(accum, accum + delta);
                accum = accum + delta;
            }
        }
        public static void BuildLineAsLineStrip(Vector a, Vector b)
        {
            Gl.glDisable(Gl.GL_LIGHTING);
            Gl.glBegin(Gl.GL_LINES);
            Gl.glVertex3d(a.GetX(), a.GetY(), a.GetZ());
            Gl.glVertex3d(b.GetX(), b.GetY(), b.GetZ());
            Gl.glEnd();
            Gl.glEnable(Gl.GL_LIGHTING);
        }


        protected override void ActualizeEquations()
        {
            Vector v = new Vector(1, 0, 0, 1) * GetRotationInStandardBasis().ComputeInverseMatrix();
            _parametricEquation = new LineParametricEquation(v.GetX(), v.GetY(), v.GetZ());
            _parametricEquationFormatString.SubstituteToFormatStrings(_parametricEquation, GetCenterInStandardBasis());
            ActualizeLinePlan(GetTransformationMatrixAlgebraic());
        }
        public override WithParametricEquationObject MakeCopyOfObject(KoefsAndCenterPair newKoefsAndCenter, int ID)
        {
            LineParametricEquation equation = (LineParametricEquation)newKoefsAndCenter.GetParametricEquation();
            CenterParametrization center = newKoefsAndCenter.GetCenterParametrization();
            Line line = new Line(equation, center, this.GetColor(),this.GetName(), ID);
            line.CopyDrawableSlots(line, this);
            return line;
        }
        public override WithParametricEquationObject MakeCopyOfObject(int ID)
        {
            LineParametricEquation equationCopy =
                new LineParametricEquation((LineParametricEquation)_parametricEquation);
            CenterParametrization centerCopy =
                new CenterParametrization(GetCenterInStandardBasis());
            return MakeCopyOfObject(new KoefsAndCenterPair(equationCopy, centerCopy), ID);
        }
        public override bool AreEqual(GraphicObject secondObject)
        {
            LineParametricEquation firstLineParametricEquation = (LineParametricEquation)this._parametricEquation;
            LineParametricEquation secondLineParametricEquation = (LineParametricEquation)((Line)secondObject)._parametricEquation;

            double koficient = 0;

            if (!Program.IsEqualInTolerance(firstLineParametricEquation.GetUx(), 0))
                koficient = secondLineParametricEquation.GetUx()/firstLineParametricEquation.GetUx();
            else if (!Program.IsEqualInTolerance(firstLineParametricEquation.GetUy(), 0))
                koficient = secondLineParametricEquation.GetUy()/firstLineParametricEquation.GetUy();
            if (!Program.IsEqualInTolerance(firstLineParametricEquation.GetUz(), 0))
                koficient = secondLineParametricEquation.GetUz()/firstLineParametricEquation.GetUz();

            return
                Program.IsEqualInTolerance(firstLineParametricEquation.GetUx() * koficient, secondLineParametricEquation.GetUx()) &&
                Program.IsEqualInTolerance(firstLineParametricEquation.GetUy() * koficient, secondLineParametricEquation.GetUy()) &&
                Program.IsEqualInTolerance(firstLineParametricEquation.GetUz() * koficient, secondLineParametricEquation.GetUz());
        }

    }
}
