﻿using System;
using Tao.OpenGl;
using Geometry.GraphicObjects.Equations;
using Geometry.GGraphic.Forms.Detail.FormatString;
using Geometry.BaseClasses;
using GMath;

namespace Geometry.GraphicObjects.Planar
{
    /// <summary>
    /// Rovnice roviny v prostoru:
    ///     ax+by+cz+d=0;
    /// parametricka rovnice:
    ///     x = u;
    ///     y = v;
    ///     z = -(a*u + b*v +d)/c;
    /// </summary>
    public class Plan:WithCommonEquationObject
    {

        public static string _planCommonFormatString = "L[!Ax + !By + !Cz + !D = 0]";
        public static string _parametricXFormatString = "L[x = !m + !u" + U.Index1Char + U.Times + "s + !v" + U.Index1Char + U.Times + "t]";
        public static string _parametricYFormatString = "L[y = !n + !u" + U.Index2Char + U.Times + "s + !v" + U.Index2Char + U.Times + "t]";
        public static string _parametricZFormatString = "L[z = !o + !u" + U.Index3Char + U.Times + "s + !v" + U.Index3Char + U.Times + "t]";

        public static string _planCanonicCommonFormatString = "L[z = 0]";
        public static string _canonicParametricXFormatString = "L[x = 1" + U.Times + "s + 0" + U.Times + "t]";
        public static string _canonicParametricYFormatString = "L[y = 0" + U.Times + "s + 1" + U.Times + "t]";
        public static string _canonicParametricZFormatString = "L[z = 0";

        public static string _wordDescriptionFirstCase = "Rovina";
        public static string _wordDescriptionSecondCase = "Roviny";
        public static string _wordDescriptionFourthCase = "Rovinu";

        private int _normalSign = 1;

        private static Vector canonicFirstDirectionVector = new Vector(1, 0, 0, 1);
        private static Vector canonicSecondDirectionVector = new Vector(0, 1, 0, 1);
        public static Vector canonicalNormalVector = new Vector(0, 0, -1, 1);

        /// <summary>
        /// Rovina z = 0;
        /// </summary>
        public Plan(GColor color, string name, int ID) 
        {
            base._color = color;
            base._name = name;
            base._ID = ID;
            base._commonEquation = new PlanEquation(0, 0, 1, 0);
            Vector center = new Vector();
            base._parametricEquation = GetParametricEquationFromCommonEquation((PlanEquation)_commonEquation, center);
            base.SetWordDescriptionFirstCase(_wordDescriptionFirstCase);
            base.SetWordDescriptionSecondCase(_wordDescriptionSecondCase);
            base.SetWordDescriptionFourthCase(_wordDescriptionFourthCase);
            base._commonEquationFormatSring = new CommonEquationFormatString(_planCommonFormatString);
            base._parametricEquationFormatString = new ParametricEquationsFormatString(_parametricXFormatString,
                _parametricYFormatString, _parametricZFormatString);
            SetTransformationMatrixAlgebraic(new TransformationMatrix(4));
            SetNormalSign();
        }

        public Plan(PlanEquation commonEquation, GColor color, string name, int ID) 
        {
            base._color = color;
            base._name = name;
            base._ID = ID;
            base._commonEquation = commonEquation;
            base._transformationAlgebraic = ComputeRotationMatrix((PlanEquation)_commonEquation);
            Vector center = GetCenterFromCommonEquation((PlanEquation)_commonEquation, _transformationAlgebraic);
            CenterParametrization centerInStandardBasis = new CenterParametrization(center);
            base._parametricEquation = GetParametricEquationFromCommonEquation((PlanEquation)_commonEquation, center);
            base.SetWordDescriptionFirstCase(_wordDescriptionFirstCase);
            base.SetWordDescriptionSecondCase(_wordDescriptionSecondCase);
            base.SetWordDescriptionFourthCase(_wordDescriptionFourthCase);
            base._commonEquationFormatSring = new CommonEquationFormatString(_planCommonFormatString);
            base._parametricEquationFormatString = new ParametricEquationsFormatString(_parametricXFormatString,
                _parametricYFormatString, _parametricZFormatString);
            SetTransformationMatrixAlgebraic(GetTransformationMatrixAlgebraic(_transformationAlgebraic, centerInStandardBasis));            
            SetNormalSign();
        }

        public Plan(PlanParametricEquation parametricEquation, CenterParametrization centerInSceneBasis,
            GColor color, string name, int ID)
        {
            base._color = color;
            base._name = name;
            base._ID = ID;
            base._parametricEquation = parametricEquation;
            base._commonEquation = GetCommonEquationFromParametricEquation(parametricEquation, centerInSceneBasis);
            base.SetWordDescriptionFirstCase(_wordDescriptionFirstCase);
            base.SetWordDescriptionSecondCase(_wordDescriptionSecondCase);
            base.SetWordDescriptionFourthCase(_wordDescriptionFourthCase);
            base._commonEquationFormatSring = new CommonEquationFormatString(_planCommonFormatString);
            base._parametricEquationFormatString = new ParametricEquationsFormatString(_parametricXFormatString,
                _parametricYFormatString, _parametricZFormatString);
            SetTransformationMatrixAlgebraic(GetTransformationMatrixAlgebraic(
                ComputeRotationMatrix(parametricEquation), centerInSceneBasis));            
            SetNormalSign();
        }



        private void SetNormalSign()
        {
            double determinant = _transformationAlgebraic.Determinant4();
            _normalSign = Math.Sign(determinant);
        }

        private TransformationMatrix ComputeRotationMatrix(PlanParametricEquation pEquation) 
        {
            Vector x = new Vector(pEquation.GetUx(), pEquation.GetUy(), pEquation.GetUz());
            Vector y = new Vector(pEquation.GetVx(), pEquation.GetVy(), pEquation.GetVz());
            Vector z = Vector.CrossProduct(x ,y).NornVector();
            return new TransformationMatrix(x, y, z).ComputeInverseMatrix();
        }
        private TransformationMatrix ComputeRotationMatrix(PlanEquation commonEquation) 
        {
            Vector u = new Vector(commonEquation.GetA(), commonEquation.GetB(), commonEquation.GetC());
            double alpha = Vector.Angle(canonicalNormalVector, u) * 180 / Math.PI; // porovnamm normaly, najdu uhel ktery otoci normalu...
            Vector v = Vector.CrossProduct(canonicalNormalVector, u);
            if (v.IsZero())
                return new TransformationMatrix(4);
            // Rotace daná obecnou rovnicí roviny není jednoznačná.
            // Aby rovina nebyla zbytečně točená okolo její normály, využiju služeb ConicSection a rotaci kolem normály odtraním.
            TransformationMatrix rotation = TransformationMatrix.GetRotationMatrixAroundAAxe(v.NornVector(), alpha);
            ConicSections.ConicSection.Matrixes matrixes = ConicSections.ConicSection.DecomposeTransformationMatrix(rotation);
            return matrixes.planTransformation;
        }
        private Vector GetCenterFromCommonEquation(PlanEquation equation, TransformationMatrix rotation) 
        {
            Vector standardCenter;
            if (!Program.IsEqualInTolerance(equation.GetA(), 0))
            {
                standardCenter = new Vector(-equation.GetD() / equation.GetA(), 0, 0);
            }
            else if (!Program.IsEqualInTolerance(equation.GetB(), 0))
            {
                standardCenter = new Vector(0, -equation.GetD() / equation.GetB(), 0);
            }
            else 
            {
                standardCenter = new Vector(0, 0, -equation.GetD() / equation.GetC());
            }
            // aby rovina neobsahovala žádné posunutí ve směrech x a y vzhledem ke kanonické bázi roviny,
            // převedu stanovený střed vzhledem ke standardní bázi do kanonické báze roviny.
            // těmto souřadnicím středu potřebné posonutí znuluju a převedu opět do standardní báze.
            Vector canonicCenter = standardCenter * rotation;
            canonicCenter.SetX(0);
            canonicCenter.SetY(0);
            return canonicCenter * rotation.ComputeInverseMatrix();
        }
        /// <summary>
        /// Potřebuji najít tři body roviny.
        /// Dosadím tedy x a y a dopočítám z.
        /// Vektory AB a AC musí být nezávislé.
        /// </summary>
        private PlanParametricEquation GetParametricEquationFromCommonEquation(
            PlanEquation commonEquation, Vector center)
        {
            if (!Program.IsEqualInTolerance(commonEquation.GetC(),0))
                return CIsNotZero(commonEquation, center);
            if (!Program.IsEqualInTolerance(commonEquation.GetB(),0))
                return BIsNotZero(commonEquation, center);
            if (!Program.IsEqualInTolerance(commonEquation.GetA(),0))
                return AIsNotZero(commonEquation, center);
            throw new Exception("Vsechny prommenne v rovnici jsou nulove.");
        }

        private PlanParametricEquation AIsNotZero(PlanEquation commonEquation, Vector center)
        {
            Vector B =
                new Vector((commonEquation.GetB() + commonEquation.GetC() + commonEquation.GetD()) / -commonEquation.GetA(), 1, 1);
            Vector C =
                    new Vector((commonEquation.GetB() - commonEquation.GetC() + commonEquation.GetD()) / -commonEquation.GetA(), 1, -1);

            Vector u = B - center;
            Vector v = C - center;
            return new PlanParametricEquation(u.GetX(), u.GetY(), u.GetZ(), v.GetX(), v.GetY(), v.GetZ());
        }
        private PlanParametricEquation BIsNotZero(PlanEquation commonEquation, Vector center)
        {
            Vector B =
                new Vector(1, (commonEquation.GetA() + commonEquation.GetC() + commonEquation.GetD()) / -commonEquation.GetB(), 1);
            Vector C =
                    new Vector(1, (commonEquation.GetA() - commonEquation.GetC() + commonEquation.GetD()) / -commonEquation.GetB(), -1);

            Vector u = B - center;
            Vector v = C - center;
            return new PlanParametricEquation(u.GetX(), u.GetY(), u.GetZ(), v.GetX(), v.GetY(), v.GetZ());
        }
        private PlanParametricEquation CIsNotZero(PlanEquation commonEquation, Vector center) 
        {
            Vector B =
                new Vector(1, 1, (commonEquation.GetA() + commonEquation.GetB() + commonEquation.GetD()) / -commonEquation.GetC());
            Vector C =
                    new Vector(1, -1, (commonEquation.GetA() - commonEquation.GetB() + commonEquation.GetD()) / -commonEquation.GetC());

            Vector u = B - center;
            Vector v = C - center;
            return new PlanParametricEquation(u.GetX(), u.GetY(), u.GetZ(), v.GetX(), v.GetY(), v.GetZ());
        }



        private PlanEquation GetCommonEquationFromParametricEquation(PlanParametricEquation parametricEquation, 
            CenterParametrization centerParametrization) 
        {
            Vector u = new Vector(parametricEquation.GetUx(), parametricEquation.GetUy(), parametricEquation.GetUz());
            Vector v = new Vector(parametricEquation.GetVx(), parametricEquation.GetVy(), parametricEquation.GetVz());
            Vector normal = -1*Vector.CrossProduct(u, v);
            Vector center = centerParametrization.GetReprezentant();
            double d = - (
                normal.GetX() * center.GetX() + 
                normal.GetY() * center.GetY() + 
                normal.GetZ() * center.GetZ());
            return new PlanEquation(normal.GetX(), normal.GetY(), normal.GetZ(), d);
        }





        protected override void BuildParametricDisplayList(Scene scene, RelativeQuality relativeQuality) 
        {            
            _displayListIdentificator =  Gl.glGenLists(1);
            Gl.glNewList(_displayListIdentificator, Gl.GL_COMPILE);
            Gl.glNormal3d(0,0,_normalSign);
            SuperPlan.BuildPlanDisplayList(scene, 
                GetClippingLeft(scene),GetClippingRight(scene), 
                GetClippingBack(scene), GetClippingFront(scene));
            Gl.glEndList();
        }

        private void ShowPlanDirectionVectorAndNormal(Scene scene) 
        {
            Vector u = new Vector(_parametricEquation.GetUx(), _parametricEquation.GetUy(), _parametricEquation.GetUz());
            Vector v = new Vector(_parametricEquation.GetVx(), _parametricEquation.GetVy(), _parametricEquation.GetVz());
            Vector line = Vector.CrossProduct(u, v);
            //Line line0 = new Line(new LineParametricEquation(
            //    line.GetX(), line.GetY(), line.GetZ()),
            //    _centerInStandardBasis, "", new Color(0, 1, 1));
            //Line line1 = new Line(new LineParametricEquation(
            //    _paramericEquation.GetUx(), _paramericEquation.GetUy(), _paramericEquation.GetUz()),
            //    _centerInStandardBasis, "", new Color(0, 1, 1));
            //Line line2 = new Line(new LineParametricEquation(
            //    _paramericEquation.GetVx(), _paramericEquation.GetVy(), _paramericEquation.GetVz()),
            //    _centerInStandardBasis, "", new Color(0, 1, 1));
            //line1.ReBuildDisplayList(scene);
            //line2.ReBuildDisplayList(scene);
            //line0.ReBuildDisplayList(scene);
            //scene.AddObject(line1);
            //scene.AddObject(line2);
            //scene.AddObject(line0);
        }




        protected override void ActualizeEquations()
        {
            Vector u = canonicFirstDirectionVector * GetRotationInStandardBasis().ComputeInverseMatrix();
            Vector v = canonicSecondDirectionVector * GetRotationInStandardBasis().ComputeInverseMatrix();
            _parametricEquation = new PlanParametricEquation(
                u.GetX(), u.GetY(), u.GetZ(),
                v.GetX(), v.GetY(), v.GetZ());
            _commonEquation = 
                GetCommonEquationFromParametricEquation((PlanParametricEquation)_parametricEquation, GetCenterInStandardBasis());
            _commonEquationFormatSring.SubstituteToFormatString(_commonEquation);
            _parametricEquationFormatString.SubstituteToFormatStrings(_parametricEquation, GetCenterInStandardBasis());
        }



        public override WithParametricEquationObject MakeCopyOfObject(KoefsAndCenterPair newKoefsAndCenter, int ID)
        {
            PlanParametricEquation equation = (PlanParametricEquation)newKoefsAndCenter.GetParametricEquation();
            CenterParametrization center = newKoefsAndCenter.GetCenterParametrization();
            Plan plan = new Plan(equation, center, _color, _name, ID);
            plan.CopyDrawableSlots(plan, this);
            return plan;
        }
        public override WithCommonEquationObject MakeCopyOfObject(CommonEquation commonEquation, int ID)
        {
            PlanEquation equation = (PlanEquation)commonEquation;
            Plan plan = new Plan(equation, _color, _name, ID);
            plan.CopyDrawableSlots(plan, this);
            return plan;
        }
        public override WithParametricEquationObject MakeCopyOfObject(int ID)
        {
            PlanParametricEquation equationCopy =
                new PlanParametricEquation((PlanParametricEquation)_parametricEquation);
            CenterParametrization centerCopy =
                new CenterParametrization(GetCenterInStandardBasis());
            return MakeCopyOfObject(new KoefsAndCenterPair(equationCopy, centerCopy), ID);
        }



        public Vector GetCanonicFirstDirectionVector() { return new Vector(canonicFirstDirectionVector); }
        public Vector GetCanonicSecondDirectionVector() { return new Vector(canonicSecondDirectionVector); }
        public Vector GetCanonicNormalVector() { return new Vector(canonicalNormalVector); }

        public Vector GetFirstDirectionVector() 
        { 
            return canonicFirstDirectionVector * GetRotationInStandardBasis().ComputeInverseMatrix(); 
        }
        public Vector GetSecondDirectionVector()
        {
            return canonicSecondDirectionVector * GetRotationInStandardBasis().ComputeInverseMatrix();
        }

    }
}
