﻿using System;
using System.Collections.Generic;
using GMath;
using Geometry.GraphicObjects.Matrixes;
using Geometry.GraphicObjects.Equations;

namespace Geometry.GraphicObjects.Quadrics
{
    /// <summary>
    /// Převod na kanonicky tvar probiha nasledovne:
    /// 1. Z matice kvadriky určím její charakteristický equation.
    /// 2. Určením kořenů tohoto equationu obdržím vlastní čísla kvadriky
    /// - charakteristický equation je equation 3.stupně.
    /// - hlavní čísla jsou tedy řešením kubické rovnice.
    /// -- z vlastností kvadrik a jejich ch. pol. předpokládám, že řešení jsou reálná.
    /// - vlastní čísla uspořádám do seznamu tak, aby byla nejprve kladna, zaporna a pak nulova.
    /// 3. Pro každé vlastní číslo vypočítám parametrizaci vlastního směru, příslušného danému vlastnímu číslu.
    /// - parametrizace je řešením soustavy 3.rovnic o 3 neznámých.
    /// - vždy obsahuje alespoň jeden parametr.
    /// -- podle násobnosti vlastního čísla provedu test zda je příslušný počet rovnic po úpravě znulován,
    ///     jinak tu nejnulovější rovnici znuluju.
    /// 4. Upravím parametrizace tak, aby po dosazení obdržené vektory znulovaly nejen příslušné členy 
    ///     v matici kvadraticke formy kvadriky, ale i členy g,h,k nebo l.
    /// - to delam v pripade, že je dana parametrizace prislusna nulovemu vlasnimu cislu a
    ///     - jedna se o kvadriku s libovolnym stredem.
    ///     - nebo se jedna o kvadriku s vrcholem, a parametrizace je maximalne druhá ze tří. (k nenuluju).
    /// 5. Dosazuji do parametrizací:
    ///     - pokud má příslušné vl. cislo jednoduhou násobnost dosadím libovolně tak, aby nebyl vysledný vektor nulový. 
    ///     - pokud má 2. násobnost, pak mu přísluší 2.parametrizace, zvolím primárně tu upravenou pokud taková je.
    ///     - do zvolené parametrizace libolobně substituuji, tak aby nebyl výsledný vektor nulový.
    ///     - provedu skalárni soucin ziskaneho vektoru a druhe parametrizace a polozim rovno nule. (kolmost)
    ///     -? vyjádřím nenulový parametr a dosadím do druhé parametrizace.
    ///     - do této parametrizace, pak dosadím tak, aby vysledny vektor nebyl nulový.
    /// </summary>
    public class QuadricRecognizer
    {
        private GColor _color;
        private string _name;
        private QuadricEquation _originalEquation; // double[10];
        private int _ID;


        private struct OwnNumberStruct 
        {
            public double ownNumber; // samotne vlastni cislo
            public int multiplicity; // nasobnost vlastniho cisla
            public bool error;      // indikuje selhani pri hledani vlastnich cisel
            public OwnNumberStruct(double ownNumber, int multiplicity) 
            {
                this.ownNumber = ownNumber;
                this.multiplicity = multiplicity;
                error = false;
            }
            public OwnNumberStruct(bool error) 
            {
                ownNumber = 0;
                multiplicity = 3;
                this.error = error;
            }
        }
        private struct OwnVectorStruct 
        {
            public OwnVectorParametrization parametrization; // parametrizace vlastniho vektoru
            public bool extended; // udává, zda byla parametrizace upravena
        }
        private struct OwnPairStruct
        {
            public OwnNumberStruct ownNumber; // vlastni cislo
            public List<OwnVectorStruct> parametrizations; // parametrizace příslušné danému číslu
        }        
        private struct ParametersValuesStruct 
        {
            public double rValue;
            public double sValue;
            public double tValue;

            public ParametersValuesStruct(double r, double s,double t) 
            {
                rValue = r;
                sValue = s;
                tValue = t;
            }
        }


        public QuadricRecognizer(QuadricEquation equation, GColor color, string name, int ID)
        {
            _ID = ID;
            _color = color;
            _name = name;
            _originalEquation = equation;
        }



        ///4x2 - 9y2 + z2 + 36 = 0 - dvoudílný hyp.
        ///x2 + 4y2 + z2 - 2x = 0 - elipsoid v 1,0,0 
        ///7x2 + 6y2 + 5z2 − 4xy − 4yz − 22x + 24y + 2z + 30 = 0. - elipsoid (1,2,2),(2,1-2),(-2,2,1) (1,-2,1);
        /// 
        /// <summary>
        /// Dodelat reseni soustavy rovnic
        /// Převede 10 hodnot equationu kvadriky na kanonicky tvar.
        /// Vytvori pri tom matici posunuti a rotace. 
        /// Pak vyzkousi vsechny rovnice kvadrik,
        /// pokud se nektera shoduje s kanonickym equationem prida se kvadrika do sceny.
        /// </summary>
        public Quadric RecognizeQuadric()
        {
            List<Quadric> recognizedQuadrics = new List<Quadric>();
            List<double> numbers = new List<double>() { 1, 1000, 0.001, 1000000, 0.000001 };
            Quadric temp;
            _originalEquation = (QuadricEquation)_originalEquation.Normalize();
            //_originalEquation.DivideBy(_originalEquation.GetA());
            foreach (double d in numbers)
            {
                try
                {
                    temp = MakeCanonicQuadric(new QuadricEquation(d * _originalEquation));
                    if (temp == null)
                        temp = MakeCanonicQuadric(new QuadricEquation(-d * _originalEquation));
                }
                catch (Exception) 
                {
                    continue;
                }
                if (temp != null)
                {
                    return temp;
                    //recognizedQuadrics.Add(temp);
                }
            }
            return GetBestOutputQuadric(recognizedQuadrics);
        }

        private Quadric GetBestOutputQuadric(List<Quadric> recognizedQuadrics)
        {
            if (recognizedQuadrics.Count == 0)
                return new EmptySet(_color, _name, _ID);
            try
            {
                Type quadricType = GetTypeOfRecognizedQuadric(recognizedQuadrics);
                int index = 0;
                double minWorseDifference = double.MaxValue;
                double temp2;
                double temp;
                for (int i = 0; i < recognizedQuadrics.Count; i++)
                {
                    if (recognizedQuadrics[i].GetType() != quadricType)
                        continue;
                    temp = GetWorseDifference(recognizedQuadrics[i].GetCommonEquation(), _originalEquation);
                    temp2 = GetSubstitutionsDifference(recognizedQuadrics[i].GetCommonEquation(), _originalEquation);
                    if (temp < minWorseDifference)
                    {
                        index = i;
                        minWorseDifference = temp;
                    }
                }
                return recognizedQuadrics[index];
            }
            catch (Exception)
            {
                return recognizedQuadrics[0];
            }
        }
        private Type GetTypeOfRecognizedQuadric(List<Quadric> recognizedQuadrics)
        {
            List<KeyValuePair<Type, int>> list = new List<KeyValuePair<Type, int>>();
            for (int i = 0; i < recognizedQuadrics.Count; i++)
            {
                Type type = recognizedQuadrics[i].GetType();
                int j = 0;
                for (j = 0; j < list.Count; j++)
                    if (list[j].Key == type)
                    {
                        list[j] = new KeyValuePair<Type, int>(type, list[j].Value + 1);
                        break;
                    }
                if (j == list.Count)
                    list.Add(new KeyValuePair<Type, int>(type, 1));
            }
            Type temp = null;
            int maxCount = 0;
            for (int j = 0; j < list.Count; j++)
                if (list[j].Value > maxCount)
                {
                    maxCount = list[j].Value;
                    temp = list[j].Key;
                }
            return temp;
        }
        private double GetWorseDifference(QuadricEquation a, QuadricEquation inputQuadric)
        {
            double d = 0;
            int index = 0;
            List<double> koefs = inputQuadric.GetValues();
            for (int i = 0; i < koefs.Count; i++)
            {
                index = i;
                d = Math.Max(Math.Abs(koefs[i]), d);
            }
            QuadricEquation aCopy = a.DivideBy(a.Get(index) / d);
            QuadricEquation bCopy = inputQuadric.DivideBy(Math.Sign(inputQuadric.Get(index)));//inputQuadric.Get(index));
            double worseDifference = 0;
            for (int i = 0; i < aCopy.GetValues().Count; i++)
            {
                if (Math.Abs(aCopy.Get(i) - bCopy.Get(i)) > worseDifference)
                    worseDifference = Math.Abs(aCopy.Get(i) - bCopy.Get(i));
            }
            return worseDifference;
        }
        private double GetSubstitutionsDifference(QuadricEquation a, QuadricEquation inputQuadric)
        {
            double d = 0;
            List<double> koefs = inputQuadric.GetValues();
            int index = 0;
            for (int i = 0; i < koefs.Count; i++)
            {
                d = Math.Max(Math.Abs(koefs[i]), d);
                index = i;
            }
            QuadricEquation aCopy = a.DivideBy(a.Get(index) / d);
            QuadricEquation bCopy = inputQuadric.DivideBy(Math.Sign(inputQuadric.Get(index)));//inputQuadric.Get(index));
            double substitutionToA = 0;
            double substitutionToInput = 0;
            Vector v = new Vector(0, 0, 0);
            for (int x = -10; x <= 10; x += 10)
                for (int y = -10; y <= 10; y += 10)
                    for (int z = -10; z <= 10; z += 10)
                    {
                        v.SetValues(x, y, z);
                        substitutionToA += aCopy.Substitute(v);
                        substitutionToInput += bCopy.Substitute(v);
                    }
            return Math.Abs(substitutionToInput - substitutionToA);
        }


        /// <summary>
        /// Vraci determinant matice equationu, upraveny na kubickou rci.
        /// Vraceny seznam ma 3 cleny (x^3 + a*x^2 + b*x + c = 0).
        /// </summary>
        /// <returns></returns>
        private QuadricCharacteristicEquation Det(QuadricMatrix quadricMatrix)
        {
            double a = quadricMatrix.GetA();
            double b = quadricMatrix.GetB();
            double c = quadricMatrix.GetC();
            double d = quadricMatrix.GetD();
            double e = quadricMatrix.GetE();
            double f = quadricMatrix.GetF();
            double lambda2 = -(a + b + c);
            double lambda = -(e * e + f * f + d * d - a * c - b * c - a * b);
            double absMember = -(c * a * b + 2 * d * f * e - e * e * b - f * f * a - d * d * c);
            return new QuadricCharacteristicEquation(1, lambda2, lambda, absMember); // abs member je spatny?
        }
        /// <summary>
        /// Resi kubickou rovnici urcenou equationem ax3 + bx2 +cx +d = 0; (x3 povazuji za 1).
        /// </summary>
        /// <param name="equation"></param>
        /// <returns></returns>
        private List<OwnNumberStruct> GetOwnNumbers(QuadricMatrix quadricMatrix)
        {
            QuadricCharacteristicEquation cubicEquation = Det(quadricMatrix);
            List<double> simpleOwnNumbers = cubicEquation.Solve();
            if (simpleOwnNumbers.Count != 3)
            {
                return new List<OwnNumberStruct>() { new OwnNumberStruct(true) }; //indikuje chybovy stav
            }
            //for (int i = 0; i < simpleOwnNumbers.Count; i++)
            //{
            //    if (Program.IsEqualInToleranceLow(simpleOwnNumbers[i], 0)) 
            //    {
            //        simpleOwnNumbers[i] = 0;                    
            //    }   
            //}
            List<OwnNumberStruct> ownNumbers = new List<OwnNumberStruct>();
            while (simpleOwnNumbers.Count != 0) 
            {
                OwnNumberStruct o = new OwnNumberStruct(simpleOwnNumbers[0], 0);
                for (int i = 0; i < simpleOwnNumbers.Count; i++) 
                {
                    if (Program.IsEqualInToleranceLow(o.ownNumber, simpleOwnNumbers[i]))
                    {
                        simpleOwnNumbers.RemoveAt(i--);
                        o.multiplicity++;
                    }
                }
                ownNumbers.Add(o);
            }
            return ownNumbers;
        }
        private List<OwnNumberStruct> SortOwnNumbers(List<OwnNumberStruct> ownNumbers) 
        {
            // netreba tridit, zavoven je osetren chybovy stav. 
            // OwnNumbers sice nemuze byt null, ale muze obsahovat jedno neplatne vlastni cislo.
            if (ownNumbers.Count == 1)
                return ownNumbers;
            List<OwnNumberStruct> retval = new List<OwnNumberStruct>();
            foreach (OwnNumberStruct ownNumber in ownNumbers)
                if (ownNumber.ownNumber > 0 && !Program.IsEqualInTolerance(ownNumber.ownNumber, 0))
                    retval.Add(ownNumber); // nejprve vlozim kladna vlastni cisla
            foreach (OwnNumberStruct ownNumber in ownNumbers)
                if (ownNumber.ownNumber < 0 && !Program.IsEqualInTolerance(ownNumber.ownNumber, 0))
                    retval.Add(ownNumber); // potom vlozim zaporna vlastni cisla
            foreach (OwnNumberStruct ownNumber in ownNumbers)
                if (Program.IsEqualInTolerance(ownNumber.ownNumber, 0))
                    retval.Add(ownNumber); // nakonec vlozim nulova vlastni cisla
            return retval;
        }

        /// <summary>
        /// Z obecne rovnice kvadriky spocitam vlastni cisla, 
        /// discriminant a determinant matice A44 a urcim kanonickou rovnici.
        /// </summary>
        /// <param name="matrix"></param>
        /// <returns></returns>
        public Quadric MakeCanonicQuadric(QuadricEquation quadricEquation)
        {
            QuadricMatrix quadricMatrix = new QuadricMatrix(quadricEquation);
            List<OwnNumberStruct> ownNumbers = SortOwnNumbers(GetOwnNumbers(quadricMatrix));
            if (ownNumbers.Count == 1 && (ownNumbers[0].error || Program.IsEqualInTolerance(ownNumbers[0].ownNumber, 0)))
                return null;
            List<Vector> ownVectorsPermutation = GetOwnVectors(ownNumbers, quadricMatrix);
            TransformationMatrix rotationMatrix = GetRotationMatrix(ownVectorsPermutation, quadricMatrix);
            TransformationMatrix toCanonicMatrix;
            QuadricEquation canonizedQuadricEquation;
            if (rotationMatrix != null)
            {
                CenterParametrization center = GetCenterParametrization(quadricMatrix);
                if (center == null)
                {
                    center =  GetPeakParametrization(quadricMatrix, rotationMatrix);
                    toCanonicMatrix = GetTransformationMatrix(rotationMatrix, center);
                    canonizedQuadricEquation = new QuadricEquation(GetQuadricInBasis(quadricMatrix, toCanonicMatrix));
                }
                else
                {
                    TransformationMatrix toCanonicComputed = GetTransformationMatrix(new TransformationMatrix(rotationMatrix), center);
                    CenterParametrization centerSquared = GetCenterParametrization(quadricMatrix, rotationMatrix);
                    TransformationMatrix toCanonicSquared = GetTransformationMatrix(new TransformationMatrix(rotationMatrix), centerSquared);
                    QuadricEquation computed = new QuadricEquation(GetQuadricInBasis(quadricMatrix, toCanonicComputed));
                    QuadricEquation squared = new QuadricEquation(GetQuadricInBasis(quadricMatrix, toCanonicSquared));
                    computed = (QuadricEquation)computed.Normalize();
                    squared = (QuadricEquation)squared.Normalize();
                    double d1 = Math.Max(Math.Max(Math.Abs(computed.GetG()), Math.Abs(computed.GetH())), Math.Abs(computed.GetK()));
                    double d2 = Math.Max(Math.Max(Math.Abs(squared.GetG()), Math.Abs(squared.GetH())), Math.Abs(squared.GetK()));
                    // pokud jsou oba koeficienty L desetinna cisla, ale computed je dokonce i v toleranci nulovosti,
                    // a je vice nulovy jako squared nebo je nejmensi koeficient squared moc velky a zaroven je 
                    // nejmensi koeficient computed mensi.
                    if ((Math.Abs(squared.GetL()) < 0 && Program.IsEqualInTolerance(computed.GetL(), 0) &&
                        Math.Abs(computed.GetL()) < Math.Abs(squared.GetL()) || (!Program.IsEqualInToleranceLow(d2, 0) && d1 < d2)))
                    {
                        toCanonicMatrix = toCanonicComputed;
                        canonizedQuadricEquation = computed;
                    }
                    else
                    {
                        toCanonicMatrix = toCanonicSquared;
                        canonizedQuadricEquation = squared;
                    }
                }
                return ClasifyQuadricEquation(canonizedQuadricEquation, toCanonicMatrix);
            }
            else
            {
                return null;
            }
        }

        private QuadricMatrix GetQuadricInBasis(QuadricMatrix quadricMatrix, Matrix tranformationMatrix)
        {
            return quadricMatrix.ApplyTransformation(tranformationMatrix);
        }
        private TransformationMatrix GetRotationMatrix(List<Vector> ownVectors, QuadricMatrix quadricMatrix)
        {
            foreach (Vector v in ownVectors)
                if (v == null)
                    return null;
            return new TransformationMatrix(ownVectors[0], ownVectors[1], ownVectors[2]);
        }
        private TransformationMatrix GetTransformationMatrix(TransformationMatrix rotationMatrix, CenterParametrization centerParametrization)
        {
            Vector center = centerParametrization.GetReprezentant();
            rotationMatrix.SetTranslation(center);
            return rotationMatrix;
        }


        /// <summary>
        /// 
        /// </summary>        
        private List<Vector> GetOwnVectors(List<OwnNumberStruct> ownNumbers, QuadricMatrix quadricMatrix)
        {
            List<OwnPairStruct> ownPairs = new List<OwnPairStruct>();
            // pro kazde vlastni cislo vypocitam prislusnou parametrizaci vlastniho vektoru.
            foreach (OwnNumberStruct ownNumber in ownNumbers)
            {
                OwnVectorParametrization owp = GetOwnVectorParametrization(quadricMatrix, ownNumber);
                OwnPairStruct ownPairStruct = new OwnPairStruct();
                ownPairStruct.ownNumber = ownNumber;
                ownPairStruct.parametrizations = new List<OwnVectorStruct>() {};
                // podle nasobnosti vlastniho cisla, pridam prislusny pocet kopii vlastniho vektoru
                for (int i = 0; i < ownNumber.multiplicity; i++)
                {
                    OwnVectorStruct ownVectorStruct = new OwnVectorStruct();
                    ownVectorStruct.extended = false;
                    ownVectorStruct.parametrization = new OwnVectorParametrization(owp);
                    ownPairStruct.parametrizations.Add(ownVectorStruct);
                }
                ownPairs.Add(ownPairStruct);
            }
            bool existsCenter = GetCenterParametrization(quadricMatrix) != null;
            ExtendOwnPairs(ownPairs, quadricMatrix, existsCenter);
            List<Vector> ownVectors = GetOwnVectorsFromParametrizations(quadricMatrix, ownPairs, existsCenter);
            return ownVectors;
        }
        /// <summary>
        /// 
        /// </summary>
        private void ExtendOwnPairs(List<OwnPairStruct> pairs, QuadricMatrix quadricMatrix, bool existsCenter)
        {
            int count = 0;
            for (int i = 0; i < pairs.Count; i++)
            {
                if (Program.IsEqualInTolerance(pairs[i].ownNumber.ownNumber, 0))
                    for (int j = 0; j < pairs[i].ownNumber.multiplicity; j++)
                    {
                        // pokud se nejedna o PVP nebo jedna ale nejde o posledni vektor.
                        if (false || (!existsCenter && count < 2))
                        {
                            pairs[i].parametrizations[j] = ExtendOwnVector(quadricMatrix, pairs[i].parametrizations[j]);
                        }
                        count++;
                    }
                else count += pairs[i].ownNumber.multiplicity;
            }
        }
        /// <summary>
        /// 
        /// </summary>
        private OwnVectorParametrization GetOwnVectorParametrization(QuadricMatrix quadricMatrix, OwnNumberStruct ownNumber)
        {
            double on = ownNumber.ownNumber;
            Equation row1 = new Equation(quadricMatrix.GetA() - on, quadricMatrix.GetD(), quadricMatrix.GetE());
            Equation row2 = new Equation(quadricMatrix.GetD(), quadricMatrix.GetB() - on, quadricMatrix.GetF());
            Equation row3 = new Equation(quadricMatrix.GetE(), quadricMatrix.GetF(), quadricMatrix.GetC() - on);

            OwnVectorEquationSystem equationSystem = new OwnVectorEquationSystem(row1, row2, row3, ownNumber.multiplicity);
            OwnVectorParametrization owp = equationSystem.Solve();
            return owp;
        }
        /// <summary>
        /// 
        /// </summary>        
        private OwnVectorStruct ExtendOwnVector(QuadricMatrix quadricMatrix, OwnVectorStruct ownVectorStruct) 
        {
            //return own;
            if (ownVectorStruct.parametrization == null)
                return ownVectorStruct;
            OwnVectorParametrization own = ownVectorStruct.parametrization;
            double a = quadricMatrix.GetA();
            double b = quadricMatrix.GetB();
            double c = quadricMatrix.GetC();
            double d = quadricMatrix.GetD();
            double e = quadricMatrix.GetE();
            double f = quadricMatrix.GetF();
            double g = quadricMatrix.GetG();
            double h = quadricMatrix.GetH();
            double k = quadricMatrix.GetK();
            ParametrizationRST M = a * own.GetX() + d * own.GetY() + e * own.GetZ();
            ParametrizationRST N = d * own.GetX() + b * own.GetY() + f * own.GetZ();
            ParametrizationRST P = e * own.GetX() + f * own.GetY() + c * own.GetZ();
            if (!M.IsZero() || !N.IsZero() || !P.IsZero())
                return ownVectorStruct;
            ParametrizationRST constant = g * own.GetX() + h * own.GetY() + k * own.GetZ();
            if (constant.ContainsRParameter())
            {
                constant = constant * (-1 / constant.GetR());
                constant.SetR(0);
                own.SubstituteForR(constant);
                ownVectorStruct.extended = true;
            }else
            if (constant.ContainsSParameter())
            {
                constant = constant * (-1 / constant.GetS());
                constant.SetS(0);
                own.SubstituteForS(constant);
                ownVectorStruct.extended = true;                
            } else
            if (constant.ContainsTParameter())
            {
                constant = constant * (-1 / constant.GetT());
                constant.SetT(0);
                own.SubstituteForT(constant);
                ownVectorStruct.extended = true;                
            }
            return ownVectorStruct;
        }  
        /// <summary>
        /// Zarucim aby vektory nebyly linearne zavisle, pak je orthonormalizuji.
        /// </summary>
        /// <param name="parametrization"></param>
        /// <returns></returns>
        private List<Vector> GetOwnVectorsFromParametrizations(QuadricMatrix quadricMatrix, List<OwnPairStruct> parametrizations, bool existsCenter)
        {
            List<Vector> ownVectors = new List<Vector>();

            for (int i = 0; i < parametrizations.Count; i++) 
            {
                if (parametrizations[i].ownNumber.multiplicity == 1)
                        ownVectors.Add(SubstituteSoThatVectorIsNotZero(parametrizations[i].parametrizations[0].parametrization).NornVector());
                if (parametrizations[i].ownNumber.multiplicity == 2)
                {
                    OwnVectorParametrization theFirst;
                    OwnVectorParametrization theSecond;
                    if (parametrizations[i].parametrizations[0].extended) // prednost dostane ten upraveny vektor
                    {
                        theFirst = parametrizations[i].parametrizations[0].parametrization;
                        theSecond = parametrizations[i].parametrizations[1].parametrization;
                    }else
                    {
                        theFirst = parametrizations[i].parametrizations[1].parametrization;
                        theSecond = parametrizations[i].parametrizations[0].parametrization;
                    }
                    Vector v1 = SubstituteSoThatVectorIsNotZero(theFirst).NornVector();
                    Vector v2 = SubstituteSoThatVectorIsNotZeroAndOrtogonal(v1, theSecond).NornVector();
                    if (parametrizations[i].parametrizations[0].extended) 
                    {
                        ownVectors.Add(v1);
                        ownVectors.Add(v2);
                    }else
                    {
                        ownVectors.Add(v2);
                        ownVectors.Add(v1);
                    }
                }
                if (parametrizations[i].ownNumber.multiplicity == 3) 
                {
                    ownVectors.Add(new Vector(1, 0, 0));
                    ownVectors.Add(new Vector(0, 1, 0));
                    ownVectors.Add(new Vector(0, 0, 1));
                }
            }
            return Vector.OrthoNormalize(ownVectors);
        }
        /// <summary>
        /// Předpokládá 1 nebo 2 parametry.
        /// </summary>
        private Vector SubstituteSoThatVectorIsNotZero(OwnVectorParametrization owp) 
        {
            ParametrizationRST x = owp.GetX();
            ParametrizationRST y = owp.GetY();
            ParametrizationRST z = owp.GetZ();

            ParametrizationRST parametrizationWithParameter = null;

            if (x.ContainsParameter())
                parametrizationWithParameter = new ParametrizationRST(x);
            if (y.ContainsParameter())
                parametrizationWithParameter = new ParametrizationRST(y);
            if (z.ContainsParameter())
                parametrizationWithParameter = new ParametrizationRST(z);
            ParametersValuesStruct values =  GetValuesOfParametersSoThatBiggestOneRests(parametrizationWithParameter);
            owp.SubstituteForR(values.rValue);
            owp.SubstituteForS(values.sValue);
            owp.SubstituteForT(values.tValue);
            return new Vector(owp.GetX().GetConstant(), owp.GetY().GetConstant(), owp.GetZ().GetConstant());
        }
        /// <summary>
        /// 
        /// </summary>
        private Vector SubstituteSoThatVectorIsNotZeroAndOrtogonal(Vector vector, OwnVectorParametrization owp)
        {
            ParametrizationRST x = owp.GetX() * vector.GetX();
            ParametrizationRST y = owp.GetY() * vector.GetY();
            ParametrizationRST z = owp.GetZ() * vector.GetZ();

            ParametrizationRST equation = x + y + z;
            OwnVectorParametrization copy = new OwnVectorParametrization(owp);

            if (equation.ContainsRParameter())
                copy.SubstituteForR(equation.ExpressR());
            else if (equation.ContainsSParameter())
                copy.SubstituteForS(equation.ExpressS());
            else if (equation.ContainsTParameter())
                copy.SubstituteForT(equation.ExpressT());
            return SubstituteSoThatVectorIsNotZero(copy);
        }
        /// <summary>
        /// 
        /// </summary>        
        private ParametersValuesStruct GetValuesOfParametersSoThatBiggestOneRests(ParametrizationRST parametrization)
        {
            double rValue = Math.Abs(parametrization.GetR());
            //return new ParametersValuesStruct();
            double sValue = Math.Abs(parametrization.GetS());
            double tValue = Math.Abs(parametrization.GetT());
            if (rValue > sValue)
                if (rValue > tValue) // r je nejvetsi
                    return new ParametersValuesStruct(
                        GetOneOrMinusOne(-parametrization.GetConstant() / parametrization.GetR()), 0, 0);
                else // t je nejvetsi
                    return new ParametersValuesStruct(
                        0, 0, GetOneOrMinusOne(-parametrization.GetConstant() / parametrization.GetT()));
            else
                if (sValue > tValue) // s je nejvetsi
                    return new ParametersValuesStruct(
                        0, GetOneOrMinusOne(-parametrization.GetConstant() / parametrization.GetS()), 0);
                else // t je nejvetsi
                    return new ParametersValuesStruct(
                        0, 0, GetOneOrMinusOne(-parametrization.GetConstant() / parametrization.GetT()));
        }
        /// <summary>
        /// Z cisel 1 a -1 vraci to, ktere je vzdalenejsi od cisla number;
        /// </summary>
        private double GetOneOrMinusOne(double number) 
        {
            // z cisel 1 a -1 vraci to vzdalenejsi od number
            if (Math.Abs(number - 1) > Math.Abs(number + 1))
                return 1;
            else return -1;
        }

        private CenterParametrization GetCenterParametrization(QuadricMatrix quadricMatrix, TransformationMatrix rotation)
        {
            QuadricMatrix matrixRotated = quadricMatrix.ApplyTransformation(rotation);
            try
            {
                double a = matrixRotated.GetA();
                double b = matrixRotated.GetB();
                double c = matrixRotated.GetC();
                double g = matrixRotated.GetG() * 2.0;
                double h = matrixRotated.GetH() * 2.0;
                double k = matrixRotated.GetK() * 2.0;

                double m = 0;
                double n = 0;
                double o = 0;
                ParametrizationRST x, y, z;
                if (!Program.IsEqualInTolerance(a, 0))
                    m = (g / (2 * a));
                if (!Program.IsEqualInTolerance(b, 0))
                    n = (h / (2 * b));
                if (!Program.IsEqualInTolerance(c, 0))
                    o = (k / (2 * c));
                x = new ParametrizationRST(-m, 0, 0, 0);
                y = new ParametrizationRST(-n, 0, 0, 0);
                z = new ParametrizationRST(-o, 0, 0, 0);
                return new CenterParametrization(x, y, z) * rotation;
            }
            catch (Exception) { return new CenterParametrization(); };
        }
        private CenterParametrization GetCenterParametrization(QuadricMatrix quadricMatrix) 
        {
            Equation row1 = new Equation
                (quadricMatrix.GetA(), quadricMatrix.GetD(), quadricMatrix.GetE(), quadricMatrix.GetG());
            Equation row2 = new Equation
                (quadricMatrix.GetD(), quadricMatrix.GetB(), quadricMatrix.GetF(), quadricMatrix.GetH());
            Equation row3 = new Equation
                (quadricMatrix.GetE(), quadricMatrix.GetF(), quadricMatrix.GetC(), quadricMatrix.GetK());
            CenterEquationSystem equationSystem = new CenterEquationSystem(row1, row2, row3);
            return equationSystem.Solve();
        }
        private CenterParametrization GetPeakParametrization(QuadricMatrix quadricMatrix, TransformationMatrix rotation)
        { 
            QuadricMatrix matrixRotated = quadricMatrix.ApplyTransformation(rotation);
            try
            {
                double a = matrixRotated.GetA();
                double b = matrixRotated.GetB();
                double g = matrixRotated.GetG() * 2.0;
                double h = matrixRotated.GetH() * 2.0;
                double k = matrixRotated.GetK() * 2.0;
                double l = matrixRotated.GetL();
                int bSign = Math.Sign(b);
                double m, n, o;
                ParametrizationRST x, y, z;
                if (a == 0 || k == 0)
                    return new CenterParametrization();
                if (b != 0)
                {
                    m = (g / (2 * a));
                    n = (h / (2 * b));
                    o = (l - ((g * g) / (4 * a)) - ((h * h) / (4 * b))) / k;
                    //= -(l - (m*m) - (n*n)) / k;                    
                    x = new ParametrizationRST(-m, 0, 0, 0);
                    y = new ParametrizationRST(-n, 0, 0, 0);
                    z = new ParametrizationRST(-o, 0, 0, 0);
                    return new CenterParametrization(x, y, z) * rotation;
                }
                m = (g / (2 * a));
                n = 0;
                //o = (l - (m * m)) / k;
                o = (l - (g * g) / (4 * a)) / k;
                x = new ParametrizationRST(-m, 0, 0, 0);
                y = new ParametrizationRST(0, 1, 0, 0);
                z = new ParametrizationRST(-o, 0, 0, 0);
                return new CenterParametrization(x, y, z) * rotation;
            }
            catch (Exception) { return new CenterParametrization(); };
        }        



        private Quadric ClasifyQuadricEquation(QuadricEquation equation, TransformationMatrix matrix)
        {
            Quadric temp;
            equation = new QuadricEquation(equation.Normalize());
            bool isCanonized = true;
            //for (int i = 0; i < equations.Count; i++)
            {
                isCanonized = true;
                isCanonized = isCanonized && Program.IsEqualInToleranceLow(equation.GetD(), 0);
                isCanonized = isCanonized && Program.IsEqualInToleranceLow(equation.GetE(), 0);
                isCanonized = isCanonized && Program.IsEqualInToleranceLow(equation.GetF(), 0);
                if (!isCanonized)
                    return null;
                temp = IsParaboloidOrParabolicCylinder(equation, matrix);
                if (temp != null)
                    return temp;
                temp = ClasificationHelp(equation, matrix);
                if (temp != null)
                    return temp;
            }
            return null;
        }
        /// <summary>
        /// Podelim equation koeficientem -L pokud neni nulovy
        /// </summary>
        /// <returns></returns>
        private QuadricEquation NormalizeEquation(QuadricEquation equation) 
        {                
            if (!Program.IsEqualInTolerance(equation.GetL(), 0)) // deleni absolutnim clenem 
            {
                for (int j = 0; j < QuadricEquation.MembersCount; j++)
                    equation.Set(j, equation.Get(j) / -equation.GetL());
            }
            return equation;
        }
        private Quadric IsParaboloidOrParabolicCylinder(QuadricEquation equation, TransformationMatrix transformation) 
        {
            Quadric temp;
            if ((temp = IsParabolicCylinder(equation, transformation)) != null)
                return temp;
            if ((temp = IsHyperbolicParaboloid(equation, transformation)) != null)
                return temp;
            if ((temp = IsElipticParaboloid(equation, transformation)) != null)
                return temp;
            return null;
        }
        private Quadric ClasificationHelp(QuadricEquation equation, TransformationMatrix transformation)
        {
            Quadric temp;

            if ((temp = IsElipsoid(equation, transformation)) != null)
                return temp;
            if ((temp = IsCone(equation, transformation)) != null)
                return temp;
            if ((temp = IsElipticCylinder(equation, transformation)) != null)
                return temp;
            if ((temp = IsHyperbolicCylinder(equation, transformation)) != null)
                return temp;
            if ((temp = IsHyperboloid1(equation, transformation)) != null)
                return temp;
            if ((temp = IsHyperboloid2(equation, transformation)) != null)
                return temp;
            if ((temp = IsDoublePlan(equation, transformation)) != null)
                return temp;
            if ((temp = IsParallelPlans(equation, transformation)) != null)
                return temp;
            if ((temp = IsLine(equation, transformation)) != null)
                return temp;
            if ((temp = IsCrossingPlans(equation, transformation)) != null)
                return temp;
            if ((temp = IsImaginarCone(equation, transformation)) != null)
                return temp;
            if ((temp = IsImaginarElipsoid(equation, transformation)) != null)
                return temp;
            if ((temp = IsImaginarElipticCylinder(equation, transformation)) != null)
                return temp;
            if ((temp = IsImaginarParallelPlans(equation, transformation)) != null)
                return temp;
            return null;
        } 



        /// <summary>
        /// a*a + b*b + c*c -1 =0;
        /// Muzu ignorovat konstanty G,H,K, protoze jdou doplnit na ctverec a asi jde o mirnou nepresnost.
        /// </summary>
        public Quadric IsElipsoid(QuadricEquation equation, TransformationMatrix transformation)
        {
            bool isElipsoid = true;
            double a = equation.GetA();
            double b = equation.GetB();
            double c = equation.GetC();
            isElipsoid = isElipsoid && a > 0 && !IsCloseToZero(a);
            isElipsoid = isElipsoid && b > 0 && !IsCloseToZero(b);
            isElipsoid = isElipsoid && c > 0 && !IsCloseToZero(c); 
            //isElipsoid = isElipsoid && Program.IsEqualInTolerance(equation.GetG(), 0);
            //isElipsoid = isElipsoid && Program.IsEqualInTolerance(equation.GetH(), 0);
            //isElipsoid = isElipsoid && Program.IsEqualInTolerance(equation.GetK(), 0);
            isElipsoid = isElipsoid && !IsCloseToZero(equation.GetL()) && equation.GetL() < 0;
            if (isElipsoid)
            {
                equation = NormalizeEquation(equation);
                a = equation.GetA();
                b = equation.GetB();
                c = equation.GetC();

                a = Math.Sqrt(1.0 / a);
                b = Math.Sqrt(1.0 / b);
                c = Math.Sqrt(1.0 / c);
                return new Elipsoid(_originalEquation, new QuadricParametricEquation(a,b,c,0),
                    transformation, _color, _name, _ID);
            }
            return null;
        }
        /// <summary>
        /// a*a + b*b + c*c +1 =0;
        /// Muzu ignorovat konstanty G,H,K, protoze jdou doplnit na ctverec a asi jde o mirnou nepresnost.
        /// </summary>
        public Quadric IsImaginarElipsoid(QuadricEquation equation, TransformationMatrix transformation)
        {
            bool isElipsoid = true;
            double a = equation.GetA();
            double b = equation.GetB();
            double c = equation.GetC();
            isElipsoid = isElipsoid && a > 0 && !IsCloseToZero(a);
            isElipsoid = isElipsoid && b > 0 && !IsCloseToZero(b);
            isElipsoid = isElipsoid && c > 0 && !IsCloseToZero(c);

            //isElipsoid = isElipsoid && Program.IsEqualInTolerance(equation.GetG(), 0);
            //isElipsoid = isElipsoid && Program.IsEqualInTolerance(equation.GetH(), 0);
            //isElipsoid = isElipsoid && Program.IsEqualInTolerance(equation.GetK(), 0);
            isElipsoid = isElipsoid && !IsCloseToZero(equation.GetL()) && equation.GetL() > 0;
            if (isElipsoid)
            {
                equation = NormalizeEquation(equation);
                a = equation.GetA();
                b = equation.GetB();
                c = equation.GetC();

                a = Math.Sqrt(1.0 / -a);
                b = Math.Sqrt(1.0 / -b);
                c = Math.Sqrt(1.0 / -c);
                return new ImaginarElipsoid(_originalEquation,new QuadricParametricEquation(a,b,c,0),
                    transformation, _color, _name, _ID);
            }
            return null;
        }
        /// <summary>
        /// a*a + b*b - c*c = 0; 
        /// Muzu ignorovat konstanty G,H,K, protoze jdou doplnit na ctverec a asi jde o mirnou nepresnost.
        /// </summary>
        public Quadric IsCone(QuadricEquation equation, TransformationMatrix transformation)
        {
            bool isCone = true;
            double a = equation.GetA();
            double b = equation.GetB();
            double c = equation.GetC();
            isCone = isCone && a > 0 && !IsCloseToZero(a);
            isCone = isCone && b > 0 && !IsCloseToZero(b);
            isCone = isCone && c < 0 && !IsCloseToZero(c);

            //isCone = isCone && Program.IsEqualInTolerance(equation.GetG(), 0);
            //isCone = isCone && Program.IsEqualInTolerance(equation.GetH(), 0);
            //isCone = isCone && Program.IsEqualInTolerance(equation.GetK(), 0);
            isCone = isCone && Program.IsEqualInTolerance(equation.GetL(), 0);
            if (isCone)
            {
                a = Math.Sqrt(1.0 / a);
                b = Math.Sqrt(1.0 / b);
                c = Math.Sqrt(-1.0 / c);
                return new Cone(_originalEquation, new QuadricParametricEquation(a, b, c, 0),
                    transformation, _color, _name, _ID);
            }
            return null;
        }
        /// <summary>
        /// a*a + b*b + c*c = 0; 
        /// Muzu ignorovat konstanty G,H,K, protoze jdou doplnit na ctverec a asi jde o mirnou nepresnost.
        /// </summary>
        public Quadric IsImaginarCone(QuadricEquation equation, TransformationMatrix transformation)
        {
            bool isCone = true;
            double a = equation.GetA();
            double b = equation.GetB();
            double c = equation.GetC();
            isCone = isCone && a > 0 && !IsCloseToZero(a);
            isCone = isCone && b > 0 && !IsCloseToZero(b);
            isCone = isCone && c > 0 && !IsCloseToZero(c);

            //isCone = isCone && Program.IsEqualInTolerance(equation.GetG(), 0);
            //isCone = isCone && Program.IsEqualInTolerance(equation.GetH(), 0);
            //isCone = isCone && Program.IsEqualInTolerance(equation.GetK(), 0);
            isCone = isCone && Program.IsEqualInTolerance(equation.GetL(), 0);
            if (isCone)
            {
                a = Math.Sqrt(1.0 / a);
                b = Math.Sqrt(1.0 / b);
                c = Math.Sqrt(1.0 / c);
                return new ImaginarCone(_originalEquation, new QuadricParametricEquation(a, b, c, 0),
                    transformation, _color, _name, _ID);
            }
            return null;
        }
        /// <summary>
        /// a*a + b*b -1 = 0;
        /// Muzu ignorovat konstanty G,H protoze jdou doplnit na ctverec a asi jde o mirnou nepresnost.
        /// </summary>
        public Quadric IsElipticCylinder(QuadricEquation equation, TransformationMatrix transformation)
        {
            bool isCylinder = true;
            double a = equation.GetA();
            double b = equation.GetB();
            isCylinder = isCylinder && a > 0 && !IsCloseToZero(a);
            isCylinder = isCylinder && b > 0 && !IsCloseToZero(b);
            isCylinder = isCylinder && Program.IsEqualInTolerance(equation.GetC(), 0);

            //isCylinder = isCylinder && Program.IsEqualInTolerance(equation.GetG(), 0);
            //isCylinder = isCylinder && Program.IsEqualInTolerance(equation.GetH(), 0);
            isCylinder = isCylinder && Program.IsEqualInTolerance(equation.GetK(), 0);
            isCylinder = isCylinder && !IsCloseToZero(equation.GetL()) && equation.GetL() < 0;
            if (isCylinder)
            {
                equation = NormalizeEquation(equation);
                a = equation.GetA();
                b = equation.GetB();

                a = Math.Sqrt(1.0 / a);
                b = Math.Sqrt(1.0 / b);
                return new ElipticCylinder(_originalEquation, new QuadricParametricEquation(a, b, 0, 0),
                    transformation, _color, _name, _ID);
            }
            return null;
        }
        /// <summary>
        /// a*a + b*b + 1 = 0;
        /// Muzu ignorovat konstanty G,H protoze jdou doplnit na ctverec a asi jde o mirnou nepresnost.
        /// </summary>
        public Quadric IsImaginarElipticCylinder(QuadricEquation equation, TransformationMatrix transformation)
        {
            bool isCylinder = true;
            double a = equation.GetA();
            double b = equation.GetB();
            isCylinder = isCylinder && a > 0 && !IsCloseToZero(a);
            isCylinder = isCylinder && b > 0 && !IsCloseToZero(b);
            isCylinder = isCylinder && Program.IsEqualInTolerance(equation.GetC(), 0);

            //isCylinder = isCylinder && Program.IsEqualInTolerance(equation.GetG(), 0);
            //isCylinder = isCylinder && Program.IsEqualInTolerance(equation.GetH(), 0);
            isCylinder = isCylinder && Program.IsEqualInTolerance(equation.GetK(), 0);
            isCylinder = isCylinder && !IsCloseToZero(equation.GetL()) && equation.GetL() > 0;
            if (isCylinder)
            {
                equation = NormalizeEquation(equation);
                a = equation.GetA();
                b = equation.GetB();

                a = Math.Sqrt(1.0 / -a);
                b = Math.Sqrt(1.0 / -b);
                return new ImaginarElipticCylinder(_originalEquation, new QuadricParametricEquation(a, b, 0, 0),
                    transformation, _color, _name, _ID);
            }
            return null;
        }
        /// <summary>
        /// a*a - b*b -1 = 0;
        /// Muzu ignorovat konstanty G,H protoze jdou doplnit na ctverec a asi jde o mirnou nepresnost.
        /// </summary>
        public Quadric IsHyperbolicCylinder(QuadricEquation equation, TransformationMatrix transformation)
        {
            bool isCylinder = true;
            double a = equation.GetA();
            double b = equation.GetB();
            isCylinder = isCylinder && a > 0 && !IsCloseToZero(a);
            isCylinder = isCylinder && b < 0 && !IsCloseToZero(b);
            isCylinder = isCylinder && Program.IsEqualInTolerance(equation.GetC(), 0);

            //isCylinder = isCylinder && Program.IsEqualInTolerance(equation.GetG(), 0);
            //isCylinder = isCylinder && Program.IsEqualInTolerance(equation.GetH(), 0);
            isCylinder = isCylinder && Program.IsEqualInTolerance(equation.GetK(), 0);
            isCylinder = isCylinder && !IsCloseToZero(equation.GetL()) && equation.GetL() < 0;
            if (isCylinder)
            {
                equation = NormalizeEquation(equation);
                a = equation.GetA();
                b = equation.GetB();

                a = Math.Sqrt(1.0 / a);
                b = Math.Sqrt(-1.0 / b);
                return new HyperbolicCylinder(_originalEquation, new QuadricParametricEquation(a, b, 0, 0),
                    transformation, _color, _name, _ID);
            }
            return null;
        }
        /// <summary>
        /// a*a +2p = 0;
        /// Muzu ignorovat konstanty G protoze jdou doplnit na ctverec a asi jde o mirnou nepresnost.
        /// </summary>
        public Quadric IsParabolicCylinder(QuadricEquation equation, TransformationMatrix transformation)
        {
            bool isCylinder = true;
            double a = equation.GetA();
            double k = equation.GetK();
            isCylinder = isCylinder && a > 0 && !IsCloseToZero(a);
            isCylinder = isCylinder && !IsCloseToZero(k);            
            isCylinder = isCylinder && Program.IsEqualInTolerance(equation.GetB(), 0);
            isCylinder = isCylinder && Program.IsEqualInTolerance(equation.GetC(), 0);

            //isCylinder = isCylinder && Program.IsEqualInToleranceLow(equation.GetG(), 0);
            //isCylinder = isCylinder && Program.IsEqualInToleranceLow(equation.GetH(), 0);
            //isCylinder = isCylinder && Program.IsEqualInToleranceLow(equation.GetL(), 0);
            if (isCylinder)
            {
                double fi = 90;
                TransformationMatrix xRotation = TransformationMatrix.GetXRotationMatrix(fi);
                equation.SetH(equation.GetK());
                equation.SetK(0);
                a = Math.Sqrt(1 / a);
                transformation = TransformationMatrix.GetZRotationMatrix(-180)* xRotation * transformation;
                double p = equation.GetH() / 2; // H ~!!!!!!!!
                if (p < 0)
                {
                    transformation = TransformationMatrix.GetZRotationMatrix(180) * transformation;
                    p = -p;
                }
                QuadricMatrix ms = new QuadricMatrix(_originalEquation).ApplyTransformation(transformation);
                return new ParabolicCylinder(_originalEquation, new QuadricParametricEquation(a, 0, 0, p),
                    transformation, _color, _name, _ID);
            }
            return null;
        }
        /// <summary>
        /// a*a + b*b - c*c -1 = 0;    
        /// Muzu ignorovat konstanty G,H,K, protoze jdou doplnit na ctverec a asi jde o mirnou nepresnost.
        /// </summary>
        public Quadric IsHyperboloid1(QuadricEquation equation, TransformationMatrix transformation)
        {
            bool isHyperboloid1 = true;
            double a = equation.GetA();
            double b = equation.GetB();
            double c = equation.GetC();
            isHyperboloid1 = isHyperboloid1 && a > 0 && !IsCloseToZero(a);
            isHyperboloid1 = isHyperboloid1 && b > 0 && !IsCloseToZero(b);
            isHyperboloid1 = isHyperboloid1 && c < 0 && !IsCloseToZero(c);

            //isHyperboloid1 = isHyperboloid1 && Program.IsEqualInTolerance(equation.GetG(), 0);
            //isHyperboloid1 = isHyperboloid1 && Program.IsEqualInTolerance(equation.GetH(), 0);
            //isHyperboloid1 = isHyperboloid1 && Program.IsEqualInTolerance(equation.GetK(), 0);
            isHyperboloid1 = isHyperboloid1 && !IsCloseToZero(equation.GetL()) && equation.GetL() < 0;
            if (isHyperboloid1)
            {
                equation = NormalizeEquation(equation);
                a = equation.GetA();
                b = equation.GetB();
                c = equation.GetC();

                a = Math.Sqrt(1.0 / a);
                b = Math.Sqrt(1.0 / b);
                c = Math.Sqrt(-1.0 / c);
                return new Hyperboloid1(_originalEquation, new QuadricParametricEquation(a, b, c, 0),
                    transformation, _color, _name, _ID);
            }
            return null;
        }
        /// <summary>
        /// -a*a - b*b + c*c -1 = 0
        /// Muzu ignorovat konstanty G,H,K, protoze jdou doplnit na ctverec a asi jde o mirnou nepresnost.
        /// </summary>
        public Quadric IsHyperboloid2(QuadricEquation equation, TransformationMatrix transformation)
        {
            bool isHyperboloid2 = true;
            double a = equation.GetA();
            double b = equation.GetB();
            double c = equation.GetC();
            isHyperboloid2 = isHyperboloid2 && a > 0 && !IsCloseToZero(a);
            isHyperboloid2 = isHyperboloid2 && b > 0 && !IsCloseToZero(b);
            isHyperboloid2 = isHyperboloid2 && c < 0 && !IsCloseToZero(c);

            //isHyperboloid2 = isHyperboloid2 && Program.IsEqualInTolerance(equation.GetG(), 0);
            //isHyperboloid2 = isHyperboloid2 && Program.IsEqualInTolerance(equation.GetH(), 0);
            //isHyperboloid2 = isHyperboloid2 && Program.IsEqualInTolerance(equation.GetK(), 0);
            isHyperboloid2 = isHyperboloid2 && !IsCloseToZero(equation.GetL()) && equation.GetL() > 0;
            if (isHyperboloid2)
            {
                equation = NormalizeEquation(equation);
                a = equation.GetA();
                b = equation.GetB();
                c = equation.GetC();

                a = Math.Sqrt(-1.0 / a);
                b = Math.Sqrt(-1.0 / b);
                c = Math.Sqrt(1.0 / c);
                return new Hyperboloid2(_originalEquation, new QuadricParametricEquation(a, b, c, 0),
                    transformation, _color, _name, _ID);
            }
            return null;
        }
        /// <summary>
        /// x*x/a*a - y*y/b*b -2kz = 0;
        /// Muzu ignorovat konstanty G,H, protoze jdou doplnit na ctverec a asi jde o mirnou nepresnost.
        /// </summary>
        public Quadric IsHyperbolicParaboloid(QuadricEquation equation, TransformationMatrix transformation)
        {
            bool isParaboloid = true;
            double a = equation.GetA();
            double b = equation.GetB();
            double k = equation.GetK();
            isParaboloid = isParaboloid && a > 0 && !IsCloseToZero(a);
            isParaboloid = isParaboloid && b < 0 && !IsCloseToZero(b);
            isParaboloid = isParaboloid && k > 0 && !IsCloseToZero(k);
            isParaboloid = isParaboloid && Program.IsEqualInTolerance(equation.GetC(), 0);

            //isParaboloid = isParaboloid && Program.IsEqualInToleranceLow(equation.GetG(), 0);
            //isParaboloid = isParaboloid && Program.IsEqualInToleranceLow(equation.GetH(), 0);
            //isParaboloid = isParaboloid && Program.IsEqualInToleranceLow(equation.GetL(), 0);
            if (isParaboloid)
            {
                a = Math.Sqrt(1.0 / a);
                b = Math.Sqrt(-1.0 / b);
                return new HyperbolicParaboloid(_originalEquation, new QuadricParametricEquation(a, b, 0, k / 2.0),
                    transformation, _color, _name, _ID); ;
            }
            return null;
        }
        /// <summary>
        /// x*x/a*a + y*y/b*b +2kz = 0;  
        /// Muzu ignorovat konstanty G,H, protoze jdou doplnit na ctverec a asi jde o mirnou nepresnost.
        /// </summary>
        public Quadric IsElipticParaboloid(QuadricEquation equation, TransformationMatrix transformation)
        {
            bool isParaboloid = true;
            double a = equation.GetA();
            double b = equation.GetB();
            double k = equation.GetK();
            isParaboloid = isParaboloid && a > 0 && !IsCloseToZero(a);
            isParaboloid = isParaboloid && b > 0 && !IsCloseToZero(b);
            isParaboloid = isParaboloid && !IsCloseToZero(k);
            isParaboloid = isParaboloid && Program.IsEqualInTolerance(equation.GetC(), 0);

            //isParaboloid = isParaboloid && Program.IsEqualInToleranceLow(equation.GetG(), 0);
            //isParaboloid = isParaboloid && Program.IsEqualInToleranceLow(equation.GetH(), 0);
            //isParaboloid = isParaboloid && Program.IsEqualInToleranceLow(equation.GetL(), 0);
            if (isParaboloid)
            {
                a = Math.Sqrt(1.0 / a);
                b = Math.Sqrt(1.0 / b);
                if (k < 0)
                    transformation = TransformationMatrix.GetXRotationMatrix(180) * transformation;
                k = Math.Abs(k);
                return new ElipticParaboloid(_originalEquation, new QuadricParametricEquation(a, b, 0, k/2.0),
                    transformation, _color, _name, _ID);
            }
            return null;
        }
        /// <summary>
        /// x*x/a*a + y*y/b*b = 0;
        /// Muzu ignorovat konstanty G,H protoze jdou doplnit na ctverec a asi jde o mirnou nepresnost.
        /// </summary>
        public Quadric IsLine(QuadricEquation equation, TransformationMatrix transformation)
        {
            bool isLine = true;
            double a = equation.GetA();
            double b = equation.GetB();
            isLine = isLine && a > 0 && !IsCloseToZero(a);
            isLine = isLine && b > 0 && !IsCloseToZero(b);
            isLine = isLine && Program.IsEqualInTolerance(equation.GetC(), 0);

            //isLine = isLine && Program.IsEqualInTolerance(equation.GetG(), 0);
            //isLine = isLine && Program.IsEqualInTolerance(equation.GetH(), 0);
            isLine = isLine && Program.IsEqualInTolerance(equation.GetK(), 0);
            isLine = isLine && Program.IsEqualInTolerance(equation.GetL(), 0);
            if (isLine)
            {
                a = Math.Sqrt(1.0 / a);
                b = Math.Sqrt(1.0 / b);
                return new ImaginarCrossingPlans(_originalEquation, new QuadricParametricEquation(a, b, 0, 0),
                    transformation, _color, _name, _ID);
            }
            return null;
        }
        /// <summary>
        /// x*x/a*a - y*y/b*b = 0;
        /// Muzu ignorovat konstanty G,H protoze jdou doplnit na ctverec a asi jde o mirnou nepresnost.
        /// </summary>
        public Quadric IsCrossingPlans(QuadricEquation equation, TransformationMatrix transformation)
        {
            bool isPlan = true;
            double a = equation.GetA();
            double b = equation.GetB();
            isPlan = isPlan && a > 0 && !IsCloseToZero(a);
            isPlan = isPlan && b < 0 && !IsCloseToZero(b);
            isPlan = isPlan && Program.IsEqualInTolerance(equation.GetC(), 0);

            //isPlan = isPlan && Program.IsEqualInTolerance(equation.GetG(), 0);
            //isPlan = isPlan && Program.IsEqualInTolerance(equation.GetH(), 0);
            isPlan = isPlan && Program.IsEqualInTolerance(equation.GetK(), 0);
            isPlan = isPlan && Program.IsEqualInTolerance(equation.GetL(), 0);
            if (isPlan)
            {
                a = Math.Sqrt(1.0 / a);
                b = Math.Sqrt(-1.0 / b);
                return new CrossingPlans(_originalEquation, new QuadricParametricEquation(a, b, 0, 0),
                    transformation, _color, _name, _ID);
            }
            return null;
        }
        /// <summary>
        /// x*x/a*a -1 =0;
        /// Muzu ignorovat konstanty G protoze jdou doplnit na ctverec a asi jde o mirnou nepresnost.
        /// </summary>
        public Quadric IsParallelPlans(QuadricEquation equation, TransformationMatrix transformation)
        {
            bool isPlan = true;
            double a = equation.GetA();
            isPlan = isPlan && a > 0 && !IsCloseToZero(a);
            isPlan = isPlan && Program.IsEqualInTolerance(equation.GetB(), 0);
            isPlan = isPlan && Program.IsEqualInTolerance(equation.GetC(), 0);

            //isPlan = isPlan && Program.IsEqualInTolerance(equation.GetG(), 0);
            isPlan = isPlan && Program.IsEqualInTolerance(equation.GetH(), 0);
            isPlan = isPlan && Program.IsEqualInTolerance(equation.GetK(), 0);
            isPlan = isPlan && !IsCloseToZero(equation.GetL()) && equation.GetL() < 0;
            if (isPlan)
            {
                equation = NormalizeEquation(equation);
                a = equation.GetA();

                a = Math.Sqrt(1.0 / a);
                return new ParallelPlans(_originalEquation, new QuadricParametricEquation(a, 0, 0, 0),
                    transformation, _color, _name, _ID);
            }
            return null;
        }
        /// <summary>
        /// x*x/a*a +1 =0;
        /// Muzu ignorovat konstanty G protoze jdou doplnit na ctverec a asi jde o mirnou nepresnost.
        /// </summary>
        public Quadric IsImaginarParallelPlans(QuadricEquation equation, TransformationMatrix transformation)
        {
            bool isPlan = true;
            double a = equation.GetA();
            isPlan = isPlan && a > 0 && !IsCloseToZero(a);
            isPlan = isPlan && Program.IsEqualInTolerance(equation.GetB(), 0);
            isPlan = isPlan && Program.IsEqualInTolerance(equation.GetC(), 0);
            //isPlan = isPlan && Program.IsEqualInTolerance(equation.GetG(), 0);
            isPlan = isPlan && Program.IsEqualInTolerance(equation.GetH(), 0);
            isPlan = isPlan && Program.IsEqualInTolerance(equation.GetK(), 0);
            isPlan = isPlan && !IsCloseToZero(equation.GetL()) && equation.GetL() > 0;
            if (isPlan)
            {
                equation = NormalizeEquation(equation);
                a = equation.GetA();

                a = Math.Sqrt(1.0 / -a);
                return new ImaginarParallelPlans(_originalEquation, new QuadricParametricEquation(a, 0, 0, 0),
                    transformation, _color, _name, _ID);
            }
            return null;
        }
        /// <summary>
        /// x*x/a*a =0;
        /// Muzu ignorovat konstanty G protoze jdou doplnit na ctverec a asi jde o mirnou nepresnost.
        /// </summary>
        public Quadric IsDoublePlan(QuadricEquation equation, TransformationMatrix transformation)
        {
            bool isPlan = true;
            double a = equation.GetA();
            isPlan = isPlan && a > 0 && !IsCloseToZero(a);
            isPlan = isPlan && Program.IsEqualInTolerance(equation.GetB(), 0);
            isPlan = isPlan && Program.IsEqualInTolerance(equation.GetC(), 0);
            //isPlan = isPlan && Program.IsEqualInTolerance(equation.GetG(), 0);
            isPlan = isPlan && Program.IsEqualInTolerance(equation.GetH(), 0);
            isPlan = isPlan && Program.IsEqualInTolerance(equation.GetK(), 0);
            isPlan = isPlan && Program.IsEqualInTolerance(equation.GetL(), 0);
            if (isPlan)
            {
                a = Math.Sqrt(1.0 / a);
                return new DoublePlan(_originalEquation, new QuadricParametricEquation(a, 0, 0, 0),
                    transformation, _color, _name, _ID);
            }
            return null;
        }



        private bool IsCloseToZero(double d) 
        {
            double tolerance = 0.00001;
            return Math.Abs(d) < tolerance;
        }

        //private Quadric InvariantClasification() 
        //{
        //    Matrix matrix = new QuadricMatrix(_originalEquation);
        //    double determinant = matrix.Determinant4();
        //    Matrix subMatrix = matrix.GetSubMatrix(3,3);
        //    double subDeterminant = subMatrix.Determinant3();
        //    if (Program.IsEqualInTolerance(determinant, 0))
        //        return IsSingular(subDeterminant);
        //    return IsRegular(subDeterminant);
        //}
        //private Quadric IsRegular(double subDeterminant) 
        //{
        //    if (Program.IsEqualInTolerance(subDeterminant, 0))
        //        return IsRegularNonCentral();
        //    return IsRegularCentral();
        //}
        //private Quadric IsSingular(double subDeterminant) 
        //{
        //    if (Program.IsEqualInTolerance(subDeterminant, 0))
        //        return IsSingularNonCentral();
        //    return IsSingularCentral();
        //}
        //private Quadric IsRegularCentral() 
        //{
        //    // imaginární elipsoid, elipsoid, hyperboloid2, hyperboloid1.
        //}
        //private Quadric IsRegularNonCentral()
        //{
        //    // hyperbolický paraboloid, eliptický paraboloid.
        //}
        //private Quadric IsSingularCentral()
        //{
        //    // kuželová plocha, imaginární kuželová plocha.
        //}
        //private Quadric IsingularNonCentral()
        //{  
        //    // imaginární eliptická válcová plocha, eliptická válcová plocha, 
        //    // imaginární různoběžné roviny, různoběžné roviny, hyperbolická válcová plocha,
        //    // parabolická válcová plocha, paralelní roviny, dvojnásobná rovina
        //    // imaginární paralelní roviny.
        //}


        /////////////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////////////
    }
}