﻿using System;
using System.Collections.Generic;
using System.Linq;
using Tao.OpenGl;
using Geometry.GraphicObjects.Matrixes;
using Geometry.GraphicObjects.Equations;
using Geometry.BaseClasses;
using GMath;

namespace Geometry.GraphicObjects.Quadrics
{
    /// <summary>
    /// Kuzelová plocha.
    /// Kanonická rovnice:      x*x/a*a + y*y/b*b - z*z/c*c = 0;
    /// 
    /// Parametrická Rovnice:   x = m + a*t*cos(u).
    ///                         y = n + a*t*sin(u).
    ///                         z = p + c*t. (u = (0,2PI), t=R).
    /// </summary>
    public class Cone : Quadric
    {
        private double _a;
        private double _b;
        private double _c;
        private double _baseQuality = 20;

        public static string canonicFormatString = "E[x-!m;!a]L[+]E[y-!n;!b]L[-]E[z-!o;!c]L[ = 0]";
        public static string parametricXFormatString = "L[x =!m+!a" + U.Times + "v" + U.Times + "cos(u)]";
        public static string parametricYFormatString = "L[y =!n+!b" + U.Times + "v" + U.Times + "sin(u)]";
        public static string parametricZFormatString = "L[z =!o+!c" + U.Times + "v]";
        public static string _wordDescriptionFirstCase = "Kuželová plocha";
        public static string _wordDescriptionSecondCase = "Kuželové plochy";
        public static string _wordDescriptionFourthCase = "Kuželovou plochu";

        public Cone(CommonEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix, GColor color, string name, int ID)
            : base(commonEquation, parametricEquation, toCanonicTransformationMatrix, color, name,
                        _wordDescriptionFirstCase, _wordDescriptionSecondCase, _wordDescriptionFourthCase,
                        canonicFormatString,
                        parametricXFormatString, parametricYFormatString, parametricZFormatString, ID)
        {
            _a = parametricEquation.GetA();
            _b = parametricEquation.GetB();
            _c = parametricEquation.GetC();
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        ///////////////////////////////////////////////////////////////////////////////////////// 

        /// <summary>
        /// a*v*cos(u)
        /// </summary>
        protected double ComputeXCoordinateAt(double u, double v)
        {
            return _a*Math.Cos(u)*v;
        }
        /// <summary>
        /// b*sin(u)*v
        /// </summary>
        protected double ComputeYCoordinateAt(double u, double v)
        {
            return Math.Sin(u) * v * _b ;
        }
        /// <summary>
        /// c*v
        /// </summary>
        protected double ComputeZCoordinateAt(double u, double v)
        {
            return _c * v;
        }

        /// <summary>
        /// Dosadi do parametricke rovnice u a v a vraci souradnice bodu.
        /// </summary>
        /// <param name="u"></param>
        /// <param name="v"></param>
        /// <returns></returns>
        protected Vector ComputeCoordinatesAt(double u, double v)
        {
            double x = ComputeXCoordinateAt(u, v);
            double y = ComputeYCoordinateAt(u, v);
            double z = ComputeZCoordinateAt(u, v);

            return new Vector(x, y, z);
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        ///////////////////////////////////////////////////////////////////////////////////////// 

        protected Vector ComputeNormalAt(Vector vertex)
        {
            double x = -2 * vertex.GetX() / (_a * _a);
            double y = -2 * vertex.GetY() / (_b * _b);
            double z = 2 * vertex.GetZ() / (_c * _c);
            if (x == 0 && y == 0 && z == 0)
                z = -1;
            Vector normal = new Vector(x, y, z);
            normal.NornVector();
            return normal * _normalSign;
        }

        protected void SetVertexAndNormal(double u, double v)
        {
            Vector vertex = ComputeCoordinatesAt(u, v);
            Vector normal = ComputeNormalAt(vertex);
            Gl.glNormal3d(normal.GetX(), normal.GetY(), normal.GetZ());
            Gl.glVertex3d(vertex.GetX(), vertex.GetY(), vertex.GetZ());
        }


        /////////////////////////////////////////////////////////////////////////////////////////
        ///////////////////////////////////////////////////////////////////////////////////////// 

        protected override void BuildParametricDisplayList(Scene scene, RelativeQuality relativeQuality)
        {
            double cback = GetClippingBack(scene);
            double cfront = GetClippingFront(scene);
            double cleft = GetClippingLeft(scene);
            double cright = GetClippingRight(scene);
            double ctop = GetClippingTop(scene);
            double cbottom = GetClippingBottom(scene);

            _displayListIdentificator = Gl.glGenLists(1);
            Gl.glNewList(_displayListIdentificator, Gl.GL_COMPILE);
            Gl.glEnable(Gl.GL_LIGHTING);

            Interval leftRight = new Interval(cleft / _a, cright / _a); // interval u pro ktere je hyperbola mezi -x a x
            Interval topBottom = new Interval(cbottom / _b, ctop / _b); // interval u pro ktere je hyperbola mezi -y a y
            Interval leftRight2 = new Interval(-cleft / _a, -cright / _a); // interval u pro ktere je hyperbola mezi -x a x
            Interval topBottom2 = new Interval(-cbottom / _b, -ctop / _b); // interval u pro ktere je hyperbola mezi -y a y

            Interval rest1 = Interval.UnionOfIntersectingIntervals(leftRight, topBottom);
            Interval rest2 = Interval.UnionOfIntersectingIntervals(leftRight2, topBottom2);
            Interval rest = Interval.UnionOfIntersectingIntervals(rest1, rest2);
            Interval backFront = new Interval(cback / _c, cfront / _c); // interval u pro ktere je hyperbola mezi -z a z

            double u0 = ComputeZCoordinateAt(0, 0);

            double uStep = Math.PI / (_baseQuality * (scene.GetBaseQuality() + relativeQuality));
            int linesPerInterval = 50;
            double vStep = backFront.GetSize() / (linesPerInterval * (scene.GetBaseQuality() + relativeQuality));

            if (u0 < backFront.GetA() || u0 > backFront.GetB()) // ve scene je jen leva cast nebo jen prava cast
            {
                Interval interval = Interval.Intersect(backFront, rest);
                double vMin = interval.GetA();
                double vMax = interval.GetB();
                for (double v = vMin; v < vMax; v += vStep)
                {
                    Gl.glBegin(Gl.GL_TRIANGLE_STRIP);

                    if (v > vMax - vStep)
                        v = vMax - vStep;
                    for (double u = 0; u <= Math.PI * 2 + uStep; u += uStep)
                    {
                        if (u > Math.PI * 2 + uStep)
                            u = Math.PI * 2 + uStep;
                        SetVertexAndNormal(u, v);
                        SetVertexAndNormal(u, v + vStep);
                        if (u == Math.PI * 2 + uStep)
                            break;
                    }
                    Gl.glEnd();
                    if (v == vMax - vStep)
                        break;
                }
            }
            if (backFront.Belongs(u0)) // ve scene jsou obe casti
            {
                double vMax = new List<double>() { cleft / _a, -cleft / _a, ctop / _b, -ctop / _b }.Max();
                vMax = Math.Min(vMax, cfront / _c);
                Gl.glBegin(Gl.GL_TRIANGLE_FAN);
                SetVertexAndNormal(0, 0);
                for (double u = Math.PI * 2; u >= -uStep; u -= uStep)
                {
                    if (u < uStep)
                        u = 0;
                    SetVertexAndNormal(u, -vStep);
                    if (u == 0)
                        break;
                }
                Gl.glEnd();

                for (double v = vStep; v < vMax; v += vStep)
                {
                    Gl.glBegin(Gl.GL_TRIANGLE_STRIP);

                    if (v > vMax - vStep)
                        v = vMax - vStep;
                    for (double u = 0; u <= Math.PI * 2 + uStep; u += uStep)
                    {
                        if (u > Math.PI * 2 + uStep)
                            u = Math.PI * 2 + uStep;
                        SetVertexAndNormal(u, v);
                        SetVertexAndNormal(u, v + vStep);
                        if (u == Math.PI * 2 + uStep)
                            break;
                    }
                    Gl.glEnd();
                    if (v == vMax - vStep)
                        break;
                }
                Gl.glPushMatrix();
                Gl.glRotated(180, 1, 0, 0);
                double vMin = new List<double>() { cleft / _a, -cleft / _a, ctop / _b, -ctop / _b }.Min();
                vMin = Math.Max(vMin, cback / _c);
                Gl.glBegin(Gl.GL_TRIANGLE_FAN);
                SetVertexAndNormal(0, 0);
                for (double u = Math.PI * 2; u >= -uStep; u -= uStep)
                {
                    if (u < uStep)
                        u = 0;
                    SetVertexAndNormal(u, -vStep);
                    if (u == 0)
                        break;
                }
                Gl.glEnd();
                Gl.glPopMatrix();
                for (double v = vMin; v < 0; v += vStep)
                {
                    Gl.glBegin(Gl.GL_TRIANGLE_STRIP);
                    if (v > -vStep*2)
                        v = -vStep*2;
                    for (double u = 0; u <= Math.PI * 2 + uStep; u += uStep)
                    {
                        if (u > Math.PI * 2 + uStep)
                            u = Math.PI * 2 + uStep;
                        SetVertexAndNormal(u, v);
                        SetVertexAndNormal(u, v + vStep);
                        if (u == Math.PI * 2 + uStep)
                            break;
                    }
                    Gl.glEnd();
                    if (v == -vStep*2)
                        break;
                }
            }
            Gl.glEndList();
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        ///////////////////////////////////////////////////////////////////////////////////////// 


        protected override Quadric CallConstructor(CommonEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix, GColor color, string name, int ID)
        {
            return new Cone(commonEquation, parametricEquation, toCanonicTransformationMatrix, color, name, ID);
        }

        public override QuadricMatrix GetCanonicMatrix(QuadricParametricEquation parametricEquation)
        {
            double A = 1 / (parametricEquation.GetA() * parametricEquation.GetA());
            double B = 1 / (parametricEquation.GetB() * parametricEquation.GetB());
            double C = -1 / (parametricEquation.GetC() * parametricEquation.GetC());
            QuadricEquation newCanonicEquation = new QuadricEquation(A, B, C, 0, 0, 0, 0, 0, 0, 0);
            return new QuadricMatrix(newCanonicEquation);
        }
    }
}
