﻿using System;
using Tao.OpenGl;
using Geometry.GraphicObjects.Matrixes;
using Geometry.GraphicObjects.Equations;
using Geometry.BaseClasses;
using GMath;

namespace Geometry.GraphicObjects.Quadrics
{
    /// <summary>
    /// Kuzelová plocha.
    /// Kanonická rovnice:      x*x/a*a + y*y/b*b + z*z/c*c = 0;
    /// 
    /// Parametrická Rovnice:   x = m + a*t*cos(u).
    ///                         y = n + a*t*sin(u).
    ///                         z = p + c*t. (u = (0,2PI), t=R).
    /// </summary>
    public class ImaginarCone : Quadric
    {
        private double _a;
        private double _b;
        private double _c;
        private const double _radius = 0.5;
        private double _baseQuality = 20;

        public static string canonicFormatString = "E[x-!m;!a]L[+]E[y-!n;!b]L[+]E[z-!o;!c]L[ = 0]";
        public static string parametricXFormatString = "L[x =!m]";
        public static string parametricYFormatString = "L[y =!n]";
        public static string parametricZFormatString = "L[z =!o]";
        public static string _wordDescriptionFirstCase = "Imaginární kuželová plocha";
        public static string _wordDescriptionSecondCase = "Imaginární kuželové plochy";
        public static string _wordDescriptionFourthCase = "Imaginární kuželovou plochu";


        public ImaginarCone(CommonEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix,
            GColor color, string name, int ID)
            : base(commonEquation, parametricEquation, toCanonicTransformationMatrix,
                        color, name,
                        _wordDescriptionFirstCase, _wordDescriptionSecondCase, _wordDescriptionFourthCase,
                        canonicFormatString,
                        parametricXFormatString, parametricYFormatString, parametricZFormatString, ID)
        {
            _a = parametricEquation.GetA();
            _b = parametricEquation.GetB();
            _c = parametricEquation.GetC();
        }


        protected double ComputeXCoordinateAt(double u, double v)
        {
            return Math.Cos(v) * Math.Sin(u) * _radius;
        }

        protected double ComputeYCoordinateAt(double u, double v)
        {
            return Math.Sin(v) * _radius;
        }

        protected double ComputeZCoordinateAt(double u, double v)
        {
            return Math.Cos(v) * Math.Cos(u) * _radius;
        }

        /// <summary>
        /// Dosadi do parametricke rovnice u a v a vraci souradnice bodu.
        /// </summary>
        /// <param name="u"></param>
        /// <param name="v"></param>
        /// <returns></returns>
        protected Vector ComputeCoordinatesAt(double u, double v)
        {
            double x = ComputeXCoordinateAt(u, v);
            double y = ComputeYCoordinateAt(u, v);
            double z = ComputeZCoordinateAt(u, v);

            return new Vector(x, y, z);
        }


        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////

        private void SetVextexAndHisNormal(double u, double v)
        {
            Vector v1 = ComputeCoordinatesAt(u, v);
            Vector normal = ComputeNormalAt(v1);
            Gl.glNormal3d(normal.GetX(), normal.GetY(), normal.GetZ());
            Gl.glVertex3d(v1.GetX(), v1.GetY(), v1.GetZ());
        }

        protected Vector ComputeNormalAt(Vector vertex)
        {
            double x = 2 * vertex.GetX() / (_a * _a);
            double y = 2 * vertex.GetY() / (_b * _b);
            double z = 2 * vertex.GetZ() / (_c * _c);
            Vector normal = new Vector(x, y, z);
            normal.NornVector();
            return normal * _normalSign;
        }

        private void ShowNormal(double u, double v, double delta)
        {
            Vector v1 = ComputeCoordinatesAt(u, v);
            Vector normal = ComputeNormalAt(v1);
            Gl.glVertex3d(v1.GetX(), v1.GetY(), v1.GetZ());
            Gl.glVertex3d(v1.GetX() + normal.GetX(), v1.GetY() + normal.GetY(), v1.GetZ() + normal.GetZ());
        }

        protected override void BuildParametricDisplayList(Scene scene, RelativeQuality relativeQuality)
        {
            _displayListIdentificator = Gl.glGenLists(1);
            Gl.glNewList(_displayListIdentificator, Gl.GL_COMPILE);

            double step = Math.PI / (_baseQuality * (scene.GetBaseQuality() + relativeQuality));
            // horni pulka
            for (double v = Math.PI / 2; v > -Math.PI / 2; v -= step)
            {
                if (v < -Math.PI / 2 + step)
                    v = -Math.PI / 2 + step;
                Gl.glBegin(Gl.GL_TRIANGLE_STRIP);
                for (double u = 0; u < Math.PI * 2 + step; u += step)
                {
                    if (u > Math.PI * 2)
                        u = Math.PI * 2;
                    SetVextexAndHisNormal(u, v);
                    SetVextexAndHisNormal(u, v - step);
                    if (u == Math.PI * 2)
                        break;
                }
                Gl.glEnd();
                if (v == -Math.PI / 2 + step)
                    break;
            }
            Gl.glEndList();
        }

        protected override Quadric CallConstructor(CommonEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix, GColor color, string name, int ID)
        {
            return new ImaginarCone(commonEquation, parametricEquation, toCanonicTransformationMatrix, color, name, ID);
        }

        public override QuadricMatrix GetCanonicMatrix(QuadricParametricEquation parametricEquation)
        {
            double A = 1 / (parametricEquation.GetA() * parametricEquation.GetA());
            double B = 1 / (parametricEquation.GetB() * parametricEquation.GetB());
            double C = 1 / (parametricEquation.GetC() * parametricEquation.GetC());
            QuadricEquation newCanonicEquation = new QuadricEquation(A, B, C, 0, 0, 0, 0, 0, 0, 0);
            return new QuadricMatrix(newCanonicEquation);
        }
    }
}
