﻿using System;
using Tao.OpenGl;
using Geometry.GraphicObjects.Matrixes;
using Geometry.GraphicObjects.Equations;
using Geometry.BaseClasses;
using GMath;

namespace Geometry.GraphicObjects.Quadrics
{
    /// <summary>
    /// Eliptická válcová plocha.
    /// Kanonická rovnice:       x*x/a*a + y*y/b*b -1 = 0; 
    /// Parametricka Rovnice:    x = m + a*cos(u).
    ///                          y = n + b*sin(u).
    ///                          z = p + v. (u = (0,2PI), v=R).
    /// </summary>
    public class ElipticCylinder : Quadric
    {
        private double _a;
        private double _b;
        private int _linesPerInterval = 20;
        private int _stripesPerInterval = 5;

        public static string canonicFormatString = "E[x-!m;!a]L[+]E[y-!n;!b]L[ = 1]";
        public static string parametricXFormatString = "L[x = !m+!a" + U.Times + "cos(u)]";
        public static string parametricYFormatString = "L[y = !n+!b" + U.Times + "sin(u)]";
        public static string parametricZFormatString = "L[z =v]";
        public static string _wordDescriptionFirstCase = "Eliptická válcová plocha";
        public static string _wordDescriptionSecondCase = "Eliptické válcové plochy";
        public static string _wordDescriptionFourthCase = "Eliptickou válcovou plochu";

        public ElipticCylinder(CommonEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix,
            GColor color, string name, int ID)
            : base(commonEquation, parametricEquation, toCanonicTransformationMatrix,
                        color, name,
                        _wordDescriptionFirstCase, _wordDescriptionSecondCase, _wordDescriptionFourthCase,
                        canonicFormatString,
                        parametricXFormatString, parametricYFormatString, parametricZFormatString, ID) 
        {
            _a = parametricEquation.GetA();
            _b = parametricEquation.GetB();
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////

        /// <summary>
        /// a*cos(u)
        /// </summary>
        protected double ComputeXCoordinateAt(double u, double v)
        {
            return _a * Math.Cos(u);
        }
        /// <summary>
        /// b*sin(u)
        /// </summary>
        protected double ComputeYCoordinateAt(double u, double v)
        {
            return _b * Math.Sin(u);
        }
        /// <summary>
        /// v
        /// </summary>
        protected double ComputeZCoordinateAt(double u, double v)
        {
            return v;
        }

        /// <summary>
        /// Dosadi do parametricke rovnice u a v a vraci souradnice bodu.
        /// </summary>
        /// <param name="u"></param>
        /// <param name="v"></param>
        /// <returns></returns>
        protected Vector ComputeCoordinatesAt(double u, double v)
        {
            double x = ComputeXCoordinateAt(u, v);
            double y = ComputeYCoordinateAt(u, v);
            double z = ComputeZCoordinateAt(u, v);

            return new Vector(x, y, z);
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////

        protected Vector ComputeNormalAt(Vector vertex)
        {
            double x = 2 * vertex.GetX() / (_a * _a);
            double y = 2 * vertex.GetY() / (_b * _b);
            double z = 0;
            Vector normal = new Vector(x, y, z);
            normal.NornVector();
            return normal * -_normalSign;
        }

        protected void SetVertexAndNormal(double u, double v)
        {
            Vector vertex = ComputeCoordinatesAt(u, v);
            Vector normal = ComputeNormalAt(vertex);
            Gl.glNormal3d(normal.GetX(), normal.GetY(), normal.GetZ());
            Gl.glVertex3d(vertex.GetX(), vertex.GetY(), vertex.GetZ());
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////

        protected override void BuildParametricDisplayList(Scene scene, RelativeQuality relativeQuality)
        {
            _displayListIdentificator = Gl.glGenLists(1);
            Gl.glNewList(_displayListIdentificator, Gl.GL_COMPILE);

            double back = GetClippingBack(scene);
            double front = GetClippingFront(scene);

            double uStep = Math.PI / (_linesPerInterval * (scene.GetBaseQuality() + relativeQuality));
            int stripesPerInterval = _stripesPerInterval * (scene.GetBaseQuality() + relativeQuality);
            double vStep = (front - back) / stripesPerInterval;

            for (double v = back; v < front; v += vStep)
            {
                if (v > front - vStep)
                    v = front - vStep;
                Gl.glBegin(Gl.GL_TRIANGLE_STRIP);
                 
                for (double u = 0; u < Math.PI * 2 + uStep; u += uStep)
                {
                    if (u > Math.PI * 2)
                        u = Math.PI * 2;
                    SetVertexAndNormal(u,v);
                    SetVertexAndNormal(u,v+vStep);
                    if (u == Math.PI * 2)
                        break;
                }
                Gl.glEnd();
                if (v == front - vStep)
                    break;
            }

            Gl.glEndList();
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////

        protected override Quadric CallConstructor(CommonEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix, GColor color, string name, int ID)
        {
            return new ElipticCylinder(commonEquation, parametricEquation, toCanonicTransformationMatrix, color, name, ID);
        }

        public override QuadricMatrix GetCanonicMatrix(QuadricParametricEquation parametricEquation)
        {
            double A = 1 / (parametricEquation.GetA() * parametricEquation.GetA());
            double B = 1 / (parametricEquation.GetB() * parametricEquation.GetB());
            QuadricEquation newCanonicEquation = new QuadricEquation(A, B, 0, 0, 0, 0, 0, 0, 0, -1);
            return new QuadricMatrix(newCanonicEquation);
        }
    }
}

