﻿using System;
using System.Collections.Generic;
using Tao.OpenGl;
using Geometry.GraphicObjects.Matrixes;
using Geometry.GraphicObjects.Equations;
using Geometry.BaseClasses;
using GMath;

namespace Geometry.GraphicObjects.Quadrics
{
    /// <summary>
    /// Hyperbolická válcová plocha.
    /// Kanonická rovnice:          x*x/a*a - y*y/b*b -1 =0;
    /// 
    /// Parametrická rovnice: 
    ///                 x	=	a*sinh(u)	
    ///                 y	=	b*cosh(u)	
    ///                 z	=	v.
    ///                 u = (-oo,oo); v = (-oo,oo);
    /// </summary>
    public class HyperbolicCylinder:SimpleRulledQuadric
    {
        private double _a;
        private double _b;
        private int _stripesPerIterval = 5;
        private int _linesPerInterval = 50;

        public static string canonicFormatString = "E[x-!m;!a]L[-]E[y-!n;!b]L[ = 1]";
        public static string parametricXFormatString = "L[x =!m" + U.PM + "!a" + U.Times + "cosh(u)]";
        public static string parametricYFormatString = "L[y =!n+!b" + U.Times + "sinh(u)]";
        public static string parametricZFormatString = "L[z =v]";
        public static string _wordDescriptionFirstCase = "Hyperbolická válcová plocha";
        public static string _wordDescriptionSecondCase = "Hyperbolické válcové plochy";
        public static string _wordDescriptionFourthCase = "Hyperbolickou válcovou plochu";

        const int LEFT = 0;
        const int RIGHT = 1;
        public struct IntervalPair
        {
            public Interval Interval;
            public int Part;
            public IntervalPair(Interval interval, int part)
            {
                Interval = interval;
                Part = part;
            }
        }
        public struct VectorPair 
        {
            public List<Vector> Vectors;
            public int Part;
            public VectorPair(List<Vector> vectors, int part)
            {
                Vectors = vectors;
                Part = part;
            }
        }

        public HyperbolicCylinder(CommonEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix,
            GColor color, string name, int ID)
            : base(commonEquation, parametricEquation, toCanonicTransformationMatrix,
                        color, name,
                        _wordDescriptionFirstCase, _wordDescriptionSecondCase, _wordDescriptionFourthCase,
                        canonicFormatString,
                        parametricXFormatString, parametricYFormatString, parametricZFormatString, ID)
        {
            _a = parametricEquation.GetA();
            _b = parametricEquation.GetB();
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        ///////////////////////////////////////////////////////////////////////////////////////// 

        /// <summary>
        /// a*Sinh(u)
        /// </summary>
        protected double ComputeXCoordinateAt(double u, double v)
        {
            return _a * Math.Cosh(u);           
            
        }
        /// <summary>
        /// b*Cosh(u)
        /// </summary>
        protected double ComputeYCoordinateAt(double u, double v)
        {            
            return _b * Math.Sinh(u);
        }
        /// <summary>
        /// v
        /// </summary>
        protected double ComputeZCoordinateAt(double u, double v)
        {
            return v;
        }

        /// <summary>
        /// Dosadi do parametricke rovnice u a v a vraci souradnice bodu.
        /// </summary>
        /// <param name="u"></param>
        /// <param name="v"></param>
        /// <returns></returns>
        protected Vector ComputeCoordinatesAt(double u, double v)
        {
            double x = ComputeXCoordinateAt(u, v);
            double y = ComputeYCoordinateAt(u, v);
            double z = ComputeZCoordinateAt(u, v);

            return new Vector(x, y, z);
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        ///////////////////////////////////////////////////////////////////////////////////////// 

        /// <summary>
        /// Normala prislusna prava casti.
        /// </summary>
        /// <param name="vertex"></param>
        /// <returns></returns>
        protected Vector ComputeNormalAt(Vector vertex)
        {
            double x = -2 * vertex.GetX() / (_a * _a);
            double y = 2 * vertex.GetY() / (_b * _b);
            double z = 0;
            Vector normal = new Vector(x, y, z);
            normal.NornVector();
            return normal * _normalSign;
        }

        private void SetVertexAndHisNormal(Vector vertex, int part) 
        {
            Vector normal = ComputeNormalAt(vertex);
            if (part == LEFT)
            {
                Gl.glNormal3d(normal.GetX(), -normal.GetY(), normal.GetZ());
                Gl.glVertex3d(-vertex.GetX(), vertex.GetY(), vertex.GetZ());
            }
            else
            {
                Gl.glNormal3d(normal.GetX(), normal.GetY(), normal.GetZ());
                Gl.glVertex3d(vertex.GetX(), vertex.GetY(), vertex.GetZ());
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        ///////////////////////////////////////////////////////////////////////////////////////// 

        private List<VectorPair> GetPoints(Scene scene, RelativeQuality relativeQuality)
        {
            List<IntervalPair> clippingIntervalsPairs = GetClippingIntervals(scene);
            List<VectorPair> points = new List<VectorPair>();
            double uMin;
            double uMax;
            double step; // = Math.PI / (_baseQuality * (scene.GetBaseQuality() + relativeQuality));
            int linesPerInterval = _linesPerInterval * (scene.GetBaseQuality() + relativeQuality);
            foreach (IntervalPair p in clippingIntervalsPairs)
            {
                if (p.Interval.IsEmpty())
                    continue;
                step = (p.Interval.Size() / linesPerInterval);
                uMin = p.Interval.GetA();
                uMax = p.Interval.GetB();
                List<Vector> hyperbolaPart = new List<Vector>();
                for (double u = uMin; u <= uMax; u += 0)
                {
                    hyperbolaPart.Add(new Vector(ComputeXCoordinateAt(u, 0), ComputeYCoordinateAt(u,0), ComputeZCoordinateAt(u, 0)));
                    if (u == uMax)
                        break;
                    u = Math.Min(u + step, uMax);
                }
                points.Add(new VectorPair(hyperbolaPart, p.Part));
            }
            return points;
        }
        private List<IntervalPair> GetClippingIntervals(Scene scene)
        {
            double cleft = GetClippingLeft(scene);
            double cright = GetClippingRight(scene);
            double ctop = GetClippingTop(scene);
            double cbottom = GetClippingBottom(scene);

            double temp = ArcCosh(cright / _a);
            Interval right = new Interval(-temp, temp); // "u" ve kterych hyperbola protina pravou hranici
            temp = ArcCosh(cleft / _a); // muze byt NAN
            Interval left = new Interval(-temp, temp); // "u" ve kterych hyperbola protina levou hranici

            Interval bottomTop = new Interval(ArcSinh(cbottom / _b), ArcSinh(ctop / _b)); // interval u pro ktere je hyperbola mezi up a down

            double leftPeak = -ComputeXCoordinateAt(0, 0);
            double rightPeak = ComputeXCoordinateAt(0, 0);

            List<IntervalPair> pairs = new List<IntervalPair>();

            // leva hyperbola je uvnitr
            if (left.GetA() != double.NaN || (bottomTop.GetA() != double.NaN && bottomTop.GetB() != double.NaN))
            {
                if (new Interval(cleft, cright).Belongs(leftPeak))
                {
                    pairs.Add(new IntervalPair(Interval.Intersect(left, bottomTop), LEFT)); // vrchol leve hyperboly je uvnitr
                }
                else if (leftPeak > cright)
                {
                    List<Interval> tempInterval = Interval.MinusSmallerFromBigger(left, right);
                    foreach (Interval i in tempInterval)
                        pairs.Add(new IntervalPair(Interval.Intersect(i, bottomTop), LEFT)); // leva hyperbola je rozdelena na dve
                }
            }
            if (right.GetA() != double.NaN || (bottomTop.GetA() != double.NaN && bottomTop.GetB() != double.NaN))
            {
                if (new Interval(cleft, cright).Belongs(rightPeak))
                {
                    pairs.Add(new IntervalPair(Interval.Intersect(right, bottomTop), RIGHT)); // vrchol prave hyperbooly je uvnitr
                }
                else if (rightPeak < cleft)
                {
                    List<Interval> tempInterval = Interval.MinusSmallerFromBigger(left, right);
                    foreach (Interval i in tempInterval)
                        pairs.Add(new IntervalPair(Interval.Intersect(i, bottomTop), RIGHT)); // prava hyperbola je rozdelena na dve
                }
            }
            return pairs;
        }
        protected override void BuildParametricDisplayList(Scene scene, RelativeQuality relativeQuality)
        {
            List<VectorPair> points = GetPoints(scene, relativeQuality);

            _displayListIdentificator = Gl.glGenLists(1);
            Gl.glNewList(_displayListIdentificator, Gl.GL_COMPILE);
            
            double back = GetClippingBack(scene);
            double front = GetClippingFront(scene);
            int stripesPerIterval = _stripesPerIterval * (scene.GetBaseQuality() + relativeQuality);
            double vStep = (front - back) / stripesPerIterval;

            for (double v = back; v < front; v += vStep)
            {
                if (v > front - vStep)
                    v = front - vStep;

                foreach (VectorPair subPoints in points)
                {
                    int part = subPoints.Part;
                    Gl.glBegin(Gl.GL_TRIANGLE_STRIP);
                    for (int i = 0; i < subPoints.Vectors.Count; i++)
                    {
                        Vector coordinate = subPoints.Vectors[i];
                        coordinate.SetZ(v);
                        SetVertexAndHisNormal(coordinate, part);
                        coordinate.SetZ(v + vStep);
                        SetVertexAndHisNormal(coordinate, part);
                    }
                    Gl.glEnd();
                }
                if (v == front - vStep)
                    break;
            }

            Gl.glEndList();
        }




        protected override Quadric CallConstructor(CommonEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix, GColor color, string name, int ID)
        {
            return new HyperbolicCylinder(commonEquation, parametricEquation, toCanonicTransformationMatrix, color, name, ID);
        }

        public override QuadricMatrix GetCanonicMatrix(QuadricParametricEquation parametricEquation)
        {
            double A = 1 / (parametricEquation.GetA() * parametricEquation.GetA());
            double B = -1 / (parametricEquation.GetB() * parametricEquation.GetB());
            QuadricEquation newCanonicEquation = new QuadricEquation(A, B, 0, 0, 0, 0, 0, 0, 0, -1);
            return new QuadricMatrix(newCanonicEquation);
        }

        public QuadricMatrix GetCanonicMatrix() 
        {
            double A = 1 / (_a*_a);
            double B = -1 / (_b * _b);
            QuadricEquation newCanonicEquation = new QuadricEquation(A, B, 0, 0, 0, 0, 0, 0, 0, -1);
            return new QuadricMatrix(newCanonicEquation);
        }
    }
}
