﻿using System;
using System.Collections.Generic;
using Tao.OpenGl;
using Geometry.GraphicObjects.Matrixes;
using Geometry.GraphicObjects.Equations;
using Geometry.BaseClasses;
using GMath;

namespace Geometry.GraphicObjects.Quadrics
{
    /// <summary>
    /// kanonicky tvar: a*a -+z/p = 0;
    /// parametricky tvar?:
    ///     x = u
    ///     y = u*u*p/a*a
    ///     z = v;
    /// </summary>
    public class ParabolicCylinder:SimpleRulledQuadric
    {
        private double _p;
        private double _a;
        private int _linesPerInterval = 50;
        private int _stripesPerInterval = 5;

        public static string canonicFormatString = "E[x-!m;!a]L[+]L[2" + U.Times + "!p" + U.Times + "(y-!n)=0]";
        public static string parametricXFormatString = "L[x = !m + u]";
        public static string parametricYFormatString = "L[y = !n - ]P[u;!a;!p]";
        public static string parametricZFormatString = "L[z = v]";
        public static string _wordDescriptionFirstCase = "Parabolická válcová plocha";
        public static string _wordDescriptionSecondCase = "Parabolické válcové plochy";
        public static string _wordDescriptionFourthCase = "Parabolickou válcovou plochu";


        public ParabolicCylinder(CommonEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix,
            GColor color, string name, int ID)
            : base(commonEquation, parametricEquation, toCanonicTransformationMatrix,
                        color, name,
                        _wordDescriptionFirstCase, _wordDescriptionSecondCase, _wordDescriptionFourthCase,
                        canonicFormatString,
                        parametricXFormatString, parametricYFormatString, parametricZFormatString, ID)
        {
            _a = parametricEquation.GetA();
            _p = parametricEquation.GetP();
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////

        /// <summary>
        /// u
        /// </summary>
        protected double ComputeXCoordinateAt(double u, double v)
        {
            return u;
        }
        /// <summary>
        /// (u*u)/2p;
        /// </summary>
        protected double ComputeYCoordinateAt(double u, double v)
        {
            return -(u * u) / ((_a*_a) * (2 * _p));
        }
        /// <summary>
        /// v
        /// </summary>
        protected double ComputeZCoordinateAt(double u, double v)
        {
            return v;
        }

        /// <summary>
        /// Dosadi do parametricke rovnice u a v a vraci souradnice bodu.
        /// </summary>
        /// <param name="u"></param>
        /// <param name="v"></param>
        /// <returns></returns>
        protected Vector ComputeCoordinatesAt(double u, double v)
        {
            double x = ComputeXCoordinateAt(u, v);
            double y = ComputeYCoordinateAt(u, v);
            double z = ComputeZCoordinateAt(u, v);

            return new Vector(x, y, z);
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////

        protected Vector ComputeNormalAt(Vector vertex)
        {
            double x = 2 * vertex.GetX() / (_a * _a);
            double y = 2 * _p;
            double z = 0;
            Vector normal = new Vector(x, y, z);
            normal.NornVector();
            return normal * _normalSign * Math.Sign(_p);
        }

        protected void SetVertexAndNormal(double u, double v)
        {
            Vector vertex = ComputeCoordinatesAt(u, v);
            Vector normal = ComputeNormalAt(vertex);
            Gl.glNormal3d(normal.GetX(), normal.GetY(), normal.GetZ());
            Gl.glVertex3d(vertex.GetX(), vertex.GetY(), vertex.GetZ());
        }


        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////

        protected override void BuildParametricDisplayList(Scene scene, RelativeQuality relativeQuality)
        {
            double cback = GetClippingBack(scene);
            double cfront = GetClippingFront(scene);
            double cleft = GetClippingLeft(scene);
            double cright = GetClippingRight(scene);
            double ctop = GetClippingTop(scene);
            double cbottom = GetClippingBottom(scene);

            int stripesPerInterval = _stripesPerInterval * (scene.GetBaseQuality() + relativeQuality);
            int linesPerInterval = _linesPerInterval * (scene.GetBaseQuality() + relativeQuality);            
            double vStep = (cfront - cback) / stripesPerInterval;
            double uStep;
            List<Interval> uIntervals = GetUIntervals(scene, relativeQuality);

            double uMin, uMax;
            _displayListIdentificator = Gl.glGenLists(1);
            Gl.glNewList(_displayListIdentificator, Gl.GL_COMPILE);


            for (double v = cback; v < cfront; v += vStep)
            {
                if (v > cfront - vStep)
                    v = cfront - vStep;
                foreach (Interval uInterval in uIntervals)
                {
                    if (uInterval.IsEmpty())
                        continue;
                    uStep = uInterval.GetSize() / linesPerInterval;
                    uMin = uInterval.GetA();
                    uMax = uInterval.GetB();
                    Gl.glBegin(Gl.GL_TRIANGLE_STRIP);
                    for (double u = uMin; u < uMax + uStep; u += uStep)
                    {
                        if (u > uMax)
                            u = uMax;
                        SetVertexAndNormal(u, v);
                        SetVertexAndNormal(u, v + vStep);
                        if (u == uMax)
                            break;
                    }
                    Gl.glEnd();
                }
                if (v == cfront - vStep)
                    break;
            }

            Gl.glEndList();
        }
        private List<Interval> GetUIntervals(Scene scene, RelativeQuality relativeQuality)
        {
            double cleft = GetClippingLeft(scene);
            double cright = GetClippingRight(scene);
            double ctop = GetClippingTop(scene);
            double cbottom = GetClippingBottom(scene);
            double c = _a * _a * 2 * _p;

            Interval xInterval = new Interval(cleft, cright);
            List<Interval> uIntervals = new List<Interval>();

            QuadraticEquation topIntervalEquation = new QuadraticEquation(1, 0, ctop * c);
            QuadraticEquation bottomIntervalEquation = new QuadraticEquation(1, 0, cbottom * c);
            List<double> topIntervalSolutions = topIntervalEquation.Solve();
            List<double> bottomIntervalSolutions = bottomIntervalEquation.Solve();
            if (topIntervalSolutions.Count != 2 && bottomIntervalSolutions.Count != 2) // parabola je mimo ctverec
            {
            }
            if (bottomIntervalSolutions.Count == 2 && topIntervalSolutions.Count == 2) // parabola je uvnitr ale vrchol ne
            {
                Interval topInterval = new Interval(topIntervalSolutions[0], topIntervalSolutions[1]);
                Interval bottomInterval = new Interval(bottomIntervalSolutions[0], bottomIntervalSolutions[1]);
                List<Interval> yIntervals = Interval.MinusSmallerFromBigger(topInterval, bottomInterval);
                uIntervals = Interval.Intersect(yIntervals, xInterval);
            }
            if (bottomIntervalSolutions.Count == 2 && topIntervalSolutions.Count != 2) // parabola uvnitr (kladne _p)
            {
                Interval yInterval = new Interval(bottomIntervalSolutions[0], bottomIntervalSolutions[1]);
                uIntervals = new List<Interval>() { Interval.Intersect(yInterval, xInterval) };
            }
            if (topIntervalSolutions.Count == 2 && bottomIntervalSolutions.Count != 2) // parabola uvnitr (zaporne _p)
            {
                Interval yInterval = new Interval(topIntervalSolutions[0], topIntervalSolutions[1]);
                uIntervals = new List<Interval>() { Interval.Intersect(yInterval, xInterval) };
            }
            return uIntervals;
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////

        protected override Quadric CallConstructor(CommonEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix, GColor color, string name, int ID)
        {
            return new ParabolicCylinder(commonEquation, parametricEquation, toCanonicTransformationMatrix, color, name, ID);
        }

        public override QuadricMatrix GetCanonicMatrix(QuadricParametricEquation parametricEquation)
        {
            double A = 1 / (parametricEquation.GetA() * parametricEquation.GetA());
            double P = 2 * parametricEquation.GetP();
            QuadricEquation newCanonicEquation = new QuadricEquation(A, 0, 0, 0, 0, 0, 0, P, 0, 0);
            return new QuadricMatrix(newCanonicEquation);
        }
    }
}
