﻿using System;
using Tao.OpenGl;
using Geometry.GraphicObjects.Matrixes;
using Geometry.GraphicObjects.Equations;
using Geometry.BaseClasses;
using GMath;

namespace Geometry.GraphicObjects.Quadrics
{
    /// <summary>
    /// Rotacni Elipsoid.
    /// Rovnice:    x = m + a*cos(v)*cos(u).
    ///             y = n + a*cos(v)*sin(u).
    ///             z = p + b*sin(v). (u = (0,2PI), t=(-PI/2,PI/2)).
    /// </summary>
    public class Elipsoid : Quadric
    {
        private double _a;
        private double _b;
        private double _c;
        private double _elipsoidQuality = 30;

        public static string canonicFormatString = "E[x-!m;!a]L[+]E[y-!n;!b]L[+]E[z-!o;!c]L[ = 1]";
        public static string parametricXFormatString = "L[x=!m+!a" + U.Times + "cos(v)" + U.Times + "cos(u)]";
        public static string parametricYFormatString = "L[y=!n+!b" + U.Times + "cos(v)" + U.Times + "sin(u)]";
        public static string parametricZFormatString = "L[z=!o+!c" + U.Times + "sin(v)]";
        public static string _wordDescriptionFirstCase = "Elipsoid";
        public static string _wordDescriptionSecondCase = "Elipsoidu";
        public static string _wordDescriptionFourthCase = "Elipsoid";


        public Elipsoid(CommonEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix, GColor color, string name, int ID)
            : base(commonEquation, parametricEquation, toCanonicTransformationMatrix, color, name,
                        _wordDescriptionFirstCase, _wordDescriptionSecondCase, _wordDescriptionFourthCase,
                        canonicFormatString,
                        parametricXFormatString, parametricYFormatString, parametricZFormatString, ID) 
        {
            _a = parametricEquation.GetA();
            _b = parametricEquation.GetB();
            _c = parametricEquation.GetC();
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////

        /// x = m + a*cos(v)*cos(u).
        protected double ComputeXCoordinateAt(double u, double v)
        {
            return _a * Math.Cos(v) * Math.Cos(u);
        }
        /// y = n + b*cos(v)*sin(u).
        protected double ComputeYCoordinateAt(double u, double v)
        {
            return _b * Math.Cos(v) * Math.Sin(u);
        }
        /// z = p + c*sin(v)
        protected double ComputeZCoordinateAt(double u, double v)
        {
            return _c * Math.Sin(v);
        }

        /// <summary>
        /// Dosadi do parametricke rovnice u a v a vraci souradnice bodu.
        /// </summary>
        /// <param name="u"></param>
        /// <param name="v"></param>
        /// <returns></returns>
        protected Vector ComputeCoordinatesAt(double u, double v) 
        {
            double x = ComputeXCoordinateAt(u, v);
            double y = ComputeYCoordinateAt(u, v);
            double z = ComputeZCoordinateAt(u, v);

            return new Vector(x, y, z);
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////

        private void SetVextexAndHisNormal(double u, double v)
        {
            Vector v1 = ComputeCoordinatesAt(u, v);
            Vector normal = GetNormalAt(v1);
            Gl.glNormal3d(normal.GetX(), normal.GetY(), normal.GetZ());
            Gl.glVertex3d(v1.GetX(), v1.GetY(), v1.GetZ());
        }

        private Vector GetNormalAt(Vector v)
        {
            double x = (2 / (_a * _a)) * v.GetX();
            double y = (2 / (_b * _b)) * v.GetY();
            double z = (2 / (_c * _c)) * v.GetZ();
            Vector normal = new Vector(x, y, z);
            normal.NornVector();
            return normal * -_normalSign;
        }

        private void ShowNormal(double u, double v, double delta)
        {
            Vector v1 = ComputeCoordinatesAt(u, v);
            Vector normal = GetNormalAt(v1);
            Gl.glVertex3d(v1.GetX(), v1.GetY(), v1.GetZ());
            Gl.glVertex3d(v1.GetX() + normal.GetX(), v1.GetY() + normal.GetY(), v1.GetZ() + normal.GetZ());
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////

        private double GetvStep(double u, double v, double vStep, double minDeviation, double maxDeviation) 
        {
            Vector a = ComputeCoordinatesAt(u, v);
            Vector b = ComputeCoordinatesAt(u, v + vStep);
            double angle = Math.Abs(Vector.Angle(a, b));
            if (angle < minDeviation)
                return vStep * 2.0;
            if (angle > maxDeviation)
                return vStep / 2.0;
            return vStep;
        }

        private double GetuStep(double u, double v, double uStep, double minDeviation, double maxDeviation)
        {
            Vector a = ComputeCoordinatesAt(u, v);
            Vector b = ComputeCoordinatesAt(u + uStep, v);
            double angle = Math.Abs(Vector.Angle(a, b));
            if (angle < minDeviation)
                return uStep * 2.0;
            if (angle > maxDeviation)
                return uStep / 2.0;
            return uStep;
        }

        protected override void BuildParametricDisplayList(Scene scene, RelativeQuality relativeQuality) 
        {
            _displayListIdentificator = Gl.glGenLists(1);
            //if (_displayListIdentificator == 3)
            //    _displayListIdentificator = _displayListIdentificator;
            Gl.glNewList(_displayListIdentificator, Gl.GL_COMPILE);
            //Gl.glPolygonMode(Gl.GL_FRONT_AND_BACK, Gl.GL_LINE);
            //double delta = 0.1;
            double step = Math.PI / (_elipsoidQuality * (scene.GetBaseQuality() + relativeQuality));
            double minDeviation = Math.PI / 20;
            double maxDeviation = Math.PI / 6;
            double vStep = GetvStep(0, Math.PI / 2, step, minDeviation, maxDeviation);
            double uStep = GetuStep(0, Math.PI / 2, step, minDeviation, maxDeviation);
            // horni pulka
            for (double v = -Math.PI / 2; v < Math.PI; v += vStep)
            {
                if (v > Math.PI / 2 - vStep)
                    v = Math.PI / 2 - vStep;
                Gl.glBegin(Gl.GL_TRIANGLE_STRIP);
                for (double u = 0; u < Math.PI * 4 + uStep; u += uStep)
                {
                    if (u > Math.PI * 2)
                        u = Math.PI * 2;
                    SetVextexAndHisNormal(u, v);
                    SetVextexAndHisNormal(u, v + vStep);
                    if (u == Math.PI * 2)
                        break;              
                }
                Gl.glEnd();
                if (v == Math.PI / 2 - vStep)
                    break;
            }
            Gl.glEndList();
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////

        protected override Quadric CallConstructor(CommonEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix, GColor color, string name, int ID)
        {
            return new Elipsoid(commonEquation, parametricEquation, toCanonicTransformationMatrix, color, name, ID);
        }



        public override QuadricMatrix GetCanonicMatrix(QuadricParametricEquation parametricEquation)
        {
            double A = 1 / (parametricEquation.GetA() * parametricEquation.GetA());
            double B = 1 / (parametricEquation.GetB() * parametricEquation.GetB());
            double C = 1 / (parametricEquation.GetC() * parametricEquation.GetC());
            QuadricEquation newCanonicEquation = new QuadricEquation(A, B, C, 0, 0, 0, 0, 0, 0, -1);
            return new QuadricMatrix(newCanonicEquation);
        }

    }
}
