﻿using System;
using Tao.OpenGl;
using Geometry.GraphicObjects.Matrixes;
using Geometry.GraphicObjects.Equations;
using GMath;
using Geometry.BaseClasses;

namespace Geometry.GraphicObjects.Quadrics
{
    /// <summary>
    /// Rotacni Jednodilny Hyperboloid
    /// Rovnice:
    ///             x = m + a*cosh(v)*cos(u).
    ///             y = n + b*cosh(v)*sin(u);
    ///             z = o + c*sinh(v);,
    ///             u = 0,2PI, v = (-oo,oo);
    ///             
    ///  stara:     x = m + a*(cos(u)/cos(t).
    ///             y = n + b*(cos(u)/sin(t).
    ///             z = p + c*tan(t). (u = (0,2PI), t=(-PI/2,PI/2).
    /// </summary>
    public class Hyperboloid1 : Quadric
    {
        private double _a;
        private double _b;
        private double _c;
        private double _uSlices = 20;
        private double _vSlices = 20;
        public static string canonicFormatString = "E[x-!m;!a]L[+]E[y-!n;!b]L[-]E[z-!o;!c]L[ = 1]";
        public static string parametricXFormatString = "L[x=!m+!a" + U.Times + "cosh(v)" + U.Times + "cos(u)]";
        public static string parametricYFormatString = "L[y=!n+!b" + U.Times + "cosh(v)" + U.Times + "sin(u)]";
        public static string parametricZFormatString = "L[z=!o+!c" + U.Times + "sinh(v)]";
        public static string _wordDescriptionFirstCase = "Jednodílný hyperbloid";
        public static string _wordDescriptionSecondCase = "Jednodílného hyperboloidu";
        public static string _wordDescriptionFourthCase = "Jednodílný hyperboloid";

        public Hyperboloid1(CommonEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix, GColor color, string name, int ID)
            : base(commonEquation, parametricEquation, toCanonicTransformationMatrix,
                        color, name,
                        _wordDescriptionFirstCase, _wordDescriptionSecondCase, _wordDescriptionFourthCase,
                        canonicFormatString,
                        parametricXFormatString, parametricYFormatString, parametricZFormatString, ID)
        {
            _a = parametricEquation.GetA();
            _b = parametricEquation.GetB();
            _c = parametricEquation.GetC();
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////

        /// <summary>
        /// a*cosh(v)*cos(u).
        /// </summary>
        protected double ComputeXCoordinateAt(double u, double v)
        {
            return _a * (Math.Cosh(v) * Math.Cos(u));
        }
        /// <summary>
        /// b*cosh(v)*sin(u).
        /// </summary>
        protected double ComputeYCoordinateAt(double u, double v)
        {
            return _b * (Math.Cosh(v) * Math.Sin(u));
        }
        /// <summary>
        /// c*sinh(v).
        /// </summary>
        protected double ComputeZCoordinateAt(double u, double v)
        {
            return _c * Math.Sinh(v);
        }

        /// <summary>
        /// Dosadi do parametricke rovnice u a v a vraci souradnice bodu.
        /// </summary>
        /// <param name="u"></param>
        /// <param name="v"></param>
        /// <returns></returns>
        protected Vector ComputeCoordinatesAt(double u, double v)
        {
            double x = ComputeXCoordinateAt(u, v);
            double y = ComputeYCoordinateAt(u, v);
            double z = ComputeZCoordinateAt(u, v);

            return new Vector(x, y, z);
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////

        protected Vector ComputeNormalAt(Vector vertex)
        {
            double x = 2 * vertex.GetX() / (_a * _a);
            double y = 2 * vertex.GetY() / (_b * _b);
            double z = -2 * vertex.GetZ() / (_c * _c);
            Vector normal = new Vector(x, y, z);
            normal.NornVector();
            return normal * -_normalSign;
        }

        protected void SetVertexAndNormal(double u, double v)
        {
            Vector vertex = ComputeCoordinatesAt(u, v);
            Vector normal = ComputeNormalAt(vertex);
            Gl.glNormal3d(normal.GetX(), normal.GetY(), normal.GetZ());
            Gl.glVertex3d(vertex.GetX(), vertex.GetY(), vertex.GetZ());
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////

        protected override void BuildParametricDisplayList
            (Scene scene, RelativeQuality relativeQuality)
        {
            double cback = GetClippingBack(scene);
            double cfront = GetClippingFront(scene);
            double cleft = GetClippingLeft(scene);
            double cright = GetClippingRight(scene);
            double ctop = GetClippingTop(scene);
            double cbottom = GetClippingBottom(scene);

            _displayListIdentificator = Gl.glGenLists(1);
            Gl.glNewList(_displayListIdentificator, Gl.GL_COMPILE);
            Gl.glPushMatrix();
            Gl.glEnable(Gl.GL_LIGHTING);

            double temp1 = Math.Max(Math.Abs(ArcCosh(cleft / _a)), Math.Abs(ArcCosh(cright / _a)));
            Interval rest1 = new Interval(-temp1, temp1); // interval u pro ktere je hyperbola mezi up a down
            double temp2 = Math.Max(Math.Abs(ArcCosh(cbottom / _b)), Math.Abs(ArcCosh(ctop / _b)));
            Interval rest2 = new Interval(-temp2, temp2);

            Interval rest = Interval.UnionOfIntersectingIntervals(rest1, rest2);

            Interval backFront = new Interval(ArcSinh(cback / _c), ArcSinh(cfront / _c)); 
            // interval u pro ktere je hyperbola mezi back a front

            Interval interval = Interval.Intersect(backFront, rest);


            double back = interval.GetA();
            double front = interval.GetB();


            double uStep = (Math.PI / _uSlices) / (scene.GetBaseQuality() + relativeQuality);
            double vStep = ((front - back) / _vSlices) / ((scene.GetBaseQuality() + relativeQuality));

            for (double v = back; v < front; v += vStep)
            {
                if (v > front - vStep)
                    v = front - vStep;
                Gl.glBegin(Gl.GL_TRIANGLE_STRIP);

                for (double u = 0; u < Math.PI * 2 + uStep; u += uStep)
                {
                    if (u > Math.PI * 2)
                        u = Math.PI * 2;
                    SetVertexAndNormal(u, v);
                    SetVertexAndNormal(u, v + vStep);
                    if (u == Math.PI * 2)
                        break;
                }
                Gl.glEnd();
                if (v == front - vStep)
                    break;                
            }
            Gl.glPopMatrix();
            Gl.glEndList();
        }


        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////


        protected override Quadric CallConstructor(CommonEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix, GColor color, string name, int ID)
        {
            return new Hyperboloid1(commonEquation, parametricEquation, toCanonicTransformationMatrix, color, name, ID);
        }

        public override QuadricMatrix GetCanonicMatrix(QuadricParametricEquation parametricEquation)
        {
            double A = 1 / (parametricEquation.GetA() * parametricEquation.GetA());
            double B = 1 / (parametricEquation.GetB() * parametricEquation.GetB());
            double C = -1 / (parametricEquation.GetC() * parametricEquation.GetC());
            QuadricEquation newCanonicEquation = new QuadricEquation(A, B, C, 0, 0, 0, 0, 0, 0, -1);
            return new QuadricMatrix(newCanonicEquation);
        }
    }
}
