﻿using System;
using System.Collections.Generic;
using Tao.OpenGl;
using Geometry.GraphicObjects.Matrixes;
using Geometry.GraphicObjects.Equations;
using Geometry.BaseClasses;
using GMath;

namespace Geometry.GraphicObjects.Quadrics
{
    /// <summary>
    /// Rotacni Jednodilny Hyperboloid
    /// Rovnice:    
    ///             x = m + a*sinh(u)*cos(v);
    ///             y = n + b*sinh(u)*sin(v);
    ///             z = o + c*cosh(u);
    /// 
    ///      stara: x = m + a*(cos(u)*tan(t).
    ///             y = n + b*(sin(u)*tan(t).
    ///             z = p + c/cos(t). (u = (0,2PI), t=(-PI/2,PI/2).
    /// </summary>
    public class Hyperboloid2 : Quadric
    {
        private double _a;
        private double _b;
        private double _c;
        private double _vSlices = 10;
        private double _uSlices = 10;

        public static string canonicFormatString = "E[x-!m;!a]L[+]E[y-!n;!b]L[-]E[z-!o;!c]L[ = -1]";
        public static string parametricXFormatString = "L[x=!m+!a" + U.Times + "sinh(v)" + U.Times + "cos(u)]";
        public static string parametricYFormatString = "L[y=!n+!b" + U.Times + "sinh(v)" + U.Times + "sin(u)]";
        public static string parametricZFormatString = "L[z=!o+!c" + U.Times + "cosh(v)]";
        public static string _wordDescriptionFirstCase = "Dvoudílný hyperboloid";
        public static string _wordDescriptionSecondCase = "Dvoudílného hyperboloidu";
        public static string _wordDescriptionFourthCase = "Dvoudílný hyperboloid";

        const int BACK = 0;
        const int FRONT = 1;
        public struct Pair
        {
            public Interval Interval;
            public int Part;
            public Pair(Interval interval, int part)
            {
                Interval = interval;
                Part = part;
            }
        }

        public Hyperboloid2(CommonEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix,
            GColor color, string name, int ID)
            : base(commonEquation, parametricEquation, toCanonicTransformationMatrix,
                        color, name,
                        _wordDescriptionFirstCase, _wordDescriptionSecondCase, _wordDescriptionFourthCase,
                        canonicFormatString,
                        parametricXFormatString, parametricYFormatString, parametricZFormatString, ID)
        {
            _a = parametricEquation.GetA();
            _b = parametricEquation.GetB();
            _c = parametricEquation.GetC();
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////

        /// <summary>
        /// a*sinh(v)*cos(u).
        /// </summary>
        protected double ComputeXCoordinateAt(double u, double v)
        {
            return _a * (Math.Sinh(v) * Math.Cos(u));
        }
        /// <summary>
        /// b*sinh(v)*sin(u).
        /// </summary>
        protected double ComputeYCoordinateAt(double u, double v)
        {
            return _b * (Math.Sinh(v) * Math.Sin(u));
        }
        /// <summary>
        /// c*cosh(v).
        /// </summary>
        protected double ComputeZCoordinateAt(double u, double v)
        {
            return _c * Math.Cosh(v);
        }

        /// <summary>
        /// Dosadi do parametricke rovnice u a v a vraci souradnice bodu.
        /// </summary>
        /// <param name="u"></param>
        /// <param name="v"></param>
        /// <returns></returns>
        protected Vector ComputeCoordinatesAt(double u, double v)
        {
            double x = ComputeXCoordinateAt(u, v);
            double y = ComputeYCoordinateAt(u, v);
            double z = ComputeZCoordinateAt(u, v);

            return new Vector(x, y, z);
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////

        protected Vector ComputeNormalAt(Vector vertex)
        {
            double x = 2 * vertex.GetX() / (_a * _a);
            double y = 2 * vertex.GetY() / (_b * _b);
            double z = -2 * vertex.GetZ() / (_c * _c);
            Vector normal = new Vector(x, y, z);
            normal.NornVector();
            return normal * -_normalSign;
        }

        protected void SetVertexAndNormal(double u, double v) 
        {
            Vector vertex = ComputeCoordinatesAt(u, v);
            Vector normal = ComputeNormalAt(vertex);
            Gl.glNormal3d(normal.GetX(), normal.GetY(), normal.GetZ());
            Gl.glVertex3d(vertex.GetX(), vertex.GetY(), vertex.GetZ());
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////

        protected  override void BuildParametricDisplayList(Scene scene, RelativeQuality relativeQuality)
        {
            if (Gl.glIsList(_displayListIdentificator) == Gl.GL_TRUE)
                Gl.glDeleteLists(_displayListIdentificator, 1);

            _displayListIdentificator = Gl.glGenLists(1);
            Gl.glNewList(_displayListIdentificator, Gl.GL_COMPILE);
            Gl.glPushMatrix();
            Gl.glEnable(Gl.GL_LIGHTING);

            Build(scene, relativeQuality);

            Gl.glPopMatrix();
            Gl.glEndList();
        }

        private void Build(Scene scene, RelativeQuality relativeQuality) 
        {
            List<Pair> pairs = GetClippingIntervals(scene);
            int quality = scene.GetBaseQuality() + relativeQuality;
            
            double back;
            double front; 
            double uStep;
            double vStep; 
            foreach (Pair p in pairs)
            {
                if (p.Interval.GetA() < 0)
                {
                    if (p.Interval.GetB() < 0)
                        continue;
                    back = 0;
                    front = Math.Max(Math.Abs(p.Interval.GetA()), Math.Abs(p.Interval.GetB()));
                }
                else
                {
                    back = p.Interval.GetA();
                    front = p.Interval.GetB();
                }
                uStep = (Math.PI / _uSlices) / (quality);
                vStep = ((front - back) / _vSlices) / (quality);
                Gl.glPushMatrix();
                if (p.Part == BACK)
                    Gl.glRotated(180, 0, 1, 0);

                for (double v = back; v < front; v += vStep)
                {
                    Gl.glBegin(Gl.GL_TRIANGLE_STRIP);
                    for (double u = 0; u < Math.PI * 2 + uStep; u += uStep)
                    {
                        if (u > Math.PI * 2)
                            u = Math.PI * 2;
                        SetVertexAndNormal(u, v);
                        SetVertexAndNormal(u, Math.Min(v + vStep, front));
                        if (u == Math.PI * 2)
                            break;
                    }
                    Gl.glEnd();
                    if (v + vStep > front)
                        break;
                }
                Gl.glPopMatrix();
            }
        }

        private List<Pair> GetClippingIntervals(Scene scene)
        {
            double cback = GetClippingBack(scene);
            double cfront = GetClippingFront(scene);
            double cleft = GetClippingLeft(scene);
            double cright = GetClippingRight(scene);
            double ctop = GetClippingTop(scene);
            double cbottom = GetClippingBottom(scene);

            Interval leftRight = new Interval(ArcSinh(cleft / _a), ArcSinh(cright / _a));
            Interval bottomTop = new Interval(ArcSinh(cbottom / _b), ArcSinh(ctop / _b));

            Interval leftRightBottomTop = Interval.UnionOfIntersectingIntervals(leftRight, bottomTop);

            Interval back = new Interval(-ArcCosh(cback / _c), ArcCosh(cback / _c));
            Interval front = new Interval(-ArcCosh(cfront / _c), ArcCosh(cfront / _c));

            double backPeak = -ComputeZCoordinateAt(0, 0);
            double frontPeak = ComputeZCoordinateAt(0, 0);

            List<Pair> pairs = new List<Pair>();

            // zadni hyperbola je uvnitr
            if (back.GetA() != double.NaN || (leftRightBottomTop.GetA() != double.NaN && leftRightBottomTop.GetB() != double.NaN))
            {
                if (new Interval(cback, cfront).Belongs(backPeak))
                {
                    pairs.Add(new Pair(Interval.Intersect(back ,leftRightBottomTop), BACK)); // vrchol zadni hyperboly je uvnitr
                }
                else if (backPeak > cfront)
                {
                    Interval tempInterval = GetCutHyperbolaInterval(back, front, leftRightBottomTop);
                    pairs.Add(new Pair(tempInterval, BACK)); // zadni hyperbola je rozdelena na dve
                }
            }
            if (front.GetA() != double.NaN || (leftRightBottomTop.GetA() != double.NaN && leftRightBottomTop.GetB() != double.NaN))
            {
                if (new Interval(cback, cfront).Belongs(frontPeak))
                {
                    pairs.Add(new Pair(Interval.Intersect(front, leftRightBottomTop), FRONT)); // vrchol predni hyperbooly je uvnitr
                }
                else if (frontPeak < cback)
                {
                    Interval tempInterval = GetCutHyperbolaInterval(back, front, leftRightBottomTop);
                    pairs.Add(new Pair(tempInterval, FRONT)); // predni hyperbola je rozdelena na dve
                }
            }
            return pairs;
        }

        /// <summary>
        ///  back.getA a back.getB jsou simetrické stejně jako 
        ///  front.getA a front.getB
        ///  
        /// Protoze je funkce Cosh sudá, je jedno jestli dodám inteval (-x, -y) nebo (x, y)
        /// Aby byla hyperbola uvnitř leftRightBotomTop a zároveň uvnitř front i Back, potom provedu průnik
        /// LeftRightBottomTop a front - back.
        /// Dostanu dva intervaly, vemu ten větší.
        /// 
        /// </summary>
        private Interval GetCutHyperbolaInterval(Interval front, Interval back, Interval leftRightBottomTop) 
        {
            Interval x = new Interval(front.GetA(), back.GetA());
            Interval y = new Interval(front.GetB(), back.GetB());

            Interval xLeftRightBottomTop = Interval.Intersect(x, leftRightBottomTop);
            Interval yLeftRightBottomTop = Interval.Intersect(y, leftRightBottomTop);

            // vezmu ten vetsi a převrátím na kladný
            // protože tam není vrvhol, není přechod přes nulu.
            if (xLeftRightBottomTop.Size() > yLeftRightBottomTop.Size())
                return new Interval(Math.Abs(xLeftRightBottomTop.GetA()), Math.Abs(xLeftRightBottomTop.GetB()));
            else
                return new Interval(Math.Abs(yLeftRightBottomTop.GetA()), Math.Abs(yLeftRightBottomTop.GetB()));
        }


        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////

        protected override Quadric CallConstructor(CommonEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix, GColor color, string name, int ID)
        {
            return new Hyperboloid2(commonEquation, parametricEquation, toCanonicTransformationMatrix, color, name, ID);
        }

        public override QuadricMatrix GetCanonicMatrix(QuadricParametricEquation parametricEquation)
        {
            double A = -1 / (parametricEquation.GetA() * parametricEquation.GetA());
            double B = -1 / (parametricEquation.GetB() * parametricEquation.GetB());
            double C = 1 / (parametricEquation.GetC() * parametricEquation.GetC());
            QuadricEquation newCanonicEquation = new QuadricEquation(A, B, C, 0, 0, 0, 0, 0, 0, -1);
            return new QuadricMatrix(newCanonicEquation);
        }
    }
}
