﻿using System;
using System.Collections.Generic;
using Tao.OpenGl;
using Geometry.GraphicObjects.Matrixes;
using Geometry.GraphicObjects.Equations;
using Geometry.BaseClasses;
using GMath;

namespace Geometry.GraphicObjects.Quadrics
{
    /// <summary>
    /// x*x/a*a - y*y/b*b +2kz = 0;
    /// 
    /// Parametrizace hyb.paraboloidu: 
    ///     g(u,v) = (u,v,u*u/a*a - v*v/b*b);,
    ///     kde z/c = y*y/b*b - x*x/a*a, typuju vynasobit c. 
    /// </summary>
    public class HyperbolicParaboloid:SimpleRulledQuadric
    {
        private double _a;
        private double _b;
        private double _p;
        private double _vSlices = 64;
        private double _uSlices = 64;

        public static string canonicFormatString = "E[x-!m;!a]L[-]E[y-!n;!b]L[+]L[2" + U.Times + "!p"
                + U.Times + "(z-!o)=0]";
        public static string parametricXFormatString = "L[x=!m+u]";
        public static string parametricYFormatString = "L[y=!n+v]";
        static string parametricZFormatString = "L[z=!o+]P[u;!a;!p]L[-]P[v;!b;!p]";
        public static string _wordDescriptionFirstCase = "Hyperbolický paraboloid";
        public static string _wordDescriptionSecondCase = "Hyperbolického paraboloidu";
        public static string _wordDescriptionFourthCase = "Hyperbolický paraboloid";

        protected struct Pair 
        {
            public Interval vSmaller;
            public Interval vBigger;
            public Interval vSmallerNormalized;
            public Interval vBiggerNormalized;
            public Pair(Interval vSmallerArg, Interval vBiggerArg) 
            {
                vSmaller = vSmallerArg;
                vBigger = vBiggerArg;
                vSmallerNormalized = null;
                vBiggerNormalized = null;
            }
        }
        public struct StepPair
        {
            public List<Interval> vIntervals;
            public Interval safetyUInterval;
            public Interval safetyVInterval;
            public double uStep;
            public double vStep;
        }

        /// <summary>
        ///  x*x/a*a - y*y/b*b +2z = 0;
        /// </summary>
        public HyperbolicParaboloid(CommonEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix,
            GColor color, string name, int ID)
            : base(commonEquation, parametricEquation, toCanonicTransformationMatrix,
                        color, name,
                        _wordDescriptionFirstCase, _wordDescriptionSecondCase, _wordDescriptionFourthCase,
                        canonicFormatString,
                        parametricXFormatString, parametricYFormatString, parametricZFormatString, ID)
        {
            _a = parametricEquation.GetA();
            _b = parametricEquation.GetB();
            _p = parametricEquation.GetP();
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////

        /// <summary>
        /// u
        /// </summary>
        protected double ComputeXCoordinateAtOLD(double u, double v)
        {
            return u;
        }
        /// <summary>
        /// v
        /// </summary>
        protected double ComputeYCoordinateAtOLD(double u, double v)
        {
            return v;
        }
        /// <summary>
        /// (u*u/a*a + v*v/b*b)/2k;
        /// </summary>
        protected double ComputeZCoordinateAtOLD(double u, double v)
        {
            return -((u * u) / (_a * _a) - (v * v) / (_b * _b)) / (2 * _p);
        }

        ////////////////////////////////////////////////////////////////////////////////////////

        /// <summary>
        /// Dosadi do parametricke rovnice u a v a vraci souradnice bodu.
        /// </summary>
        /// <param name="u"></param>
        /// <param name="v"></param>
        /// <returns></returns>
        protected Vector ComputeCoordinatesAt(double u, double v)
        {
            double x = ComputeXCoordinateAtOLD(u, v);
            double y = ComputeYCoordinateAtOLD(u, v);
            double z = ComputeZCoordinateAtOLD(u, v);

            return new Vector(x, y, z);
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////

        /// <summary>
        /// x2/a2 - y2/b2 - 2kz = 0 
        /// </summary>
        private void SetVextexAndHisNormal(double u, double v)
        {
            Vector v1 = ComputeCoordinatesAt(u, v);
            Vector normal = ComputeNormalAt(v1);
            Gl.glNormal3d(normal.GetX(), normal.GetY(), normal.GetZ());
            Gl.glVertex3d(v1.GetX(), v1.GetY(), v1.GetZ());
        }

        /// <summary>
        /// x2/a2 + y2/b2 +- 2kz = 0 
        /// </summary>
        protected Vector ComputeNormalAt(Vector vertex)
        {
            double x = 2 * vertex.GetX() / (_a * _a);
            double y = -2 * vertex.GetY() / (_b * _b);
            double z = 2 * _p;
            Vector normal = new Vector(x, y, z);
            normal.NornVector();
            return normal * _normalSign * Math.Sign(_p)*-1;
        }

        /// <summary>
        /// x2/a2 - y2/b2 - 2kz = 0 
        /// </summary>
        private void ShowNormal(double u, double v, double delta)
        {
            Vector v1 = ComputeCoordinatesAt(u, v);
            Vector normal = ComputeNormalAt(v1);
            Gl.glVertex3d(v1.GetX(), v1.GetY(), v1.GetZ());
            Gl.glVertex3d(v1.GetX() + normal.GetX(), v1.GetY() + normal.GetY(), v1.GetZ() + normal.GetZ());
        }

        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////

        bool negativeP = false;
        protected override void BuildParametricDisplayList(Scene scene, RelativeQuality relativeQuality)
        {
            _displayListIdentificator = Gl.glGenLists(1);
            Gl.glNewList(_displayListIdentificator, Gl.GL_COMPILE);
            if (_p < 0)
            {
                negativeP = true;
                double temp = _a;
                _a = _b;
                _b = temp;
                _p = -_p;
                TransformationMatrix transformation = TransformationMatrix.GetXRotationMatrix(180);
                Gl.glMultMatrixd(transformation.ToDoubleArray());
            }
            else { negativeP = false; }
            double qualityKoeficient = scene.GetBaseQuality() + relativeQuality;
            Build(qualityKoeficient, scene);           
            Gl.glEndList();
            if (negativeP)
            {
                double temp = _a;
                _a = _b;
                _b = temp;
                _p = -_p;
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////


        private void Build(double qualityKoeficient, Scene scene)
        {
            StepPair stepPair = GetSteps(scene, qualityKoeficient);
            double vStep = stepPair.vStep;
            double uStep = stepPair.uStep;
            List<Interval> vIntervals = stepPair.vIntervals;

            foreach (Interval vInterval in vIntervals)
            {
                double vMin = vInterval.GetA();
                double vMax = vInterval.GetB();

                double v = vMin;
                for (int i = 0; i < int.MaxValue; i++)
                {
                    List<Pair> uIntervals = GetUIntervals(v, Math.Min(vMax, v + vStep), uStep, scene);
                    if (uIntervals.Count == 3) // vyplnim trojuhelnik, ktery vznika na prechodu mezi stredni a krajni casti.
                    {
                        Gl.glBegin(Gl.GL_TRIANGLE_STRIP);
                        SetVextexAndHisNormal(uIntervals[2].vSmaller.GetA(), v);
                        SetVextexAndHisNormal(uIntervals[2].vBigger.GetA(), v+vStep);
                        SetVextexAndHisNormal(uIntervals[2].vSmaller.GetB(), v);
                        SetVextexAndHisNormal(uIntervals[2].vBigger.GetB(), v+vStep);
                        Gl.glEnd();
                        uIntervals.RemoveAt(2);
                    }
                    foreach (Pair p in uIntervals)
                    {
                        if (p.vBigger.IsEmpty() || p.vSmaller.IsEmpty() || p.vBiggerNormalized.IsEmpty() || p.vSmallerNormalized.IsEmpty())
                            continue;
                        double uMin = Math.Min(p.vSmallerNormalized.GetA(), p.vBiggerNormalized.GetA()) - uStep;
                        double uMax = Math.Max(p.vSmallerNormalized.GetB(), p.vBiggerNormalized.GetB()) + uStep;
                        Gl.glBegin(Gl.GL_TRIANGLE_STRIP);
                        double u = uMin;
                        for (int j = 0; i < int.MaxValue; j++)
                        {
                            SetVextexAndHisNormal(Math.Min(Math.Max(p.vSmaller.GetA(), u), p.vSmaller.GetB()), v);
                            SetVextexAndHisNormal(Math.Min(Math.Max(p.vBigger.GetA(), u), p.vBigger.GetB()), Math.Min(v + vStep, vMax));
                            if (u >= uMax)
                                break;
                            u = uMin + uStep * j;
                        }
                        Gl.glEnd();
                    }
                    if (v >= vMax)
                        break;
                    v = vMin + vStep * i;
                }
            }
        }
        /// <summary>
        /// Pro dané v vypočítám intervaly parametru U, pro které je paraboloiod uvnitř krychle.
        /// </summary>
        private List<Pair> GetUIntervals(double vSmaller, double vBigger, double uStep, Scene scene)
        {
            double cfront = GetClippingFront(scene);
            double cback = GetClippingBack(scene);
            double cleft = GetClippingLeft(scene);
            double cright = GetClippingRight(scene);


            if (negativeP)
            {
                cleft = Math.Min(-GetClippingBottom(scene), -GetClippingTop(scene));
                cright = Math.Max(-GetClippingBottom(scene), -GetClippingTop(scene));
                double temp = cfront;
                cfront = Math.Max(-temp, -cback);
                cback = Math.Min(-temp, -cback);
            }
            double a = _a * _a * 2.0 * _p;
            double b = _b * _b * 2.0 * _p;
            Interval leftRight = new Interval(cleft, cright);

            double backv1 = Math.Sqrt((vSmaller * vSmaller / b - cback) * a); // dva body, ve kterych protina bottom
            double backv2 = Math.Sqrt((vBigger * vBigger / b - cback) * a); // dva body, ve kterych protina bottom                             
            double frontv1 = Math.Sqrt((vSmaller * vSmaller / b - cfront) * a); // dva body, ve kterych protina top
            double frontv2 = Math.Sqrt((vBigger * vBigger / b - cfront) * a); // dva body, ve kterych protina top

            List<Pair> retval = new List<Pair>();

            if (double.IsNaN(frontv1) && double.IsNaN(frontv2)) // jde o prostřední část
            {
                retval.Add(new Pair(new Interval(-backv1, backv1), new Interval(-backv2, backv2)));
            }
            else
                // přechod mezi prostřední a krajní částí.
                // kdyby se zmenila podminka v prechozim na nebo, 
                // pak by byl vykreslen paraboloid temer stejne, ale v prechodu, by se objevilo nenavazovani.
                // Protoze krajni cast konci na nenormovane hodnote, U. Ale stredni cast jede vzdy po normovanem U. 
                // Proto rozdelim prechodovou cast na dve pulky jako by to byla krajni cast.
                // Tim, ale vznikne dirka, kterou vyplni posledni unterval hodnot U.
                if (double.IsNaN(frontv1)) // jde o prostřední část, přechod
                {
                    retval.Add(new Pair(new Interval(-backv1, 0), new Interval(-backv2, -frontv2)));
                    retval.Add(new Pair(new Interval(0, backv1), new Interval(frontv2, backv2)));
                    retval.Add(new Pair(new Interval(0, 0), new Interval(-frontv2, frontv2)));
                }
                else
                    if (double.IsNaN(frontv2)) // jde o prostřední část, přechod
                    {
                        retval.Add(new Pair(new Interval(-backv1, -frontv1), new Interval(-backv2, 0)));
                        retval.Add(new Pair(new Interval(frontv1, backv1), new Interval(0, backv2)));
                        retval.Add(new Pair(new Interval(-frontv1, frontv1), new Interval(0, 0)));
                    }
                    else
                    {
                        retval.Add(new Pair(new Interval(-backv1, -frontv1), new Interval(-backv2, -frontv2)));
                        retval.Add(new Pair(new Interval(frontv1, backv1), new Interval(frontv2, backv2)));
                    }
            for (int i = 0; i < retval.Count; i++) 
            {
                retval[i] = Intersect(retval[i], leftRight);
                retval[i] = Normalize(retval[i], uStep);
            }
            return retval;
        }

        private Pair Intersect(Pair pair, Interval leftRight) 
        {
            pair.vBigger = Interval.Intersect(pair.vBigger, leftRight);
            pair.vSmaller = Interval.Intersect(pair.vSmaller, leftRight);
            return pair;
        }
        private Pair Normalize(Pair pair, double uStep)
        {
            pair.vSmallerNormalized = ApplyUstepNormalize(pair.vSmaller, uStep);
            pair.vBiggerNormalized = ApplyUstepNormalize(pair.vBigger, uStep);
            return pair;
        }
        private Interval ApplyUstepNormalize(Interval interval, double uStep)
        {
            if (interval.IsEmpty())
                return interval;
            double uMin = interval.GetA();
            double uMax = interval.GetB();
            int sign = Math.Sign(uMin);
            uMin = Math.Floor(uMin * sign / uStep) * uStep * sign;
            sign = Math.Sign(uMax);
            uMax = Math.Floor(uMax * sign / uStep) * uStep * sign;
            return new Interval(uMin, uMax);
        }


        private StepPair GetSteps(Scene scene, double qualityKoeficient) 
        {
            StepPair stepPair;
            if (_a < _b)
            {
                // je hodne svisle-placaty, je lepsi zjistit intervaly V a přitom i ohraničit intervaly U.
                stepPair = GetVStep(scene, qualityKoeficient); 
                stepPair = GetUStep(scene, qualityKoeficient, stepPair);
            }
            else
            {
                // je hodne vodorovně-placaty, je lepsi zjistit intervaly U a přitom i ohraničit intervaly V.
                stepPair = GetUStep(scene, qualityKoeficient);
                stepPair = GetVStep(scene, qualityKoeficient, stepPair);
            }
            return stepPair;
        }

        /// <summary>
        /// Bottom a top omezují hodnoty kterých může nabývat V.
        /// Projedu tyto hodnoty po nějakém delta a určím v kolika bodech V existuje U, které je uvnitř krychle.
        /// Na základě toho vrátím vStep
        /// </summary>
        private StepPair GetVStep(Scene scene, double qualityKoeficient)
        {
            double cbottom = GetClippingBottom(scene);
            double ctop = GetClippingTop(scene);
            double cleft = GetClippingLeft(scene);
            double cright = GetClippingRight(scene);
            double cfront = GetClippingFront(scene);
            double cback = GetClippingBack(scene);

            if (negativeP)
            {
                cleft = Math.Min(-GetClippingBottom(scene), -GetClippingTop(scene));
                cright = Math.Max(-GetClippingBottom(scene), -GetClippingTop(scene));
                cbottom = GetClippingLeft(scene);
                ctop = GetClippingRight(scene);
                double temp = cfront;
                cfront = Math.Max(-temp, -cback);
                cback = Math.Min(-temp, -cback);
            }

            double a = _a * _a * 2.0 * _p;
            double b = _b * _b * 2.0 * _p;

            // intervaly V, ve kterych je paraboloid uvnitr krychle, 
            // protože je paraboloid více svisle-placaty, tak bych měl vždy nějaký najít
            List<Interval> vIntervals = new List<Interval>() { new Interval(cbottom, ctop) };
            Interval safetyUInterval = new Interval(cleft, cright);
            double vMin = cbottom;
            double vMax = ctop;
            double vStep = (((vMax - vMin) / _vSlices)) / qualityKoeficient;
            int count = 0; // pocet v ve kterych existuje u, ktere je uvnitr krychle
            double extremMostBackv1 = 0;
            List<Interval> futureVIntervals = new List<Interval>();
            foreach (Interval vInterval in vIntervals)
            {
                double backv1, frontv1;
                bool wasIn = false;
                vMin = vInterval.GetA();
                vMax = vInterval.GetB();
                for (double v = vMin; v < double.MaxValue; v += 0)
                {
                    backv1 = Math.Sqrt((v * v / b - cback) * a); // dva body, ve kterych protina back                            
                    frontv1 = Math.Sqrt((v * v / b - cfront) * a); // dva body, ve kterych protina front                            
                    if ((!double.IsNaN(backv1) && ((backv1 >= cleft && backv1 <= cright) || (-backv1 >= cleft && -backv1 <= cright))) ||
                        (!double.IsNaN(frontv1) && ((frontv1 >= cleft && frontv1 <= cright) || (-frontv1 >= cleft && -frontv1 <= cright))))
                    {
                        count++;
                        if (!wasIn) // pokud nebyl minule uvnitr zacnu novy interval, jinak rozsirim stavajici
                            futureVIntervals.Add(new Interval(v - vStep, v + vStep)); // aby byl ten interval, ve kterém pak testuji pro jistotu o něco větší.
                        else futureVIntervals[futureVIntervals.Count - 1].SetB(v + vStep); // aby byl ten interval, ve kterém pak testuji pro jistotu o něco větší.
                        wasIn = true;
                    }
                    else
                        wasIn = false;
                    if (!double.IsNaN(backv1)) // pokud je paraboloid moc rovinaty, protina back a front moc daleko. 
                    {
                        extremMostBackv1 = Math.Max(backv1, extremMostBackv1);
                        safetyUInterval.SetA(Math.Max(-extremMostBackv1, cleft));
                        safetyUInterval.SetB(Math.Min(extremMostBackv1, cright));
                    }
                    v += vStep;
                    if (v >= vMax)
                        break;
                }
            }
            if (futureVIntervals.Count != 0) // pokud je paraboloid moc rovinaty, protina back a front moc daleko.
                vIntervals = futureVIntervals;

            StepPair retval = new StepPair();
            retval.vStep = vStep;
            retval.safetyUInterval = safetyUInterval;
            retval.vIntervals = vIntervals;
            return retval;
        }
        /// <summary>
        /// Left a Right omezují hodnoty kterých může nabývat U.
        /// Projedu tyto hodnoty po nějakém delta a určím v kolika bodech U existuje V, které je uvnitř krychle.
        /// Na základě toho vrátím uStep
        /// </summary>
        private StepPair GetUStep(Scene scene, double qualityKoeficient)
        {
            double cbottom = GetClippingBottom(scene);
            double ctop = GetClippingTop(scene);
            double cleft = GetClippingLeft(scene);
            double cright = GetClippingRight(scene);
            double cfront = GetClippingFront(scene);
            double cback = GetClippingBack(scene);

            if (negativeP)
            {
                cleft = Math.Min(-GetClippingBottom(scene), -GetClippingTop(scene));
                cright = Math.Max(-GetClippingBottom(scene), -GetClippingTop(scene));
                cbottom = GetClippingLeft(scene);
                ctop = GetClippingRight(scene);
                double temp = cfront;
                cfront = Math.Max(-temp, -cback);
                cback = Math.Min(-temp, -cback);
            }

            double a = _a * _a * 2.0 * _p;
            double b = _b * _b * 2.0 * _p;

            List<Interval> uIntervals = new List<Interval>() { new Interval(cleft, cright) }; // intervaly U, ve kterych je paraboloid uvnitr krychle
            Interval safetyVInterval = new Interval(cbottom, ctop);
            double uMin = cleft;
            double uMax = cright;
            double uStep = (((uMax - uMin) / _uSlices)) / qualityKoeficient;
            int count = 0; // pocet u ve kterych existuje v, ktere je uvnitr krychle
            double extremeMostBackv1 = 0;
            //while (count < 10)
            count = 0;
            List<Interval> futureUIntervals = new List<Interval>();
            foreach (Interval uInterval in uIntervals)
            {
                double backv1, frontv1;
                bool wasIn = false;
                uMin = uInterval.GetA();
                uMax = uInterval.GetB();
                for (double u = uMin; u < double.MaxValue; u += 0)
                {
                    backv1 = Math.Sqrt((u * u / a + cback) * b); // dva body, ve kterych protina back                            
                    frontv1 = Math.Sqrt((u * u / a + cfront) * b); // dva body, ve kterych protina front                            
                    if ((!double.IsNaN(backv1) && ((backv1 >= cbottom && backv1 <= ctop) || (-backv1 >= cbottom && -backv1 <= ctop))) ||
                        (!double.IsNaN(frontv1) && ((frontv1 >= cbottom && frontv1 <= ctop) || (-frontv1 >= cbottom && -frontv1 <= ctop))))
                    {
                        count++;
                        if (!wasIn) // pokud nebyl minule uvnitr zacnu novy interval, jinak rozsirim stavajici
                            futureUIntervals.Add(new Interval(u - uStep, u + uStep)); // aby byl ten interval, ve kterém pak testuji pro jistotu o něco větší.
                        else futureUIntervals[futureUIntervals.Count - 1].SetB(u + uStep); // aby byl ten interval, ve kterém pak testuji pro jistotu o něco větší.
                        wasIn = true;
                    }
                    else wasIn = false;
                    if (!double.IsNaN(frontv1))// pokud je paraboloid moc rovinaty, protina back a front moc daleko, proto toto neni v tom druhem IFU
                    {
                        extremeMostBackv1 = Math.Max(frontv1, extremeMostBackv1);

                        double temp = Math.Sqrt(((u - uStep) * (u - uStep) / a + cfront) * b); // dva body, ve kterych protina front 
                        if (!double.IsNaN(temp))
                            extremeMostBackv1 = Math.Max(temp, extremeMostBackv1);
                        temp = Math.Sqrt(((u + uStep) * (u + uStep) / a + cfront) * b); // dva body, ve kterych protina front 
                        if (!double.IsNaN(temp))
                            extremeMostBackv1 = Math.Max(temp, extremeMostBackv1);

                        safetyVInterval.SetA(Math.Max(-extremeMostBackv1, cbottom));
                        safetyVInterval.SetB(Math.Min(extremeMostBackv1, ctop));
                    }
                    u += uStep;
                    if (u >= uMax)
                        break;
                }
            }

            StepPair retval = new StepPair();
            retval.uStep = uStep;
            retval.safetyVInterval = safetyVInterval;
            return retval;
        }

        /// <summary>
        /// Bottom a top omezují hodnoty kterých může nabývat V.
        /// Projedu tyto hodnoty po nějakém delta a určím v kolika bodech V existuje U, které je uvnitř krychle.
        /// Na základě toho vrátím vStep
        /// </summary>
        private StepPair GetVStep(Scene scene, double qualityKoeficient, StepPair stepPair)
        {
            double a = _a * _a * 2.0 * _p;
            double b = _b * _b * 2.0 * _p;
 
            double vMin = stepPair.safetyVInterval.GetA();
            double vMax = stepPair.safetyVInterval.GetB();
            double vStep = (((vMax - vMin) / _vSlices)) / qualityKoeficient;

            stepPair.vStep = vStep;
            stepPair.vIntervals = new List<Interval>() { stepPair.safetyVInterval };
            return stepPair;
        }
        /// <summary>
        /// Left a Right omezují hodnoty kterých může nabývat U.
        /// Projedu tyto hodnoty po nějakém delta a určím v kolika bodech U existuje V, které je uvnitř krychle.
        /// Na základě toho vrátím uStep
        /// </summary>
        private StepPair GetUStep(Scene scene, double qualityKoeficient, StepPair stepPair)
        {
            double a = _a * _a * 2.0 * _p;
            double b = _b * _b * 2.0 * _p;

            double uMin = stepPair.safetyUInterval.GetA();
            double uMax = stepPair.safetyUInterval.GetB(); ;
            double uStep = (((uMax - uMin) / _uSlices))/qualityKoeficient;
            stepPair.uStep = uStep;
            return stepPair;
        }



        /////////////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////////////


        protected override Quadric CallConstructor(CommonEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix, GColor color, string name, int ID)
        {
            return new HyperbolicParaboloid(commonEquation, parametricEquation, toCanonicTransformationMatrix, color, name, ID);
        }

        public override QuadricMatrix GetCanonicMatrix(QuadricParametricEquation parametricEquation)
        {
            double A = 1 / (parametricEquation.GetA() * parametricEquation.GetA());
            double B = -1 / (parametricEquation.GetB() * parametricEquation.GetB());
            double P = 2 * parametricEquation.GetP();
            QuadricEquation newCanonicEquation = new QuadricEquation(A, B, 0, 0, 0, 0, 0, 0, P, 0);
            return new QuadricMatrix(newCanonicEquation);
        }
    }
}
