﻿using System;
using System.Collections.Generic;
using Tao.OpenGl;
using Geometry.GraphicObjects.Planar;
using Geometry.GraphicObjects.Equations;
using Geometry.GraphicObjects.Matrixes;
using Geometry.BaseClasses;
using GMath;

namespace Geometry.GraphicObjects.Quadrics
{
    /// <summary>
    /// x*x/a*a - y*y/b*b = 0;
    /// </summary>
    public class CrossingPlans:SimpleRulledQuadric
    {
        private double _a;
        private double _b;
        public static string canonicFormatString = "E[x-!m;!a]L[-]E[y-!n;!b]L[=0]";
        public static string parametricXFormatString = "L[x=!m+u]";
        public static string parametricYFormatString = "L[y=!n" + U.PM + "]F[L[u" + U.Times + "!b];L[!a]]";
        public static string parametricZFormatString = "L[z=v]";

        // parametrizace
        public static string _wordDescriptionFirstCase = "Různoběžné roviny";
        public static string _wordDescriptionSecondCase = "Různoběžných rovin";
        public static string _wordDescriptionFourthCase = "Různoběžné roviny";

        //  x*x*a*a - y*y*b*b = 0
        // a = lambda1Abs
        // b = lambda2Abs
        public CrossingPlans(CommonEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix,
            GColor color, string name, int ID)
            : base(commonEquation, parametricEquation, toCanonicTransformationMatrix,
                        color, name,
                        _wordDescriptionFirstCase, _wordDescriptionSecondCase, _wordDescriptionFourthCase,
                        canonicFormatString,
                        parametricXFormatString, parametricYFormatString, parametricZFormatString, ID) 
        {
            _a = parametricEquation.GetA();
            _b = parametricEquation.GetB();
        }

        protected override void BuildParametricDisplayList(Scene scene, RelativeQuality relativeQuality) 
        {
            _displayListIdentificator = Gl.glGenLists(1);
            Gl.glNewList(_displayListIdentificator, Gl.GL_COMPILE);
            Gl.glPushMatrix();
            BuildPlan(_a, _b, scene);
            Gl.glPopMatrix();
            Gl.glPushMatrix();
            BuildPlan(_a, -_b, scene);
            Gl.glPopMatrix();

            Gl.glEndList();
        }

        public void BuildPlan(double a, double b, Scene scene)
        {
            Vector normal = Vector.CrossProduct(new Vector(a, b, 0), new Vector(0, 0, 1));
            Gl.glNormal3d(normal.GetX(), normal.GetY(), normal.GetZ());

            double cleft = GetClippingLeft(scene);
            double cright = GetClippingRight(scene);
            double cback = GetClippingBack(scene);
            double cfront = GetClippingFront(scene);
            double cbottom = GetClippingBottom(scene);
            double ctop = GetClippingTop(scene);

            Interval xInterval = new Interval(cleft, cright);
            Interval yInterval = new Interval(cbottom * a / b, ctop * a / b);

            Vector p1;
            Vector p2;
            if (xInterval.Size() < yInterval.Size())
            {
                p1 = new Vector(cleft, (b / a) * cleft, 0);
                p2 = new Vector(cright, (b / a) * cright, 0);
            }
            else
            {
                p1 = new Vector((a / b) * cbottom, cbottom, 0);
                p2 = new Vector((a / b) * ctop, ctop, 0);
            }
            double vMin = cback;
            double vMax = cfront;

            SuperPlan.BuildPlanDisplayList(scene, p1, p2, vMin, vMax);
        }

        public void BuildPlanOldBad(double a, double b, Scene scene) 
        {
            double planAngle = Math.Atan(b / a) * 180 / Math.PI;
            Gl.glRotated(planAngle, 0, 0, 1); // rotace prvni roviny o prislusny uhel;
            Gl.glRotated(90, 1, 0, 0); // rovina z = 0, se zmeni na y = 0;
            Gl.glNormal3d(0, 0, _normalSign);

            double cleft = GetClippingLeft(scene);
            double cright = GetClippingRight(scene);
            double cback = GetClippingBack(scene);
            double cfront = GetClippingFront(scene);
            double cbottom = GetClippingBottom(scene);
            double ctop = GetClippingTop(scene);

            double uMin;
            double uMax;
            if (b > 0)
            {
                uMin = Math.Max(cleft*b/a, cbottom*b/a);
                uMax = Math.Min(cright*b/a, ctop*b/a);
            }
            else
            {
                uMin = -Math.Min(cleft*b, ctop*a);
                uMax = -Math.Max(cright*b, cbottom*a);
            }

            double vMin = cback;
            double vMax = cfront;

            SuperPlan.BuildPlanDisplayList(scene, uMin, uMax, vMin, vMax);
        }

        protected override Quadric CallConstructor(CommonEquation commonEquation, ParametricEquation parametricEquation,
            TransformationMatrix toCanonicTransformationMatrix, GColor color, string name, int ID)
        {
            return new CrossingPlans(commonEquation, parametricEquation, toCanonicTransformationMatrix, color, name, ID);
        }

        public override QuadricMatrix GetCanonicMatrix(QuadricParametricEquation parametricEquation)
        {
            double A = 1 / (parametricEquation.GetA() * parametricEquation.GetA());
            double B = -1 / (parametricEquation.GetB() * parametricEquation.GetB());
            QuadricEquation newCanonicEquation = new QuadricEquation(A, B, 0, 0, 0, 0, 0, 0, 0, 0);
            return new QuadricMatrix(newCanonicEquation);
        }
    }
}
