﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Geometry.GraphicObjects.Quadrics;
using Geometry.GraphicObjects.ConicSections;
using Geometry.GraphicObjects;
using Geometry.GraphicObjects.Intersections;
using Geometry.GGraphic; 

namespace Geometry
{
    /// <summary>
    /// Jelikoz vytvarime inverzni operace z tech puvodnich zamenou old a new,
    /// Pak se NESMI menit hodnoty old a new v HistoryItem.
    /// Z toho duvodu se itemy v historii neupravuji, ale vymenuji za nove.
    /// </summary>
    public class History
    {
        List<IHistoryItem> _historyList;
        int _historyPosition = -1; //ukazuje na aktualni item, ktery lze undo.
        ManipulationManager _manipulationManager;

        public History(ManipulationManager manipulationManager) 
        {
            _historyList = new List<IHistoryItem>();
            _manipulationManager = manipulationManager;
        }



        public void AddToHistoryOLD(IHistoryItem item) 
        {
            _historyPosition++;
            _historyList.Insert(_historyPosition, item);
            for (int i = _historyPosition + 1; i < _historyList.Count; i++)
                _historyList.RemoveAt(i);
        }
        /// <summary>
        /// 1) X| u  v  w
        /// 2) X|-u -v -w
        /// 3) X|-u -v -w w v u
        /// 4) X -u -v -w w v u Z|
        /// </summary>
        /// <param name="item"></param>
        public void AddToHistory(IHistoryItem item)
        {
            List<IHistoryItem> newList = new List<IHistoryItem>();
            for (int i = 0; i <= _historyPosition; i++)
                newList.Add(_historyList[i]);
            for (int i = _historyPosition + 1; i < _historyList.Count; i++)
                newList.Add(_historyList[i].GetInverseHistoryItem());
            for (int i = _historyList.Count - 1; i > _historyPosition; i--)
                newList.Add(_historyList[i]);
            newList.Add(item);
            _historyList = newList;
            _historyPosition = _historyList.Count - 1;
        }
        /// <summary>
        /// Volana po undo nebo redo
        /// Jelikoz vytvarime inverzni operace z tech puvodnich zamenou old a new,
        /// Pak se NESMI menit hodnoty old a new v HistoryItem.
        /// Z toho duvodu se itemy v historii neupravuji, ale vymenuji za nove.
        /// </summary>
        private void AddInverseHistoryItem(int index, IHistoryItem oldHistoryItem) 
        {
            IHistoryItem inverse = oldHistoryItem.GetInverseHistoryItem();
            _historyList.RemoveAt(index);
            _historyList.Insert(index, inverse);
        }



        public void Undo()
        {
            if (_historyPosition > -1 && _historyList.Count != 0)
            {
                IHistoryItem item = _historyList.ElementAt(_historyPosition);
                _historyPosition--;
                AddInverseHistoryItem(_historyPosition + 1, item);
                ApplyHistoryItem(item);
            }
        }
        public void Redo()
        {
            if (_historyList != null && _historyList.Count > _historyPosition + 1)
            {
                _historyPosition++;
                IHistoryItem item = _historyList.ElementAt(_historyPosition);
                AddInverseHistoryItem(_historyPosition, item);
                ApplyHistoryItem(item);
            }
        }


        private void ApplyHistoryItem(IHistoryItem item) 
        {
            if (item is HistoryItem)
                ApplyHistoryItem((HistoryItem)item);
            else ApplyHistoryItem((CombinedHistoryItem)item);
        }
        private void ApplyHistoryItem(HistoryItem item)
        {
            switch (item.GetInverseHistoryOperation())
            {
                case HistoryOperation.ChangeName: ApplyChangeName(item); break;
                case HistoryOperation.ChangeColor: ApplyChangeColor(item); break;
                case HistoryOperation.DeleteObject: ApplyDeleteObject(item); break;
                case HistoryOperation.SwitchObjects: ApplySwitchObjects(item); break;
                case HistoryOperation.UnDeleteObject: ApplyUnDeleteObject(item); break;
                case HistoryOperation.ChangeObjectVisibility: ApplyChangeObjectVisibility(item); break;
                case HistoryOperation.ChangeObjectAxesVisibility: ApplyChangeObjectAxesVisibility(item); break;
            }
        }
        private void ApplyHistoryItem(CombinedHistoryItem item)
        {
            foreach (HistoryItem i in item.GetHistoryItems())
                ApplyHistoryItem(i);
        }

        
        
        private void ApplyChangeName(HistoryItem item)
        {
            GraphicObject drawable = _manipulationManager.GetObjectWithID(item.GetSubjectID());
            string oldName = (string)item.GetOldValue();
            _manipulationManager.AplyChangeName(drawable, oldName);
        }
        private void ApplyChangeColor(HistoryItem item) 
        {
            GraphicObject drawable = _manipulationManager.GetObjectWithID(item.GetSubjectID());
            GColor oldcolor = (GColor)item.GetOldValue();
            _manipulationManager.ApplyChangeColor(drawable, oldcolor);
        }
        private void ApplyUnDeleteObject(HistoryItem item) 
        {
            GraphicObject drawable = _manipulationManager.GetObjectWithID(item.GetSubjectID());
            _manipulationManager.ApplyUnDeleteObject(drawable);
        }
        private void ApplySwitchObjects(HistoryItem item)
        {
            GraphicObject newDrawable = _manipulationManager.GetObjectWithID((int)item.GetNewValue());
            GraphicObject oldDrawable = _manipulationManager.GetObjectWithID((int)item.GetOldValue());
            _manipulationManager.ApplySwitchObjects(newDrawable, oldDrawable);
        }
        private void ApplyDeleteObject(HistoryItem item) 
        {
            GraphicObject drawable = _manipulationManager.GetObjectWithID(item.GetSubjectID());
            _manipulationManager.ApplyDeleteObject(drawable);
        }
        private void ApplyChangeObjectVisibility(HistoryItem item) 
        {
            GraphicObject drawable = _manipulationManager.GetObjectWithID(item.GetSubjectID());
            bool oldVisible = (bool)item.GetOldValue();
            _manipulationManager.ApplyHideObject(drawable, oldVisible);
        }
        private void ApplyChangeObjectAxesVisibility(HistoryItem item) 
        {
            GraphicObject graphicObject = _manipulationManager.GetObjectWithID(item.GetSubjectID());
            WithParametricEquationObject drawable = (WithParametricEquationObject)graphicObject;
            bool oldVisible = (bool)item.GetOldValue();
            _manipulationManager.ApplyHideObjectsAxes(drawable, oldVisible);            
        }




        public int GetUndoCount()
        {
            if (_historyList != null)
            {
                return _historyPosition + 1;
            }
            else
            {
                return 0;
            }
        }
        public int GetRedoCount()
        {
            if (_historyList != null)
            {
                return _historyList.Count - (_historyPosition + 1);
            }
            else
            {
                return 0;
            }
        }



        /// <summary>
        /// Jen pro potřeby ukládání.
        /// Nemodifikovat!
        /// </summary>
        public List<IHistoryItem> GetHistoryItems() { return _historyList; }
        /// <summary>
        /// Jen pro potřeby ukládání.
        /// Nepooužívat!
        /// </summary>
        public int GetHistoryPosition() { return _historyPosition; }
        public void LoadHistory(List<IHistoryItem> savedHistoryItems, int historyPosition)
        {
            foreach (IHistoryItem historyItem in savedHistoryItems)
            {
                _historyList.Add(historyItem);
            }
            _historyPosition = historyPosition;
            for (int i = 0; i <= historyPosition; i++)
                if (_historyList[i] is HistoryItem)
                {
                    HistoryItem item = (HistoryItem)_historyList[i];
                    if (!((item.GetHistoryOperation() == HistoryOperation.AddObject ||
                        item.GetHistoryOperation() == HistoryOperation.UnDeleteObject) &&
                        _manipulationManager.GetObjectWithID(item.GetSubjectID()) is IIntersection)) // veskere pruniky jsou jiz pridany
                        ApplyHistoryItem(_historyList[i].GetInverseHistoryItem());
                }
                else                 
                    if (_historyList[i] is CombinedHistoryItem)
                    {
                        CombinedHistoryItem item = (CombinedHistoryItem)_historyList[i];
                        ApplyHistoryItem(_historyList[i].GetInverseHistoryItem());
                    }
                
            for (int i = historyPosition + 1; i < _historyList.Count; i++)
                ApplyHistoryItem(_historyList[i]);
            for (int i = _historyList.Count - 1; i > historyPosition; i--)
                ApplyHistoryItem(_historyList[i].GetInverseHistoryItem());
            _manipulationManager.ControlUndoRedoButtons();
        }
    }
}
