﻿using System;
using System.Collections.Generic;
using System.Xml;

using Geometry.GraphicObjects.Equations;
using Geometry.GraphicObjects;
using Geometry.GraphicObjects.Planar;
using Geometry.GraphicObjects.ConicSections;
using Geometry.GraphicObjects.Quadrics;
using Geometry.GraphicObjects.Intersections;
using GMath;

namespace Geometry.LoadSave
{
    class LoadManager
    {
        private XmlDocument _document;
        private string _baseFilePath;
        List<GraphicObject> _graphicObjects;

        public LoadManager(string filePath)
        {
            try
            {
                _baseFilePath = filePath;
                //Path.Combine(AppDomain.CurrentDomain.BaseDirectory, name);
                _document = new XmlDocument();
                _document.Load(_baseFilePath);
            }
            catch (Exception e) { throw e; }
        }


        public List<GraphicObject> LoadObjects() 
        {
            XmlNode sceneXML = _document.SelectSingleNode("document").SelectSingleNode("scene");
            _graphicObjects = new List<GraphicObject>();
            XmlNodeList objectsXML = sceneXML.SelectSingleNode("objects").SelectNodes("graphicObject");
            foreach (XmlNode objectXML in objectsXML) 
            {
                _graphicObjects.Add(LoadGraphicObject(objectXML));
            }
            return _graphicObjects;
        }

        private GraphicObject LoadGraphicObject(XmlNode objectXML) 
        {
            XmlNode typeXML = objectXML.SelectSingleNode("type");
            GraphicObject graphicObject;
            switch (typeXML.InnerText)
            {
                case "axes": graphicObject = LoadAxes(objectXML); break;
                case "point": graphicObject = LoadPoint(objectXML); break;
                case "line": graphicObject = LoadLine(objectXML); break;
                case "plan": graphicObject = LoadPlan(objectXML); break;
                case "conic": graphicObject = LoadConic(objectXML); break;
                case "quadric": graphicObject = LoadQuadric(objectXML); break;
                case "intersection": graphicObject = LoadIntersection(objectXML); break;            
                default: throw new Exception("Grafický objekt má neznámý typ.");
            }
            graphicObject.SetVisible(LoadVisible(objectXML));
            if (graphicObject is WithParametricEquationObject) 
                ((WithParametricEquationObject)graphicObject).SetAxesVisible(LoadAxesVisible(objectXML));
            return graphicObject;
        }

        private GraphicObject LoadAxes(XmlNode objectXML) 
        {
            int ID = LoadID(objectXML);
            return new Axes(ID);
        }

        private GraphicObject LoadPoint(XmlNode objectXML)
        {
            int ID = LoadID(objectXML);
            string name = LoadName(objectXML);
            GColor color = LoadColor(objectXML);
            TransformationMatrix openGLtransformation = LoadTransformationMatrix(objectXML).ComputeInverseMatrix();
            Vector center = openGLtransformation.GetTranslation();
            return new GPoint(center, color, name, ID);
        }

        private GraphicObject LoadLine(XmlNode objectXML)
        {
            int ID = LoadID(objectXML);
            string name = LoadName(objectXML);
            GColor color = LoadColor(objectXML);            
            ParametricEquation parametricEquation = LoadParametricEquation(objectXML);
            LineParametricEquation lineEquation = new LineParametricEquation(parametricEquation.CopyValues(parametricEquation));
            TransformationMatrix openGLtransformation = LoadTransformationMatrix(objectXML).ComputeInverseMatrix();
            CenterParametrization center = new CenterParametrization(openGLtransformation.GetTranslation());
            return new Line(lineEquation, center, color, name, ID);
        }

        private GraphicObject LoadPlan(XmlNode objectXML)
        {
            int ID = LoadID(objectXML);
            string name = LoadName(objectXML);
            GColor color = LoadColor(objectXML);
            CommonEquation equation = LoadCommonEquation(objectXML);
            PlanEquation planEquation = new PlanEquation(equation.CopyValues(equation));
            return new Plan(planEquation, color, name, ID);
        }

        private GraphicObject LoadConic(XmlNode objectXML)
        {
            int ID = LoadID(objectXML);
            string name = LoadName(objectXML);
            GColor color = LoadColor(objectXML);
            CommonEquation equation = LoadCommonEquation(objectXML);
            ConicEquation conicEquation = new ConicEquation(equation.CopyValues(equation));
            TransformationMatrix transformationMatrix = LoadTransformationMatrix(objectXML);
            if (!TransformationMatrix.IsTransformationMatrix(transformationMatrix))
                throw new Exception("Uložená matice není transformační matice");
            return  new ConicRecognizer(conicEquation, color, name, ID).Recognize(transformationMatrix);
        }

        private GraphicObject LoadQuadric(XmlNode objectXML)
        {
            int ID = LoadID(objectXML);
            string name = LoadName(objectXML);
            GColor color = LoadColor(objectXML);
            CommonEquation commonEquation = LoadCommonEquation(objectXML);
            QuadricEquation quadricEquation = new QuadricEquation(commonEquation.CopyValues(commonEquation));
            return new QuadricRecognizer(quadricEquation, color, name, ID).RecognizeQuadric();
        }

        private GraphicObject LoadIntersection(XmlNode objectXML)
        {
            int ID = LoadID(objectXML);
            string name = LoadName(objectXML);
            GColor color = LoadColor(objectXML);

            XmlNodeList objectsIDsXML = objectXML.SelectNodes("objectID");
            int objectAID = int.Parse(objectsIDsXML[0].InnerText);
            int objectBID = int.Parse(objectsIDsXML[1].InnerText);
            GraphicObject objectA = null;
            GraphicObject objectB = null;
            foreach (GraphicObject graphicObject in _graphicObjects) 
            {
                if (graphicObject.GetID() == objectAID)
                    objectA = graphicObject;
                if (graphicObject.GetID() == objectBID)
                    objectB = graphicObject;
            }

            if (objectA == null || objectB == null)
                throw new Exception();
            return new IntersectionRecognizer(objectA, objectB, color, name, ID).Recognize();
        }




        public List<IHistoryItem> LoadHistoryItems(List<GraphicObject> graphicObjects)
        {
            _graphicObjects = graphicObjects;
            List<IHistoryItem> historyItems = new List<IHistoryItem>();
            XmlNode historyXML = _document.SelectSingleNode("document").SelectSingleNode("history");
            XmlNodeList historyItemsXML = historyXML.SelectNodes("historyItem");
            foreach (XmlNode historyItemXML in historyItemsXML)
            {
                historyItems.Add(LoadIHistoryItem(historyItemXML));    
            }
            return historyItems;
        }

        public int LoadHistoryPosition() 
        {
            XmlNode historyPositionXML = _document.SelectSingleNode("document")
                .SelectSingleNode("history").SelectSingleNode("historyPosition");
            return int.Parse(historyPositionXML.InnerText);
        }


        private IHistoryItem LoadIHistoryItem(XmlNode historyItemXML)
        {
            if (historyItemXML.SelectSingleNode("historyItemType").InnerText == "combinedHistoryItem")
                return LoadCombinedHistoryItem(historyItemXML);
            return LoadHistoryItem(historyItemXML);
        }

        private CombinedHistoryItem LoadCombinedHistoryItem(XmlNode historyItemXML)
        {
            HistoryOperation historyOperation =
                ConvertStringOnHistoryOperation(historyItemXML.SelectSingleNode("historyOperation").InnerText);
            List<HistoryItem> historyItems = new List<HistoryItem>();
            XmlNodeList historyItemsXML = historyItemXML.SelectNodes("historyItem");
            foreach (XmlNode itemXML in historyItemsXML)
            {
                historyItems.Add(LoadHistoryItem(itemXML));
            }
            return new CombinedHistoryItem(historyItems, historyOperation);
        }

        private HistoryItem LoadHistoryItem(XmlNode historyItemXML)
        {
            HistoryOperation historyOperation =
                ConvertStringOnHistoryOperation(historyItemXML.SelectSingleNode("historyOperation").InnerText);
            switch (historyOperation)
            {
                case HistoryOperation.AddObject: return LoadAddObjectHistoryOperation(historyItemXML, historyOperation);
                case HistoryOperation.DeleteObject: return LoadDeleteObjectHistoryOperation(historyItemXML, historyOperation);
                case HistoryOperation.UnDeleteObject: return LoadUnDeleteObjectHistoryOperation(historyItemXML, historyOperation);
                case HistoryOperation.SwitchObjects: return LoadSwitchObjectsHistoryOperation(historyItemXML, historyOperation);
                case HistoryOperation.ChangeColor: return LoadChangeColorHistoryOperation(historyItemXML, historyOperation);
                case HistoryOperation.ChangeName: return LoadChangeNameHistoryOperation(historyItemXML, historyOperation);
                case HistoryOperation.ChangeObject: return LoadChangeObjectHistoryOperation(historyItemXML, historyOperation);
                case HistoryOperation.ChangeObjectAxesVisibility: return LoadChangeObjectAxesVisibilityHistoryOperation(historyItemXML, historyOperation);
                case HistoryOperation.ChangeObjectVisibility: return LoadChangeObjectVisibilityHistoryOperation(historyItemXML, historyOperation);
            }
            throw new Exception();
        }

        private HistoryItem LoadAddObjectHistoryOperation(XmlNode historyItemXML, 
            HistoryOperation historyOperation) 
        {
            return LoadAddOrDeleteHistoryOperation(historyItemXML, historyOperation);
        }

        private HistoryItem LoadDeleteObjectHistoryOperation(XmlNode historyItemXML,
            HistoryOperation historyOperation)
        {
            return LoadAddOrDeleteHistoryOperation(historyItemXML, historyOperation);
        }

        private HistoryItem LoadUnDeleteObjectHistoryOperation(XmlNode historyItemXML,
            HistoryOperation historyOperation)
        {
            return LoadAddOrDeleteHistoryOperation(historyItemXML, historyOperation);
        }

        private HistoryItem LoadAddOrDeleteHistoryOperation(XmlNode historyItemXML,
            HistoryOperation historyOperation) 
        {
            int subjectID = int.Parse(historyItemXML.SelectSingleNode("subject").SelectSingleNode("ID").InnerText);
            return new HistoryItem(subjectID, historyOperation, null, null);
        }

        private HistoryItem LoadChangeNameHistoryOperation(XmlNode historyItemXML,
            HistoryOperation historyOperation)
        {
            return LoadChangeObjectPropertyHistoryOperation(historyItemXML, historyOperation);
        }

        private HistoryItem LoadChangeColorHistoryOperation(XmlNode historyItemXML, 
            HistoryOperation historyOperation)
        {
            return LoadChangeObjectPropertyHistoryOperation(historyItemXML, historyOperation);
        }

        private HistoryItem LoadChangeObjectVisibilityHistoryOperation(XmlNode historyItemXML,
            HistoryOperation historyOperation)
        {
            return LoadChangeObjectPropertyHistoryOperation(historyItemXML, historyOperation);
        }

        private HistoryItem LoadChangeObjectAxesVisibilityHistoryOperation(XmlNode historyItemXML,
            HistoryOperation historyOperation)
        {
            return LoadChangeObjectPropertyHistoryOperation(historyItemXML, historyOperation);
        }

        private HistoryItem LoadChangeObjectPropertyHistoryOperation(XmlNode historyItemXML,
            HistoryOperation historyOperation)
        {
            int subjectID = int.Parse(historyItemXML.SelectSingleNode("subject").SelectSingleNode("ID").InnerText);            
            object oldValue = LoadHistoryArgument(historyItemXML.SelectSingleNode("oldValue"));
            object newValue = LoadHistoryArgument(historyItemXML.SelectSingleNode("newValue"));
            return new HistoryItem(subjectID, historyOperation, oldValue, newValue);
        }

        private HistoryItem LoadChangeObjectHistoryOperation(XmlNode historyItemXML,
            HistoryOperation historyOperation)
        {
            object oldValue = LoadHistoryArgument(historyItemXML.SelectSingleNode("oldValue"));
            object newValue = LoadHistoryArgument(historyItemXML.SelectSingleNode("newValue"));
            return new HistoryItem(HistoryItem.NOTaObject, historyOperation, oldValue, newValue);
        }

        private HistoryItem LoadSwitchObjectsHistoryOperation(XmlNode historyItemXML, 
            HistoryOperation historyOperation)
        {
            object oldValue = LoadHistoryArgument(historyItemXML.SelectSingleNode("oldValue"));
            object newValue = LoadHistoryArgument(historyItemXML.SelectSingleNode("newValue"));
            return new HistoryItem(HistoryItem.NOTaObject, historyOperation, oldValue, newValue);
        }


        private object LoadHistoryArgument(XmlNode historyItemArgumentXML)
        {
            switch (historyItemArgumentXML.FirstChild.Name)
            {
                case "boolean": return LoadBoolean(historyItemArgumentXML);
                case "color": return LoadColor(historyItemArgumentXML);
                case "name": return LoadName(historyItemArgumentXML);
                case "center": return LoadCenter(historyItemArgumentXML);
                case "transformationMatrix": return LoadTransformationMatrix(historyItemArgumentXML);
                case "ID": return LoadID(historyItemArgumentXML);
                case "": return null;
                default: throw new Exception();
            }
        }

        private HistoryOperation ConvertStringOnHistoryOperation(string historyItemOperationString) 
        {
            switch (historyItemOperationString)
            {
                case "addObject": return HistoryOperation.AddObject;
                case "deleteObject": return HistoryOperation.DeleteObject;
                case "unDeleteObject": return HistoryOperation.UnDeleteObject;
                case "changeColor": return HistoryOperation.ChangeColor;
                case "switchObjects": return HistoryOperation.SwitchObjects;
                case "changeName": return HistoryOperation.ChangeName;
                case "changeObject": return HistoryOperation.ChangeObject;
                case "changeObjectAxesVisibility": return HistoryOperation.ChangeObjectAxesVisibility;
                case "changeObjectVisibility": return HistoryOperation.ChangeObjectVisibility;
                default: throw new Exception("Neznámá history operace.");
            }
        }



        private CommonEquation LoadCommonEquation(XmlNode subjectXML)
        {
            List<double> values = new List<double>();
            XmlNodeList xmlEquation = subjectXML.SelectSingleNode("commonEquation").SelectNodes("value");
            foreach (XmlNode xmlValue in xmlEquation)
                values.Add(GetNumber(xmlValue.InnerText));
            return new CommonEquation(values);
        }

        private ParametricEquation LoadParametricEquation(XmlNode subjectXML)
        {
            List<double> values = new List<double>();
            XmlNodeList xmlEquation = subjectXML.SelectSingleNode("parametricEquation").SelectNodes("value");
            foreach (XmlNode xmlValue in xmlEquation)
                values.Add(GetNumber(xmlValue.InnerText));
            return new ParametricEquation(values);
        }

        private TransformationMatrix LoadTransformationMatrix(XmlNode subjectXML) 
        {
            XmlNode widthXML = subjectXML.SelectSingleNode("transformationMatrix").SelectSingleNode("width");
            XmlNode heightXML = subjectXML.SelectSingleNode("transformationMatrix").SelectSingleNode("height");
            int width = int.Parse(widthXML.InnerText);
            int height = int.Parse(heightXML.InnerText);
            double[] values = new double[width*height];
            XmlNodeList valuesXML = subjectXML.SelectSingleNode("transformationMatrix").SelectNodes("value");
            int i = 0;
            foreach (XmlNode valueXML in valuesXML) 
            {
                values[i++] = GetNumber(valueXML.InnerText);
            }
            return new TransformationMatrix(width, height, values);
        }

        private Vector LoadCenter(XmlNode subjectXML)
        {
            XmlNode centerXML = subjectXML.SelectSingleNode("center");
            XmlNode xXML = centerXML.SelectSingleNode("x");
            XmlNode yXML = centerXML.SelectSingleNode("y");
            XmlNode zXML = centerXML.SelectSingleNode("z");
            double x = GetNumber(xXML.InnerText);
            double y = GetNumber(yXML.InnerText);
            double z = GetNumber(zXML.InnerText);
            return new Vector(x, y, z);
        }

        private GColor LoadColor(XmlNode subjectXML) 
        {
            XmlNode colorXML = subjectXML.SelectSingleNode("color");

            double r = GetNumber(colorXML.SelectSingleNode("r").InnerText);
            double g = GetNumber(colorXML.SelectSingleNode("g").InnerText);
            double b = GetNumber(colorXML.SelectSingleNode("b").InnerText);
            if (!IsColorComponetValid(r) || !IsColorComponetValid(g) || !IsColorComponetValid(b))
                throw new Exception();
            return new GColor(r, g, b);
        }
        private bool IsColorComponetValid(double component) 
        {
            return component >= 0 && component <= 1;
        } 

        private string LoadName(XmlNode subjectXML) 
        {
            XmlNode nameXML = subjectXML.SelectSingleNode("name");
            return nameXML.InnerText;
        }

        private bool LoadBoolean(XmlNode subjectXML) 
        {
            XmlNode boolXML = subjectXML.SelectSingleNode("boolean");
            return bool.Parse(boolXML.InnerText);
        }

        private int LoadID(XmlNode subjectXML) 
        {
            XmlNode IDXML = subjectXML.SelectSingleNode("ID");
            return int.Parse(IDXML.InnerText);
        }

        private bool LoadVisible(XmlNode subjectXML)
        {
            XmlNode visibleXML = subjectXML.SelectSingleNode("visible");
            return bool.Parse(visibleXML.InnerText);
        }

        private bool LoadAxesVisible(XmlNode subjectXML)
        {
            XmlNode axesVisibleXML = subjectXML.SelectSingleNode("axesVisible");
            return bool.Parse(axesVisibleXML.InnerText);
        }

        private double GetNumber(string stringNumber) 
        {
            if (System.Globalization.CultureInfo.CurrentUICulture.NumberFormat.NumberDecimalSeparator == ".")
                stringNumber = stringNumber.Replace(",", ".");
            if (System.Globalization.CultureInfo.CurrentUICulture.NumberFormat.NumberDecimalSeparator == ",")
                stringNumber = stringNumber.Replace(".", ",");
            return double.Parse(stringNumber);
        }
    }
}