﻿using System;
using System.Collections.Generic;
using System.Xml;
using System.IO;

using Geometry.GraphicObjects;
using Geometry.GraphicObjects.Planar;
using Geometry.GraphicObjects.ConicSections;
using Geometry.GraphicObjects.Quadrics;  
using Geometry.GraphicObjects.Equations;
using Geometry.GraphicObjects.Intersections;
using GMath;

namespace Geometry.LoadSave
{
    class SaveManager
    {
        private XmlDocument _document;
        private string _baseFilePath;

        public SaveManager(string baseFilePath)
        {
            try
            {
                _baseFilePath = baseFilePath;
                //_baseFilePath = Path.Combine(AppDomain.CurrentDomain.BaseDirectory, name);
                _document = new XmlDocument();
                //_document.Load(_baseFilePath);
            }
            catch (FileNotFoundException e) { Console.WriteLine("xml dokument daneho nazvu nenalezen."); throw e; }
        }



        private void SaveDocument()
        {
            using (StreamWriter writer = new StreamWriter
                (_baseFilePath, false))
            //prekladac vi ze objekt ma platnost v ramci bloku usiing
            {
                _document.Save(writer);
                writer.Close();
            }
        }

        public void Save(List<GraphicObject> graphicObjects, History history) 
        {
            XmlNode documentXML = _document.CreateElement("document");
            SaveObjects(graphicObjects, documentXML);
            SaveHistory(history, documentXML);
            _document.AppendChild(documentXML);
            SaveDocument();
        }

        private void SaveObjects(List<GraphicObject> graphicObjects, XmlNode documentXML)
        {
            XmlNode sceneXML = _document.CreateElement("scene");
            XmlNode objectsXML = _document.CreateElement("objects");

            for (int i = 0; i < graphicObjects.Count; i++)
            {
                SaveGraphicObject(graphicObjects[i], objectsXML);
            }
            sceneXML.AppendChild(objectsXML);
            documentXML.AppendChild(sceneXML);
        }

        private void SaveGraphicObject(GraphicObject graphicObject, XmlNode objects)
        {
            XmlNode graphicObjectXML = _document.CreateElement("graphicObject");

            SaveType(graphicObject, graphicObjectXML);
            SaveID(graphicObject.GetID(), graphicObjectXML);
            SaveName(graphicObject.GetName(), graphicObjectXML);
            SaveColor(graphicObject.GetColor(), graphicObjectXML);
            SaveVisible(graphicObject.GetVisible(), graphicObjectXML);
            if (graphicObject is IIntersection)
                SaveIntersection(((IIntersection)graphicObject), graphicObjectXML);
            else
                if (graphicObject is WithCommonEquationObject)
                {
                    SaveAxesVisible(((WithParametricEquationObject)graphicObject).GetAxesVisible(), graphicObjectXML);
                    SaveCommonEquation(((WithCommonEquationObject)graphicObject).GetCommonEquation(), graphicObjectXML);
                    SaveTransformationMatrix(((WithParametricEquationObject)graphicObject).GetTransformationMatrixAlgebraic(), graphicObjectXML);
                }
                else if (graphicObject is WithParametricEquationObject)
                {
                    SaveAxesVisible(((WithParametricEquationObject)graphicObject).GetAxesVisible(), graphicObjectXML);
                    SaveParametricEquation(((WithParametricEquationObject)graphicObject).GetParametricEquation(), graphicObjectXML);
                    SaveTransformationMatrix(((WithParametricEquationObject)graphicObject).GetTransformationMatrixAlgebraic(), graphicObjectXML);
                }
            objects.AppendChild(graphicObjectXML);
        }



        private void SaveIntersection(IIntersection intersection, XmlNode subject)
        {
            XmlNode objectAIDXML = _document.CreateElement("objectID");
            XmlNode objectBIDXML = _document.CreateElement("objectID");
            objectAIDXML.InnerText = intersection.GetObjectA().GetID().ToString();
            objectBIDXML.InnerText = intersection.GetObjectB().GetID().ToString();
            subject.AppendChild(objectAIDXML);
            subject.AppendChild(objectBIDXML);
        }

        private void SaveType(GraphicObject graphicObject, XmlNode subjectXML) 
        {
            string value = "";
            if (graphicObject is Axes) value = "axes"; 
            if (graphicObject is GPoint) value = "point"; 
            if (graphicObject is Line) value = "line"; 
            if (graphicObject is Plan) value = "plan"; 
            if (graphicObject is ConicSection) value = "conic"; 
            if (graphicObject is Quadric) value = "quadric";
            if (graphicObject is IIntersection) value = "intersection";
            if (value == "") throw new Exception("Grafický objekt má neznámý typ.");
            

            XmlNode typeXML = _document.CreateElement("type");
            typeXML.InnerText = value;
            subjectXML.AppendChild(typeXML);
        }



        private void SaveCommonEquation(CommonEquation equation, XmlNode subjectXML) 
        {
            XmlNode equationXML = _document.CreateElement("commonEquation");

            int count = equation.GetMembersCount();
            for (int i = 0; i < count; i++) 
            {
                XmlNode temp = _document.CreateElement("value");
                temp.InnerText = equation.Get(i).ToString();
                equationXML.AppendChild(temp);
            }
            subjectXML.AppendChild(equationXML);
        }

        private void SaveParametricEquation(ParametricEquation equation, XmlNode subjectXML) 
        {
            XmlNode equationXML = _document.CreateElement("parametricEquation");
            for (int i = 0; i < equation.GetMembersCount(); i++) 
            {
                XmlNode value = _document.CreateElement("value");
                value.InnerText = equation.Get(i).ToString();
                equationXML.AppendChild(value);
            }
            subjectXML.AppendChild(equationXML);    
        }

        private void SaveTransformationMatrix(TransformationMatrix matrix, XmlNode subjectXML) 
        {
            XmlNode matrixXML = _document.CreateElement("transformationMatrix");
            XmlNode widthXML = _document.CreateElement("width");
            XmlNode heightXML = _document.CreateElement("height");
            widthXML.InnerText = matrix.GetWidth().ToString();
            heightXML.InnerText = matrix.GetHeight().ToString();
            matrixXML.AppendChild(widthXML);
            matrixXML.AppendChild(heightXML);
            for (int h = 0; h < matrix.GetHeight(); h++)
                for (int w = 0; w < matrix.GetWidth(); w++) 
                {
                    XmlNode temp = _document.CreateElement("value");
                    temp.InnerText = matrix.Get(h, w).ToString();
                    matrixXML.AppendChild(temp);
                }
            subjectXML.AppendChild(matrixXML);
        }

        private void SaveCenter(Vector center, XmlNode subjectXML) 
        {
            XmlNode centerXML = _document.CreateElement("center");
            XmlNode xXML = _document.CreateElement("x");
            XmlNode yXML = _document.CreateElement("y");
            XmlNode zXML = _document.CreateElement("z");
            xXML.InnerText = center.GetX().ToString();
            yXML.InnerText = center.GetY().ToString();
            zXML.InnerText = center.GetZ().ToString();
            centerXML.AppendChild(xXML);
            centerXML.AppendChild(yXML);
            centerXML.AppendChild(zXML);
            subjectXML.AppendChild(centerXML);
        }

        private void SaveColor(GColor color, XmlNode subjectXML) 
        {
            XmlNode colorXML = _document.CreateElement("color");
            XmlNode rXML = _document.CreateElement("r");
            XmlNode gXML = _document.CreateElement("g");
            XmlNode bXML = _document.CreateElement("b");
            rXML.InnerText = color.GetRed().ToString();
            gXML.InnerText = color.GetGreen().ToString();
            bXML.InnerText = color.GetBlue().ToString();
            colorXML.AppendChild(rXML);
            colorXML.AppendChild(gXML);
            colorXML.AppendChild(bXML);
            subjectXML.AppendChild(colorXML);
        }

        private void SaveName(string name, XmlNode subjectXML) 
        {
            XmlNode nameXML = _document.CreateElement("name");
            nameXML.InnerText = name;
            subjectXML.AppendChild(nameXML);
        }

        private void SaveID(int ID, XmlNode subjectXML)
        {
            XmlNode IDXML = _document.CreateElement("ID");
            IDXML.InnerText = ID.ToString();
            subjectXML.AppendChild(IDXML);
        }

        private void SaveVisible(bool visible, XmlNode subjectXML)
        {
            XmlNode visibleXML = _document.CreateElement("visible");
            visibleXML.InnerText = visible.ToString();
            subjectXML.AppendChild(visibleXML);
        }

        private void SaveAxesVisible(bool axesVisible, XmlNode subjectXML)
        {
            XmlNode axesVisibleXML = _document.CreateElement("axesVisible");
            axesVisibleXML.InnerText = axesVisible.ToString();
            subjectXML.AppendChild(axesVisibleXML);
        }

        private void SaveBoolean(bool value, XmlNode subjectXML)
        {
            XmlNode valueXML = _document.CreateElement("boolean");
            valueXML.InnerText = value.ToString();
            subjectXML.AppendChild(valueXML);
        }




        private void SaveHistory(History history, XmlNode documentXML)
        {
            XmlNode historyXML = _document.CreateElement("history");
            XmlNode historyPositionXML = _document.CreateElement("historyPosition");
            historyPositionXML.InnerText = history.GetHistoryPosition().ToString();
            historyXML.AppendChild(historyPositionXML);
            foreach (IHistoryItem historyItem in history.GetHistoryItems()) 
            {
                if (historyItem is HistoryItem)
                    SaveHistoryItem((HistoryItem)historyItem, historyXML);
                else SaveHistoryItem((CombinedHistoryItem)historyItem, historyXML);
            }
            documentXML.AppendChild(historyXML);
        }

        private void SaveHistoryItem(CombinedHistoryItem historyItem, XmlNode historyXML)
        {
            XmlNode historyItemXML = _document.CreateElement("historyItem");
            XmlNode historyItemTypeXML = _document.CreateElement("historyItemType");
            historyItemTypeXML.InnerText = "combinedHistoryItem";
            historyItemXML.AppendChild(historyItemTypeXML);

            XmlNode historyOperationXML = _document.CreateElement("historyOperation");
            string historyOperationString = ConvertOnString(historyItem.GetHistoryOperation());
            historyOperationXML.InnerText = historyOperationString;
            historyItemXML.AppendChild(historyOperationXML);

            foreach (HistoryItem item in historyItem.GetHistoryItems())
                SaveHistoryItem(item, historyItemXML);

            historyXML.AppendChild(historyItemXML);
        }

        private void SaveHistoryItem(HistoryItem historyItem, XmlNode historyXML) 
        {
            XmlNode historyItemXML = _document.CreateElement("historyItem");
            XmlNode historyItemTypeXML = _document.CreateElement("historyItemType");
            historyItemTypeXML.InnerText = "historyItem";
            historyItemXML.AppendChild(historyItemTypeXML);

            XmlNode historyOperationXML = _document.CreateElement("historyOperation");
            string historyOperationString = ConvertOnString(historyItem.GetHistoryOperation());
            historyOperationXML.InnerText = historyOperationString;
            historyItemXML.AppendChild(historyOperationXML);


            if (historyItem.GetSubjectID() != HistoryItem.NOTaObject)
            {
                XmlNode subjectXML = _document.CreateElement("subject");
                SaveID(historyItem.GetSubjectID(), subjectXML);
                historyItemXML.AppendChild(subjectXML);
            }
            
            XmlNode oldValueXML = _document.CreateElement("oldValue");
            SaveHistoryArgument(historyItem.GetOldValue(), oldValueXML);
            historyItemXML.AppendChild(oldValueXML);

            XmlNode newValueXML = _document.CreateElement("newValue"); 
            SaveHistoryArgument(historyItem.GetNewValue(), newValueXML);
            historyItemXML.AppendChild(newValueXML);

            historyXML.AppendChild(historyItemXML);
        }

        private void SaveHistoryArgument(object argument, XmlNode historyItemArgumentXML) 
        {
            if (argument is bool) SaveBoolean((bool)argument, historyItemArgumentXML);
            if (argument is GColor) SaveColor((GColor)argument, historyItemArgumentXML);
            if (argument is string) SaveName((string)argument, historyItemArgumentXML);
            if (argument is Vector) SaveCenter((Vector)argument, historyItemArgumentXML);
            if (argument is TransformationMatrix) SaveTransformationMatrix((TransformationMatrix)argument, historyItemArgumentXML);
            if (argument is int) SaveID((int)argument, historyItemArgumentXML);
        }

        private string ConvertOnString(HistoryOperation historyOperation) 
        {
            switch (historyOperation) 
            {
                case HistoryOperation.AddObject: return "addObject";
                case HistoryOperation.DeleteObject: return "deleteObject";
                case HistoryOperation.UnDeleteObject: return "unDeleteObject";
                case HistoryOperation.ChangeColor: return "changeColor";
                case HistoryOperation.ChangeName: return "changeName";
                case HistoryOperation.ChangeObject: return "changeObject";
                case HistoryOperation.SwitchObjects: return "switchObjects";
                case HistoryOperation.ChangeObjectAxesVisibility: return "changeObjectAxesVisibility";
                case HistoryOperation.ChangeObjectVisibility: return "changeObjectVisibility";
                default: throw new Exception("Neznámá history operace."); 
            }
        }



        private string ConvertOnString(double number) 
        {
            string stringNumber = number.ToString();
            if (stringNumber.Contains("."))
                stringNumber.Replace(".", ",");
            return stringNumber;
        }
    }
}