﻿using System;
using System.Collections.Generic;
using System.Windows.Forms;
using Geometry.BaseClasses;
using Geometry.GGraphic.Forms.AddForms;
using Geometry.GGraphic.Forms.Detail;
using Geometry.GGraphic.Forms.Detail.QQIntersection;
using Geometry.GGraphic.Forms.Detail.Dialogs;
using Geometry.GraphicObjects;
using Geometry.GraphicObjects.ConicSections;
using Geometry.GraphicObjects.Equations;
using Geometry.GraphicObjects.Intersections;
using Geometry.GraphicObjects.Planar;
using Geometry.GraphicObjects.Quadrics;
using GMath;
using Tao.Platform.Windows;

namespace Geometry
{
    public class ManipulationManager
    {
        private History _history;
        private List<GraphicObject> _graphicObjects;
        private List<IDialog> _graphicObjectDialogs;
        private List<GraphicObjectDetail> _graphicObjectDetails;
        private Form1 _form;
        private AddManager _addManager;

        private int _lastSaveHistoryCount;
        private int _lastSaveHistoryPosition;

        public ManipulationManager(Form1 form) 
        {
            _form = form;
            _addManager = new AddManager(form, this);
            _history = new History(this);
            _graphicObjects = new List<GraphicObject>();
            _graphicObjectDialogs = new List<IDialog>();
            _graphicObjectDetails = new List<GraphicObjectDetail>();
            _lastSaveHistoryCount = _history.GetHistoryItems().Count;
            _lastSaveHistoryPosition = _history.GetHistoryPosition();
            Axes axes = new Axes(0);
            ApplyAddObject(axes);
        }

        public GraphicObject GetObjectWithID(int id) 
        {
            if (id < _graphicObjects.Count && id >= 0)
                return _graphicObjects[id];
            throw new Exception("Index out of range");
        }
        public int GetNewGraphicObjectID()
        {
            return _graphicObjects.Count;
        }
        public AddManager GetAddManager() 
        {
            return _addManager;
        }
        public History GetHistory()
        {
            return _history;
        }
        public void AddToHistory(IHistoryItem item)
        {
            _form.EnableUndo();
            _history.AddToHistory(item);
            _form.DisableRedo();
        }
        private void AddGraphicObject(GraphicObject graphicObject)
        {
            _graphicObjects.Add(graphicObject);
        }
        public void ControlUndoRedoButtons()
        {
            if (_history.GetUndoCount() == 0)
                _form.DisableUndo();
            else
                _form.EnableUndo();

            if (_history.GetRedoCount() != 0)
                _form.EnableRedo();
            else
                _form.DisableRedo();
        }
        /// <summary>
        /// Vrací seznam všech objektů, které kdy byly přidány nebo vznikly modifikací objektů
        /// </summary>
        /// <returns></returns>
        public List<GraphicObject> GetGraphicObjects()
        {
            return _graphicObjects;
        }
        /// <summary>
        /// Vrací seznam všech objektů, které kdy byly přidány nebo vznikly modifikací objektů.
        /// Avšak, pokud je otevřen Interaktivní dialog, potom simuluje na objektech provedení storno.
        /// Díky tomu, že průniky se ukládají jen jako informace o tom, že A a B mají průnik, není třeba je aktualizovat.
        /// </summary>
        /// <returns></returns>        
        internal List<GraphicObject> GetGraphicObjectsToSave()
        {
            List<GraphicObject> tempObjects = new List<GraphicObject>();
            foreach (GraphicObject o in _graphicObjects)
                tempObjects.Add(o);
            AngleDialog dialog;
            for (int i = 0; i < tempObjects.Count; i++)
            {
                GraphicObject o = tempObjects[i];
                if ((dialog = CheckAngleDialogOpened(o)) != null)
                {
                    WithParametricEquationObject mustBeParametricObject = (WithParametricEquationObject)o;
                    WithParametricEquationObject copy = mustBeParametricObject.MakeCopyOfObject(o.GetID());
                    copy.SetRotationInStandardBasis(dialog.GetInitialRotation());
                    tempObjects[i] = copy;
                }
            }
            return tempObjects;
        }
        public void LoadNewScene(List<GraphicObject> loadGraphicObjects, List<IHistoryItem> loadHistoryItems, int historyPosition)
        {
            ClearScene();
            _graphicObjects = loadGraphicObjects;
            _history.LoadHistory(loadHistoryItems, historyPosition);
            foreach (GraphicObject graphicObject in _graphicObjects)
                graphicObject.BuildDisplayList(_form.GetScene(), graphicObject.GetRelativeQuality());
        }
        public void ClearScene()
        {
            int count = _graphicObjectDetails.Count;
            for (int i = 0; i < count; i++)
                _graphicObjectDetails[0].Close();
            count = _graphicObjectDialogs.Count;
            for (int i = 0; i < count; i++)
                _graphicObjectDialogs[0].Close();
            _graphicObjectDetails = new List<GraphicObjectDetail>();            
            _graphicObjectDialogs = new List<IDialog>();
            _form.GetScene().DeleteObjects();
            _form.ClearListViewItems();
            _graphicObjects = new List<GraphicObject>();
            _history = new History(this);
            EnableInterActiveColisingActions();
            _form.DisableRedo();
            _form.DisableUndo();
            Axes axes = new Axes(0);
            ApplyAddObject(axes);
        }

        public void UndoClick()
        {
            _history.Undo();
            if (_history.GetUndoCount() == 0)
            {
                _form.DisableUndo();
            }
            if (_history.GetRedoCount() != 0)
            {
                _form.EnableRedo();
            }
        }
        public void RedoClick()
        {
            _history.Redo();
            if (_history.GetUndoCount() != 0)
            {
                _form.EnableUndo();
            }
            if (_history.GetRedoCount() == 0)
            {
                _form.DisableRedo();
            }
        }

        public void ListViewItemDeleteDown(ListView listView) 
        {
            if (listView.SelectedItems.Count < 1)
                return;
            List<GraphicObject> graphicObjectsToDelete = new List<GraphicObject>();
            foreach (ListViewItem item in listView.SelectedItems)
                if (!(item.Tag is Axes))
                    graphicObjectsToDelete.Add((GraphicObject)(item.Tag));
            DeleteObjectInvoked(graphicObjectsToDelete);
        }

        public void ListViewItemDoubleClick(ListViewItem selectedItem)
        {
            GraphicObject selectedGraphicObject = ((GraphicObject)selectedItem.Tag);
            GraphicObject detailedGraphicObject = null;
            foreach (GraphicObjectDetail detail in _graphicObjectDetails)
                if (detail.GetGraphicObject().GetID() == selectedGraphicObject.GetID())
                {
                    detailedGraphicObject = detail.GetGraphicObject();
                    break;
                }
            if (detailedGraphicObject == null)
                ShowGraphicObjectDetail((GraphicObject)selectedItem.Tag);
        }
        private void ShowGraphicObjectDetail(GraphicObject drawable)
        {
            GraphicObjectDetail detail = null;
            if (drawable is Quadric)            
                detail = new QuadricDetail(_form, (Quadric)drawable);            
            if (drawable is ConicSection)            
                detail = new ConicDetail(_form, (ConicSection)drawable);            
            if (drawable is Plan)            
                detail = new PlanDetail(_form, (Plan)drawable);            
            if (drawable is Line)
                detail = new LineDetail(_form, (Line)drawable);            
            if (drawable is GPoint)            
                detail = new PointDetail(_form, (GPoint)drawable);            
            if (drawable is IIntersection)
                detail = new IntersectionDetail(_form, (IIntersection)drawable);
            if (detail != null) 
            {
                if (EditDialogOpen(drawable))
                    detail.DisableEditButtons();
                detail.Show(_form);
                SetDetailLocation(detail);
                _form.DisableItemFromListBox(drawable);
                _graphicObjectDetails.Add(detail);
                detail.Rescale();
            }            
        }
        private void SetDetailLocation(GraphicObjectDetail detail)
        {
            System.Drawing.Rectangle resolution = System.Windows.Forms.Screen.PrimaryScreen.Bounds;
            int x = Math.Min(_form.Location.X + _form.GetListBoxLocationX(), resolution.Width - detail.Width);
            int y = Math.Min(_form.Location.Y + _form.GetListBoxLocationY(), resolution.Height - detail.Height);
            detail.Location = new System.Drawing.Point(x, y);
        }
        private void SetDialogLocation(IDialog detail)
        {
            System.Drawing.Rectangle resolution = System.Windows.Forms.Screen.PrimaryScreen.Bounds;
            int x = Math.Min(_form.Location.X + _form.GetListBoxLocationX(), resolution.Width - detail.Width);
            int y = Math.Min(_form.Location.Y + _form.GetListBoxLocationY(), resolution.Height - detail.Height);
            detail.Location = new System.Drawing.Point(x, y);
        }
        public void CloseGraphicObjectDetail(GraphicObject drawable)
        {
            foreach (GraphicObjectDetail detail in _graphicObjectDetails)
                if (detail.GetGraphicObject().GetID() == drawable.GetID())
                {
                    detail.Close();
                    return;
                }
        }
        /// <summary>
        /// Smaze detail ze seznamu detailu a povoli otevreni detailu objektu.
        /// </summary>
        public void DetailClosed(GraphicObjectDetail graphicObjectDetail)
        {
            _graphicObjectDetails.Remove(graphicObjectDetail);
            _form.Activate();
            _form.EnableItemFromListBox(graphicObjectDetail.GetGraphicObject());
        }

        private void DisableEditButtons(GraphicObject graphicObject) 
        {
            for (int i = 0; i < _graphicObjectDetails.Count; i++)
                if (_graphicObjectDetails[i].GetGraphicObject().GetID() == graphicObject.GetID())
                {
                    _graphicObjectDetails[i].DisableEditButtons();
                }
        }
        private void EnableEditButtons(GraphicObject graphicObject)
        {
            for (int i = 0; i < _graphicObjectDetails.Count; i++)
                if (_graphicObjectDetails[i].GetGraphicObject().GetID() == graphicObject.GetID())
                {
                    _graphicObjectDetails[i].EnableEditButtons();
                }
        }
        /// <summary>
        /// Ve všech otevřených detailech vymění grafický objekt.
        /// Současně danému detailu aktualizuje obsah.
        /// </summary>
        private void SwapDetailsAllContents(GraphicObject oldGraphicObject, GraphicObject newGraphicObject)
        {
            for (int i = 0; i < _graphicObjectDetails.Count; i++)
                if (_graphicObjectDetails[i].GetGraphicObject().GetID() == oldGraphicObject.GetID())
                {
                    _graphicObjectDetails[i].GenerateContents(newGraphicObject);
                }
        }
        /// <summary>
        /// Ve všech otevřených detailech vymění grafický objekt.
        /// Současně danému detailu aktualizuje obsah.
        /// </summary>
        private void SwapDetailsDrawableContents(GraphicObject oldGraphicObject, GraphicObject newGraphicObject)
        {
            for (int i = 0; i < _graphicObjectDetails.Count; i++)
                if (_graphicObjectDetails[i].GetGraphicObject().GetID() == oldGraphicObject.GetID())
                {
                    _graphicObjectDetails[i].GenerateDrawableContents(newGraphicObject);
                }
        }
        public void SetObjectsQualityInvoked(BaseQuality quality)
        {
            _form.SetProgressBarText("Probíhá přetváření všech objektů.");
            Scene scene = _form.GetScene();
            scene.SetBaseQuality(quality);
            RebuildAllGraphicObjects();
            _form.SetProgressBarText("");
            _form.RenderScene();
        }
        public void SetMSAAInvoked() 
        {
            _form.SetProgressBarText("Probíhá nastavování vzorkování.");
            RebuildAllGraphicObjects();
            _form.SetProgressBarText("");
            _form.RenderScene();
        }
        public void DeleteAllDisplayLists()
        {
            Scene scene = _form.GetScene();
            List<GraphicObject> allGraphicObjects = new List<GraphicObject>();
            foreach (GraphicObject g in _graphicObjects)
                allGraphicObjects.Add(g);
            bool add;
            for (int i = 0; i < scene.GetObjectsCount(); i++)
            {
                add = true;
                for (int j = 0; j < allGraphicObjects.Count; j++)
                {
                    if (scene.ObjectAt(i) == allGraphicObjects[j]) // aby se nemazaly objekty dvakrat, jinak se ale mažou všechny.
                        add = false;
                }
                if (add)
                    allGraphicObjects.Add(scene.ObjectAt(i));
            }
            for (int i = 0; i < allGraphicObjects.Count; i++)
                allGraphicObjects[i].DeleteDisplayList();
        }
        private void RebuildAllGraphicObjects() 
        {
            Scene scene = _form.GetScene();
            List<GraphicObject> allGraphicObjects = new List<GraphicObject>();
            foreach (GraphicObject g in _graphicObjects)
                allGraphicObjects.Add(g);
            bool add;
            for (int i = 0; i < scene.GetObjectsCount(); i++)
            {
                add = true;
                for (int j = 0; j < allGraphicObjects.Count; j++)
                {
                    if (scene.ObjectAt(i) == allGraphicObjects[j]) // aby se nebuildovaly objekty dvakrat, jinak se ale buildují všechny.
                        add = false;
                }
                if (add)
                    allGraphicObjects.Add(scene.ObjectAt(i));
            }
            for (int i = 0; i < allGraphicObjects.Count; i++)
                allGraphicObjects[i].BuildDisplayList(scene, allGraphicObjects[i].GetRelativeQuality());
        }

        private bool EditDialogOpen(GraphicObject graphicObject) 
        {
            foreach (IDialog dialog in _graphicObjectDialogs)
                if (dialog.GetGraphicObject().GetID() == graphicObject.GetID())
                    return true;
            return false;
        }
        private void ShowEditDialog(IDialog dialog, GraphicObject graphicObject)
        {
            DisableEditButtons(graphicObject);
            _graphicObjectDialogs.Add(dialog);
            dialog.Show(_form);
            dialog.StartPosition = FormStartPosition.Manual;
            SetDialogLocation(dialog);
            DisableInterActiveColisingActions();
        }
        private void CloseDialogs(GraphicObject graphicObject)
        {
            for (int i = _graphicObjectDialogs.Count - 1; i > -1; i--)
                if (_graphicObjectDialogs[i].GetGraphicObject().GetID() == graphicObject.GetID())
                    _graphicObjectDialogs[i].Close();
        }
        private void RemoveDialogs(GraphicObject graphicObject) 
        {
            for (int i = _graphicObjectDialogs.Count - 1; i > -1; i--)
                if (_graphicObjectDialogs[i].GetGraphicObject().GetID() == graphicObject.GetID())
                    _graphicObjectDialogs.RemoveAt(i--);
            EnableEditButtons(graphicObject);
            if (!CheckInterActiveDialogOpened())
                EnableInterActiveColisingActions();
        }
        private void RemoveDialog(IDialog dialog) 
        {
            _graphicObjectDialogs.Remove(dialog);
            EnableEditButtons(dialog.GetGraphicObject());
            if (!CheckInterActiveDialogOpened())
                EnableInterActiveColisingActions();
        }
        /// <summary>
        /// Ve všech otevřených editačních dialozích vymění grafický objekt.
        /// Současně danému dialogu aktualizuje obsah.
        /// </summary>
        private void SwapDialogsObjects(GraphicObject oldGraphicObject, GraphicObject newGraphicObject) 
        {
            for (int i = 0; i < _graphicObjectDialogs.Count; i++)
                if (_graphicObjectDialogs[i].GetGraphicObject().GetID() == oldGraphicObject.GetID())
                {
                    _graphicObjectDialogs[i].GenerateContents(newGraphicObject);                  
                }
        }

        public void ChangeStandardCenterInvoked(WithParametricEquationObject graphicObject)
        {
            CenterParametrization center = new CenterParametrization(graphicObject.GetCenterInStandardBasis());
            EditStandardCenterDialog editCenter = new EditStandardCenterDialog(this, graphicObject);
            ShowEditDialog(editCenter, graphicObject);
        }
        public void ChangeCanonicCenterInvoked(WithCanonicEquationObject graphicObject) 
        {
            CenterParametrization center = new CenterParametrization(graphicObject.GetCenterInCanonicBasis());
            EditCanonicCenterDialog editCenter = new EditCanonicCenterDialog(this, graphicObject);
            ShowEditDialog(editCenter, graphicObject);
        }
        public void EditStandardCenterClosed(EditStandardCenterDialog editCenter) 
        {
            RemoveDialog(editCenter);
            if (editCenter.DialogResult != DialogResult.OK)
                return;
            WithParametricEquationObject graphicObject = (WithParametricEquationObject)editCenter.GetGraphicObject();
            CenterParametrization newStandardCenter = editCenter.GetCenter();
            WithParametricEquationObject objectCopy = graphicObject.MakeCopyOfObject(GetNewGraphicObjectID());
            objectCopy.SetCenterInStandardBasis(newStandardCenter);
            ChangeObjectInvoked(graphicObject, objectCopy);
        }
        public void EditCanonicCenterClosed(EditCanonicCenterDialog editCenter)
        {
            RemoveDialog(editCenter);
            if (editCenter.DialogResult != DialogResult.OK)
                return;
            WithCanonicEquationObject graphicObject = (WithCanonicEquationObject)editCenter.GetGraphicObject();
            CenterParametrization newStandardCenter = graphicObject.ConvertCanonicToStandardBasis(editCenter.GetCenter());
            WithParametricEquationObject objectCopy = graphicObject.MakeCopyOfObject(GetNewGraphicObjectID());
            objectCopy.SetCenterInStandardBasis(newStandardCenter);
            ChangeObjectInvoked(graphicObject, objectCopy);
        }

        public void ChangeParametricEquationInvoked(WithParametricEquationObject graphicObject) 
        {
            EditParametricEquationDialog edit = new EditParametricEquationDialog(this, graphicObject);
            ShowEditDialog(edit, graphicObject);
        }
        public void EditParametricEquationsClosed(EditParametricEquationDialog editParametricEquation) 
        {
            RemoveDialog(editParametricEquation);
            if (editParametricEquation.DialogResult != DialogResult.OK)
                return;
            WithParametricEquationObject oldGraphicObject = (WithParametricEquationObject)editParametricEquation.GetGraphicObject();
            WithParametricEquationObject newGraphicObject =
                oldGraphicObject.MakeCopyOfObject(editParametricEquation.GetKoefsAndCenterPair(), GetNewGraphicObjectID());
            ChangeObjectInvoked(oldGraphicObject, newGraphicObject);
        }

        public void ChangeCanonicEquationInvoked(WithCanonicEquationObject graphicObject)
        {
            EditCanonicEquationDialog edit = new EditCanonicEquationDialog(this, graphicObject);
            ShowEditDialog(edit, graphicObject);
        }
        public void EditCanonicEquationClosed(EditCanonicEquationDialog editCanonicEquation)
        {
            RemoveDialog(editCanonicEquation);
            if (editCanonicEquation.DialogResult != DialogResult.OK)
                return;
            WithCanonicEquationObject oldGraphicObject = (WithCanonicEquationObject)editCanonicEquation.GetGraphicObject();
            WithParametricEquationObject newGraphicObject =
                oldGraphicObject.MakeCopyOfObject(editCanonicEquation.GetKoefsAndCenterPair(), GetNewGraphicObjectID());
            ChangeObjectInvoked(oldGraphicObject, newGraphicObject);
        }

        public void ChangeCommonEquationInvoked(WithCommonEquationObject graphicObject) 
        {
            EditCommonEquationDialog edit = new EditCommonEquationDialog(this, graphicObject);
            ShowEditDialog(edit, graphicObject);
        }
        public void EditCommonEquationClosed(EditCommonEquationDialog editCommonEquation)
        {
            RemoveDialog(editCommonEquation);
            if (editCommonEquation.DialogResult != DialogResult.OK)
                return;
            WithCommonEquationObject oldGraphicObject = (WithCommonEquationObject)editCommonEquation.GetGraphicObject();
            WithCommonEquationObject newGraphicObject =
                ((WithCommonEquationObject)oldGraphicObject).
                MakeCopyOfObject(editCommonEquation.GetCommonEquation(), GetNewGraphicObjectID());
            ChangeObjectInvoked(oldGraphicObject, newGraphicObject);
        }

        private Plan GetThePlan(GraphicObject graphicObject)
        {
            if (graphicObject is ConicSection)
                return ((ConicSection)graphicObject).GetConicPlan();
            if (graphicObject is Line)
                return ((Line)graphicObject).GetLinePlan();
            throw new Exception("Snazim se vytvorit detail rovnice objektu, ktery rovnici nema nebo se na nej zapomnelo.");
        }
        internal void EditObjectsPlanParametricalyInvoked(WithParametricEquationObject graphicObject)
        {
            Plan plan = GetThePlan(graphicObject);
            EditObjectsPlanParametricaly edit = new EditObjectsPlanParametricaly(this, graphicObject, plan);
            ShowEditDialog(edit, graphicObject);
        }
        internal void EditObjectsPlanParametricalyClosed(EditObjectsPlanParametricaly dialog)
        {
            RemoveDialog(dialog);
            if (dialog.DialogResult != DialogResult.OK)
                return;
            KoefsAndCenterPair koefsAndCenterPair = dialog.GetKoefsAndCenterPair();
            WithParametricEquationObject graphicObject = (WithParametricEquationObject)dialog.GetGraphicObject();
            Plan plan = GetThePlan(graphicObject);
            Plan newPlan = (Plan)plan.MakeCopyOfObject(koefsAndCenterPair, plan.GetID());
            TransformationMatrix planTransformationMatrix = newPlan.GetTransformationMatrixAlgebraic();
            WithParametricEquationObject newGraphicObject = graphicObject.MakeCopyOfObject(GetNewGraphicObjectID());
            TransformationMatrix objectsPlanTransformation = plan.GetTransformationMatrixAlgebraic();
            TransformationMatrix objectsTransformation = 
                objectsPlanTransformation.ComputeInverseMatrix() * graphicObject.GetTransformationMatrixAlgebraic();
            newGraphicObject.SetTransformationMatrixAlgebraic(planTransformationMatrix * objectsTransformation);
            ChangeObjectInvoked(graphicObject, newGraphicObject);
        }

        internal void EditObjectsPlanCommonlyInvoked(WithParametricEquationObject graphicObject)
        {
            Plan plan = GetThePlan(graphicObject);
            EditObjectsPlanCommonly edit = new EditObjectsPlanCommonly(this, graphicObject, plan);
            ShowEditDialog(edit, graphicObject);
        }
        internal void EditObjectsPlanCommonlyClosed(EditObjectsPlanCommonly dialog)
        {
            RemoveDialog(dialog);
            if (dialog.DialogResult != DialogResult.OK)
                return;
            CommonEquation commonEquation = dialog.GetCommonEquation();
            WithParametricEquationObject graphicObject = (WithParametricEquationObject)dialog.GetGraphicObject();
            Plan plan = GetThePlan(graphicObject);
            Plan newPlan = (Plan)plan.MakeCopyOfObject(commonEquation, plan.GetID());
            TransformationMatrix planTransformationMatrix = newPlan.GetTransformationMatrixAlgebraic();
            WithParametricEquationObject newGraphicObject = graphicObject.MakeCopyOfObject(GetNewGraphicObjectID());
            TransformationMatrix objectsPlanTransformation = GetThePlan(graphicObject).GetTransformationMatrixAlgebraic();
            TransformationMatrix objectsTransformation = 
                objectsPlanTransformation.ComputeInverseMatrix() * graphicObject.GetTransformationMatrixAlgebraic();
            newGraphicObject.SetTransformationMatrixAlgebraic(planTransformationMatrix * objectsTransformation);
            ChangeObjectInvoked(graphicObject, newGraphicObject);
        }

        internal void ChangeObjectsRotationInvoked(WithParametricEquationObject withParametricEquationObject)
        {
            AngleDialog angleDialog = new AngleDialog(withParametricEquationObject, this,
                                     withParametricEquationObject is WithCanonicEquationObject);
            ShowEditDialog(angleDialog, withParametricEquationObject);
        }
        internal void SetObjectsStandardRotation(AngleDialog dialog, RotationAngles angles)
        {
            WithParametricEquationObject graphicObject = (WithParametricEquationObject)dialog.GetGraphicObject();
            TransformationMatrix newStandardRotation = new TransformationMatrix(angles);
            ApplyOrSetNewRotationToObject(graphicObject, newStandardRotation, false, dialog.GetUseLowerQuality());
        }
        internal void ApplyObjectsStandardRotation(AngleDialog dialog, RotationAngles angles)
        {
            WithParametricEquationObject graphicObject = (WithParametricEquationObject)dialog.GetGraphicObject();
            TransformationMatrix newStandardRotation = new TransformationMatrix(angles);
            ApplyOrSetNewRotationToObject(graphicObject, newStandardRotation, true, dialog.GetUseLowerQuality());
        }
        internal void ApplyObjectsCanonicRotation(AngleDialog dialog, RotationAngles angles)
        {
            WithCanonicEquationObject graphicObject = (WithCanonicEquationObject)dialog.GetGraphicObject();
            TransformationMatrix oldStandardRotation = graphicObject.GetRotationInStandardBasis();
            TransformationMatrix newStandardRotation = new TransformationMatrix(angles, oldStandardRotation);
            ApplyOrSetNewRotationToObject(graphicObject, newStandardRotation, true, dialog.GetUseLowerQuality());
        }
        private void ApplyOrSetNewRotationToObject(WithParametricEquationObject graphicObject,
            TransformationMatrix newStandardRotation, bool apply, bool useLowerQuality)
        {
            if (apply)
                graphicObject.ApplyRotationInStandardBasis(newStandardRotation); // zmenim rotaci objektu            
            else
                graphicObject.SetRotationInStandardBasis(newStandardRotation); // zmenim rotaci objektu           
            RelativeQuality relativeQuality;
            if (!useLowerQuality)
                relativeQuality = Scene.QualityRelativeNormal;
            else relativeQuality = Scene.QualityRelativeLow;
            graphicObject.BuildDisplayListIfNecessary(_form.GetScene(), relativeQuality);
            RecomputeIntersectionsOf(graphicObject, graphicObject, false, relativeQuality);
            SwapDetailsAllContents(graphicObject, graphicObject);
            //SwapDialogsObjects(graphicObject, graphicObject);
            _form.RenderScene();
        }
        internal void EditObjectsRotationClosed(AngleDialog angleDialog)
        {
            RemoveDialog(angleDialog);
            WithParametricEquationObject graphicObject = (WithParametricEquationObject)angleDialog.GetGraphicObject();
            TransformationMatrix initialRotation = angleDialog.GetInitialRotation();
            TransformationMatrix actualRotation = graphicObject.GetRotationInStandardBasis();
            graphicObject.SetRotationInStandardBasis(initialRotation);
            graphicObject.BuildDisplayListIfNecessary(_form.GetScene(), Scene.QualityRelativeNormal);
            /// DisplayListy se maží ručně.
            /// Každý objekt je mazán, jen pokud se mění úroveň MSAA nebo kvalita, nebo se ukončuje program nebo AddIntersection.
            /// Dočasné průniky se smažou jakmile jsou nahrazeny dalším průnikem. Tyto průniky vznikají jen při rotaci.
            /// Prvním otočením se smaže i displayList průniku, který je uložen mezi grafickými objetky, protože se ale po uzavření dialogu rotace, 
            /// kvůli možným přidaným průnikům během rotace aktializují původní objekty průniku, tak to nevadí.
            ActualizeDuringRotatingAddedIntersections(graphicObject);
            if (angleDialog.DialogResult != DialogResult.OK)
            {
                //ApplyRotationStorno(graphicObject, initialRotation, actualRotation);
                ApplySwitchObjects(graphicObject,graphicObject);
            }
            else 
            {
                WithParametricEquationObject newGraphicObject = graphicObject.MakeCopyOfObject(GetNewGraphicObjectID());
                newGraphicObject.SetRotationInStandardBasis(actualRotation);
                ChangeObjectInvoked(graphicObject, newGraphicObject);
            }
        }
        private void ActualizeDuringRotatingAddedIntersections(GraphicObject rotatedObject) 
        {
            for (int i = 0; i < _graphicObjects.Count; i++) 
            {
                if (_graphicObjects[i] is IIntersection) 
                {
                    IIntersection intersection = (IIntersection)_graphicObjects[i];
                    if (rotatedObject.GetID() == intersection.GetObjectA().GetID())
                    {
                        _graphicObjects[i] = intersection.RecomputeIntersection(_form.GetScene(),
                            rotatedObject, intersection.GetObjectB(), intersection.GetID());                        
                    }
                    else if (rotatedObject.GetID() == intersection.GetObjectB().GetID())
                    {
                        _graphicObjects[i] = intersection.RecomputeIntersection(_form.GetScene(),
                            intersection.GetObjectA(), rotatedObject, intersection.GetID());
                    }
                    else continue;
                    _graphicObjects[i].CopyDrawableSlots(_graphicObjects[i], intersection);
                    _graphicObjects[i].BuildDisplayListIfNecessary(_form.GetScene(), Scene.QualityRelativeNormal);
                    _graphicObjects[i].SetDeleted(intersection.GetDeleted());
                    _form.GetScene().ChangeObject(intersection, _graphicObjects[i]);
                    SwapDetailsAllContents(intersection, _graphicObjects[i]);
                }
            }
        }


        public void ChangeNameInvoked(GraphicObject graphicObject) 
        {
            ChangeNameDialog changeNameDialog = new ChangeNameDialog(this, graphicObject);
            ShowEditDialog(changeNameDialog, graphicObject);
        }
        public void ChangeNameClosed(ChangeNameDialog changeNameDialog)
        {
            RemoveDialog(changeNameDialog);
            if (changeNameDialog.DialogResult == System.Windows.Forms.DialogResult.OK)
            {
                GraphicObject graphicObject = changeNameDialog.GetGraphicObject();
                string oldName = graphicObject.GetName();
                string name = changeNameDialog.GetName();
                HistoryItem item = new HistoryItem(graphicObject.GetID(),
                                                    HistoryOperation.ChangeName,
                                                    oldName, name);
                AddToHistory(item);                                   // pridam item do historie
                AplyChangeName(graphicObject, name);
            }
        }
        public void ChangeColorInvoked(List<GraphicObject> graphicObjects)
        {
            if (graphicObjects.Count < 1)
                return;
            ColorDialog colorDialog = new ColorDialog();
            colorDialog.FullOpen = true;
            colorDialog.Color = Program.ColorToSystemColor(graphicObjects[0].GetColor());
            HistoryOperation colorOperation = HistoryOperation.ChangeColor;
            if (colorDialog.ShowDialog() == DialogResult.OK)
            {
                GColor newColor = Program.SystemColorToColor(colorDialog.Color);
                List<HistoryItem> historyItems = new List<HistoryItem>();
                foreach(GraphicObject graphicObject in graphicObjects)
                {
                    historyItems.Add(new HistoryItem(
                        graphicObject.GetID(), colorOperation, graphicObject.GetColor(), newColor));
                    ApplyChangeColor(graphicObject, newColor);
                }
                CombinedHistoryItem item = new CombinedHistoryItem(
                                   historyItems,
                                   HistoryOperation.ChangeColor);
                AddToHistory(item);
            }
        }
        public void AddObjectInvoked(GraphicObject graphicObject) 
        {
            HistoryItem item =
                new HistoryItem(
                    graphicObject.GetID(),
                    HistoryOperation.AddObject,
                    null, null);
            AddToHistory(item);
            ApplyAddObject(graphicObject);
        }
        public void DeleteObjectInvoked(GraphicObject graphicObject) 
        {
            HistoryItem delete =
                new HistoryItem(
                graphicObject.GetID(),
                HistoryOperation.DeleteObject,
                null, null);
            AddToHistory(delete);
            ApplyDeleteObject(graphicObject);
        }
        public void DeleteObjectInvoked(List<GraphicObject> graphicObjects)
        {
            List<HistoryItem> historyItems = new List<HistoryItem>();
            HistoryOperation historyOperation = HistoryOperation.DeleteObject;
            foreach (GraphicObject graphicObject in graphicObjects)
            {
                historyItems.Add(new HistoryItem(
                    graphicObject.GetID(), historyOperation, null, null));
                ApplyDeleteObject(graphicObject);
            }
            CombinedHistoryItem item = new CombinedHistoryItem(
                               historyItems,
                               historyOperation);
            AddToHistory(item);
        }
        public void ChangeObjectInvoked(GraphicObject oldGraphicObject, GraphicObject newGraphicObject) 
        {
            HistoryItem item =
                new HistoryItem(
                    HistoryItem.NOTaObject,
                    HistoryOperation.ChangeObject,
                    oldGraphicObject.GetID(), newGraphicObject.GetID()
                    );
            AddToHistory(item);
            ApplyChangeObject(oldGraphicObject, newGraphicObject);
        }
        public void HideObjectsAxesInvoked(GraphicObject graphicObject, bool hide) 
        {
            HistoryItem item =
                new HistoryItem(
                graphicObject.GetID(),
                HistoryOperation.ChangeObjectAxesVisibility,
                !((WithParametricEquationObject)graphicObject).GetAxesVisible(), hide);
            AddToHistory(item);
            ApplyHideObjectsAxes(graphicObject, hide);
        }
        public void HideObjectAxesInvoked(List<GraphicObject> graphicObjects, bool hide)
        {
            List<HistoryItem> historyItems = new List<HistoryItem>();
            HistoryOperation historyOperation = HistoryOperation.ChangeObjectAxesVisibility;
            foreach (GraphicObject graphicObject in graphicObjects)
            {
                historyItems.Add(new HistoryItem(
                    graphicObject.GetID(), historyOperation, 
                    !((WithParametricEquationObject)graphicObject).GetAxesVisible(), hide));
                ApplyHideObjectsAxes(graphicObject, hide);
            }
            CombinedHistoryItem item = new CombinedHistoryItem(
                               historyItems,
                               historyOperation);
            AddToHistory(item);
        }        
        public void HideObjectInvoked(GraphicObject graphicObject, bool hide) 
        {
            HistoryItem item =
                new HistoryItem(
                graphicObject.GetID(),
                HistoryOperation.ChangeObjectVisibility,
                !graphicObject.GetVisible(), hide);
            AddToHistory(item);
            ApplyHideObject(graphicObject, hide);
        }
        public void HideObjectInvoked(List<GraphicObject> graphicObjects, bool hide)
        {
            List<HistoryItem> historyItems = new List<HistoryItem>();
            HistoryOperation historyOperation = HistoryOperation.ChangeObjectVisibility;
            foreach (GraphicObject graphicObject in graphicObjects)
            {
                historyItems.Add(new HistoryItem(
                    graphicObject.GetID(), historyOperation, !graphicObject.GetVisible(), hide));
                ApplyHideObject(graphicObject, hide);
            }
            CombinedHistoryItem item = new CombinedHistoryItem(
                               historyItems,
                               historyOperation);
            AddToHistory(item);
        }
        public void AddIntersectionInvoked(List<GraphicObject> graphicObjects) 
        {
            if (graphicObjects.Count != 2)
                throw new Exception("Nelze určit průnik " + graphicObjects.Count.ToString() + " objektů");

            GraphicObject a = graphicObjects[0];
            GraphicObject b = graphicObjects[1];
            IntPtr form1hRC = _form.GetCurrenthRC();
            IntPtr form1hDC = _form.GetCurrentDC();
            Gdi.PIXELFORMATDESCRIPTOR form1pfd = _form.GetCurrentPixelFormat();
            GColor color = Configuration.NewQuadricsIntersectionColor;
            string name = Configuration.NewQuadricsIntersectionName;
            AddIntersection addIntersectionDialog =
                new AddIntersection(_form, form1hDC, form1hRC, form1pfd, color, name);
            addIntersectionDialog.SetSelectedObjects(a,b);
            addIntersectionDialog.ShowDialog();
        }

        public void AplyChangeName(GraphicObject graphicObject, string name)
        {
            graphicObject.ChangeName(name);                         // potřeba kvuli historii (objekt z historie)
            foreach (GraphicObject o in _graphicObjects)
                if (o.GetID() == graphicObject.GetID())
                    o.ChangeName(name);                             // zmenim jmeno objektu
            for (int i = 0; i < _form.GetScene().GetObjectsCount(); i++)
            {
                GraphicObject o = _form.GetScene().ObjectAt(i);
                if (o.GetID() == graphicObject.GetID())
                    o.ChangeName(name);                             // zmenim jmeno objektu       
            }
            _form.ChangeItemFromListBox(graphicObject, graphicObject);// aktualizuji popisek objektu v listview
            ChangeIntersectionsNamesWith(graphicObject);              // aktualizuji take popisky pruniků s tímto objektem 
            SwapDetailsDrawableContents(graphicObject, graphicObject);// aktualizuji detaily tohoto objektu
            //SwapDialogsObjects(graphicObject, graphicObject);
            _form.Refresh();
        }
        public void ApplyChangeColor(GraphicObject graphicObject, GColor color)
        {
            graphicObject.ChangeColor(color);                           // potřeba kvuli historii (objekt z historie)
            foreach (GraphicObject o in _graphicObjects)
                if (o.GetID() == graphicObject.GetID())
                    o.ChangeColor(color);                               // zmenim barvu objektu
            for (int i = 0; i < _form.GetScene().GetObjectsCount(); i++)
            {
                GraphicObject o = _form.GetScene().ObjectAt(i);
                if (o.GetID() == graphicObject.GetID())
                    o.ChangeColor(color);                               // zmenim barvu objektu
            }
            SwapDetailsDrawableContents(graphicObject, graphicObject);// aktualizuji detaily tohoto objektu
            //SwapDialogsObjects(graphicObject, graphicObject);
            _form.Refresh();
        }
        public void ApplyAddObject(GraphicObject graphicObject) 
        {
            graphicObject.BuildDisplayList(_form.GetScene(), Scene.QualityRelativeNormal);

            AddGraphicObject(graphicObject);
            _form.GetScene().AddObject(graphicObject);
            _form.AddItemToListBox(graphicObject);
            _form.RenderScene();
        }
        public void ApplyDeleteObject(GraphicObject graphicObject) 
        {
            graphicObject.SetDeleted(true);
            graphicObject.SetInScene(false);
            for (int i = 0; i < _form.GetScene().GetObjectsCount(); i++)
            {
                GraphicObject o = _form.GetScene().ObjectAt(i);
                if (o.GetID() == graphicObject.GetID())
                    _form.GetScene().DeleteObject(o);           // smazu aktualni objekt
            }
            foreach (GraphicObject o in _graphicObjects)
                if (o.GetID() == graphicObject.GetID())
                    o.SetDeleted(true);                         // zmenim jmeno objektu

            _form.DeleteItemFromListBox(graphicObject);
            CloseGraphicObjectDetail(graphicObject);
            CloseDialogs(graphicObject);
            RemoveDialogs(graphicObject);
            RemoveIntersectionsWith(graphicObject);
            _form.Refresh();
        }
        public void ApplyUnDeleteObject(GraphicObject graphicObject) 
        {
            //if (graphicObject is IIntersection)
            //    return;
            graphicObject.SetDeleted(false);
            graphicObject.SetInScene(true);
            for (int i = 0; i < _graphicObjects.Count; i++)
            {
                GraphicObject o = _graphicObjects[i];
                if (o.GetID() == graphicObject.GetID())
                    _form.GetScene().AddObject(o);      // pridam aktualni objekt
            }
            _form.AddItemToListBox(graphicObject);
            AddIntersectionsWith(graphicObject);
            _form.Refresh();
        }
        public void ApplyChangeObject(GraphicObject oldGraphicObject, GraphicObject newGraphicObject) 
        {
            newGraphicObject.BuildDisplayList(_form.GetScene(), Scene.QualityRelativeNormal);
            AddGraphicObject(newGraphicObject);
            _form.GetScene().ChangeObject(oldGraphicObject, newGraphicObject);
            _form.ChangeItemFromListBox(oldGraphicObject, newGraphicObject);
            RecomputeIntersectionsOf(oldGraphicObject, newGraphicObject, true, Scene.QualityRelativeNormal);
            SwapDetailsAllContents(oldGraphicObject, newGraphicObject);
            SwapDialogsObjects(oldGraphicObject, newGraphicObject);
            _form.Refresh();
        }
        public void ApplySwitchObjects(GraphicObject oldGraphicObject, GraphicObject newGraphicObject) 
        {
            _form.GetScene().ChangeObject(oldGraphicObject, newGraphicObject);
            _form.ChangeItemFromListBox(oldGraphicObject, newGraphicObject);
            SwapDetailsAllContents(oldGraphicObject, newGraphicObject);
            SwapDialogsObjects(oldGraphicObject, newGraphicObject);
            SwapIntersectionsWith(oldGraphicObject, newGraphicObject);
            //RemoveIntersectionsWith(oldGraphicObject);
            //AddIntersectionsWith(newGraphicObject);
            _form.Refresh();
        }
        public void ApplyHideObjectsAxes(GraphicObject graphicObject, bool hide) 
        {
            WithParametricEquationObject drawable = (WithParametricEquationObject)graphicObject;
            drawable.SetAxesVisible(!hide);                             // potřeba kvuli historii (objekt z historie)

            foreach (GraphicObject o in _graphicObjects)
                if (o.GetID() == graphicObject.GetID() && o is WithParametricEquationObject)
                    ((WithParametricEquationObject)o).SetAxesVisible(!hide);  // zmeni viditelnost os objektu
            for (int i = 0; i < _form.GetScene().GetObjectsCount(); i++)
            {
                GraphicObject o = _form.GetScene().ObjectAt(i);
                if (o.GetID() == graphicObject.GetID() && o is WithParametricEquationObject)
                    ((WithParametricEquationObject)o).SetAxesVisible(!hide);  // zmeni viditelnost os objektu
            } 
            SwapDetailsDrawableContents(graphicObject, graphicObject);
            _form.Refresh();

        }
        public void ApplyHideObject(GraphicObject graphicObject, bool hide) 
        {
            graphicObject.SetVisible(!hide);                        // potřeba kvuli historii (objekt z historie)

            foreach (GraphicObject o in _graphicObjects)
                if (o.GetID() == graphicObject.GetID())
                    o.SetVisible(!hide);  // zmeni viditelnost objektu
            for (int i = 0; i < _form.GetScene().GetObjectsCount(); i++)
            {
                GraphicObject o = _form.GetScene().ObjectAt(i);
                if (o.GetID() == graphicObject.GetID())
                    o.SetVisible(!hide);  // zmeni viditelnost objektu
            } 
            SwapDetailsDrawableContents(graphicObject, graphicObject);
            _form.Refresh();        
        }


        ///<summary>
        /// AddToFormsObjects slouží pro indikaci, zda se má nový průnik přidat do seznamu objektů a dát mu nové ID.
        /// V podstatě znamená rozdil mezi Changing a Changed.
        /// 
        /// Dočasné průniky se smažou jakmile jsou nahrazeny dalším průnikem. Tyto průniky vznikají jen při rotaci.
        /// Prvním otočením se smaže i displayList průniku, který je uložen mezi grafickými objetky, protože se ale po uzavření dialogu rotace, 
        /// kvůli možným přidaným průnikům během rotace aktializují původní objekty průniku, tak to nevadí.
        ///</summary>
        private void IntersectionChanged(
            GraphicObject oldIntersection, GraphicObject newIntersection,
            bool addToFormsObjects, RelativeQuality relativeQuality)
        {
            if (addToFormsObjects)
            {
                //((QuadricAndQuadricIntersection)newIntersection)._temp = false;
                AddGraphicObject(newIntersection);
            }
            else
            {
                //((QuadricAndQuadricIntersection)newIntersection)._temp = true;
                oldIntersection.DeleteDisplayList();
            }
            newIntersection.CopyDrawableSlots(newIntersection, oldIntersection);
            _form.GetScene().ChangeObject(oldIntersection, newIntersection);
            _form.ChangeItemFromListBox(oldIntersection, newIntersection);
            SwapDetailsAllContents(oldIntersection, newIntersection);
            SwapDialogsObjects(oldIntersection, newIntersection);
            newIntersection.BuildDisplayList(_form.GetScene(), relativeQuality);
        }

        /// <summary>
        /// Pokud meneny object oldDrawable je v nejakem pruniku.
        /// Pak tento prunik prepocitam.
        /// NewDrawable je object zmeneny.
        /// AddToFormsObjects slouží pro indikaci, zda se má nový průnik přidat do seznamu objektů a dát mu nové ID.
        /// V podstatě znamená rozdil mezi Changing a Changed.
        /// </summary>
        public void RecomputeIntersectionsOf(
            GraphicObject oldDrawable, GraphicObject newDrawable, bool addToFormsObjects, RelativeQuality relativeQuality)
        {
            List<GraphicObject> objectsCopy = new List<GraphicObject>();
            for (int i = 0; i < _form.GetScene().GetObjectsCount(); i++)
            {
                objectsCopy.Add(_form.GetScene().ObjectAt(i));
            }
            for (int i = 0; i < objectsCopy.Count; i++)
            {
                if (objectsCopy[i] is IIntersection)
                {
                    RecomputeIntersection(
                        oldDrawable, newDrawable, (IIntersection)objectsCopy[i], addToFormsObjects, relativeQuality);
                }
            }
        }
        /// <summary>
        /// Mam dan prunik a kvadriku jez byla zmenena.
        /// Pokud byla soucasti tohoto pruniku, pak jej prepocitam.
        /// AddToFormsObjects slouží pro indikaci, zda se má nový průnik přidat do seznamu objektů a dát mu nové ID.
        /// V podstatě znamená rozdil mezi Changing a Changed.
        /// </summary>
        private void RecomputeIntersection(
            GraphicObject oldDrawable, GraphicObject newDrawable, IIntersection intersection,
            bool addToFormsObjects, RelativeQuality relativeQuality)
        {
            GraphicObject a = intersection.GetObjectA();
            GraphicObject b = intersection.GetObjectB();
            bool intersectionChanged = false;
            if (a.GetID() == oldDrawable.GetID())
            {
                a = newDrawable;
                intersectionChanged = true;
            }
            if (b.GetID() == oldDrawable.GetID())
            {
                b = newDrawable;
                intersectionChanged = true;
            }
            if (intersectionChanged)
            {
                int newID;
                if (addToFormsObjects) // indikuje, že se zmenil prunik a uz se menit nebude, akce byla pridana do historie
                    newID = GetNewGraphicObjectID(); // v tom pripade je novy prunik pridan do seznamu objektu s novym ID
                else newID = intersection.GetID(); // jinak neni pridan do seznamu objektu, a zustane mu stare ID.
                IntersectionChanged(
                    intersection,
                    intersection.RecomputeIntersection(_form.GetScene(), a, b, newID), addToFormsObjects, relativeQuality);
            }
        }



        public void AddIntersection(GraphicObject a, GraphicObject b, GColor color, string name)
        {
            int ID = GetNewGraphicObjectID();
            _form.SetProgressBarText("Probíhá generování průniku.");
            IntersectionRecognizer intersectionRecognizer = new IntersectionRecognizer(a, b, color, name, ID);
            IIntersection intersection = intersectionRecognizer.Recognize();
            _form.SetProgressBarText("");
            AddObjectInvoked(intersection);
        }



        /// <summary>
        /// GraphicObjectu se zmenilo jmeno, pokud je soucasti nejakeho pruniku, 
        /// pak tomuto pruniku aktualizuji popis.
        /// </summary>
        private void ChangeIntersectionsNamesWith(GraphicObject renamedGraphicObject)
        {
            List<GraphicObject> graphicObjects = GetGraphicObjects();
            for (int i = 0; i < graphicObjects.Count; i++)
            {
                if (graphicObjects[i] is IIntersection)
                {
                    IIntersection intersection = (IIntersection)graphicObjects[i];
                    GraphicObject objectA = intersection.GetObjectA();
                    GraphicObject objectB = intersection.GetObjectB();
                    if (renamedGraphicObject.GetID() == objectA.GetID() ||
                        renamedGraphicObject.GetID() == objectB.GetID())
                    {
                        _form.ChangeItemFromListBox(intersection, intersection);
                        SwapDialogsObjects(intersection, intersection);
                        SwapDetailsDrawableContents(intersection, intersection);
                    }
                }
            }
        }
        /// <summary>
        /// Smazany object se vratil do sceny, vratim tam tedy i pruniky kterych se ucastni.
        /// Je treba zkontrolovat, zda ma tento prunik oba tvorici objekty ve scene.
        /// </summary>
        private void AddIntersectionsWith(GraphicObject addedGraphicObject)
        {
            List<GraphicObject> graphicObjects = GetGraphicObjects();
            for (int i = 0; i < graphicObjects.Count; i++)
            {
                if (graphicObjects[i] is IIntersection)
                {
                    IIntersection intersection = (IIntersection)graphicObjects[i];
                    GraphicObject objectA = intersection.GetObjectA();
                    GraphicObject objectB = intersection.GetObjectB();
                    if (((addedGraphicObject.GetID() == objectA.GetID() && objectB.GetIsInScene()) ||
                         addedGraphicObject.GetID() == objectB.GetID() && objectA.GetIsInScene()) &&
                        !intersection.GetDeleted())
                    {
                        _form.AddItemToListBox(intersection);
                        _form.GetScene().AddObject(intersection);
                    }
                }
            }
        }
        /// <summary>
        /// Object byl smazan ze sceny, odstanim tedy i vsechny pruniky s timto objektem.
        /// Odstranim jej ze sceny, listview a zavru jeho detail.
        /// </summary>
        /// <param name="removedGraphicObject"></param>
        private void RemoveIntersectionsWith(GraphicObject removedGraphicObject)
        {
            List<GraphicObject> graphicObjects = new List<GraphicObject>();
            for (int i = 0 ; i < _form.GetScene().GetObjectsCount(); i++)
            {
                graphicObjects.Add(_form.GetScene().ObjectAt(i));
            }
            for (int i = 0; i < graphicObjects.Count; i++)
            {
                if (graphicObjects[i] is IIntersection)
                {
                    IIntersection intersection = (IIntersection)graphicObjects[i];
                    GraphicObject objectA = intersection.GetObjectA();
                    GraphicObject objectB = intersection.GetObjectB();
                    if (removedGraphicObject.GetID() == objectA.GetID() ||
                        removedGraphicObject.GetID() == objectB.GetID())
                    {
                        _form.DeleteItemFromListBox(intersection);
                        _form.GetScene().DeleteObject(intersection);
                        CloseGraphicObjectDetail(intersection);
                    }
                }
            }
        }
        private void SwapIntersectionsWith(GraphicObject removedGraphicObject, GraphicObject addedGraphicObject)
        {
            List<GraphicObject> graphicObjects = new List<GraphicObject>();
            for (int i = 0; i < _form.GetScene().GetObjectsCount(); i++)
            {
                graphicObjects.Add(_form.GetScene().ObjectAt(i));
            }
            for (int i = 0; i < graphicObjects.Count; i++)
            {
                if (graphicObjects[i] is IIntersection)
                {
                    IIntersection oldIntersection = (IIntersection)graphicObjects[i];
                    GraphicObject oldObjectA = oldIntersection.GetObjectA();
                    GraphicObject oldObjectB = oldIntersection.GetObjectB();
                    //najdu ve scene pruniky se smazanym objektem
                    if (removedGraphicObject.GetID() == oldObjectA.GetID() ||
                        removedGraphicObject.GetID() == oldObjectB.GetID())
                    {
                        GraphicObject secondIntersectionObject; // objekt, se kterym je removed v pruniku. (i added)
                        if (removedGraphicObject.GetID() == oldObjectA.GetID())
                            secondIntersectionObject = oldObjectB;
                        else secondIntersectionObject = oldObjectA;
                        for (int j = 0; j < _graphicObjects.Count; j++)
                        {
                            if (_graphicObjects[j] is IIntersection)// z nej ziskam ID 2. objektu pruniku
                            {
                                IIntersection newIntersection = (IIntersection)_graphicObjects[j];
                                GraphicObject newObjectA = newIntersection.GetObjectA();
                                GraphicObject newObjectB = newIntersection.GetObjectB();
                                // pokud se jedna o prunik noveho objektu a druheho objektu z prvniho pruniku.
                                if (!newIntersection.GetDeleted() &&
                                    !newIntersection.GetIsInScene() &&
                                    ((newObjectA.GetID() == addedGraphicObject.GetID() &&
                                    newObjectB.GetID() == secondIntersectionObject.GetID()) ||
                                    (newObjectA.GetID() == secondIntersectionObject.GetID() &&
                                    newObjectB.GetID() == addedGraphicObject.GetID())))
                                {
                                    _form.GetScene().ChangeObject(oldIntersection, newIntersection);
                                    _form.ChangeItemFromListBox(oldIntersection, newIntersection);
                                    SwapDetailsAllContents(oldIntersection, newIntersection);
                                }
                            }
                        }
                    }
                }
            }
            // melo by byt provedeno pouze pri nacitani sceny.
            for (int j = 0; j < _graphicObjects.Count; j++) // puvodni objekt nemel prunik, novy ano.
            {
                if (_graphicObjects[j] is IIntersection)
                {
                    IIntersection newIntersection = (IIntersection)_graphicObjects[j];
                    GraphicObject newObjectA = newIntersection.GetObjectA();
                    GraphicObject newObjectB = newIntersection.GetObjectB();
                    if (((newObjectA.GetID() == addedGraphicObject.GetID() && newObjectB.GetIsInScene()) ||
                         (newObjectB.GetID() == addedGraphicObject.GetID() && newObjectA.GetIsInScene())) &&
                         !newIntersection.GetDeleted() && !newIntersection.GetIsInScene())
                    {
                        _form.GetScene().AddObject(newIntersection);
                        _form.AddItemToListBox(newIntersection);
                    }
                }
            }
        }



        internal bool CheckInterActiveDialogOpened()
        {
            foreach (IDialog dialog in _graphicObjectDialogs)
                if (dialog is IInterActive)
                    return true;
            return false;
        }
        /// <summary>
        /// Vraci otevreny dialog rotace, jinak null
        /// </summary>
        /// <param name="graphicObject"></param>
        /// <returns></returns>
        internal AngleDialog CheckAngleDialogOpened(GraphicObject graphicObject)
        {
            foreach (IDialog dialog in _graphicObjectDialogs)
                if (dialog is AngleDialog && dialog.GetGraphicObject().GetID() == graphicObject.GetID())
                    return (AngleDialog)dialog;
            return null;
        }
        internal void DisableInterActiveColisingActions() 
        {
            _form.DisableInterActiveColisingActions();
        }
        internal void EnableInterActiveColisingActions() 
        {
            _form.EnableInterActiveColisingActions();
        }

        public bool SaveNeeded() 
        {
            return _lastSaveHistoryCount != _history.GetHistoryItems().Count ||
                _lastSaveHistoryPosition != _history.GetHistoryPosition(); 
        }
        public void SetSaveOcurred() 
        {
            _lastSaveHistoryCount = _history.GetHistoryItems().Count;
            _lastSaveHistoryPosition = _history.GetHistoryPosition();
        }


        internal static void FailOccured(Exception e)
        {
            MessageBox.Show(e.Message + "\n, stav programu bude uložen");            
        }
    }
}
